#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#ifdef UNIX
#include <unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/kd.h>
#include <sys/time.h>
#include <dlfnc.h>
#endif
#include "prolog.h"
#include "cdi.h"	// interface to user program
#include "cdiSedacLib.h"

#define WAIT_TIME_NS 1000000  // 1ms
int nSedacLines = N_SEDAC_LINES;
int SedacSimulate = 1;

__attribute__((constructor)) void cdiSedacMain(void)
{
    int cc = 0;

    if ((cc=cdiRegisterBusFilter("SEDAC",filterSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusInitialization("SEDAC",initSedacForCDI)) != 0 ) goto err;
    if ((cc=cdiRegisterBusHandler("SEDAC",execSedacBus)) != 0 ) goto err;
    if ((cc=cdiRegisterBusCleanup("SEDAC",exitSedac)) != 0 ) goto err;
    if ((cc=cdiRegisterBusScanner("SEDAC",sedacScanFcn)) != 0 ) goto err;
err:
    if (cc)
    {
      cdilog("%s : failure in bus registration SEDAC",erlst[cc]);
    }
    return;
}
SEDAC_STRUCT sedacInfo[N_SEDAC_LINES] = {
	{-1, -1}, {-1, -1}, {-1, -1}, {-1, -1},  {-1, -1}, {-1, -1}, {-1, -1}, {-1, -1}
};

char *scanCommand[] = {"getcommands","getlines","getcrates","getmodules","readmodules","writemodules"};
enum scommand {getcommands, getlines, getcrates, getmodules, readmodules, writemodules, nCommands};

int sedacScanFcn(char *target,char *result,int resultBufferSize)
{
  char *c, lclTarget[256];
  int i, j, cnt, ret, slen = 0;
  int nData, data[10];
  short rData, wData;
  short line = -1, crate = -1, subaddr = -1;
  CdiLineInfoBlk *busHandle;

  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  // parse target according to "commandString:<line>:<crate>:<subaddr>"; e.g. NULL or "command:1" or "command:1:48" or "command:1:48:16" 
  if (target == NULL || strlen(target) == 0 ) target = "getcommands";
  strncpy(lclTarget,target,255);
  strlwr(lclTarget);
  // get command string
  if ((c=strtok(lclTarget,":")) == NULL) return 0;  // nothing to do
  for (j=0; j<nCommands; j++) if ( strcmp(scanCommand[j], c) == 0 ) break;
  if ( j == nCommands ) j = getcommands; // return the available commands as 'help'
  for ( nData=0; (c=strtok(NULL,":")) != NULL; nData++ ) data[nData] = strtol(c,NULL,0); //atoi(c);

  
  if (j == getcommands)
  {
    _snprintf(result,resultBufferSize,"availabled command strings:\n");
    for (i=0; i<nCommands; i++)
    {
      if ((slen = strlen(result)) > resultBufferSize - 12) break;
      strncat(result,scanCommand[i],resultBufferSize-slen);
      if (i <nCommands-1) strcat(result,",");
    }
    return 0;
  }
  
  if ( j == getlines )
  {
    _snprintf(result,resultBufferSize,"Sedac lines configured:\n");
    for (i=0, cnt=0; i<N_SEDAC_LINES; i++)
    {
      if ( sedacInfo[i].sedac_fd < 0 ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
       _snprintf(&result[slen],resultBufferSize-slen,"%d",i+1);
      cnt++;
    }
    return 0;
  }
  
  line = data[0];
  if ( nData < 1 || line < 1 || line > nSedacLines || sedacInfo[line-1].sedac_fd < 0) return no_such_line;

  if ((busHandle=cdiGetBusHandle(sedacInfo[line-1].cdiLine)) == NULL)
  {
    cdilog("sedpc: Wrong cdi_line: %d for busscan() call\n",sedacInfo[line-1].cdiLine);
    return no_such_line;;
  }

  if ( j == getcrates )
  {
    _snprintf(result,resultBufferSize,"crates configured on line %d:\n",line);
    cdiMutexLock(&busHandle->lineMutex);
    for (crate=0, cnt=0; crate<32; crate++)
    {
      ret = sedac(line, crate, 253, READ, &rData, 1);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      _snprintf(&result[slen],resultBufferSize-slen,"%d",crate);
      cnt++;
    }
    cdiMutexUnlock(&busHandle->lineMutex);
    return 0;
  }
  else if ( j == getmodules )
  {
    crate = data[1];
    if ( nData < 2 || crate < 0 || crate >= 32 ) return device_not_connected;
    _snprintf(result,resultBufferSize,"Subaddress on line %d crate %d was found:\n",line,crate);
    cdiMutexLock(&busHandle->lineMutex);
    for (subaddr=0, cnt=0; subaddr<256; subaddr++)
    {
      ret = sedac(line, crate, subaddr, READ, &rData, 1);
      if ( ret ) continue;
      if ((slen = strlen(result)) > resultBufferSize - 1) break;
      if ( cnt > 0 )
      {
        _snprintf(&result[slen],resultBufferSize-slen,",");
        if ((slen = strlen(result)) > resultBufferSize - 1) break;
      }
      _snprintf(&result[slen],resultBufferSize-slen,"%d",subaddr);
      cnt++;
    }
    cdiMutexUnlock(&busHandle->lineMutex);
    return 0;
  }
  crate = data[1];
  subaddr = data[2];
  if ( nData < 3 || crate < 0 || crate >= 32 || subaddr < 0 || subaddr > 255) return device_not_connected;
  if ( j == readmodules )
  {
    _snprintf(result,resultBufferSize,"Read Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    cdiMutexLock(&busHandle->lineMutex);
    ret = sedac(line, crate, subaddr, READ, &rData, 1);
    cdiMutexUnlock(&busHandle->lineMutex);
    if ((slen = strlen(result)) < resultBufferSize - 18) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",rData,ret);
  }
  else if ( j == writemodules )
  {
    wData = data[3];
    if ( nData < 4 ) return argument_list_error;
    _snprintf(result,resultBufferSize,"Write Data on line %d crate %d subaddr %d:\n",line,crate,subaddr);
    cdiMutexLock(&busHandle->lineMutex);
    ret = sedac(line, crate, subaddr, WRITE, &wData, 1);
    cdiMutexUnlock(&busHandle->lineMutex);
    if ((slen = strlen(result)) < resultBufferSize - 18) _snprintf(&result[slen],resultBufferSize-slen,"%d <return code %d>",wData,ret);
  }  
  return 0;
}

int filterSedacBus(int *parameter, int accessFlag)
{
  if (parameter == NULL) return argument_list_error;
  return 0;
}
int initSedacForCDI(int busLine, int cdiLine, int numberDevices, char *parameterList)
{ int ret;

  ret = initSedac(busLine);
  if ( ret >= 0 ) sedacInfo[busLine - 1].cdiLine = cdiLine;
  return ret;
}
int doSedacRequest(CdiRequestInfoBlk *pReq, int step)
{ int access, ret, length, writeOffset, done;
  void *pOut;

  cdiMutexLock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  if ( step == 1 ) /* first request */
    access = (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag && pReq->accessFlag != cdiReadReadFlag) ? WRITE : READ;
  else if ( step == 2 )/* second */
    access = (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteWriteFlag ) ? WRITE : READ;
  else
    access = pReq->accessFlag == cdiWriteReadReadFlag ? READ : WRITE;

  if ( access == WRITE )
  {
    if ( pReq->accessFlag == cdiWriteWriteFlag || pReq->accessFlag == cdiWriteReadWriteFlag )
    {
      if ( step == 1 )
      {
        length = sizeof(short);
        pOut = (void *)pReq->pOutData;
        writeOffset = sedacSubOffsetSecond;
      }
      else
      {
        length = (pReq->accessFlag == cdiWriteReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : (pReq->outLength - sizeof(short));
        pOut = (void *)((char *)pReq->pOutData + sizeof(short));
        writeOffset = sedacSubOffsetThird;
      }
    }
    else
    {
      length = (pReq->accessFlag == cdiReadWriteFlag && pReq->conditionFlag == anyData) ? sizeof(short) : pReq->outLength;
      pOut = (void *)pReq->pOutData;
      writeOffset = sedacSubOffsetSecond;
    }
  }
  else
  {
    done = 0;
    if ( pReq->accessFlag == cdiWriteReadReadFlag || pReq->accessFlag == cdiReadReadFlag )
    {
      if ( (pReq->accessFlag == cdiWriteReadReadFlag && step == 2) || (pReq->accessFlag == cdiReadReadFlag && step == 1 ) )
      {
        length = pReq->inLength < pReq->pDev->dataOuputSize ? pReq->inLength : pReq->pDev->dataOuputSize;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
        done = 1;
      }
      else if ( step == 3 )
      {
        if ( pReq->inLength <= pReq->pDev->dataOuputSize ) return pReq->errorCode;;
        length = pReq->inLength < (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize) ? (pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize - pReq->inLength) : pReq->pDev->dataOuput2ndSize;
        pOut = (void *)((char *)pReq->pInData + pReq->pDev->dataOuputSize);
        writeOffset = sedacSubOffsetThird;
        done = 1;
      }
    }
    if ( done == 0 )
    {
        length = pReq->inLength;
        pOut = (void *)pReq->pInData;
        writeOffset = sedacSubOffset;
    }
  }

  ret = sedac((u16)((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busLine, 
          (u16)pReq->pDev->cdiAddr[caAddrIndex],
//          (u16)(pReq->pDev->cdiSubAddr + pReq->devParameters[step == 1 ? sedacSubOffset : sedacSubOffsetSecond]), 
          (u16)(pReq->pDev->cdiAddr[subAddrIndex] + pReq->devParameters[writeOffset]), 
          (int)access,
          (u16 *)pOut,
          (int)length / sizeof(short) );
  if ( ret != 0 )
  {
    pReq->errorCode = ret < 0 ? -ret : sedac_error;
    if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
    if ( ret > 0 )
    {
      pReq->deviceError = ret;    // return from hardware
      pReq->pDev->numberDeviceError++;
    }
    pReq->active = finishRequest; /* finish request */
  }

  cdiMutexUnlock(&((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->lineMutex);
  return pReq->errorCode;
}

/********************************************************/
// if sync call, data not pass filter then repeat Transmit
int checkData(CdiRequestInfoBlk *pReq)
{
  struct timespec tv;

  /* check return condition */
  checkReturnCondition(pReq, (char *)pReq->bufLastData, pReq->pInData, pReq->inLength, 0);
  if ( pReq->dataReturnCondition == DATA_PASS_FILTER)
    memcpy( pReq->bufLastData, pReq->pInData, pReq->inLength < 8 ? pReq->inLength : 8);

 	if ( pReq->conditionFlag > anyData && pReq->dataReturnCondition != DATA_PASS_FILTER )
	{
    if ( pReq->callback == NULL )
    {
      if ( getTimeMilliseconds() - pReq->requestTime >= pReq->timeout )
      {
         pReq->errorCode = invalid_data;	// data back for sync call
         if ( pReq->perror ) *(pReq->perror) = pReq->errorCode;
      }
      else
      {
        tv.tv_sec = 0; tv.tv_nsec = WAIT_TIME_NS; // delay 1ms
        nanosleep( &tv, NULL );    
        return 1;
      }
    }
    pReq->active = finishRequest;
  }
  return 0;
}
/********************************************************/
/* only call once with pReq->active == transRequest */
/********************************************************/
void execSedacBus(CdiRequestInfoBlk *pReq)
{ 
  /* do first part; */
  do
  {
    doSedacRequest(pReq, 1);
    if ( pReq->active == finishRequest ) return;
    /* check return condition */
    if (pReq->accessFlag != cdiReadFlag && pReq->accessFlag != cdiReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if ( pReq->active == finishRequest ) return;
  if ( pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
  {
    pReq->active = finishRequest;
    return;
  }

  /* do second part */
  do
  {
    doSedacRequest(pReq, 2);
    if ( pReq->active == finishRequest ) return;
    /* check return condition */
    if (pReq->accessFlag != cdiWriteReadFlag && pReq->accessFlag != cdiWriteReadWriteFlag) break;
    if ( checkData(pReq) == 0 ) break;
  } while (TRUE);

  if (pReq->accessFlag != cdiWriteReadWriteFlag && pReq->accessFlag != cdiWriteReadReadFlag)
  {
    pReq->active = finishRequest; /* finish request */
    return;
  }

  /* do third part */
  doSedacRequest(pReq, 3);
  pReq->active = finishRequest; /* finish request */
}

/* ============  init routine  =================================== */
int initSedac(int ln)
{ char fname[80];
  SEDAC_STRUCT *pSedac;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  if (SedacSimulate) return ln;		// ln number as file handle!

  pSedac = &sedacInfo[ln];
  if ( pSedac->sedac_fd >= 0 )
  {
    cdilog("sedac already opened! sedac line: %d!\n", ln+1);
    return pSedac->sedac_fd;
  }

  sprintf(fname,"/dev/sedac%d",ln);
  if(( pSedac->sedac_fd = open(fname, O_RDWR) ) >= 0 )
  {
    pSedac->da = NULL;
    pSedac->dataSize = 0;	
    pSedac->sedacSimulationCount = 0;

    cdilog("Open pktr sedac line OK! line: %d\n",ln+1);	
    return pSedac->sedac_fd;
  }

  cdilog( "initSedac for PKTR card open() error! line: %d; strerror: %s", ln+1, strerror(errno) );
  fprintf(stderr, "%s: %s; error Number: %d\n", fname, strerror(errno), errno);
  return -file_not_open;
}

// ============  exit routine  ===================================
int exitSedac(int ln)
{ SEDAC_STRUCT *pSedac;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  if (SedacSimulate) return 0;

  cdilog("close sedac line! line number: %d", ln+1);

  pSedac = &sedacInfo[ln];
  if ( pSedac->sedac_fd < 0 ) return -file_not_open;
  
  close(pSedac->sedac_fd);
  pSedac->sedac_fd = -1;

  if ( pSedac->da )
  {
    free(pSedac->da);
    pSedac->da = NULL;
    pSedac->dataSize = 0;	
    pSedac->sedacSimulationCount = 0;
  }

  return 0;
}

// ============  getBuffer  ===================================
PSEDAC_DATA getBuffer(int ln, int number)
{ SEDAC_STRUCT *pSedac;

  pSedac = &sedacInfo[ln];
  if ( pSedac->da )
  {
    free(pSedac->da);
    pSedac->da = NULL;
  }

  pSedac->da = (PSEDAC_DATA)malloc( SEDAC_DATA_SIZE * (number + 1) );
  if ( pSedac->da != NULL )	pSedac->dataSize = number;
  else
  {
    cdilog("sedac pktr line: %d; => get buffer failed!! size: %d\n",ln+1,number);
    return NULL;
  }
  cdilog("get buffer: %x; size: %d",(int)(pSedac->da),number);
  return pSedac->da;
}
int msleep(int msec)
{
  struct timespec ts;
  if (msec <= 0) return 0;
  ts.tv_sec = msec/1000;
  ts.tv_nsec = (msec%1000) * 1000000;
  return nanosleep(&ts,NULL);
}
int execSedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords, int flag)
{ int i, ret;
  SEDAC_STRUCT *pSedac;

  if (ln<1 || ln>nSedacLines) return -illegal_line;
  ln--;

  pSedac = &sedacInfo[ln];

  if ( SedacSimulate == 0 && (pSedac->da == NULL || pSedac->dataSize < nwords)
	&& getBuffer(ln, nwords) == NULL ) return -out_of_local_memory;
 
  if ( SedacSimulate )
  {
    i = nwords * 250;
    if ( i > 4 ) msleep ( i/4 ); else msleep(1);
    for (i=0; i<nwords; i++, data++)
    {
      ret = flag ? RWF[i] : RWF[0];
      if ( ret == READ )
      *data = pSedac->sedacSimulationBuffer + pSedac->sedacSimulationCount++;
      else
      pSedac->sedacSimulationBuffer = *data;
    }
    return 0;
  }

  if ( pSedac->sedac_fd < 0 ) return -file_not_open;

  for (i=0; i<nwords; i++) 
  {
    ret = flag ? RWF[i] : RWF[0];
  	if ( ret == WRITE ) pSedac->da[i+1].Data = data[i];
    pSedac->da[i+1].ExcessMode = ret == READ ? SE_READ : SE_WRITE;
    pSedac->da[i+1].Line = (BYTE)ln;
    pSedac->da[i+1].SubAdr = flag ? subadd[i] : subadd[0];
    pSedac->da[i+1].CaAdr = flag ? crate[i] : crate[0];
  }
  ret = ioctl(pSedac->sedac_fd, nwords, pSedac->da);
  if ( ret == nwords + 1 )
  {
    for (i=0; i<nwords; i++)
      if ( (flag ? RWF[i] : RWF[0]) == READ ) data[i] = pSedac->da[i+1].Data;
    return ((PSEDAC_HEADER)pSedac->da)->NumberOfError;
  }

  if ( ret < 0 )
  {
    cdilog("pktr sedac ioclt() error: line: %d, getLastError: %s\n",ln+1, strerror(errno));
    fprintf(stderr, "ioctl error (request #Telegrams= %d): %s; return code: %d \n",nwords, strerror(errno), ret);
    return -software_error;
  }
  return -operation_timeout;
}

// ================  read/write nwords telegram to same addesses  ===================================
int sedac(u16 ln, u16 crate, u16 subadd, int RWF, u16 *data, int nwords)
{
  return execSedac(ln, &crate, &subadd, &RWF, data, nwords, 0);
}

// ================  read/write nwords telegram to different addresses ===================================
int nsedac(u16 ln, u16 *crate, u16 *subadd, int *RWF, u16 *data, int nwords)
{
  return execSedac(ln, crate, subadd, RWF, data, nwords, 1);
}

