//	SedIni.c		230703		Stadtmueller	mst3	110703
//		SedPc & Sed1Pc & Sed2Pc from Version 98
//
//		SedLib, SedIni, SedExi
//	get lib version, initialize the library, clean up sedac-access
//


#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/io.h>
#include <stdlib.h>
#include "SedPc.h"

LONG	glDatReg, glStaReg, glCtrReg;	// PortRegister
LONG	glNoLin;			// Number of Lines found
LONG	glSelLin;			// least selected Line
LONG	glLogLin[XMLN];		// logical LineNumbers
LONG	glDevAdr[XMLN];		// DeviceAddr of Lines
LONG	glModAdr[XMLN];		// ModuleAddr of Lines (from Sed1Pc)
LONG	glTelDly;			// Tel DelayTime to get Response
LONG	glTickLoop;			// ForLoops per Tick for short WaitTimes

//-------+---------+---------+---------+---------+---------+---------+---------+


// SedLib	get the version of that SedPc.lib as string
//
//	call:	SedLib (lpszVer);
//	para:	lpszVer = array to put the token string, >= 64 character

void SedLib ( CHAR *lpszVer )

{
	strcpy (lpszVer, "SedPc.lib 250703 for SedPc, Sed1Pc and Sed2Pc from Vers. 
98");

	return;
}		// end of SedLib
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedIni	Look for environment-variable SEDPC, store that values in externals.
//			 If SEDPC variable not found, take line-number 1 and address
//			 hex 378, that's LPT1 on new PC's as an default for old SedPc.
//			 New Sed1Pc doesn't have an default value. You have to set the
//			 SEDPC environment-variable.
//
//		You get access to all lines (ports). This should be disabled by calling
//		SedExi before your exit-statement.
//
//		After execution of SedIni run SedClr to make sure that you have a
//		SedPc of the right typ connected to that lines.
//
//		example for setting SEDPC:
//		export SEDPC=LLN1-BA1[.M]:LLN2-BA2[.M]:...:LLNn-BAn[.M]
//			LLN - logical line-number will be taken as an decimal number
//			BA- board-adress will be taken as an hexa number
//			M - module-number of new Sed1Pc, not given for old SedPc
//		old PC's:LPT1 = 0x3BC, LPT2 = 0x378, LPT3 = 0x278
//		new PC's:LPT1 = 0x378, LPT2 = 0x278
//
//	call: SedIni (lplFndLin, lrglLinNum, lrglDevAdd, lplErr);
//	para:	lplFndLin= number of found Sedac line-numbers
//			lrglLinNum = array with logical line-numbers
//			lrglDevAdd = array with device-addresses fit to line-numbers
//				bit15 = 0 if SedPc; = 1 if Sed1Pc
//			 	bits 14..12 = 0 if SedPc; = module-number if Sed1Pc
//			*lplErr = 0 = no error
//				bit 14 = Error in IoPermission
//				bit 12 = no such line-number
//				bit 10 = array overflow
//
//	subp: dstrtu, hstrtu

void SedIni ( LONG *lplFndLin, LONG *lrglLinNum, LONG *lrglDevAdd, LONG *lplErr 
)
{

	static	LONG	err, i, j, k, n;
	static	BYTE	c, s[10];
	static	CHAR	*lpszEnv;

	glNoLin = glSelLin = *lplErr = 0;

	lpszEnv = getenv ("SEDPC");			// Environment Variable SEDPC
	if ( lpszEnv == NULL ) {			// not found, take default
		glNoLin++;
		glLogLin[0] = 1;
		glDevAdr[0] = 0x378;
		glModAdr[0] = SEDPC;
		goto ready;
	}

	i = 0;								// index to ev-string
	j = 0;								// index to token (number-string)
	k = 0;								// index to LinNum- and DevAdd-array
	while ( k < XMLN ) {
		c = *(lpszEnv + i++);			// get next character from ev-string
		switch (c) {
		case '-': 						// token is logical line-number
			s[j] = '\0';
			if ( dstrtu(s, &glLogLin[k] ) )
				goto err1;
			*(lrglLinNum+k) = glLogLin[k];
			j = 0;
			break;
		case ':': 						// token is device-address SedPc
			s[j] = '\0';
			if ( hstrtu(s, &glDevAdr[k]) )
				goto err1;
			*(lrglDevAdd+k) = glDevAdr[k];
			j = 0;
			glModAdr[k] = SEDPC;		// old SedPc marker to module-address
			glNoLin = ++k;
			break;
		case '\0' : 					// token is device-address SedPc
			s[j] = '\0';
			if ( hstrtu(s, &glDevAdr[k]) )
				goto err1;
			*(lrglDevAdd+k) = glDevAdr[k];
			j = 0;
			glModAdr[k] = SEDPC;		// old SedPc marker to module-address
			glNoLin = ++k;
			goto ready;
		case '.': 						// token is device-address Sed1Pc
			s[j] = '\0';
			if ( hstrtu(s, &glDevAdr[k]) )
				goto err1;
			j = 0;
			c = *(lpszEnv + i++);		// get next character from ev-string
			n = c & 7;
			glModAdr[k] = 1 << n;		// module-address to lower byte
			glModAdr[k] |= SED1PC; 		// Sed1Pc marker to upper byte
			*(lrglDevAdd+k) = glDevAdr[k] | n << 12 | 0x8000;
			glNoLin = ++k;
			c = *(lpszEnv + i++);		// get next character from ev-string
			if ( c == '\0' )
				goto ready;
			else
				break;
		default : 						// put character to token
			s[j++] = c;
			s[j] = '\0';
		} // switch
	} // while

	*lplErr = 0x400;				// error, array full but ev-string not empty

ready:
	*lplFndLin = glNoLin;
	glDatReg = glDevAdr[0];
	glStaReg = glDatReg + 1;
	glCtrReg = glDatReg + 2;
	*lplErr |= err;

	for ( i = 0; i < glNoLin; i++ ) {
		if ( ioperm(glDevAdr[i], 3, 1) ) {	// get access to ports
//			perror("ioperm");
			*lplErr |= 0x4000;
	}	}

	return;

err1:
	*lplErr = 0x1000; 				// illegal character in number
	goto ready;

}		// end of SedIni
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedExi	Disable access to the lines (ports).
//
//	call: SedExi (lplErr);
//	para:	*lplErr = 0 = no error
//				bit 14 = Error in IoPermission
//
//	subp: dstrtu, hstrtu

void SedExi ( LONG *lplErr )
{

	static	LONG	i;

	for ( i = 0; i < glNoLin; i++ ) {
		if ( ioperm(glDevAdr[i], 3, 0) ) {	// give up access to ports
//			perror("ioperm");
			*lplErr |= 0x4000;
	}	}

	return;

}		// end of SedExi
//-------+---------+---------+---------+---------+---------+---------+---------+

// end of file SedIni.c
