// SedTst.c 240703  Stadtmueller    mst3    230703
//      SedPc & Sed1Pc & Sed2Pc from Version 98
//
//      SedTst, SedFt, SedEst
// routines for sedac test-telegram, fifo-test and enable self-test
//

#include <stdlib.h>
#include "SedPc.h"

//-------+---------+---------+---------+---------+---------+---------+---------+

// SedTst   enable or disable a running "TEST"-telegram
//          A TEST-telegram get also the delay-time, so an old TEST-tel
//          takes the full delay-time.
//
//  call: SedTst (lLn, lServ, lplFcn, lplSa, lplDat, lplErr);
//  para:   lLn      =  logical line-number
//          lServ    =  service of this routine
//             0     =  disable TestTel other parameter are ignored
//             1     =  enable old TestTel (F=11)
//             3     =  enable TestTel with given parameter lplFcn & lplSa & lplDat
//             4     =  enable last TestTel, other parameter are ignored
//             8     =  ask for running TestTel, lplFcn = 0 if stoped
//          lplFcn   =  function-code and crate-number
//          lplSa    =  subaddress, 0-255 valid
//          lplDat   =  data, 16 bit to sedac, only valid with function-code write
//          *lplErr  = 0 = no error
//            bit 13 = illegal SEDPC type
//            bit 12 = no such line-number
//            bit  7 = ParameterError: one parameter incorrect
//            bit  6 = Response/Busy not set; data not taken by SEDPC
//            bit  5 = Bytecount mismatch
//            bit  4 = Timeout: no response from SEDPC
//            bit  3 = LastByte marker not set or illegal function
//  subp: sedwr1, sedwr2, sedwr4, sedrd1, sedrd2,
//
//  all parameter must be given for compatibility with basic.

void SedTst ( LONG lLn, LONG lServ, LONG *lplFcn,
				LONG *lplSa, LONG *lplDat, LONG *lplErr )

{
	static LONG	dat, err, bc;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	switch ( lServ ) {

	case 0 :                        // disable TestTel
		err  = sedwr4 (0x500, 0);
		err |= sedwr0 (LASTBY);
		goto ep01;

	case 1 :                        // enable old TestTel (F=11)
		err  = sedwr1 (0x7);
		err  = sedwr4 (0xC000, 0);
		err |= sedwr0 (LASTBY);
		goto ep01;

	case 3 :                        // enable test, fcn & sa & dat set
		if ( *lplFcn & 0xFF00 || *lplSa & 0xFF00 ) {
			*lplErr = 0x80;
			goto exit;
		}
		err  = sedwr1 (0x7);
		err |= sedwr1 (*lplFcn);
		err |= sedwr1 (*lplSa);
		err |= sedwr2 (*lplDat);
		err |= sedwr0 (LASTBY);
		goto ep01;

	case 4 :                        // enable last TestTel
		err  = sedwr4 (0x600, 0);
		err |= sedwr0 (LASTBY);
		goto ep01;

	case 8 :                        // ask for running TestTel
		err  = sedwr4 (0x800, 0);
		err |= sedwr0 (LASTBY);
		goto ep02;

	default :                       // error, wrong lServ
		*lplErr = 0x80;
		goto exit;
	}

ep01:
	if ( *lplErr = err )            // error set ?
		goto exit;
	err = sedrd1 (&dat);            // get sedac-status
	if ( (dat & 0xF0) != 0x20 )     // function or byte-count error ?
		err |= 0x20;
	err |= sedrd2 (&dat);           // get DummyData
	err |= sedrd1 (&dat);           // get LastByteMarker
	if ( dat != LASTBY )
		err |= 0x8;
	*lplErr = err;
	goto exit;

ep02:
	if ( *lplErr = err )            // error set ?
		goto exit;
	err = sedrd1 (&dat);            // get sedac-status
	if ( (dat & 0xF0) != 0x30 )     // function or byte-count error ?
		err |= 0x20;
	err |= sedrd2 (&bc);            // get ByteCount
	if ( bc != 4 )
		err |= 0x20;                // ByteCount mismatch
	err |= sedrd1 (lplFcn);         // get Function and CrateNumber
	err |= sedrd1 (lplSa);          // get SubAddress
	err |= sedrd2 (lplDat);         // get Data
	err |= sedrd1 (&dat);           // get LastByteMarker
	if ( dat != LASTBY )
		err |= 0x8;                 // LastByteMarker wrong
	*lplErr = err;

exit:
	return;

}		// end SedTst
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedFt    Fifo-Test  SEDPC fifo-interface-test
//          SEDPC will copy data from in-fifo to out-fifo + LastByte.
//          Max. block-size of data is fifo-size - 8 bytes, because
//          the program need 8 more bytes to go through all pattern.
//
//  call: SedFt (lLn, lplPat, lplArr, lBl, lplDerr, lplSerr, lplErr);
//  para: lLn    = logical line-number
//        lplPat = pointer to array with test-pattern (normally 7 willi-pattern)
//        lplArr = pointer to array used for data-storage [max. fifo-size]
//					LowByte = send Data, HighByte = received Data
//        lBl    = block-lenght for test-data, max = (MAXFIF - 8)
//        lplDerr= number of data-errors found
//        lplSerr= number of status-errors found
//        *lplErr  = 0 = no error
//  subp: sedsel, SedClr, sedwr0, sedwr1, sedwr4, sedrd1, sedrd2

void SedFt  ( LONG lLn, LONG *lplPat, LONG *lplArr, LONG lBl,
							LONG *lplDerr, LONG *lplSerr, LONG *lplErr )

{
	#define SPFT    0x1600              // SpecialFunctionCode

	static LONG i, j, k, err, dat, dly;

	*lplDerr = 0;
	*lplSerr = err = 0;

	if ( err = sedsel(lLn) )            // set device-addresses to externals
		goto errsel;

	for ( j = 0; j < lBl+7; j = j+7 ) { // fill array with pattern
		for ( i = 0; i < 7; i++ )
			*(lplArr+j+i) = *(lplPat+i) & 0xFF;
	}

	for ( j = 0; j < 7; j++ ) {         // all 7 data to all locations
		err  = sedwr4 (SPFT, lBl);      // fun & BlockLength to fifo
		for ( i = 0; i < lBl; i++ )     // data to fifo
			err |= sedwr1 (*(lplArr+j+i));
		err |= sedwr0 (LASTBY);
		if ( err )                      // error ?
			goto errsta;

		err  = sedrd2 (&dat);           // get FunctionCode
		if ( err )                      // error ?
			goto errsta;
		if (dat != SPFT)
			(*lplDerr)++;
		err |= sedrd2 (&dat);           // get BlockLength
		if ( err )                      // error ?
			goto errsta;
		if (dat != lBl)
			goto errdat;
		for ( i = 0; i < lBl; i++ ) {
			err |= sedrd1 (&dat);       // get data from fifo
			if ( err )                      // error ?
				goto errsta;
			if (dat != *(lplArr+j+i))
				(*lplDerr)++;
		}
		err |= sedrd1 (&dat);
		if ( err )                      // error ?
			goto errsta;
		if (dat != LASTBY)
			(*lplDerr)++;

		continue;

errsta:
		(*lplSerr)++;                   // status-error
		SedClr (lLn, &dly, &err);
		continue;

errdat:
		(*lplDerr)++;                   // data-error, blocklength wrong
		SedClr (lLn, &dly, &err);
	}

exit:
	return;

errsel:
	(*lplSerr)++;               // status-error
	SedClr (lLn, &dly, &err);
	goto exit;

}		// end SedFt
//-------+---------+---------+---------+---------+---------+---------+---------+

// SedEst   enable SelfTest of SEDxPC at next Reset
//
//  call: SedEst (lLn, lplErr);
//  para: lLn    = logical line-number
//        *lplErr  = 0 = no error
//          bit 13 = illegal SEDPC type
//          bit 12 = no such line-number
//          bit  7 = ParameterError: one parameter incorrect
//          bit  6 = Response/Busy not set; data not taken by SEDPC
//          bit  5 = Bytecount mismatch
//          bit  4 = Timeout: no response from SEDPC
//          bit  3 = LastByte marker not set or illegal function
//  subp: sedsel, sedwr0, sedwr4, sedrd1, sedrd2,
//

void SedEst ( LONG lLn, LONG *lplErr )

{
	static LONG dat;

	if ( *lplErr = sedsel(lLn) )    // set device-addresses to externals
		goto exit;

	*lplErr  = sedwr4 (0x3700, 0);
	*lplErr |= sedwr0 (LASTBY);

	if ( *lplErr )              // error set ?
		goto exit;
	*lplErr = sedrd1 (&dat);    // get sedac-status
	if ( (dat & 0xF0) != 0x20 ) // function or byte-count error ?
		*lplErr |= 0x20;
	*lplErr |= sedrd2 (&dat);   // get DummyData
	*lplErr |= sedrd1 (&dat);   // get LastByteMarker
	if ( dat != LASTBY )
		*lplErr |= 0x8;

exit:
	return;

}		// end SedEst
//-------+---------+---------+---------+---------+---------+---------+---------+

// end of SedTst.c
