/** \file cdi.c
 * \brief basic CDI library routines
 *      
 */
#ifndef NO_DOXYGEN
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <time.h>      /* time() */
#include <memory.h>
#include <errno.h>
#include <fcntl.h>
#include <signal.h>
#include <setjmp.h>
#include <math.h>
#include "tine.h"
#include "cdi.h"
#include "tinecdi.h"
#include "cdiListener.h"
#endif /* NO_DOXYGEN */

int cdiHeapCorrupted = FALSE;
int gSimulationFlag = FALSE;
CdiTemplate *CdiTemplateTbl[MAX_NUM_TEMPLATES];
int CdiTemplateTblSize = 0;

PARAM_STRUCT *paramList = NULL;
int paramListSize = 0;

CdiLineInfoBlk *busHandle[NUMBER_LINES];
int CdiNumIndiciesPerLine[NUMBER_LINES];
int numConnectedHandles = 0;
int cdiInitFlag = 0; /* calling cdiInitialize() */

cdiRedirXRefStruct **cdiRedirXref = NULL;
int nCdiRedirectedDevices = 0;

CdiBusInfoBlk busList[NUMBER_LINES];
int numBusList = 0;

extern char cdiDeviceFilter[64];
int isPastDeviceFilter(char* str);

#define TIMEOUT_CHECK	100
#define MAX_BUSY_MS		200
#define MAX_LINUX_BUSY_MS	1//2
#define RELAX_TIME_MS	2
#define RELAX_LINUX_MS	7	// 7 ms time for one loop of all requested send, maximum 142 transfere per second(e.g can bus)
#define COMPLEX_RANGE_TOKEN " - "
#define SIMPLE_RANGE_TOKEN "-"
/*#define SEPERATOR_TOKEN ",;+" */
#define SEPERATOR_TOKEN ",;"

double applyClbr(CdiRule *rule,BYTE *data,int dataFormat);
void assignClbr(double dval,BYTE *data,int dataFormat);

CDI_EXPORT CdiLineInfoBlk *cdiGetBusHandle(int cdiLine) 
{
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return NULL;
  return busHandle[cdiLine];
}
CDI_EXPORT int cdi_GetNumberConnectedHandles(void) { return numConnectedHandles; }
CDI_EXPORT int cdi_GetNumberRegisteredBus(void) { return numBusList; }
CDI_EXPORT int cdi_GetNumberRequestsMax(int cdiLine)
{
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
  return busHandle[cdiLine]->numberActiveRequest;
}
CDI_EXPORT int cdi_GetNumberRequests(int cdiLine)
{ int i, cnt;
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
  for (i=0, cnt=0; i<busHandle[cdiLine]->numberActiveRequest; i++) if ( busHandle[cdiLine]->pReqList[i].active ) cnt++;
  return cnt;
}
CDI_EXPORT int cdi_GetNumberActiveRequests(int cdiLine)
{ int i, cnt;
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
  for (i=0, cnt=0; i<busHandle[cdiLine]->numberActiveRequest; i++)
    if ( busHandle[cdiLine]->pReqList[i].active > idleState && busHandle[cdiLine]->pReqList[i].active < removeRequest ) cnt++;
  return cnt;
}
char *lineinfFieldNames[] = 
{
  "size request table",
  "number active",
  "maximum size request table",
  "reserved",
  "request index",
  "active state",
  "polling interval",
  "request counter",
  "request time",
  "response time",
  "group id (0=not grouped)",
  "index within group",
  "group return flag",
  "group sync start",
  "async call flag",
  "has callback",
  "call condition",
  "data return condition",
  "cancel pending",
  "access flag",
  "last polling time",
  "device error",
  "error code",
  "poll flag",
  "cancel flag",
  "device ready time",
  "device number",
  "address(0)",
  "address(1)",
  "address(2)",
  "param(0)",
  "param(1)",
  "param(2)",
  "param(3)",
  "param(4)",
  "param(5)",
  "param(6)",
  "param(7)",
  "param(8)",
  "param(9)",
  "param(10)",
  "param(11)",
  "param(12)",
  "param(13)",
  "param(14)",
  "param(15)"
};
int nLineinfFieldNames = sizeof(lineinfFieldNames)/sizeof(char *);
// Hongkong's original :
CDI_EXPORT int cdi_GetActiveRequestInfor(int cdiLine, int *pBuffer, int size, int offset)
{ int i, cnt;
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
  if ( size < (NUMBER_EXEC_PARAM + 25) ) return 0;
  for (i=0, cnt=0; i<busHandle[cdiLine]->numberActiveRequest; i++)
  {
//    if ( busHandle[cdiLine]->pReqList[i].active <= idleState || busHandle[cdiLine]->pReqList[i].active >= removeRequest ) continue;
    cnt++;
    if ( cnt != offset ) continue;
    pBuffer[0] = busHandle[cdiLine]->pReqList[i].active;
    pBuffer[1] = busHandle[cdiLine]->pReqList[i].timeout;
    pBuffer[2] = busHandle[cdiLine]->pReqList[i].requestCounter;
    pBuffer[3] = busHandle[cdiLine]->pReqList[i].requestTime;
    pBuffer[4] = busHandle[cdiLine]->pReqList[i].responseTime;
    pBuffer[5] = (int) busHandle[cdiLine]->pReqList[i].groupReq;
    pBuffer[6] = busHandle[cdiLine]->pReqList[i].NgroupsReqs;
    pBuffer[7] = busHandle[cdiLine]->pReqList[i].NgroupRets;
    pBuffer[8] = busHandle[cdiLine]->pReqList[i].groupSyncStart;
    pBuffer[9] = busHandle[cdiLine]->pReqList[i].asynCallFlag;
    pBuffer[10] = (int)busHandle[cdiLine]->pReqList[i].callback;
    pBuffer[11] = busHandle[cdiLine]->pReqList[i].conditionFlag;
    pBuffer[12] = busHandle[cdiLine]->pReqList[i].dataReturnCondition;
    pBuffer[13] = busHandle[cdiLine]->pReqList[i].userCallDown;
    pBuffer[14] = busHandle[cdiLine]->pReqList[i].accessFlag;
    pBuffer[15] = busHandle[cdiLine]->pReqList[i].pollTime;
    pBuffer[16] = busHandle[cdiLine]->pReqList[i].deviceError;
    pBuffer[17] = busHandle[cdiLine]->pReqList[i].errorCode;
    pBuffer[18] = busHandle[cdiLine]->pReqList[i].pollFlag;
    pBuffer[19] = busHandle[cdiLine]->pReqList[i].cancelFlag;
    pBuffer[20] = busHandle[cdiLine]->pReqList[i].devReadyTime;
    pBuffer[21] = busHandle[cdiLine]->pReqList[i].pDev->devNumber;
    pBuffer[22] = busHandle[cdiLine]->pReqList[i].pDev->cdiAddr[caAddrIndex];
    pBuffer[23] = busHandle[cdiLine]->pReqList[i].pDev->cdiAddr[subAddrIndex];
    pBuffer[24] = busHandle[cdiLine]->pReqList[i].pDev->cdiAddr[2];
    memcpy(&pBuffer[25], busHandle[cdiLine]->pReqList[i].devParameters, NUMBER_EXEC_PARAM*sizeof(int));
    break;
  }
  return cnt;
}
void listenerCb(int id, int cc); // from cdiListener.c
void cbDsm(int id, int cc); // from cdiTineLib.c
CDI_EXPORT void cdiDumpActiveRequestInfo(int cdiLine)
{
  int i;
  CdiRequestInfoBlk *blk;
  char *dev;
  if (cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL) return;
  for (i = 0; i < busHandle[cdiLine]->numberActiveRequest; i++)
  {
    blk = &busHandle[cdiLine]->pReqList[i];
    dev = blk->pDev != NULL ? blk->pDev->devName : "unavailable";
    if (strlen(cdiDeviceFilter) && !isPastDeviceFilter(dev)) continue;
    ttyoutput("device : %.32s", dev);
    ttyoutput("active: %s",(blk->active?"TRUE":"FALSE"));
    ttyoutput("wait timeout: %d",blk->timeout);
    ttyoutput("number requests: %d", blk->requestCounter);
    ttyoutput("last request (msec since startup): %d", blk->requestTime);
    ttyoutput("last response (msec since startup): %d", blk->responseTime);
    ttyoutput("is grouped : %s", (blk->groupReq == NULL?"NO":"YES") );
    ttyoutput("grouped requests: %d", blk->NgroupsReqs);
    ttyoutput("grouped returns: %d", blk->NgroupRets);
    ttyoutput("callback : %s", blk->callback == listenerCb ? "listenerCb" : blk->callback == cbDsm ? "cbDsm" : blk->callback == NULL ? "none" : "(user's)");
    ttyoutput("monitor interval : %d", blk->pollTime);
    ttyoutput("error status: %d", blk->errorCode);
    ttyoutput("cancel flag: %d", blk->cancelFlag);
  }
  return;
}
int cdiDumpLineInfo(int i1,int i2,int i3,int i4)
{
  //int cc, i, idx = 1, line = 0;
  //int lineinf[PRP_CDIINFOR_SIZE];
  int line = 0;
  if (i1 != 0) line = i1;
  //if (i1 < 0) line = 0;
  //if (i2 != 0) idx = i2;
  //if (idx < 1) idx = 1;

  ttyoutput("%s : %d",lineinfFieldNames[0],cdi_GetNumberRequests(line));
  ttyoutput("%s : %d",lineinfFieldNames[1],cdi_GetNumberActiveRequests(line));
  ttyoutput("%s : %d",lineinfFieldNames[2],cdi_GetNumberRequestsMax(line));

  cdiDumpActiveRequestInfo(line);

  //cc = cdi_GetActiveRequestInfor(line, &lineinf[5], PRP_CDIINFOR_SIZE - 5, idx);

  //lineinf[4] = idx;
  //if ( !cc || cc != idx )
  //{
  //  lineinf[4] = 0;
	 // for (i=5; i<PRP_CDIINFOR_SIZE; i++) lineinf[i] = 0;
	 // cc = 0;
  //}
  //for (i=4; i<PRP_CDIINFOR_SIZE && i<nLineinfFieldNames; i++)
  //{
  //  ttyoutput("%s : %d",lineinfFieldNames[i],lineinf[i]);
  //}
  return 0;
}
int cdiGetNextAvailableBusType(void)
{ /* find the next free (unnamed) slot */
  int i;
  for (i=0; i<NUMBER_LINES && strlen(busList[i].busName) > 0; i++);
  if (i == NUMBER_LINES) 
  {
    cdilog("cdi bus identifiers exhausted!");
    return -resources_exhausted;
  }
  return i;
}
int cdiGetBusType(char *busName)
{ 
  int i;
  for (i=0; i<numBusList && strncmp(busName, busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) return -1;
  return busList[i].busType;
}
CDI_EXPORT int cdi_RegisterBus(
  char *busName,
  int busType,
  int (*filterBusFcn)(int *parameter, int accessFlag),
  int (*cfgBusFcn)(void),
  int (*initBusFcn)(int busline,int cdiline,int ndevices,char *paramList),
  void (*execBusFcn)(CdiRequestInfoBlk *pReq),
  int (*freeBusFcn)(int busline)
  )
{
  int i;
  for (i=0; i<numBusList && busType != busList[i].busType && strncmp(busName, busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) /* it's new !*/
  {
    if (i == NUMBER_LINES) 
    {
      cdilog("cdi bus %.16s : no more identifiers !",busName);
      return -resources_exhausted;
    }
    numBusList++;
    cdilog("cdi bus %s (id %d) registered",busName,busType);
  }
  else 
  {
    cdilog("cdi bus %s (id %d) : already registered",busName,busType);
    return -invalid_index; /*  same type index! */
  }
  strncpy(busList[i].busName,busName,CDI_BUSNAME_SIZE);
  busList[i].busType = busType;
  busList[i].filterBusFcn = filterBusFcn;
  busList[i].cfgBusFcn = cfgBusFcn;
  busList[i].initBusFcn = initBusFcn;
  busList[i].execBusFcn = execBusFcn;
  busList[i].freeBusFcn = freeBusFcn;
  return i;
}

/************************************/
CDI_EXPORT int getTimeMilliseconds(void)
{ 
  struct timeval tv; 
  gettimeofday (&tv, NULL);
  return (int)((tv.tv_sec - gCdiStartupTime) * 1000 + tv.tv_usec / 1000);
}

/********************************************/
CDI_EXPORT int getTypeLength(short dataType)
{
  if ( dataType == cdiChar ) return 1;
  else if ( dataType == cdiByte ) return 1;
  else if ( dataType == cdiShort ) return 2;
  else if ( dataType == cdiInteger ) return 4;
  else if ( dataType == cdiFloat ) return 4;
  else if ( dataType == cdiDouble ) return 8;
  else if ( dataType == cdiChar32 ) return 32;
  return 0;
}
/********************************************/
/* compare pattern or last data             */
/********************************************/
CDI_EXPORT int cdiDataCmp(CdiRequestInfoBlk *pReq, char *lastDataOrPattern, char *str1, int length1, int stringType)
{ int cmpLength, i, j;
  int delta, lastData, window, *iTmp;
  double fdelta, flastData, fwindow;
  char *plast, *pnow, last[8], now[8];		/* for rs232 string type data comparison, convert to number first!!! */
  short *sTmp;
  BYTE *bTmp;
  float *fTmp;
  double *dTmp;

  cmpLength = length1 < (int)pReq->inLength ? length1 : (int)pReq->inLength;
  if ( stringType == 0 )
  {
    plast = lastDataOrPattern;
    pnow = str1;
  }
  else		/* convert string "1234" to number */
  {
    plast = last;
    pnow = now;
    if ( pReq->inType == cdiByte )
    {
      sscanf(lastDataOrPattern, "%c", (char *)last ); 
      sscanf(str1, "%c", (char *)now ); 
    }
    else if ( pReq->inType == cdiShort )
    {
      sscanf(lastDataOrPattern, "%hd", (short *)last ); 
      sscanf(str1, "%hd", (short *)now ); 
    }
    else if ( pReq->inType == cdiInteger )
    {
      sscanf(lastDataOrPattern, "%d", (int *)last ); 
      sscanf(str1, "%d", (int *)now ); 
    }
    else if ( pReq->inType == cdiFloat )
    {
      sscanf(lastDataOrPattern, "%f", (float *)last ); 
      sscanf(str1, "%f", (float *)now ); 
    }
    else
    {
      sscanf(lastDataOrPattern, "%lf", (double *)last ); 
      sscanf(str1, "%lf", (double *)now ); 
    }
  }
  if ( pReq->conditionFlag == maskChangeWindow || pReq->conditionFlag == maskChangePercent )
  {
    if ( pReq->inType == cdiChar || pReq->inType == cdiChar32 ) return 0;    /* no tolerance for string data */
    else if ( pReq->inType == cdiByte || pReq->inType == cdiShort || pReq->inType == cdiInteger )
    {
      if ( pReq->inType == cdiByte )
      {
        window = *(BYTE *)pReq->patternFilter;
        lastData = *(BYTE *)plast;
        delta = *(BYTE *)plast - *(BYTE *)pnow;
	    }
	    else if ( pReq->inType == cdiShort )
	    {
        window = *(short *)pReq->patternFilter;
        lastData = *(short *)plast;
        delta = *(short *)plast - *(short *)pnow;
      }
      else
      {
        window = *(int *)pReq->patternFilter;
        lastData = *(int *)plast;
        delta = *(int *)plast - *(int *)pnow;
      }
      if ( delta < 0 ) delta = -delta;
      if ( pReq->conditionFlag == maskChangeWindow )
	      return delta >= window ? 1 : 0;
      else
	      return delta * 100 >= window * lastData ? 1 : 0;			  
    }
    else	/* no mask for float and double */
    {
      if ( pReq->inType == cdiFloat )
      {
        fwindow = *(float *)pReq->patternFilter;
        flastData = *(float *)plast;
        fdelta = *(float *)plast - *(float *)pnow;
      }
      else
      {
        fwindow = *(double *)pReq->patternFilter;
        flastData = *(double *)plast;
        fdelta = *(double *)plast - *(double *)pnow;
      }
      if ( fdelta < 0 ) fdelta = -fdelta;
      if ( pReq->conditionFlag == maskChangeWindow )
	      return fdelta > fwindow ? 1 : 0;
      else
	      return fdelta * 100 > fwindow * flastData ? 1 : 0;			  
    }
  }
  else  /* compare pattern */
  {
    if ( pReq->inType == cdiChar || pReq->inType == cdiChar32 )  /* compare string */
    {
      cmpLength = length1 < (int)strlen((const char *)pReq->patternFilter) ? length1 : 
                            (int)strlen((const char *)pReq->patternFilter);
      for ( i=0; i<cmpLength; i++ ) if ( str1[i] != ((char *)pReq->patternFilter)[i] ) break;
      return strncmp(str1, (char *)pReq->patternFilter, cmpLength);
    }
    else    /* compare number */
    {
      if ( pReq->inType == cdiByte )
      {
        bTmp = (BYTE *)str1;
        for (i=0; i<length1; i+=1 )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && i < pReq->pDev->dataPatternStartIndex && i > pReq->pDev->dataPatternStopIndex ) continue;
          if ( bTmp[i] != *(BYTE *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiShort )
      {
        sTmp = (short *)str1;
        for (i=j=0; j<length1; j+=2, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( sTmp[i] != *(short *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiInteger )
      {
        iTmp = (int *)str1;
        for (i=j=0; j<length1; j+=4, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( iTmp[i] != *(int *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiFloat )
      {
        fTmp = (float *)str1;
        for (i=j=0; j<length1; j+=4, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( (int) fTmp[i] != *(int *)pReq->patternFilter ) break;
        }
      }
      else if ( pReq->inType == cdiDouble )
      {
        dTmp = (double *)str1;
        for (i=j=0; j<length1; j+=8, i++ )
        {
          if ( pReq->pDev->dataPatternStartIndex >= 0 && (i < pReq->pDev->dataPatternStartIndex || i > pReq->pDev->dataPatternStopIndex) ) continue;
          if ( (int) dTmp[i] != *(int *)pReq->patternFilter ) break;
        }
      }
      if ( (i <= pReq->pDev->dataPatternStartIndex || i <= pReq->pDev->dataPatternStopIndex) && pReq->conditionFlag == maskCompNotEqual ) return 1; /* can not compare then return notEqual */
      return j < length1 ? 1 : 0;
    }
#ifdef DONOTDELEAT
    for ( i=0; i<cmpLength; i++ ) 
	  {
      if ( pReq->hasDataMask )
      {
        if ( (str1[i] & pReq->mask[i]) != pReq->patternFilter[i] ) break; /* is mask guaranteed to be equal in length ? */
      }
      else if ( str1[i] != ((char *)pReq->patternFilter)[i] ) break;
 	  }
#endif
  }
  return 0;
}

/*******************************************************/
/* called from busplug after data comes from Hardware, set return conditionflag */
/*******************************************************/
CDI_EXPORT void checkReturnCondition(CdiRequestInfoBlk *pReq, char *lastData, char *str, int datalen, int stringType)
{ short *sTmp;
  BYTE *bTmp;
  int i, j, comFlag, *iTmp;
  float fp;

  if (pReq == NULL) return;

  /* check return data format */
  if ( stringType != 0 )
  {
    if ( pReq->inType >= cdiFloat && sscanf(str,"%f",&fp) != 1 ) return;
    else if ( pReq->inType >= cdiByte && sscanf(str,"%d",&i) != 1 ) return;
  }
  /* should use default data type not input(stringType) ??????? */
  if ( stringType == 0 && pReq->hasDataMask ) /* mask for byte, short or integer type data */
  {
    if ( pReq->inType == cdiByte )
    {
      bTmp = (BYTE *)str;
      for (i=0; i<datalen; i+=1 )
      {
        if ( pReq->pDev->dataMaskStartIndex >= 0 && i < pReq->pDev->dataMaskStartIndex && i > pReq->pDev->dataMaskStopIndex ) continue;
        (*bTmp) &= (BYTE)pReq->maskValue;
        bTmp++;
      }
    }
    else if ( pReq->inType == cdiShort )
    {
      sTmp = (short *)str;
      for (i=j=0; j<datalen; j+=2, i++ )
      {
        if ( pReq->pDev->dataMaskStartIndex >= 0 && i < pReq->pDev->dataMaskStartIndex && i > pReq->pDev->dataMaskStopIndex ) continue;
        (*sTmp) &= (short)pReq->maskValue;
        sTmp++;
      }
    }
    else if ( pReq->inType == cdiInteger )
    {
      iTmp = (int *)str;
      for (i=j=0; j<datalen; j+=4, i++ )
      {
        if ( pReq->pDev->dataMaskStartIndex >= 0 && i < pReq->pDev->dataMaskStartIndex && i > pReq->pDev->dataMaskStopIndex ) continue;
        (*iTmp) &= (int)pReq->maskValue;
        iTmp++;
      }
    }
  }

  if ( pReq->conditionFlag == anyData )
  {
    if ( stringType == 0 && (pReq->accessFlag == cdiReadWriteFlag || pReq->accessFlag == cdiWriteReadWriteFlag)) /* read then write bits */
    { /* mask the readData to the write Buffer */
      i = pReq->accessFlag == cdiReadWriteFlag ? 0 : 1;
      if ( pReq->inType == cdiByte )
      {
        if ( ((BYTE *)pReq->pOutData)[i+1] )
          ((BYTE *)pReq->pOutData)[i] |= *(BYTE *)str;
        else
          ((BYTE *)pReq->pOutData)[i] = (~((BYTE *)pReq->pOutData)[i]) & *(BYTE *)str;
      }
      else if ( pReq->inType == cdiShort )
      {
        if ( ((short *)pReq->pOutData)[i+1] )
          ((short *)pReq->pOutData)[i] |= *(short *)str;
        else
          ((short *)pReq->pOutData)[i] = (~((short *)pReq->pOutData)[i]) & *(short *)str;
      }
      else if ( pReq->inType == cdiInteger )
      {
        if ( ((int *)pReq->pOutData)[i+1] )
          ((int *)pReq->pOutData)[i] |= *(int *)str;
        else
          ((int *)pReq->pOutData)[i] = (~((int *)pReq->pOutData)[i]) & *(int *)str;
      }
    }
    pReq->dataReturnCondition = DATA_PASS_FILTER;
  }
  else if ( pReq->conditionFlag == maskChangeWindow || pReq->conditionFlag == maskChangePercent )
  {
    if ( pReq->requestCounter == 1 ) 
      pReq->dataReturnCondition = DATA_PASS_FILTER; /* first call of Async or sync */
    else 
    {
      /*if ( pReq->patternFilter == NULL ) */
      if ( pReq->hasPatternFilter == FALSE ) 
      {
        pReq->dataReturnCondition = DATA_PASS_FILTER; /* no filter => any data */
      }
      else /* data changed? compare to last one */
      {
        comFlag = cdiDataCmp(pReq,lastData,str,datalen, stringType);
        if ( comFlag != 0 ) pReq->dataReturnCondition = DATA_PASS_FILTER;
        else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK; /* data back for polling request! */
      }
    }
  }
  else
  {
    comFlag = cdiDataCmp(pReq,pReq->patternFilter,str,datalen, stringType);
	  if (cdiDebugLevel) dbglog("checkReturnCondition: %d-%d: data: %d --- %d(%d); datalen: %d; stringtype: %d; cmp: %d\n",pReq->conditionFlag, pReq->dataReturnCondition, *(short *)lastData, *(short *)str, *(short *)pReq->patternFilter, datalen, stringType, comFlag);
    if ( comFlag == 0 )
    {
      if ( pReq->conditionFlag == maskCompEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
      else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK;
    }
    else if ( pReq->conditionFlag == maskCompNotEqual ) pReq->dataReturnCondition = DATA_PASS_FILTER;
    else if ( pReq->pollFlag ) pReq->dataReturnCondition = DATA_BACK;
	  if (cdiDebugLevel) dbglog("checkReturnCondition: %d-%d: data: %d --- %d(%d); datalen: %d; stringtype: %d; cmp: %d\n",pReq->conditionFlag, pReq->dataReturnCondition, *(short *)lastData, *(short *)str, *(short *)pReq->patternFilter, datalen, stringType, comFlag);
  }
}

/************************************/
/* check if need to transfer */
/************************************/
CDI_EXPORT int checkRequestTime(CdiRequestInfoBlk *pReq)
{ int tTmp;
  if ( pReq->active == transRequest )
  {
    if ( pReq->callback == NULL || pReq->timeout == 0 ) return 1; /* sync or no timeout */
    if ( pReq->responseTime - pReq->requestTime < 0 ) return 1; /* first call for asyn */
    tTmp = getTimeMilliseconds();
    if ( tTmp - pReq->requestTime > abs(pReq->timeout) )
    {
      if ( pReq->pollFlag && ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->nextWakeupTime != 0 )
      {
        if ( tTmp - pReq->requestTime > 2 * abs(pReq->timeout) ) ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->nextWakeupTime = 0;
        else if ( tTmp + abs(pReq->timeout) - ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->nextWakeupTime < 0 ) ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->nextWakeupTime = tTmp + abs(pReq->timeout);
      }
      return 1; /* poll interval */
    }
    return 0;
  }
  else if ( pReq->active == transPoll || pReq->active == transPoll2nd )
  {
    if ( getTimeMilliseconds() - pReq->pollTime > 10 ) return 1; /* poll after 10 ms */
  }
  else if ( pReq->active == transRequest2nd ) return 1;
  return 0;
}

/************************************/
CDI_EXPORT int assignGroupInOutFormatCode(CdiRequestInfoBlk *pReq, int inFormat, int outFormat)
{ int i, j, cnt;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL )
  {
    pReq->outFormatCode = inFormat;
    pReq->inFormatCode = outFormat;
    return 1;
  }
  for ( j = 0, cnt = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      pLine->pReqList[i].outFormatCode = inFormat;
      pLine->pReqList[i].inFormatCode = outFormat;
      cnt++;
    }
  }
  return cnt;
}
/************************************/
CDI_EXPORT int checkGroupReturnFlags(CdiRequestInfoBlk *pReq, int taskIndex)
{ 
  int i, j, rc, cnt;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL ) return 0;
  if ( pReq->NgroupRets )
  {
	  cdilog("checkGroupReturnFlags called more than once!!! Number of GroupReq: %d, this Request Number: %d; pReq: 0x%x; reqIndex: %d; devNumber: %d; taskIndex: %d; callBack: 0x%x; caAddrIndex: %d; 0x%x,0x%x,0x%x\n",((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs, pReq->NgroupsReqs, pReq, pReq->reqIndex, pReq->pDev->devNumber, taskIndex, pReq->callback, pReq->pDev->cdiAddr[caAddrIndex], pReq->devParameters[0], pReq->devParameters[1], pReq->devParameters[2]);
  }
  if ( taskIndex == cancelIndex ) cdilog("checkGroupReturnFlags called from cancelRequest! Number of GroupReq: %d, this Request Number: %d; pReq: 0x%x; taskIndex: %d; callBack: 0x%x\n",((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs, pReq->NgroupsReqs, pReq, taskIndex, pReq->callback);
  pReq->NgroupRets = 1;
  cnt = 0;
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if ( pLine->pReqList[i].NgroupRets ) cnt++;
    }
  }
  rc = (((CdiRequestInfoBlk *)pReq->groupReq)->asynCallFlag == 0 && ((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs == cnt) ? 1 : 0;
  if (rc == 0 && ((CdiRequestInfoBlk*)pReq->groupReq)->errorCode != 0) rc = 1; // p.d. 2021 (try this !)
  if (isPastDeviceFilter(pReq->pDev->devName))
  {
    ttyoutput("checkGroupReturnFlags: %.32s (%d from %d returned) returning %d", pReq->pDev->devName, cnt, ((CdiRequestInfoBlk*)pReq->groupReq)->NgroupsReqs, rc);
  }
  if ( ((CdiRequestInfoBlk *)pReq->groupReq)->asynCallFlag && ((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs == cnt )
    cdilog("checkGroupReturnFlags: all back but asynCallFlag already set!! Number of GroupReq: %d, this Request Number: %d; pReq: 0x%x, taskIndex: %d, callBack: 0x%x, timeout: %d, errorCode: %d, accessFlag: %d, caAddrIndex: %d, 0x%x,0x%x,0x%x,0x%x,0x%x,0x%x,0x%x,0x%x,0x%x,0x%x,0x%x\n",((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs, pReq->NgroupsReqs, pReq, taskIndex, pReq->callback, pReq->timeout, pReq->errorCode, pReq->accessFlag, pReq->pDev->cdiAddr[caAddrIndex], pReq->devParameters[0], pReq->devParameters[1], pReq->devParameters[2], pReq->devParameters[3], pReq->devParameters[4], pReq->devParameters[5], pReq->devParameters[6], pReq->devParameters[7], pReq->devParameters[8], pReq->devParameters[9], pReq->devParameters[10]);
  return rc;
}

/************************************/
/* return value: error for single request; one of the erros for group request  */
CDI_EXPORT int getRequestErrors(CdiRequestInfoBlk *pReq)
{ int i,j;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL || ((CdiRequestInfoBlk *)pReq->groupReq)->NgrpErrors == 0 ) return pReq->errorCode;
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if ( pLine->pReqList[i].errorCode ) return pLine->pReqList[i].errorCode;
    }
  }
  return 0;
}

void setGroupFlags(CdiRequestInfoBlk *pReq, int index)
{
  if ( index & 0x1 ) pReq->userCallDown++;
  if ( index & 0x2 ) pReq->groupSyncStart = 1; /* mark next start  */
  if ( index & 0x4 )
  {
	pReq->NgroupRets = 0;     /* reset return flags  */
    pReq->devReadyTime = 0;
  }
  if ( pReq->errorCode ) ((CdiRequestInfoBlk *)pReq->groupReq)->NgrpErrors++;
}

/************************************/
/* return value: 0 for single request; number of erros for group request  */
CDI_EXPORT int setGroupReturnFlags(CdiRequestInfoBlk *pReq, int index)
{ int i, j;
  CdiRequestInfoBlk *pMainReq;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL ) return 0;
  ((CdiRequestInfoBlk *)pReq->groupReq)->NgrpErrors = 0;
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
	  if ( &pLine->pReqList[i] == pReq->groupReq )
	  {
		  pMainReq = &pLine->pReqList[i];
		  continue;
	  }
	  setGroupFlags(&pLine->pReqList[i], index);
#ifdef ADDED2010
//	  if ( &pLine->pReqList[i] != pReq ) cdiWaitSemaphore(&pLine->pReqList[i].exec_SemId,0);	// added to 2010-1
      if ( index & 0x1 ) pLine->pReqList[i].userCallDown++;
      if ( index & 0x2 ) pLine->pReqList[i].groupSyncStart = 1; /* mark next start  */
      if ( index & 0x4 )
      {
        pLine->pReqList[i].NgroupRets = 0;     /* reset return flags  */
        pLine->pReqList[i].devReadyTime = 0;
      }
      if ( pLine->pReqList[i].errorCode ) pReq->NgrpErrors++;
//	  if ( &pLine->pReqList[i] != pReq ) cdiPostSemaphore(&pLine->pReqList[i].exec_SemId);	// added to 2010-1
#endif
    }
  }
  setGroupFlags(pMainReq, index);
  return ((CdiRequestInfoBlk *)pReq->groupReq)->NgrpErrors;
}

/************************************/
CDI_EXPORT int checkDeadGroupCalls(CdiRequestInfoBlk *pReq)
{ int i, cnt, cdiLine;
  if ( pReq->groupReq == NULL || pReq->callback == NULL || pReq->groupSyncStart || pReq != pReq->groupReq ) return 0;
  if ( pReq->NgroupRets ) return 0;		// if the main group is already returned!
  cdiLine = ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine;
  for (i=0, cnt=0; i<busHandle[cdiLine]->numberActiveRequest; i++)	// find the rest
  {
	if ( &busHandle[cdiLine]->pReqList[i] == pReq ) continue;
	if ( busHandle[cdiLine]->pReqList[i].groupReq != pReq->groupReq ) continue;
    cnt++;
  }
  if ( pReq->NgroupsReqs == cnt + 1 && getTimeMilliseconds() - pReq->requestTime > abs(pReq->timeout) + 1000 ) 
  {
	setGroupReturnFlags(pReq, 0x6);
	cdilog("checkDeadGroupCalls ==> workaround for this dead group call!!! Number of GroupReq: %d, this Request Number: %d; pReq: 0x%x; reqIndex: %d; devNumber: %d; callBack: 0x%x; caAddrIndex: %d; 0x%x,0x%x,0x%x\n",((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs, pReq->NgroupsReqs, pReq, pReq->reqIndex, pReq->pDev->devNumber, pReq->callback, pReq->pDev->cdiAddr[caAddrIndex], pReq->devParameters[0], pReq->devParameters[1], pReq->devParameters[2]);
	return 1;
  }
  return 0;
}

void setReqFlagsAttachet(CdiRequestInfoBlk *pReq, int taskIndex)
{
  if ( pReq->groupReq == NULL || checkGroupReturnFlags(pReq, taskIndex) )
  {
    if (cdiDebugLevel) dbglog("post callback semaphore: task: 0x%x; task Index: %d\n",pReq, taskIndex);
    if ( pReq->groupReq == NULL ) pReq->asynCallFlag = 1;
    else ((CdiRequestInfoBlk *)pReq->groupReq)->asynCallFlag = 1;
    setGroupReturnFlags(pReq, 0x2);
    if ( pReq->groupReq == NULL ) cdiPostSemaphore(&(((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->call_SemId));
    else cdiPostSemaphore(&(((CdiLineInfoBlk *)((CdiRequestInfoBlk *)pReq->groupReq)->pDev->pcdiLine)->call_SemId));
  }
}
/************************************/
/* exec request, step by step */
/* all 4 steps arranged here! */
/************************************/
CDI_EXPORT void cdiExecRequest(CdiRequestInfoBlk *pReq, int taskIndex)
{ unsigned int i;
  if (cdiDebugLevel || isPastDeviceFilter(pReq->pDev->devName)) 
    dbglog("Call cdiExecRequest: %.32s taskIndex === %d; activeFlag: %d; cancelFlag: %d; requestNumber: 0x%x",
      pReq->pDev->devName, taskIndex, pReq->active, pReq->cancelFlag, pReq->requestNumber);
  if ( taskIndex == readTaskIndex && (pReq->active != recvRequest && pReq->active != recvRequest2nd)  ) return;  /*background task called already  */
  if ( pReq->inside ) dbglog("Inside Flag True!!!! Call cdiExecRequest: taskIndex === %d; activeFlag: %d; cancelFlag: %d; requestNumber: 0x%x\n",taskIndex, pReq->active, pReq->cancelFlag, pReq->requestNumber);
  if ( pReq->active <= idleState || pReq->active >= finishRequest ) return;
  pReq->inside = 1;
  /* send out request */
  if ( pReq->active == transRequest ) /* init transmission parameters only on first transfer */
  {
    pReq->groupSyncStart = 0;
    pReq->requestTime = getTimeMilliseconds();
    if ( pReq->responseTime - pReq->requestTime >= 0 ) pReq->responseTime = pReq->requestTime - 1;

    pReq->pollTime = pReq->requestTime;
    pReq->dataReturnCondition = 0;
    pReq->errorCode = 0;
    if (pReq->perror != NULL) *pReq->perror = 0;
    i = pReq->requestCounter++;
    if ( i && pReq->requestCounter == 0 ) pReq->requestCounter = 2; /* detect overflow */
    if (cdiDebugLevel > 3) dbglog("set request Time: %d; preq: %x\n",pReq->requestTime, pReq);
  }
  if ( pReq->active == transPoll || pReq->active == transPoll2nd ) pReq->pollTime = getTimeMilliseconds();
  else if ( pReq->active == recvRequest2nd ) pReq->active = finishRequest; /* second receive */
  else if ( pReq->active == recvRequest ) /* first receive => change to finish or 2ndSend */
  { /* data came back! */
    if (pReq->errorCode || pReq->accessFlag == cdiReadFlag || pReq->accessFlag == cdiWriteFlag)
    {
	    pReq->active = finishRequest;
	    if ( pReq->accessFlag == cdiWriteFlag && pReq->errorCode == 0 && pReq->dataReturnCondition == 0 )	// added to 2010-1
		    pReq->dataReturnCondition = DATA_PASS_FILTER; /* flag for callBack */
    }
    else if ( pReq->dataReturnCondition != DATA_BACK ) pReq->active = transRequest2nd;
    else pReq->active = finishRequest; /* data do not pass condition, then finish! */
  }

  if ( pReq->active != finishRequest ) /* only first or second request */
  {
    if ( gSimulationFlag ) 
    {
      if ( pReq->pInData && pReq->inLength > 0 ) for(i=0; i<pReq->inLength; i++) ((char *)pReq->pInData)[i] = (char)rand();
      pReq->active = finishRequest;
    }
    else
    {
      if (busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].execBusFcn != NULL)
      {
        busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].execBusFcn(pReq);
        cdiAssertHeapOkay("cdiExecRequest",pReq->pDev->devName,pReq->accessFlag);
      }
      if ( pReq->active == finishRequest && pReq->errorCode == 0 && pReq->dataReturnCondition == 0 )
        pReq->dataReturnCondition = DATA_PASS_FILTER; /* flag for callBack */
    }
  }
  if ( pReq->active != finishRequest )
  {
    pReq->inside = 0;
    return;
  }
  /* finish request!! */
  pReq->responseTime = getTimeMilliseconds(); /* avoid timeout check!!!! */
  if (cdiDebugLevel > 3) dbglog("set response time: %d, preq: %x\n",pReq->responseTime, pReq);
  if ( pReq->callback == NULL )
  {
    if ( pReq->groupReq == NULL ) cdiPostSemaphore(&pReq->sync_SemId);
    else if ( checkGroupReturnFlags(pReq, taskIndex) ) cdiPostSemaphore(&((CdiRequestInfoBlk *)pReq->groupReq)->sync_SemId);
  }
  else
  {
    if ( (pReq->timeout >= 0 && pReq->dataReturnCondition == DATA_PASS_FILTER) || pReq->errorCode )
    { 
      setReqFlagsAttachet(pReq, taskIndex);
    }
    if ( pReq->pollFlag && pReq->cancelFlag == 0 )
    {
      pReq->active = transRequest; /* start next time interval */
      /* Let background task start, avoid continuous recursive call */
      /*      if ( pReq->timeout == 0 ) cdiExecRequest(pReq, taskIndex); */
    }
  }
  pReq->inside = 0;
  return;
}

/**************************************/
int checkRestGroupElement(CdiRequestInfoBlk *pReq)
{ int j, cnt, k;
  CdiLineInfoBlk *tmpLine;

  if ( pReq->groupReq == NULL || pReq != pReq->groupReq ) return 0;   /* not group or not the first element(mainRequest) in the group */
  for ( j = 0, cnt = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    tmpLine = busHandle[j];
    for ( k = 0; k < tmpLine->numberActiveRequest; k++ )
    {
      if ( tmpLine->pReqList[k].groupReq != pReq->groupReq ) continue;
      cnt++;
    }
  }
  return cnt == 1 ? 0 : 1;    /* after the rest of the group have been removed, then the last one (mainRequest: head of group) can be removed! */
}

/**************************************/
CDI_EXPORT void freeRequestBlock(CdiLineInfoBlk *pLine)
{ int i, findActive;

  findActive = 0;
  for ( i = pLine->numberActiveRequest - 1; i >= 0 ; i-- )
  {
    if ( cdiWaitSemaphore(&pLine->pReqList[i].exec_SemId, 10) ) continue;
    if ( pLine->pReqList[i].active == removeRequest ) 
    {
      if ( checkRestGroupElement(&pLine->pReqList[i]) ) /* remove the rest of the group first then the mainRequest(head of the group)!  */
      {
        cdiPostSemaphore(&pLine->pReqList[i].exec_SemId);
        continue;
      }
      pLine->pReqList[i].userCallDown = 0;
//      pLine->pReqList[i].active = 0;

      pLine->pReqList[i].dataReturnCondition = 0;
      pLine->pReqList[i].errorCode = 0;
      pLine->pReqList[i].deviceError = 0;
      pLine->pReqList[i].cancelFlag = 0;
      pLine->pReqList[i].requestCounter = 0;

      pLine->pReqList[i].asynCallFlag = 0;
      pLine->pReqList[i].NgroupRets = 0;
      pLine->pReqList[i].NgrpErrors = 0;
      pLine->pReqList[i].groupSyncStart = 0;
      pLine->pReqList[i].devReadyTime = 0;

	  if ( pLine->pReqList[i].needsCalibration & FREE_ONCLOSE )	// 2010-4
		if ( pLine->pReqList[i].groupReq == NULL || pLine->pReqList[i].groupReq == &pLine->pReqList[i] )
			if (pLine->pReqList[i].perror != NULL)
			{
				free(pLine->pReqList[i].perror);
				pLine->pReqList[i].perror = NULL;
			}

      if (pLine->pReqList[i].freeOutDataMemoryOnCancel) 
      { 
        free(pLine->pReqList[i].pOutData); pLine->pReqList[i].pOutData = NULL; 
        pLine->pReqList[i].freeOutDataMemoryOnCancel = FALSE;
      }
      if (pLine->pReqList[i].freeInDataMemoryOnCancel) 
      { 
        free(pLine->pReqList[i].pInData); pLine->pReqList[i].pInData = NULL; 
        pLine->pReqList[i].freeInDataMemoryOnCancel = FALSE;
      }
      if (pLine->pReqList[i].pBlockBuffer) 
      {
        busList[((CdiLineInfoBlk *)pLine->pReqList[i].pDev->pcdiLine)->busIndex].blockFcn((int *)&pLine->pReqList[i],0,&(pLine->pReqList[i].pBlockBuffer),0);
        cdiAssertHeapOkay("cdi_Request",pLine->pReqList[i].pDev->devName,pLine->pReqList[i].accessFlag);
        pLine->pReqList[i].pBlockBuffer = NULL;
      }
      pLine->pReqList[i].groupReq = NULL;
      pLine->pReqList[i].active = 0;
    }
    if ( pLine->pReqList[i].active ) findActive = 1;
    if ( pLine->pReqList[i].active == 0 && findActive == 0 && i == (pLine->numberActiveRequest - 1) )	/* adjust max number */
    {
/* 2008-5 problems!!!!!      pLine->numberActiveRequest = i; */
/*      if (cdiDebugLevel) dbglog("set number active request: %d\n", pLine->numberActiveRequest); */
    }
    if ( pLine->pReqList[i].active == finishRequest && pLine->pReqList[i].userCallDown )
    {
    	if ( pLine->pReqList[i].cancelFlag == 2 )
	      pLine->pReqList[i].active = removeRequest;
      else
  	    pLine->pReqList[i].active = pLine->pReqList[i].lockRequestFlag ? idleState : removeRequest;
      pLine->pReqList[i].cancelFlag = 0;
    }
    cdiPostSemaphore(&pLine->pReqList[i].exec_SemId);
  }
}
int isCdiWriteAccess(short acc)
{
  if (acc == cdiWriteFlag) return TRUE;
  if (acc == cdiReadWriteFlag) return TRUE;
  if (acc == cdiWriteWriteFlag) return TRUE;
  if (acc == cdiWriteReadWriteFlag) return TRUE;
  return FALSE;
}
void cdiSetAlarmData(char *dev,int errcode,char *almdata)
{
  char aaddr[256], saddr[CDI_DEVICE_LONGNAME_SIZE];
  int i, iaddr[CDI_ADDR_SIZE];
  aaddr[0] = 0; saddr[0] = 0;
  if (cdi_getBusAddressParametersForDevice(dev,iaddr,saddr) == 0)
  {
    if (strlen(saddr) > 0) strncpy(aaddr,saddr,256);
    else for (i=0; i<4; i++) sprintf(&aaddr[strlen(aaddr)],"%d ",iaddr[i]);
  }
  memset(almdata,0,64);
# ifdef WIN32
  _snprintf(almdata,64,"%s : %s %s (%d)",
# elif defined(UNIX)
  snprintf(almdata,64,"%s : %s %s (%d)",
# else
  sprintf(almdata,"%s : %s %s (%d)",
# endif
    cdi_getFieldBusNameForDevice(dev),cdi_getBusNameForDevice(dev),aaddr,errcode);
}
int cdiSetAlarms(CdiRequestInfoBlk *pReq)
{
  char almdata[256], *dev;
  CdiLineInfoBlk *pL;
  int i, k;
  if (!gGenerateAlarms) return 0; /* feature disabled */
  if (pReq->perror == NULL) return 0; /* shouldn't be possible */
  if (pReq->groupReq == NULL) 
  {
    if (*pReq->perror != 0) 
    {
      pReq->nerrors++;
      if (pReq->nerrors > gIoErrorThreshold || isCdiWriteAccess(pReq->accessFlag))
      {
        dev = pReq->pDev->devName;
        cdiSetAlarmData(dev,*pReq->perror,almdata);
        SetDeviceAlarm(CDIEQM_TAG,dev,hardware_error,(BYTE *)almdata,almNEWALARM);
        SetDeviceAlarm(CDIEQM_TAG,dev,hardware_error|ALM_SYSTEM_DEFAULT,(BYTE *)almdata,almNEWALARM);
      }
    }
    else
    {
      pReq->nerrors = 0;
    }
  }
  else
  {
    for (i=0; i<numConnectedHandles; i++)	
    { /* search all devices belonging to this group */
      if ((pL=busHandle[i]) == NULL ) continue;
      for (k=0; k<pL->numberActiveRequest; k++)
      {
        if (pL->pReqList[k].groupReq != pReq->groupReq) continue;
        if ( pL->pReqList[k].perror == NULL ) continue;	// 2010-4
        if ( *pL->pReqList[k].perror != 0) 
        {
          pL->pReqList[k].nerrors++;
          if (pL->pReqList[k].nerrors > gIoErrorThreshold || isCdiWriteAccess(pL->pReqList[k].accessFlag))
          {
            dev = pL->pReqList[k].pDev->devName;
            cdiSetAlarmData(dev,*pL->pReqList[k].perror,almdata);
            SetDeviceAlarm(CDIEQM_TAG,dev,hardware_error,(BYTE *)almdata,almNEWALARM);
            SetDeviceAlarm(CDIEQM_TAG,dev,hardware_error|ALM_SYSTEM_DEFAULT,(BYTE *)almdata,almNEWALARM);
          }
        }
        else
        {
          pL->pReqList[k].nerrors = 0;
        }
      }
    }
  }
  return 0;
}
/**************************************/
CDI_EXPORT THREAD_TYPE ThreadCallback(THREAD_PARAM param)
{ 
  CdiLineInfoBlk *pLine = (CdiLineInfoBlk *)param, *pL;
  int j, k, cc, applyRules = 0, hasErrLst = 0, hasNameLst = 0, doutlen, cnt = 0;
  DTYPE dout;
  int *errLst = NULL;
  NAME32 *nameLst = NULL, *nameLstPtr;

  for(;;)
  {
    if ( cnt == 0 ) if (cdiWaitSemaphore(&pLine->call_SemId,0))
    {
      cdilog("ThreadCallback : semaphore ERROR! line: %d, busLine: %d", pLine->cdiLine, pLine->busLine);
      continue;
    }
    for ( cnt = 0, pLine->callCounter = 0; pLine->callCounter < pLine->numberActiveRequest; pLine->callCounter++ )
    {
      if ( pLine->pReqList[pLine->callCounter].asynCallFlag == 0 ) continue;
      cnt++;
      if ( cdiWaitSemaphore(&pLine->pReqList[pLine->callCounter].exec_SemId, 10) ) continue;
//      if ( pLine->pReqList[pLine->callCounter].asynCallFlag == 0 || pLine->pReqList[pLine->callCounter].cancelFlag || (pLine->pReqList[pLine->callCounter].active == 0 || pLine->pReqList[pLine->callCounter].active == removeRequest))
      if ( pLine->pReqList[pLine->callCounter].asynCallFlag == 0 )
      {
        if (isPastDeviceFilter(pLine->pReqList[pLine->callCounter].pDev->devName))
        {
          ttyoutput("ThreadCallback: %.64s is not signaled", pLine->pReqList[pLine->callCounter].pDev->devName);
        }
        cdiPostSemaphore(&pLine->pReqList[pLine->callCounter].exec_SemId);
        continue;
      } 
      /* do any calibration /reformatting here */
      if (pLine->pReqList[pLine->callCounter].needsCalibration)
      {
        applyRules = pLine->pReqList[pLine->callCounter].needsCalibration & RETURN_CALIBRATED;
        hasErrLst = pLine->pReqList[pLine->callCounter].needsCalibration & RETURN_STATUS;
        hasNameLst = pLine->pReqList[pLine->callCounter].needsCalibration & RETURN_NAMES;
        if (hasNameLst)
        { /* point to the right spot in memory to put the names ... */
          doutlen = pLine->pReqList[pLine->callCounter].NgroupsReqs;
          if (nameLst == NULL)
            nameLst = (NAME32 *)((int *)pLine->pReqList[pLine->callCounter].pInData + doutlen*2);
        }
        if (pLine->pReqList[pLine->callCounter].groupReq == NULL) 
        {
          cc = clbrRequest(&pLine->pReqList[pLine->callCounter],applyRules);
          if (nameLst != NULL)
          {
            strncpy(nameLst->name,pLine->pReqList[pLine->callCounter].pDev->devName,CDI_DEVICE_NAME_SIZE);
          }
        }
        else
        {
          nameLstPtr = nameLst;
          for (j=0; j<numConnectedHandles; j++)	/* search all devices belonging to this group */
          {
            if (busHandle[j] == NULL ) continue;
            pL = busHandle[j];
            cc = 0;
            for (k=0; k<pL->numberActiveRequest; k++)
            {
              if (pL->pReqList[k].groupReq != pLine->pReqList[pLine->callCounter].groupReq) continue;
              if ((cc=clbrRequest(&pL->pReqList[k],applyRules)) < 0) break;
              if (nameLstPtr != NULL)
              {
                strncpy(nameLstPtr->name,pL->pDevList[k].devName,CDI_DEVICE_NAME_SIZE);
                nameLstPtr++;
              }
            }
            if (cc < 0) break;
          }
        }
        if (pLine->pReqList[pLine->callCounter].errorCode == 0) pLine->pReqList[pLine->callCounter].errorCode = cc;
      }
      /* set any hardware alarms ...*/
      if ( pLine->pReqList[pLine->callCounter].cancelFlag == 0 )	// 2010-4 to avoid access the perror buffer after cdi_Cancel
      {
        cdiSetAlarms(&pLine->pReqList[pLine->callCounter]);
        if (hasErrLst)
        { /* note : cdiRemapDataSets() can remap the caller's memory ! */
          dout.dArrayLength = pLine->pReqList[pLine->callCounter].inLength / getTypeLength(pLine->pReqList[pLine->callCounter].inType);
          if ( pLine->pReqList[pLine->callCounter].groupReq ) dout.dArrayLength *= pLine->pReqList[pLine->callCounter].NgroupsReqs;
          dout.dFormat = pLine->pReqList[pLine->callCounter].inFormatCode;
          dout.data.vptr = pLine->pReqList[pLine->callCounter].pInData;
          errLst = pLine->pReqList[pLine->callCounter].perror;
          cdiRemapDataSets(&dout,dout.data.bptr,errLst,nameLst);
        }
      }
      /* call registered callback routine */
      if (isPastDeviceFilter(pLine->pReqList[pLine->callCounter].pDev->devName))
      {
        ttyoutput("ThreadCallback: %.64s is firing callback", pLine->pReqList[pLine->callCounter].pDev->devName);
      }
      pLine->pReqList[pLine->callCounter].callback(pLine->pReqList[pLine->callCounter].callBackIndex, getRequestErrors(&pLine->pReqList[pLine->callCounter]));
      pLine->pReqList[pLine->callCounter].devReadyTime = 0;
      setGroupReturnFlags(&(pLine->pReqList[pLine->callCounter]), 0x4);
      pLine->pReqList[pLine->callCounter].asynCallFlag = 0;
      if ( pLine->pReqList[pLine->callCounter].pollFlag == 0 )
      {
        pLine->pReqList[pLine->callCounter].userCallDown++;
        setGroupReturnFlags(&(pLine->pReqList[pLine->callCounter]), 0x1);
      }
      else if (hasErrLst && pLine->pReqList[pLine->callCounter].cancelFlag == 0) memset(errLst,0,dout.dArrayLength*sizeof(int));
      cdiPostSemaphore(&pLine->pReqList[pLine->callCounter].exec_SemId);
    }
  }
}

/**************************************/
int checkGroupSequence(CdiRequestInfoBlk *pReq)
{ int j, k, groupPosition;
  void *groupLinePtr;
  CdiLineInfoBlk *tmpLine;

  if ( pReq->groupReq == NULL ) return 1;
	
  // 2013-05-15 commend out this line because it should never happen from now on
  // checkDeadGroupCalls(pReq);

  if ( pReq->groupSyncStart == 0 ) return 0;
  
  /* callback not down yet! return avoid starting exec without clearing yet 2013-05-15  */
  if ( pReq->NgroupRets )
  {
  	//cdilog("NgroupRets still set!! Num GroupReq: %d, this GroupReq: %d; pReq: 0x%x; reqIndex: %d; devNumber: %d; callBack: 0x%x; caAddrIndex: %d; 0x%x,0x%x,0x%x\n",((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs, pReq->NgroupsReqs, pReq, pReq->reqIndex, pReq->pDev->devNumber, pReq->callback, pReq->pDev->cdiAddr[caAddrIndex], pReq->devParameters[0], pReq->devParameters[1], pReq->devParameters[2]);
    return 0;
  }

  if ( pReq == pReq->groupReq ) return 1;   /* mainRequest(head of the group), send first */

  // 2013-05-15 added this if condition: groupBlocked or in the same busline then check the maonGroup
  if ( (((CdiRequestInfoBlk *)pReq->groupReq)->pBlockBuffer && busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].blockFcn != NULL)
    || pReq->pDev->pcdiLine == ((CdiRequestInfoBlk *)pReq->groupReq)->pDev->pcdiLine )
    if ( ((CdiRequestInfoBlk *)pReq->groupReq)->groupSyncStart ) return 0; /* mainRequest has not been send */
  
  if ( ((CdiRequestInfoBlk *)pReq->groupReq)->pBlockBuffer && busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].blockFcn != NULL ) return 0;	/* 2009-4 block transfer for mainRequest, already done	*/ 

  /* find the seqence */
  groupPosition = pReq->NgroupsReqs;
  groupLinePtr = pReq->pDev->pcdiLine;
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    tmpLine = busHandle[j];
    for ( k = 0; k < tmpLine->numberActiveRequest; k++ )
    {
      if ( tmpLine->pReqList[k].groupReq != pReq->groupReq ) continue;
      if ( tmpLine->pReqList[k].pDev->pcdiLine != groupLinePtr ) continue;
      if ( tmpLine->pReqList[k].NgroupsReqs < groupPosition && tmpLine->pReqList[k].groupSyncStart ) return 0; /* previous element has not been sent  */
    }
  }
  return 1;
}
int cdiHasAlarms(void)
{
  return GetNumAlarms(CDIEQM_TAG,-1,NULL,NULL,NULL,NULL);
}
int cdiBusIsReady(CdiRequestInfoBlk *pReq,int taskIndex)
{
  int cc;
  if (busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].deviceReadyFcn == NULL) return TRUE;
  cc = busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].deviceReadyFcn(pReq,taskIndex);
  cdiAssertHeapOkay("cdiBusIsReady",pReq->pDev->devName,pReq->accessFlag);
  return cc;
}
// need a new parameter in the Device Block: numberConnectedRequests
// if a new Request for this device comes, this number increases by 1
// when it goes out, the number decreases.
// then one can use this number for searching: if pReq->pDev->numberConnectedRequests > 1, one should search, otherwise do not need to search!
int findSameRequest(CdiRequestInfoBlk *pReq, int taskIndex)
{ int tTmp, i;
  CdiLineInfoBlk *pLine;
  CdiRequestInfoBlk *tmpReq, *mainReq;

  /* it come in with pReq->active == transRequest */
  if ( pReq->groupReq != NULL && busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].blockFcn != NULL ) return 0;   /* a single request or no blockfunction registered */
  if ( pReq->pollFlag == 0 || pReq->callback == NULL ) return 0;  /* not an attched polling reqest  */
  if ( pReq->timeout <= 0 ) return 0;           /* if sync or no timeout */
  if ( pReq->accessFlag != cdiReadFlag ) return 0;  /* not a read only  */
  if ( pReq->requestCounter < 1 ) return 0;
  if ( pReq->inLength > 8 ) return 0;       /* a valable memory: short bufLastData[4] */
  tTmp = getTimeMilliseconds();
  /* do not need to check time + 1 second! */
/*	if ( tTmp - pReq->requestTime < abs(pReq->timeout) + 1000 ) return 0; */ /* less then timeInterval + 1 second */

  /* find if there is one with the same device name */
  pLine = busHandle[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine];
  for (i=0; i<pLine->numberActiveRequest; i++)
  {
    tmpReq = &pLine->pReqList[i];
    if ( pReq->pDev != tmpReq->pDev || pReq == tmpReq ) continue;
    if ( pReq->accessFlag != tmpReq->accessFlag || pReq->inLength != tmpReq->inLength ) continue;
    if ( pReq->conditionFlag != tmpReq->conditionFlag ) continue;
    if ( tmpReq->active <= idleState || tmpReq->active >= removeRequest ) continue; 
    if ( tmpReq->dataReturnCondition != DATA_PASS_FILTER ) continue;  /* buffer is not copied into bufLastData  */
    mainReq = (tmpReq->groupReq == NULL || busList[((CdiLineInfoBlk *)tmpReq->pDev->pcdiLine)->busIndex].blockFcn == NULL) ? tmpReq : tmpReq->groupReq;
    if ( tTmp - mainReq->responseTime < pReq->timeout )
    {
      if (cdiDebugLevel > 1) dbglog("findSameRequest: find one req having same device call: %s!", pReq->pDev->devName);
      break; /* find one */
    }
  }
  if ( i == pLine->numberActiveRequest ) return 0;
//  cdilog("findSameRequest: find one req having same device call: %s! now: %d; responce: %d; request: %d; mainResponce: %d; timeinterval:%d\n", pReq->pDev->devName, tTmp, pReq->responseTime, pReq->requestTime, mainReq->responseTime, pReq->timeout);
  /* copy data  */
  pReq->requestCounter++;
  pReq->groupSyncStart = 0;
  pReq->responseTime = mainReq->responseTime;   /* copy the time  */
  pReq->pollTime = pReq->requestTime = pReq->responseTime - 1;
  pReq->active = finishRequest;
  pReq->dataReturnCondition = tmpReq->dataReturnCondition;
  pReq->errorCode = tmpReq->errorCode;
  if (pReq->perror != NULL && tmpReq->perror != NULL ) *pReq->perror = *tmpReq->perror;
  memcpy(pReq->pInData, tmpReq->bufLastData, pReq->inLength);
  setReqFlagsAttachet(pReq, taskIndex);
  if ( pReq->cancelFlag == 0 ) pReq->active = transRequest;   /* start next time interval */
  return 1;
}
/**************************************/
int isCleaningUp = FALSE;
CDI_EXPORT THREAD_TYPE ThreadMaster(THREAD_PARAM param)
{ 
  CdiLineInfoBlk *pLine = (CdiLineInfoBlk *)param;
  int tTmp, ready, nNotReady=0, nRequested=0;
  int cStart, cStop, cDiff, cMax=0, cDump=0;
  int doClearAlarms = pLine->cdiLine == 0 ? TRUE : FALSE;
  struct timeval t_now;
  static struct timeval interim = {0, 0};
  static int chkalm_time = 0;
  int lastSleepTime, msecs, interval;
  if (interim.tv_sec == 0) gettimeofday(&interim, NULL);
  lastSleepTime = getTimeMilliseconds();
  cDump = lastSleepTime;
  pLine->nextWakeupTime = lastSleepTime + BACK_THREAD_TIME / 1000;
  while (!isCleaningUp)
  {
    gettimeofday(&t_now, NULL);
    interval = busList[pLine->busIndex].interval > 0 ? busList[pLine->busIndex].interval : 1000;
    if ((msecs=MSECS(t_now,interim)) > interval)
    { /* note: if no 'activity' then Honggong blocks on cdiWaitSemaphore() below */
      interim = t_now;
      chkalm_time += msecs;
      if (chkalm_time > 1000 && gGenerateAlarms && doClearAlarms)
      { /* clear all alarms once per second */
        ClearAlarm(CDIEQM_TAG, -1);
        chkalm_time = 0;
      }
      if (busList[pLine->busIndex].cycleFcn != NULL) busList[pLine->busIndex].cycleFcn();
    }
    pLine->threadBusy = FALSE;
    if ( pLine->activeCounter || cdiHasAlarms() )
    {
      tTmp = getTimeMilliseconds();
      ready = (pLine->nextWakeupTime != 0 && pLine->nextWakeupTime - tTmp > 0) ? 1 : 0;
      if ( ready || tTmp - lastSleepTime > MAX_BUSY_MS )
      {
        pLine->backReq.tv_sec = 0;
        if ( ready == 0 )	// after MAX_BUSY_MS busy time, make RELAX_TIME_MS ms sleep to avoid 100% CPU busy
          pLine->backReq.tv_nsec = RELAX_TIME_MS * 1000000;
        else if ( pLine->nextWakeupTime - tTmp > BACK_THREAD_TIME / 1000)
          pLine->backReq.tv_nsec = BACK_THREAD_TIME * 1000;
        else
          pLine->backReq.tv_nsec = (pLine->nextWakeupTime - tTmp) * 1000000;
        cdi_nanosleep( &pLine->backReq, NULL );
        lastSleepTime = tTmp;
      }
      else if ( pLine->nextWakeupTime == 0 && tTmp - lastSleepTime > MAX_LINUX_BUSY_MS && RELAX_LINUX_MS - cDiff > 0 )
      {		// after MAX_LINUX_BUSY_MS busy time, make RELAX_TIME_MS ms sleep to avoid High CPU last
          pLine->backReq.tv_sec = 0;
          pLine->backReq.tv_nsec = (RELAX_LINUX_MS - cDiff) * 1000000;	// 100 usec change to 500us
          cdi_nanosleep( &pLine->backReq, NULL );
          lastSleepTime = tTmp;
      }
    }
    else if (!isCleaningUp && cdiWaitSemaphore(&pLine->back_SemId,0))
    {
      cdilog("ThreadMaster : semaphore ERROR! cdi line: %d, busLine: %d", pLine->cdiLine, pLine->busLine);
      continue;
    }
    pLine->threadBusy = TRUE;
    pLine->activeCounter = 0;
    pLine->nextWakeupTime = getTimeMilliseconds() + BACK_THREAD_TIME / 1000;
    
    nNotReady = 0;
    nRequested = 0;
    cStart = getTimeMilliseconds();
    for ( pLine->backCounter = 0; pLine->threadActive && pLine->backCounter < pLine->numberActiveRequest; pLine->backCounter++ )
    {
      if ( pLine->pReqList[pLine->backCounter].active <= idleState ) continue;
      pLine->activeCounter++;
      /* check group call, callback not done yet ==> continue */
		  if ( pLine->pReqList[pLine->backCounter].groupReq )
      {
        if ( ((CdiRequestInfoBlk *)pLine->pReqList[pLine->backCounter].groupReq)->asynCallFlag )
        {
          pLine->nextWakeupTime = 0;
          continue;
        }
      } 
      else 
      {
        if ( pLine->pReqList[pLine->backCounter].asynCallFlag )
        {
          pLine->nextWakeupTime = 0;
          continue;
        }
      }

      /* check timeout */
      if ( cdiWaitSemaphore(&pLine->pReqList[pLine->backCounter].exec_SemId,10) ) continue;

      if (pLine->pReqList[pLine->backCounter].active < finishRequest && busList[((CdiLineInfoBlk *)pLine->pReqList[pLine->backCounter].pDev->pcdiLine)->busIndex].timeoutFcn != NULL)
      {
        ready = busList[((CdiLineInfoBlk *)pLine->pReqList[pLine->backCounter].pDev->pcdiLine)->busIndex].timeoutFcn(&pLine->pReqList[pLine->backCounter], 0, backTaskIndex);
        cdiAssertHeapOkay("ThreadMaster (timeoutFcn)",pLine->pReqList[pLine->backCounter].pDev->devName,pLine->pReqList[pLine->backCounter].accessFlag);
      }
      else ready = 0;
        
      pLine->backTime = getTimeMilliseconds();
      if ( ready || (pLine->pReqList[pLine->backCounter].active < finishRequest && pLine->pReqList[pLine->backCounter].responseTime - pLine->pReqList[pLine->backCounter].requestTime < 0 &&
           (pLine->pReqList[pLine->backCounter].timeout && (pLine->backTime - pLine->pReqList[pLine->backCounter].requestTime > (abs(pLine->pReqList[pLine->backCounter].timeout) + TIMEOUT_CHECK)))  ))
      {
        cdilog("timeout: %d; response %d, request %d, background %d, pReq: 0x%x, readyFlag: %d, active: %d, pollFlag: %d; DevParams: %d, %d, %d; LocalDevParam: 0x%x, 0x%x, 0x%x, 0x%x, 0x%x\n",
          pLine->pReqList[pLine->backCounter].timeout, 
          pLine->pReqList[pLine->backCounter].responseTime, 
          pLine->pReqList[pLine->backCounter].requestTime, 
          pLine->backTime, &pLine->pReqList[pLine->backCounter], ready, pLine->pReqList[pLine->backCounter].active,
          pLine->pReqList[pLine->backCounter].pollFlag, pLine->pReqList[pLine->backCounter].pDev->cdiAddr[0], pLine->pReqList[pLine->backCounter].pDev->cdiAddr[1], pLine->pReqList[pLine->backCounter].pDev->cdiAddr[2],
          pLine->pReqList[pLine->backCounter].devParameters[0], pLine->pReqList[pLine->backCounter].devParameters[1], pLine->pReqList[pLine->backCounter].devParameters[2], pLine->pReqList[pLine->backCounter].devParameters[3], pLine->pReqList[pLine->backCounter].devParameters[4]);

        if (pLine->pReqList[pLine->backCounter].active < finishRequest && busList[((CdiLineInfoBlk *)pLine->pReqList[pLine->backCounter].pDev->pcdiLine)->busIndex].timeoutFcn != NULL)
        {
          busList[((CdiLineInfoBlk *)pLine->pReqList[pLine->backCounter].pDev->pcdiLine)->busIndex].timeoutFcn(&pLine->pReqList[pLine->backCounter], 1, backTaskIndex);
          cdiAssertHeapOkay("ThreadMaster (timeoutFcn)",pLine->pReqList[pLine->backCounter].pDev->devName,pLine->pReqList[pLine->backCounter].accessFlag);
        }
        pLine->pReqList[pLine->backCounter].active = finishRequest; /* finish */
        pLine->pReqList[pLine->backCounter].errorCode = operation_timeout;
        if ( pLine->pReqList[pLine->backCounter].perror ) *(pLine->pReqList[pLine->backCounter].perror) = operation_timeout;
        pLine->pReqList[pLine->backCounter].responseTime = pLine->backTime;
        
        /* treat sync and async differently */
        if ( pLine->pReqList[pLine->backCounter].callback == NULL )
        {
          if ( pLine->pReqList[pLine->backCounter].groupReq == NULL ) cdiPostSemaphore(&pLine->pReqList[pLine->backCounter].sync_SemId);
          else if ( checkGroupReturnFlags(&pLine->pReqList[pLine->backCounter], backTaskIndex) ) cdiPostSemaphore(&((CdiRequestInfoBlk *)pLine->pReqList[pLine->backCounter].groupReq)->sync_SemId);
      	}
        else
        {
          if (isPastDeviceFilter(pLine->pReqList[pLine->backCounter].pDev->devName))
          {
            ttyoutput("ThreadMaster: signal %.32s for callback", pLine->pReqList[pLine->backCounter].pDev->devName);
          }
          setReqFlagsAttachet(&pLine->pReqList[pLine->backCounter], backTaskIndex);
          if ( pLine->pReqList[pLine->backCounter].pollFlag  && pLine->pReqList[pLine->backCounter].cancelFlag == 0 ) pLine->pReqList[pLine->backCounter].active = transRequest; /* start next time interval */
        }
        cdiPostSemaphore(&pLine->pReqList[pLine->backCounter].exec_SemId);
        continue;
      }

      if ( pLine->pReqList[pLine->backCounter].active == transPoll || pLine->pReqList[pLine->backCounter].active == transPoll2nd )
      {
        if ( checkRequestTime( &pLine->pReqList[pLine->backCounter] ) )
        {
          if (ready = cdiBusIsReady(&pLine->pReqList[pLine->backCounter],backTaskIndex))
            cdiExecRequest(&(pLine->pReqList[pLine->backCounter]), backTaskIndex);
          nRequested++;
          if ( ready == 0 )
          {
            pLine->nextWakeupTime = 0;
            nNotReady++;
          }
        }
        cdiPostSemaphore(&pLine->pReqList[pLine->backCounter].exec_SemId);
        continue;
      }
      cdiPostSemaphore(&pLine->pReqList[pLine->backCounter].exec_SemId);

      if ( pLine->pReqList[pLine->backCounter].active != transRequest ) continue;
      if ( pLine->pReqList[pLine->backCounter].callback == NULL &&   /* start the rest of sync group call */
           pLine->pReqList[pLine->backCounter].active == transRequest &&
           (pLine->pReqList[pLine->backCounter].groupReq == NULL || pLine->pReqList[pLine->backCounter].groupReq == &pLine->pReqList[pLine->backCounter]) 
         ) continue;

      if ( cdiWaitSemaphore(&pLine->pReqList[pLine->backCounter].exec_SemId,10) ) continue;
      if ( checkRequestTime( &pLine->pReqList[pLine->backCounter] ) )
      {
        ready = 1;
        if ( checkGroupSequence(&pLine->pReqList[pLine->backCounter]) )
        if (ready = cdiBusIsReady(&pLine->pReqList[pLine->backCounter],backTaskIndex))
        {
          cdiExecRequest(&(pLine->pReqList[pLine->backCounter]), backTaskIndex);
        }
        nRequested++;
        if ( ready == 0 )
        {
          pLine->nextWakeupTime = 0;
          nNotReady++;
        }
      }
      cdiPostSemaphore(&pLine->pReqList[pLine->backCounter].exec_SemId);
    }
    if ( pLine->activeCounter ) freeRequestBlock(pLine);
    cStop = getTimeMilliseconds();
    cDiff = cStop - cStart;
    if ( cDiff > cMax ) cMax = cDiff;
    if ( cStop - cDump > 2000 )
    {
      cDump = cStop;
//      cdilog("bus line: %d, cdi line: %d,loop interval: %d, max interval: %d \n",pLine->busLine, pLine->cdiLine, cDiff, cMax);
    }
  }
  return 0;
}

/**************************************/
CDI_EXPORT int disconnectFromBus ( int busType, int busLine, int fileHandle )
{
  int i, cc;
  char scratch[64];
  for (i=0; i<numBusList; i++) if (busList[i].busType == busType) 
  {
    if (busList[i].freeBusFcn != NULL) 
    {
      if ( gSimulationFlag ) return 0;
      cc = busList[i].freeBusFcn(busLine);
      sprintf(scratch,"line %d",busLine);
      cdiAssertHeapOkay("disconnectFromBus",scratch,cc);
      return cc;
    }
  }
  return -not_implemented;
}
/**************************************/
CDI_EXPORT int connectToBus ( int busType, int busLine, char *parameter, int cdiLine, int numberDevices)
{
  int i, cc;
  char scratch[256];
  for (i=0; i<numBusList; i++) if (busList[i].busType == busType) 
  {
    if (busList[i].initBusFcn != NULL) 
    {
      if ( gSimulationFlag ) return 0;
      cc = busList[i].initBusFcn(busLine,cdiLine,numberDevices,parameter);
      sprintf(scratch,"bus line %d, cdi line %d,num devices %d",busLine,cdiLine,numberDevices);
      cdiAssertHeapOkay("connectToBus",scratch,cc);
      return cc;

    }
    if (parameter != NULL) 
      cdilog("bus parameters for %.16s with no initialization handler",busList[i].busName);
  }
  return -1;
}

/**************************************/
CDI_EXPORT int checkLine(int busType, int busLine)
{ 
  int i;

  for ( i = 0; i < numConnectedHandles; i++ )
    if ( busHandle[i] != NULL && busList [ busHandle[i]->busIndex ].busType == busType && busHandle[i]->busLine == busLine ) break;
  return i < numConnectedHandles ? i : -1;
}

/**************************************/
CDI_EXPORT int findFreeLine()
{ 
  int i;
  if ( numConnectedHandles == 0 ) return 0;
  for ( i = 0; i < numConnectedHandles; i++ ) if ( busHandle[i] == NULL ) return i;
  if ( numConnectedHandles >= NUMBER_LINES ) return -1; /* slots full! */
  return numConnectedHandles;
}

/**************************************/
CDI_EXPORT int getFreeRequestBlock(CdiLineInfoBlk *pLine)
{ 
  int i;

  if ( pLine->numberActiveRequest == 0 ) return 0;
  /* find if there is a Loch */
  for ( i = 0; i < pLine->numberActiveRequest; i++ ) 
  {
    if ( pLine->pReqList[i].active == 0 ) return i;
  }
  if ( pLine->numberActiveRequest >= pLine->numberRequest ) return -resources_exhausted; /* slots full! */
  return pLine->numberActiveRequest;
}

/**************************************/
CDI_EXPORT void makeCdiHashTable(int cdiLine, int ndevices)
{
  int cc=0,n=ndevices;

  if (numConnectedHandles == 0 || busHandle == NULL) return;
  if (cdiLine < 0 || cdiLine > numConnectedHandles || busHandle[cdiLine] == NULL) return;
  if (busHandle[cdiLine]->devHashList != NULL) return; /* already have hash table */
  if (n < 1) n = busHandle[cdiLine]->numberDevices;/* get total number of devices */
  if (n < CDI_PAYOFF_HASHSIZE) return;
  if (n < CDI_SMALL_HASHSIZE) busHandle[cdiLine]->devHashSize = CDI_SMALL_HASHSIZE;
  else if (n < CDI_MEDIUM_HASHSIZE) busHandle[cdiLine]->devHashSize = CDI_MEDIUM_HASHSIZE;
/*  else if (n < CDI_LARGE_HASHSIZE) busHandle[cdiLine]->devHashSize = CDI_LARGE_HASHSIZE; */
  else busHandle[cdiLine]->devHashSize = CDI_LARGE_HASHSIZE;
  if ((busHandle[cdiLine]->devHashList=(cdiHashXRefStruct **)calloc(busHandle[cdiLine]->devHashSize,sizeof(cdiHashXRefStruct *))) == NULL) 
    ccerr(out_of_server_memory);
err:
  if (cc) cdilog("makeCdiHashTable: %s",erlst[cc]);
  return;
}
CDI_EXPORT int findDevice(char *devName, int *pCdiLine, int *pDevIndex)
{ 
  int i, j, idx, devnr = -1, cc = 0;
  cdiHashXRefStruct *xref;
  char scratch[128],*c=NULL;

  if (devName == NULL || pCdiLine == NULL || pDevIndex == NULL) ccerr(argument_list_error);
  if (strlen(devName) == 0) ccerr(device_not_connected);
  if (devName[0] == '#') /* given as a number */
  {
    strncpy(scratch,&devName[1],CDI_DEVICE_NAME_SIZE); scratch[CDI_DEVICE_NAME_SIZE] = 0;
    if ((c=strchr(scratch,'.')) != NULL)
    { /* given as #<ln>.<index> */
      *c++ = 0;
      i = atoi(scratch);
      j = atoi(c);
      if (i < 0 || i >= numConnectedHandles) ccerr(illegal_line);
      if (j < 0 || j >= busHandle[i]->numberDevices) ccerr(invalid_index);
      *pCdiLine = i;
      *pDevIndex = j;
      return busHandle[i]->pDevList[j].devNumber;
    }
    /* given as #<nr> */
    devnr = atoi(scratch);
  }
  for (i=0; i<numConnectedHandles; i++) /* search through all bus lines */
  {
    if (busHandle[i] == NULL) continue;
    if (busHandle[i]->devHashSize == 0) /* no hash table ...*/
    {
      for (j=0; j<busHandle[i]->numberDevices; j++)
      {
        if (devnr >= 0)
        {
          if (busHandle[i]->pDevList[j].devNumber != devnr) continue;
        }
        else
        {
          if (busHandle[i]->pDevList[j].devName[0] == 0) continue;
          if (strncmp(devName, busHandle[i]->pDevList[j].devName,CDI_DEVICE_NAME_SIZE)) continue;
        }
        *pCdiLine = i;
        *pDevIndex = j;
        return busHandle[i]->pDevList[j].devNumber;
      }
    }
    else
    {
      idx = (int)(ElfHash((unsigned char *)devName) % busHandle[i]->devHashSize);
      for (xref=busHandle[i]->devHashList[idx]; xref; xref=xref->nxt)
      {
        if (devnr >= 0)
        {
          if (devnr != xref->devNumber) continue;
          j = xref->cdiIndex;
        }
        else
        {
          j = xref->cdiIndex;
          if (busHandle[i]->pDevList[j].devName[0] == 0) continue;
          if (strnicmp(devName, busHandle[i]->pDevList[j].devName,CDI_DEVICE_NAME_SIZE)) continue;
        }
        *pCdiLine = i;
        *pDevIndex = j;
        return busHandle[i]->pDevList[j].devNumber;
      }
    }
  }
  cc = device_not_connected;
err:
  return -cc;
}
int isAccessAllowed(CdiDeviceInfoBlk *pDev,int accessFlag)
{
  int accessMask = 0x00;
  if (pDev == NULL) return FALSE;
  if (pDev->allowedAccess == 0 || pDev->allowedAccess == 0xffff) return TRUE;
  accessMask |= (0x01 << accessFlag);
  return (pDev->allowedAccess & accessMask);
}

/************************************/
CDI_EXPORT int checkReqWriteData(CdiRequestInfoBlk *pReq, int *paccessFlag, int returnCondition)
{ int flen, xlen, size, retFlag = 0;
  /* use default input data from the database */
  size = getTypeLength((short)pReq->pDev->dataFormat);
  if ( pReq->pDev->dataInput != NULL )
  { /* there is registered input data for this device! do I need to use it? */
    if (*paccessFlag == cdiReadFlag)
    {
      if ( pReq->pOutData == NULL ) /* no input data supplied by the caller, change RD to WRRD */
      {
        if (pReq->pDev->allowedAccess & 0x08)
        {
          if ( pReq->pDev->dataInputSize < size ) return -dimension_error;
          *paccessFlag = cdiWriteReadFlag;
          pReq->pOutData = pReq->pDev->dataInput;
          pReq->outLength = size;
          pReq->outType = pReq->pDev->dataFormat;
          retFlag = 1;
        }
        else if (pReq->pDev->allowedAccess == 0x02) return -illegal_command;
      }
    }
    else if (*paccessFlag == cdiReadReadFlag)
    {
      if ( pReq->pOutData == NULL ) /* no input data supplied by the caller, change RDRD to WRRDRD */
      {
        if (pReq->pDev->allowedAccess & 0x80)
        {
          if ( pReq->pDev->dataInputSize < size ) return -dimension_error;
          *paccessFlag = cdiWriteReadReadFlag;
          pReq->pOutData = pReq->pDev->dataInput;
          pReq->outLength = size;
          pReq->outType = pReq->pDev->dataFormat;
          retFlag = 1;
        }
        else if (pReq->pDev->allowedAccess == 0x02) return -illegal_command;
      }
    }
    else if (*paccessFlag == cdiWriteFlag)  /* change WR to WRWR */
    {
      if (pReq->pDev->allowedAccess & 0x10)
      {
        if ( pReq->pDev->dataInputSize < size ) return -dimension_error;
        *paccessFlag = cdiWriteWriteFlag;
        memcpy(pReq->cpOutData, pReq->pDev->dataInput, size);
        flen = pReq->outLength;
        if ( flen > DEF_INPUT_BYTES - size ) return -dimension_error;
        if ( flen > DEF_INPUT_BYTES - size ) flen = DEF_INPUT_BYTES - size;
        memcpy(&pReq->cpOutData[size], pReq->pOutData, flen);
        pReq->pOutData = pReq->cpOutData;
        pReq->outLength = size + flen;
        retFlag = 1;
      }
    }
    else if (*paccessFlag == cdiReadWriteFlag)  /* change RDWR to WRRDWR */
    { /* default firstWrite and SecondWrite  */
      xlen = (pReq->pDev->dataPattern && returnCondition > anyData) ? 1 : 2;
      if (pReq->pDev->dataInputSize >= size && pReq->pOutData && (pReq->pDev->dataInputSize + (int)pReq->outLength) >= size * xlen && pReq->pDev->allowedAccess & 0x20)
      {
        *paccessFlag = cdiWriteReadWriteFlag;
        memcpy(pReq->cpOutData, pReq->pDev->dataInput, pReq->pDev->dataInputSize);
        flen = pReq->outLength;
        if ( flen > DEF_INPUT_BYTES - pReq->pDev->dataInputSize ) return -dimension_error;
        if ( flen > DEF_INPUT_BYTES - pReq->pDev->dataInputSize ) flen = DEF_INPUT_BYTES - pReq->pDev->dataInputSize;
        memcpy(&pReq->cpOutData[pReq->pDev->dataInputSize], pReq->pOutData, flen);
        pReq->pOutData = pReq->cpOutData;
        pReq->outLength = pReq->pDev->dataInputSize + flen;
        retFlag = 1;
      }
    }
  }
  if ( *paccessFlag != cdiReadFlag && *paccessFlag != cdiReadReadFlag )
  { /* not a simple read access */
    if ( pReq->pDev->dataFormat >= 0 && pReq->pDev->dataFormat != pReq->outType ) return -illegal_format;

    if ( *paccessFlag == cdiWriteWriteFlag ) flen = 2;
    else if ( *paccessFlag == cdiReadWriteFlag)
    {
      if ( pReq->pDev->dataPattern == NULL || returnCondition == anyData ) flen = 2;
    }
    else if ( *paccessFlag == cdiWriteReadWriteFlag)
    {
      flen = (pReq->pDev->dataPattern == NULL || returnCondition == anyData) ? 3 : 2;
    }
    else flen = 1;

    if ( (int)pReq->outLength < size * flen ) return -dimension_error;

    if ( retFlag == 0 ) /* input from user */
    {
      if ( pReq->pDev->dataInput == NULL && 
           pReq->pDev->dataInputSize && 
           (int)pReq->outLength > pReq->pDev->dataInputSize + (flen-1) * size ) 
        return -dimension_error;
    }
    else if ( pReq->pDev->dataInputSize > 0)  /* use default input data */
    { /* do not check WriteWrite */
      if ( *paccessFlag != cdiWriteWriteFlag && (int)pReq->outLength > size * flen ) return -dimension_error;
    }
  }
  return retFlag;
}
int rvrsClbrRequest(CdiRequestInfoBlk *pReq,int applyRules)
{
  int i, flen, tlen, dlen;
  CdiDeviceInfoBlk *pDev;
  int fromFormat, toFormat;
  BYTE *fromData, *toData;
  CdiRule *rule = NULL;
  double dval;
  /* the request has input calibrated data which need to be converted to hdw bus data */
  if ((fromFormat=pReq->outType) < 0) return -illegal_format;
  flen = getTypeLength((short)fromFormat);
  dlen = pReq->outLength/flen;
  if (flen <= 0) return -illegal_format;
  fromData = pReq->pOutData;
  pDev = pReq->pDev;
  toFormat = pDev->dataFormat;
  if (toFormat < 0) return -illegal_format;
  if ((tlen=getTypeLength((short)toFormat)) <= 0) return -illegal_format;
  if (!applyRules && toFormat == fromFormat) return 0; /* nothing to do */
  /* allocate separate buffer to hold the input hdw bus data */
  toData = (BYTE *)calloc(dlen,getTypeLength((short)pDev->dataFormat));
  if (toData == NULL) return -out_of_server_memory;
  rule = applyRules ? pDev->rvrsRules : NULL;
  for (i=dlen-1; i>=0; i--)
  { /* start from the end and work backwards for in situ assignments */
    dval = applyClbr(rule,fromData+i*flen,fromFormat);
    assignClbr(dval,toData+i*tlen,toFormat);
  }
  pReq->pOutData = toData;
  pReq->outType = toFormat;
  pReq->outLength = dlen * tlen;
  pReq->freeOutDataMemoryOnCancel = TRUE;
  return 0;
}

void cdi_reverseCalibrate(CdiRequestInfoBlk *pReq,int cdiNeedsCalibration)
{
  if (cdiNeedsCalibration & INPUT_CALIBRATED)
  {
    rvrsClbrRequest(pReq,TRUE);
  }
  else if (cdiNeedsCalibration & INPUT_REFORMATED)
  {
    rvrsClbrRequest(pReq,FALSE);
  }
  return;
}

/************************************/
/* find and init requestBuffer */
/************************************/
CDI_EXPORT int init_RequestBuffer(char *devName, void *inData, short readType, int inLength, void *outData, short writeType, int outLength, void *dataPatternWindow, 
                       void *dataMask, int *parameter, int accessFlag, int returnCondition, int timeout, int *perror)
{ 
  int i, cc = 0, flen, tlen, cdiLine, cdiIndex, reqIndex, cdiNeedsCalibration = 0;
  int isDiscovery = IsDiscoveryDevice(devName);
  CdiRequestInfoBlk *pReq = NULL;

  if (isDiscovery) goto access_ok;
  /* check parameters */
  if (returnCondition & RETURN_CALIBRATED) cdiNeedsCalibration |= RETURN_CALIBRATED;
  if (returnCondition & RETURN_REFORMATED) cdiNeedsCalibration |= RETURN_REFORMATED;
  if (returnCondition & RETURN_STATUS) cdiNeedsCalibration |= RETURN_STATUS;
  if (returnCondition & RETURN_NAMES) cdiNeedsCalibration |= RETURN_NAMES;
  if (returnCondition & INPUT_CALIBRATED) cdiNeedsCalibration |= INPUT_CALIBRATED;
  if (returnCondition & FREE_ONCLOSE) cdiNeedsCalibration |= FREE_ONCLOSE;
  returnCondition &= 0x7;
  
  if ( accessFlag < cdiReadFlag || accessFlag > cdiWriteReadReadFlag ) return -argument_list_error;
  if ( returnCondition < anyData || returnCondition > maskChangePercent ) return -argument_list_error;
  if ( accessFlag != cdiReadFlag && accessFlag != cdiReadReadFlag )
  {
    if ( outData == NULL || outLength == 0 ) return -argument_list_error;
  }
  if ( accessFlag != cdiWriteFlag && accessFlag != cdiWriteWriteFlag )
  {
    if ( inData == NULL || inLength == 0 ) return -argument_list_error;
  }
  if ( inData && inLength != 0 )
    if ( inLength < getTypeLength(readType) ) return -dimension_error;

access_ok:
  if ( findDevice(devName, &cdiLine, &cdiIndex) < 0 )
  {
    if (cdiDebugLevel) dbglog("invalid device name: %s",devName);
    return -device_not_connected;
  }

  reqIndex = getFreeRequestBlock(busHandle[cdiLine]);
  if ( reqIndex < 0 )
  {
    cdilog("no free memory slot! Number of active maximum slots: %d", busHandle[cdiLine]->numberActiveRequest);
    return reqIndex;
  }

  if (cdiDebugLevel) dbglog("1. get free request Index: %d; cdiLine: %d, activeRequest: %d\n", reqIndex, cdiLine, busHandle[cdiLine]->numberActiveRequest);

  pReq = &(busHandle[cdiLine]->pReqList[reqIndex]); /* load request block */
  cdiWaitSemaphore(&pReq->exec_SemId,0);
  pReq->pDev = &(busHandle[cdiLine]->pDevList[cdiIndex]); /* assign device */
  pReq->pDev->reqFormat = readType;

  if (pReq->pDev->dataFormat != readType) cdiNeedsCalibration |= RETURN_REFORMATED;

  /* is there data to write to hardware ? */
  pReq->pOutData = outData;
  pReq->outLength = outLength;
  pReq->outType = writeType;
  /* does input data need to be reformatted ? */
  if (outLength > 0 && pReq->pDev->dataFormat != writeType) 
    cdiNeedsCalibration |= INPUT_REFORMATED;
  cdi_reverseCalibrate(pReq,cdiNeedsCalibration);

  /* check write data format and size  */
  cc = checkReqWriteData(pReq, &accessFlag, returnCondition);
  if ( cc == 0 ) 
  {
    if (!isAccessAllowed(pReq->pDev,accessFlag)) ccerr(illegal_command);
  }
  else if ( cc < 0 ) ccerr(-cc);

  /* check read format  */
  if (pReq->pDev->dataFormat >= 0 && readType >= 0)
  {
    flen = getTypeLength(readType);
    tlen = getTypeLength((short)pReq->pDev->dataFormat);
    /* data area allocated as per tlen */
    if (inLength < tlen) ccerr(illegal_format); /*dimension_error*/ 
    if (flen <= 0 || flen < tlen) ccerr(illegal_format);
    if ( pReq->pDev->dataFormat == cdiChar && readType == cdiChar32)
      pReq->inLength = inLength;
    else
      pReq->inLength = inLength*tlen/flen;
    pReq->inType = pReq->pDev->dataFormat;
  }
  else
  {
    pReq->inLength = inLength;
    pReq->inType = readType;
  }
  pReq->pInData = inData;

  /* check read size  */
  if ( pReq->pDev->dataOuputSize && (int)pReq->inLength > pReq->pDev->dataOuputSize + pReq->pDev->dataOuput2ndSize) ccerr(dimension_error);

  pReq->hasPatternFilter = FALSE;
  if (pReq->pDev->dataPattern != NULL && returnCondition == anyData && (accessFlag != cdiWriteFlag && accessFlag != cdiWriteWriteFlag) )
  {
    returnCondition = pReq->pDev->dataPatternTrigger == CM_SEEKPATTERN ? maskCompEqual : maskCompNotEqual;
  }
  if ( returnCondition > anyData )	/* check dataPatternWindow */
  {
    if (dataPatternWindow != NULL)
    {
      if (((int *)dataPatternWindow)[0] != 0 || ((int *)dataPatternWindow)[1] != 0) 
      {
        memcpy(pReq->patternFilter,dataPatternWindow,CDI_PATTERN_FILTER_SIZE);
        pReq->hasPatternFilter = TRUE;
      }
    }
    else if (pReq->pDev->dataPattern != NULL)
    {
      memcpy(pReq->patternFilter,pReq->pDev->dataPattern,CDI_PATTERN_FILTER_SIZE);
      pReq->hasPatternFilter = TRUE;
    }
    if ( pReq->hasPatternFilter == FALSE ) ccerr(argument_list_error);
  }

  pReq->hasDataMask = FALSE;
  if ( pReq->inType <= cdiInteger && pReq->inType >= cdiByte && (accessFlag != cdiWriteFlag && accessFlag != cdiWriteWriteFlag) ) /* check dataMask for reading data */
  {
    if (dataMask != NULL)
    {
      if (((int *)dataMask)[0] != -1 || ((int *)dataMask)[1] != -1) 
	    {
        memcpy(pReq->mask,dataMask,8);
        pReq->hasDataMask = TRUE;
	    }
    }
    else if (pReq->pDev->dataMask != NULL)
    {
	    memcpy(pReq->mask,pReq->pDev->dataMask,8);
	    pReq->hasDataMask = TRUE;
    }
    if ( pReq->hasDataMask == TRUE )
    {
      pReq->maskValue = *(int *)pReq->mask;
      if ( pReq->maskValue == 0 || pReq->maskValue == -1 || (pReq->inType == cdiShort && pReq->maskValue == 0xffff) ||
            (pReq->inType == cdiByte && pReq->maskValue == 0xff) || (pReq->inType == cdiInteger && pReq->maskValue == 0xffffffff)) pReq->hasDataMask = FALSE;
    }
  }

  /* wrong parameters! */
  i = busHandle[cdiLine]->busIndex;
  if (busList[i].filterBusFcn != NULL && gSimulationFlag == FALSE )
  {
    cc = busList[i].filterBusFcn(parameter != NULL ? parameter : pReq->pDev->devParameters,accessFlag);
    cdiAssertHeapOkay("init_RequestBuffer (filterBusFcn)",pReq->pDev->devName,accessFlag);
    if (cc != 0) ccerr(-cc);
  }

  if (parameter != NULL) /* use what was passed */
  {
    memcpy(pReq->devParameters, parameter, NUMBER_EXEC_PARAM * sizeof(int));
  }
  else if (pReq->pDev->devParameters != NULL) /* use the default set */
  {
    memcpy(pReq->devParameters, pReq->pDev->devParameters, NUMBER_EXEC_PARAM * sizeof(int));
  }

  pReq->accessFlag = accessFlag;
  pReq->conditionFlag = returnCondition;
  pReq->timeout = timeout; /* milliseconds */
  pReq->perror = perror;   /* user provides buffer for storing error */
  pReq->needsCalibration = cdiNeedsCalibration;

  pReq->registerTime = getTimeMilliseconds();

  pReq->requestNumber = reqIndex + (cdiLine << 16); /* upper word is cdiLine number */
  pReq->reqIndex = reqIndex;
  pReq->groupReq = NULL;
  pReq->NgroupsReqs = 0;
  pReq->groupSyncStart = 0;
  pReq->bufIndex = -1;
  pReq->active = idleState;   /* avoid to be removed!!!! */

  if ( reqIndex >= busHandle[cdiLine]->numberActiveRequest )
  {
    busHandle[cdiLine]->numberActiveRequest = reqIndex + 1; /*    busHandle[cdiLine]->numberActiveRequest++;  */
    if (cdiDebugLevel) dbglog("set number active request: %d", busHandle[cdiLine]->numberActiveRequest);
  }
err:
  if (cc) 
  {
    if (pReq != NULL && pReq->freeOutDataMemoryOnCancel) 
    {
      free(pReq->pOutData);
      pReq->freeOutDataMemoryOnCancel = FALSE;
    }
  }
  cdiPostSemaphore(&pReq->exec_SemId);
  return cc ? (-cc) : pReq->requestNumber;
}

/************************************/
CDI_EXPORT int checkCallbackIndex(void (*callBack)(int, int),int callbackID)
{ int i, j;

  for ( j = 0; j < numConnectedHandles; j++ )	/* search all device's callbackID  */
  {
	  if ( busHandle[j] == NULL ) continue;
	  for ( i = 0; i < busHandle[j]->numberActiveRequest; i++ )
	  {
	    if (busHandle[j]->pReqList[i].active == 0 || /* not active */
          busHandle[j]->pReqList[i].active == removeRequest || /* wait for remove */
          busHandle[j]->pReqList[i].cancelFlag ||	/* link already canceled	*/
          busHandle[j]->pReqList[i].callBackIndex != callbackID || /* index is different */
          busHandle[j]->pReqList[i].callback != callBack)   /* function is different */
        continue;
	    return -argument_list_error; /* ids are equal and cb fcns are equal */
	  }
  }
  return 0;
}

/************************************/
/* sync or Async call */
/************************************/
CDI_EXPORT char *cdiGetRedirectedDeviceServer(char *devName)
{
  int idx;
  cdiRedirXRefStruct *xref;
  if (devName == NULL) return NULL;
  if (gCdiRedirectedDevicesHashSize == 0) return NULL;
  idx = (int)(ElfHash((unsigned char *)devName) % gCdiRedirectedDevicesHashSize);
  for (xref=cdiRedirXref[idx]; xref; xref=xref->nxt)
  {
    if (!strncmp(devName,cdiRedirXref[idx]->devName,CDI_DEVICE_NAME_SIZE)) 
      return cdiRedirXref[idx]->devRedirectedServerName;
  }
  return NULL;
}
CDI_EXPORT int registerRequest(char *devName, void *readData, short readType, unsigned int readLength, void *writeData, short writeType, unsigned int writeLength, void *dataPatternWindow, 
                void *dataMask, int *parameter, int accessFlag, int returnCondition, int timeout, void (*callBack)(int, int), 
                int pollFlag, int *perror, int callbackID, int lockRequest)
{ 
  CdiRequestInfoBlk *pReq;
  int reqLineNumber, cdiLine, reqIndex, err;
  char *c;

  if ((c=cdiGetRedirectedDeviceServer(devName)) != NULL)
  {
    strncpy(devName,c,64);
    return -server_redirection;
  }

  if ( callBack && callbackID >= 0 && callbackID <= 0xffff )	/* check if callbackID exists */
  {
    reqLineNumber = checkCallbackIndex(callBack,callbackID);
    if ( reqLineNumber < 0 ) return reqLineNumber; /* error return! */
  }

  reqLineNumber = init_RequestBuffer(devName, readData, readType, readLength, writeData, writeType, writeLength, dataPatternWindow, dataMask, parameter, accessFlag, returnCondition, timeout,  perror);
  if ( reqLineNumber < 0 ) return reqLineNumber; /* error return! */

  cdiLine = reqLineNumber >> 16;
  reqIndex = reqLineNumber & 0xffff;
  pReq = &(busHandle[cdiLine]->pReqList[reqIndex]); /* load request block */

  cdiWaitSemaphore(&pReq->exec_SemId,0);
  pReq->callBackIndex = (callBack && callbackID >= 0 && callbackID <= 0xffff) ? callbackID : (reqLineNumber + 0x10000);
  pReq->callback = callBack;
  pReq->pollFlag = callBack == NULL ? 0 : pollFlag; /* polling with async together */
  pReq->lockRequestFlag = lockRequest;
  pReq->active = idleState;
  cdiPostSemaphore(&pReq->exec_SemId);

  if ((CdiLineInfoBlk *)pReq->pDev->pcdiLine == NULL) return -device_not_connected;
  if ( busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].loadRequestFcn != NULL )
	  if ( (err = busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].loadRequestFcn(pReq, execIndex)) < 0 ) return err;

  if (cdiDebugLevel) dbglog("preq: 0x%x callback : 0x%x; Callbackindex: %d, return Handle: 0x%x\n",pReq, callBack, pReq->callBackIndex, (reqLineNumber + 0x10000));
  return (reqLineNumber + 0x10000); /* reqLineNumber; */
}

int cdiParseDeviceString(char *devString,NAME32 *devGroup,int *groupSize)
{
  int i, len, num=0, cc=0;
  int cdiLine, index, n, istart, istop;
  char *g=NULL,*c=NULL,*g2=NULL,*cp=NULL,*rngToken=SIMPLE_RANGE_TOKEN;
  char *tmpString;

  if (devString == NULL || devGroup == NULL || groupSize == NULL || *groupSize < 1) ccerr(argument_list_error);
  if ((tmpString=(char *)alloca(cdiDevStringBufferSize+16)) == NULL) ccerr(out_of_local_memory);
  strncpy(tmpString,devString,cdiDevStringBufferSize);
  /* split the seperated groups */
  for (i=0, g=strtok_r(tmpString,SEPERATOR_TOKEN,&cp); g != NULL; i++, g=strtok_r(NULL,SEPERATOR_TOKEN,&cp))
  { /* split each group into the individual devices ... */
    while (*g == ' ') g++; /* leading blanks */
    rngToken = (g[0] == '#') ? SIMPLE_RANGE_TOKEN : COMPLEX_RANGE_TOKEN;
    if ((c=strstr(g,rngToken)) != NULL)
    {
      g2 = c + strlen(rngToken);
      *c = 0;
      while (*g2 == ' ') g2++; /* leading blanks in front of range terminator */
      for (len=(int)strlen(g2); len > 0 && g2[len-1] == ' '; g2[len-1] = 0); /* trailing blanks */
    }
    for (len=(int)strlen(g); len > 0 && g[len-1] == ' '; g[len-1] = 0); /* trailing blanks */
    if ((istart=findDevice(g,&cdiLine,&index)) < 0) 
    {
      cdilog("device %s was not found !",g);
      ccerr(device_not_connected);
    }
    if ( c == NULL ) istop = istart;
    else if ((istop=findDevice(g2,&cdiLine,&index)) < 0) 
    {
      cdilog("device %s was not found !",g2);
      ccerr(device_not_connected);
    }

    for (n=istart; ; ) /* allow high to low */
    {
      if (num < *groupSize)
#     ifdef WIN32
        _snprintf(devGroup[num].name,CDI_DEVICE_NAME_SIZE,"#%d",n);
#     else
        snprintf(devGroup[num].name,CDI_DEVICE_NAME_SIZE,"#%d",n);
#     endif

      num++;
      if ( n == istop ) break; 
      istart < istop ? (n++) : (n--);
    }
  }
  *groupSize = num;
err:
  return cc;
}

/************************************/
CDI_EXPORT int cdi_Request(char *devName, void *readData, short readType, unsigned int readLength, void *writeData, short writeType, unsigned int writeLength, void *dataPatternWindow, 
                void *dataMask, int *parameter, int accessFlag, int returnCondition, int timeout, void (*callBack)(int, int), 
                int pollFlag, int *perror, int callbackID, int lockRequest)
{ int i, j, numDevices, ret, *pInt;
  NAME32 localDevName, *pDevName;
  int rLen, wLen, tmpLen;
  CdiRequestInfoBlk *pReq;

  if (readLength <= 0) readType = -1;
  numDevices = 1;
  ret = cdiParseDeviceString(devName, &localDevName, &numDevices);
  if (cdiDebugLevel) dbglog("cdiParseDeviceString: numDevices: %d; name: %s; return: %d\n",numDevices, localDevName.name, ret);
  if ( ret ) return -ret;
  if ( numDevices == 1 )
  {
    return registerRequest(devName, readData, readType, readLength, writeData, writeType, writeLength, dataPatternWindow, 
                dataMask, parameter, accessFlag, returnCondition, timeout, callBack, pollFlag, perror, callbackID, lockRequest);
  }
  /* for group call */
  if ( (pDevName = (NAME32 *)calloc(numDevices, sizeof(NAME32))) == NULL ) return -out_of_local_memory;
  if ( (pInt = (int *)calloc(numDevices, sizeof(int))) == NULL )
  {
    free(pDevName);
    return -out_of_local_memory;
  }
  ret = cdiParseDeviceString(devName, pDevName, &numDevices);
  rLen = (readLength == 0 || readData == NULL ) ?  0 : (readLength / numDevices);
  wLen = (writeLength == 0 || writeData == NULL ) ? 0 : (writeLength / numDevices);
  tmpLen = (rLen && getTypeLength(readType) > 0) ? (rLen / getTypeLength(readType)) : 0;
  for (i=0; i<numDevices; i++)
  {
    pInt[i] = registerRequest(pDevName[i].name, 
      rLen == 0 ? NULL : &((char *)readData)[rLen * i], readType, rLen, 
      wLen == 0 ? NULL : &((char *)writeData)[wLen * i], writeType, wLen, 
      dataPatternWindow, dataMask, parameter, accessFlag, returnCondition, timeout, 
      callBack, pollFlag, perror==NULL ? NULL : &perror[i*tmpLen], i == 0 ? callbackID : -1, lockRequest);
    if ( pInt[i] < 0 ) break;
    if ( i > 0 )
    {
      ret = cdi_loadGroupRequests(pInt[0], pInt[i]);
      if ( ret ) break;
    }
  }
  if ( i < numDevices || ret ) /* error */
  {
    for (j=0; j<=i; j++) if (pInt[j]>0) cdi_CancelRequest(pInt[j]); /* error  */
    if ( ret == 0 ) ret = pInt[i] < 0 ? pInt[i] : -invalid_data; 
  }
  else
  {
    ret = 0;
/*    if (pollFlag)  */ /* only for polling request */
    {
      get_RequestBufferPointer(pInt[0], &pReq);
      if (busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].blockFcn != NULL)
      {
        ret = busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].blockFcn(pInt, numDevices, &(pReq->pBlockBuffer), 1);
        cdiAssertHeapOkay("cdi_Request",pReq->pDev->devName,pReq->accessFlag);
		if ( ret )
		{
			for (j=0; j<numDevices; j++) if (pInt[j]>0) cdi_CancelRequest(pInt[j]); /* error  */
		    if ( ret > 0 ) ret = -invalid_data; 
		}
      }
    }
	if ( ret == 0 ) ret = pInt[0];
  }
  free(pDevName);
  free(pInt);
  return ret;
}
#ifdef CHANGE_CALLBACK_ID
/************************************/
CDI_EXPORT int findReqIndex(int requestIndex)
{ 
  int i, j;

  for (i=0; i<numConnectedHandles; i++)
  {
    if ( busHandle[i] == NULL ) return -1;
    for (j=0; j<busHandle[i]->numberActiveRequest; j++)
      if ( requestIndex == busHandle[i]->pReqList[j].callBackIndex ) break;
    if ( j < busHandle[i]->numberActiveRequest ) return busHandle[i]->pReqList[j].requestNumber + 0x10000;
  }
  return -1;
}
#endif
/************************************/
/* check request index */
/************************************/
CDI_EXPORT int cdi_isValidRequestIndex(int requestIndex)
{ int cdiLine, reqIndex;

  if ( requestIndex < 0x1000 ) return FALSE;

  cdiLine = (requestIndex >> 16) - 1;
  reqIndex= (requestIndex & 0xffff);
  if (cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL || reqIndex >= busHandle[cdiLine]->numberActiveRequest ) return FALSE;
  return TRUE; /* if it is a valid requestIndex ! */
}

/************************************/
CDI_EXPORT int get_RequestBufferPointer(int requestIndex, CdiRequestInfoBlk **pReq)
{ int cdiLine, reqIndex;

  if ( cdi_isValidRequestIndex(requestIndex) == FALSE )
  {
    if (cdiDebugLevel) dbglog("invalid input: %x; (line %d, index %d)",
      requestIndex, (requestIndex >> 16) - 1, requestIndex & 0xffff);
    return -non_existent;
  }
  cdiLine = (requestIndex >> 16) - 1;
  reqIndex = requestIndex & 0xffff;
  *pReq = &(busHandle[cdiLine]->pReqList[reqIndex]); /* load request block */

  return 0;
}

/************************************/
CDI_EXPORT int getGroupRequestInfo(int requestIndex, void *buffers, int index)
{ int i, j, offset;
  CdiLineInfoBlk *pLine;
  CdiRequestInfoBlk *pReq;

  if ( buffers == NULL ) return -illegal_data_size;
  if ( (i = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return i;
  if ( pReq->groupReq == NULL )
  {
    if ( index == CDI_INFO_ERRORS )
	{
	  if ( buffers ) *(int *)buffers = pReq->errorCode;
	  else return pReq->errorCode;
	}
    else if ( index == CDI_INFO_NAMES ) strncpy(buffers, pReq->pDev->devName, CDI_DEVICE_NAME_SIZE);
    else if ( index == CDI_INFO_STATUS ) *(int *)buffers = pReq->active;
    else if ( index == CDI_INFO_NDEVS ) *(int *)buffers = 1;
    else if ( index == CDI_INFO_ERRCNT ) *(int *)buffers = pReq->pDev->numberDeviceError;
  }
  else for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belonging to this group */
  {
    if ( index == CDI_INFO_NDEVS )
    {
      *(int *)buffers = ((CdiRequestInfoBlk *)pReq->groupReq)->NgroupsReqs;
      break;
    }
    else if ( index == CDI_INFO_ERRCNT )
    {
      *(int *)buffers = ((CdiRequestInfoBlk *)pReq->groupReq)->NgrpErrors;
      break;
    }
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if (cdiDebugLevel) dbglog("getGroupRequestInfo-> device group index: %d\n",pLine->pReqList[i].NgroupsReqs);
      offset = &pLine->pReqList[i] == pReq ? 0 :  pLine->pReqList[i].NgroupsReqs;
      if ( index == CDI_INFO_ERRORS )
	  {
		if ( buffers ) ((int *)buffers)[ offset ] = pLine->pReqList[i].errorCode;
		else if ( pLine->pReqList[i].errorCode ) return pLine->pReqList[i].errorCode;
	  }
      else if ( index == CDI_INFO_NAMES ) strncpy( ((NAME32 *)buffers)[offset].name, pLine->pReqList[i].pDev->devName, CDI_DEVICE_NAME_SIZE);
      else if ( index == CDI_INFO_STATUS ) ((int *)buffers)[ offset ] = pLine->pReqList[i].active;
    }
  }
  return 0;
}

/************************************/
CDI_EXPORT int cdi_GetRequestErrors(int requestIndex, int *errorArray)
{
  return getGroupRequestInfo(requestIndex, (void *)errorArray, CDI_INFO_ERRORS);
}
/************************************/
CDI_EXPORT int cdi_GetRequestDeviceNames(int requestIndex, NAME32 *namebuffer)
{
  return getGroupRequestInfo(requestIndex, (void *)namebuffer, CDI_INFO_NAMES);
}
CDI_EXPORT int cdi_GetRequestStat(int requestIndex, int *activeStatArray)
{
  return getGroupRequestInfo(requestIndex, (void *)activeStatArray, CDI_INFO_STATUS);
}
CDI_EXPORT int cdi_GetRequestNumberDevices(int requestIndex, int *numberOfDevices)
{
  return getGroupRequestInfo(requestIndex, (void *)numberOfDevices, CDI_INFO_NDEVS);
}

/************************************/
/* first call cdi_Request for all(N), then call this rountine N times to group them */
CDI_EXPORT int loadGroupRequests(CdiRequestInfoBlk *pMainReq, CdiRequestInfoBlk *pFollowerReq)
{ int ret=0;
  if ( pMainReq == NULL || pFollowerReq == NULL ) return -invalid_index;	/* no such a request */
  if ( pFollowerReq->callback != pMainReq->callback ) return -ambiguous;	/* not same for sunc or async */
  if ( pFollowerReq->callback && pFollowerReq->pollFlag != pMainReq->pollFlag ) return -ambiguous;	/* not same for polling */
  cdiWaitSemaphore(&pMainReq->exec_SemId,0);	// added to 2010-1
  cdiWaitSemaphore(&pFollowerReq->exec_SemId,0);
  if ( pMainReq->NgroupsReqs == 0 ) /* first time call  */
  {
    pMainReq->groupReq = pMainReq;
    pMainReq->NgroupsReqs++;
  }
  /* check if already signed  */
  if ( pFollowerReq->groupReq != NULL )	ret = (pFollowerReq->groupReq != pMainReq->groupReq) ? -illegal_equipment_name : 0;
  else
  {
	pFollowerReq->groupReq = pMainReq->groupReq;  /* add this new element to group  */
	pFollowerReq->NgroupsReqs = pMainReq->NgroupsReqs;  /* index number for this group request: 1, 2, 3....  pMainReq->NgroupsReqs */
	pMainReq->NgroupsReqs++;
  }
  cdiPostSemaphore(&pFollowerReq->exec_SemId);
  cdiPostSemaphore(&pMainReq->exec_SemId);	// added to 2010-1
  return ret;
}

/************************************/
CDI_EXPORT int cdi_loadGroupRequests(int requestIndexMain, int requestIndexFollower)
{ int retCode;
  CdiRequestInfoBlk *pMainReq, *pFollowerReq;

  if ( (retCode = get_RequestBufferPointer(requestIndexMain, &pMainReq)) < 0 ) return retCode;
  if ( (retCode = get_RequestBufferPointer(requestIndexFollower, &pFollowerReq)) < 0 ) return retCode;
  return loadGroupRequests(pMainReq, pFollowerReq);
}

/************************************/
CDI_EXPORT int postBackTaskSemaphore(CdiRequestInfoBlk *pReq)
{ int i, j, ret=0, tmpret;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL )
  {
    ret = cdiPostSemaphore(&(((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->back_SemId));
    if ( ret )
      if (cdiDebugLevel) dbglog("background SemId : cdiPostSemaphore error");
    return ret;
  }

  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      /* start background task for this cdi line  */
      tmpret = cdiPostSemaphore(&(((CdiLineInfoBlk *)pLine->pReqList[i].pDev->pcdiLine)->back_SemId));
      if (tmpret)
      {
        ret = tmpret;
        if (cdiDebugLevel) dbglog("background SemId : cdiPostSemaphore error");
      }
      break;
    }
  }
  return ret;
}


/************************************/
CDI_EXPORT int cdiStopCancel(CdiRequestInfoBlk *pReq, int cancelFlag)
{ int tmp;

  cdiWaitSemaphore(&pReq->exec_SemId,0);
//  if ( cancelFlag && (pReq->needsCalibration & FREE_ONCLOSE) )	// 2010-4
//  {
//	if ( pReq->groupReq == NULL || pReq->groupReq == pReq ) if (pReq->perror != NULL) free(pReq->perror);
//    pReq->perror = NULL;
//  }
  if ( pReq->active == 0 || pReq->active > finishRequest)
  {
    if ( busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].cancelRequestFcn != NULL )
    busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].cancelRequestFcn(pReq, cancelIndex);

    cdiPostSemaphore(&pReq->exec_SemId);
    return 0;	/* nothing to do */
  }
  if (pReq->active < finishRequest && busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].timeoutFcn != NULL)
  {
    busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].timeoutFcn(pReq, 1, cancelIndex);
    cdiAssertHeapOkay("cdiStopCancel (timeoutFcn)",pReq->pDev->devName,pReq->accessFlag);
  }
  if ( pReq->callback == NULL && pReq->active > idleState && pReq->active < finishRequest )
  {
    pReq->active = finishRequest;		/* make it finish!!!!! */
    pReq->errorCode = device_busy;
    if ( pReq->perror ) *(pReq->perror) = device_busy;
    if ( pReq->groupReq == NULL ) cdiPostSemaphore(&pReq->sync_SemId);
    else if ( checkGroupReturnFlags(pReq, cancelIndex) ) cdiPostSemaphore(&((CdiRequestInfoBlk *)pReq->groupReq)->sync_SemId);
  }
  else pReq->userCallDown = 1;
  tmp = pReq->active == idleState ? 1 : 0;
  /* avoid user does: first cancel, then stop	*/
  if ( pReq->cancelFlag != 2 ) pReq->cancelFlag = cancelFlag ? 2 : 1;	/* stop: 1; cancel: 2 */
  pReq->active = finishRequest;

  /* activate background task */
  if ( tmp && postBackTaskSemaphore(pReq) ) 
  {
    if (cdiDebugLevel) dbglog("background SemId : cdiPostSemaphore error");
  }
  cdiPostSemaphore(&pReq->exec_SemId);

  if ( busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].cancelRequestFcn != NULL )
	busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].cancelRequestFcn(pReq, cancelIndex);

  return 0;
}

/************************************/
/* do rest of the group items	*/
CDI_EXPORT void cdiStopCancelGroup(CdiRequestInfoBlk *pReq, int cancelFlag)
{ int i, j;
  CdiLineInfoBlk *pLine;

  if ( pReq->groupReq == NULL ) return;
  if ( pReq->groupReq == pReq && pReq->pBlockBuffer ) cdiWaitSemaphore(&pReq->exec_SemId,0);	// added to 2010-5
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      if ( (CdiRequestInfoBlk *)&(pLine->pReqList[i]) == pReq ) continue;
      cdiStopCancel(&(pLine->pReqList[i]), cancelFlag);
    }
  }
  if ( pReq->groupReq == pReq && pReq->pBlockBuffer ) cdiPostSemaphore(&pReq->exec_SemId);	// added to 2010-5
}
void assignClbr(double dval,BYTE *data,int dataFormat)
{
  switch (dataFormat)
  {
    default: 
    case cdiChar: *(data) = (BYTE)dval; break;
    case cdiByte: *(data) = (BYTE)dval; break;
    case cdiShort: *((UINT16 *)data) = (UINT16)dval; break;
    case cdiInteger: *((UINT32 *)data) = (UINT32)dval; break;
    case cdiFloat: *((float *)data) = (float)dval; break;
    case cdiDouble: *((double *)data) = dval; break;
  }
}
void assignMsg(char *msg,BYTE *data,int dataFormat)
{
  switch (dataFormat)
  {
    default:
      return;
    case cdiChar32:
      /* must ensure that there's room to put this ! */
      strncpy((char *)data,msg,32);
      return;
  }
}
char *applyMsg(CdiRule *rule,BYTE *data,int dataFormat)
{
  CdiRule *r = NULL, *msg;
  int hasMessage = FALSE, isSigned = FALSE;
  int mval = 0;

  if (rule != NULL) isSigned = rule->isSigned;
  switch (dataFormat)
  {
    default: 
    case cdiChar: mval = *(data); break;
    case cdiByte: mval = *(data); break;
    case cdiShort: mval = isSigned ? *(short *)(data) : *(UINT16 *)(data); break;
    case cdiInteger: mval = isSigned ?  *(SINT32 *)(data) : *(UINT32 *)(data); break;
  }
  for (r=rule; r != NULL && !hasMessage; r=r->nxt)
  {
    switch (r->operation)
    {
      case cdiRuleSub: 
        mval -= (int)r->operand; 
        break;
      case cdiRuleAdd: 
        mval += (int)r->operand; 
        break;
      case cdiRuleMult: 
        mval *= (int)r->operand; 
        break;
      case cdiRuleDiv: 
        if (r->operand != 0) mval /= (int)r->operand; 
        break;
      case cdiRulePwr: 
        mval = (int)pow((double)mval,r->operand);
        break;
      case cdiRuleMod:
        mval = (mval)%((int)r->operand);
        break;
      case cdiRuleShiftRight:
        mval = (mval)>>((int)r->operand);
        break;
      case cdiRuleShiftLeft:
        mval = (mval)<<((int)r->operand);
        break;
      case cdiRuleXor:
        mval = (mval)^((int)r->operand);
        break;
      case cdiRuleNot:
        mval = ~(mval);
        break;
      case cdiRuleAnd:
        mval = (mval)&((int)r->operand);
        break;
      case cdiRuleExp: 
        mval = (int)pow(r->operand,(double)mval);
        break;
      case cdiRuleFcn:
        if (r->fcn != NULL) mval = (int)r->fcn((double)mval);
        break;
      case cdiRuleMessage:  
        hasMessage = TRUE; /* will also break the rules chain */
        msg = r;
        break;
      case cdiRuleSigned:
        isSigned = TRUE;
        break;
      default:
        break;
    }
  }
  if (!hasMessage) return NULL;
  return mval == msg->msgOnValue ? msg->msgOn : msg->msgOff;
}
double applyClbr(CdiRule *rule,BYTE *data,int dataFormat)
{
  CdiRule *r = NULL;
  double dval;
  int isSigned = FALSE;
  /* first rule sets the signed/unsigned conversion */
  if (rule != NULL) isSigned = rule->isSigned;
  switch (dataFormat)
  {
    default: 
    case cdiChar: dval = *(data); break;
    case cdiByte: dval = *(data); break;
    case cdiShort: dval = isSigned ? *(short *)(data) : *(UINT16 *)(data); break;
    case cdiInteger: dval = isSigned ? *(SINT32 *)(data) : *(UINT32 *)(data); break;
    case cdiFloat: dval = *(float *)(data); break;
    case cdiDouble: dval = *(double *)(data); break;
  }
  for (r=rule; r != NULL; r=r->nxt)
  {
    switch (r->operation)
    {
      case cdiRuleSub: 
        dval -= r->operand; 
        break;
      case cdiRuleAdd: 
        dval += r->operand; 
        break;
      case cdiRuleMult: 
        dval *= r->operand; 
        break;
      case cdiRuleDiv: 
        if (r->operand != 0) dval /= r->operand; 
        break;
      case cdiRulePwr: 
        dval = pow(dval,r->operand);
        break;
      case cdiRuleLog:
        /* no operand here ! */
        if (dval != 0) dval = log10(dval);
        break;
      case cdiRuleMod:
        dval = ((UINT32)dval)%((UINT32)r->operand);
        break;
      case cdiRuleShiftRight:
        dval = ((UINT32)dval)>>((UINT32)r->operand);
        break;
      case cdiRuleShiftLeft:
        dval = ((UINT32)dval)<<((UINT32)r->operand);
        break;
      case cdiRuleXor:
        dval = ((UINT32)dval)^((UINT32)r->operand);
        break;
      case cdiRuleNot:
        dval = ~((UINT32)dval);
        break;
      case cdiRuleAnd:
        dval = ((UINT32)dval)&((UINT32)r->operand);
        break;
      case cdiRuleOr:
        dval = ((UINT32)dval)|((UINT32)r->operand);
        break;
      case cdiRuleExp: 
        dval = pow(r->operand,dval);
        break;
      case cdiRuleFcn:
        if (r->fcn != NULL) dval = r->fcn(dval);
        break;
      case cdiRuleFix:
        return dval = r->operand;
      default: break;
    }
  }
  return dval;
}
static CdiDeviceInfoBlk *getDevPtr(CdiRequestInfoBlk *pReq,int groupIndex)
{ 
  int i, j, k = 0;
  CdiLineInfoBlk *pLine;

  if (pReq->groupReq == NULL) return pReq->pDev;
  for (j=0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for (i=0; i < pLine->numberActiveRequest; i++ )
    {
      if (pLine->pReqList[i].groupReq != pReq->groupReq) continue;
      if (k == groupIndex) return pLine->pReqList[i].pDev;
      k++;
    }
  }
  return NULL;
}
int clbrRequest(CdiRequestInfoBlk *pReq,int applyRules)
{
  int i, flen, tlen, dlen;
  CdiDeviceInfoBlk *pDev;
  int fromFormat, toFormat;
  BYTE *fromData, *toData;
  CdiRule *rule = NULL;
  double dval;
  char *msg;

  if ((fromFormat=pReq->inType) < 0) return -illegal_format;
  flen = getTypeLength((short)fromFormat);
  dlen = pReq->inLength/flen;
  if (flen <= 0) return -illegal_format;
  toData = fromData = pReq->pInData;
  pDev = pReq->pDev;
  toFormat = getCdiBaseFormat(pReq->inFormatCode);
  if (toFormat < 0) return -illegal_format;
  if ((tlen=getTypeLength((short)toFormat)) <= 0) return -illegal_format;
  if (!applyRules && toFormat == fromFormat) return 0; /* nothing to do */
  rule = applyRules ? pDev->clbrRules : NULL;
  for (i=dlen-1; i>=0; i--)
  { /* start from the end and work backwards for in situ assignments */
    if (toFormat == cdiChar32)
    {
      if ((msg=applyMsg(rule,fromData+i*flen,fromFormat)) != NULL)
      {
        assignMsg(msg,toData+i*tlen,toFormat);
      }
    }
    else
    {
      dval = applyClbr(rule,fromData+i*flen,fromFormat);
      assignClbr(dval,toData+i*tlen,toFormat);
    }
  }
  return 0;
}
CDI_EXPORT int cdi_CalibrateRequest(int requestIndex,int applyRules)
{
  int cc, j, k;
  CdiRequestInfoBlk *pReq;
  CdiLineInfoBlk *pLine;

  if ((cc=get_RequestBufferPointer(requestIndex, &pReq)) < 0) return cc;

  if (pReq->groupReq == NULL) return clbrRequest(pReq,applyRules);
  for (j=0; j<numConnectedHandles; j++)	/* search all devices belonging to this group */
  {
    if (busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for (k=0; k<pLine->numberActiveRequest; k++)
    {
      if (pLine->pReqList[k].groupReq != pReq->groupReq) continue;
      if ((cc=clbrRequest(&pLine->pReqList[k],applyRules)) < 0) return cc;
    }
  }      
  return 0;
}

/************************************/
/* cancel Async call */
/************************************/
CDI_EXPORT int cdi_RequestTerminated(int requestIndex)
{
  CdiRequestInfoBlk *pReq;
  int retCode;
  if ((retCode = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return FALSE;
  if ( !pReq->active ) return TRUE; 
  return FALSE;
}
CDI_EXPORT int cdi_CancelRequest(int requestIndex)
{
  CdiRequestInfoBlk *pReq;
  int retCode;

  if ((retCode = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return retCode;
  if (pReq->groupReq && pReq->groupReq != pReq) return -invalid_index;
  if (isPastDeviceFilter(pReq->pDev->devName))
  {
    ttyoutput("cdi_CancelRequest %.64s", pReq->pDev);
  }
  if (pReq->callback != NULL)
  {
	  cdilog("cdi cancel link : %s\n",pReq->pDev->devName);
  }
  cdiStopCancelGroup(pReq, 1);
  cdiStopCancel(pReq, 1);
//  if (pReq->needsCalibration & FREE_ONCLOSE)
//  { /* remove the error buffer */
//	cdiWaitSemaphore(&pReq->exec_SemId,0);	// added to 2010-1
//    if (pReq->perror != NULL) free(pReq->perror);
//    pReq->perror = NULL;
//	cdiPostSemaphore(&pReq->exec_SemId);	// added to 2010-1
//  }
  return 0;
}

/************************************/
CDI_EXPORT void initStartRequest(CdiRequestInfoBlk *pReq, int flag)
{
  cdiWaitSemaphore(&pReq->exec_SemId,0);	// added to 2010-1
  if ( flag & 0x1 )
  {
    pReq->dataReturnCondition = 0;
    pReq->errorCode = 0;
    pReq->deviceError = 0;
    pReq->cancelFlag = 0;
    pReq->requestCounter = 0;
    if ( pReq->perror ) *(pReq->perror) = 0;

    pReq->asynCallFlag = 0;
    pReq->userCallDown = 0;
    pReq->NgroupRets = 0;
    pReq->NgrpErrors = 0;
    pReq->groupSyncStart = 1;
    pReq->devReadyTime = 0;
  }
  if ( flag & 0x2 )
  {
    pReq->requestTime = getTimeMilliseconds();
    pReq->responseTime = pReq->requestTime - 1;
    pReq->active = transRequest;	/* mark start flag! */
  }
  cdiPostSemaphore(&pReq->exec_SemId);	// added to 2010-1
}	

/************************************/
CDI_EXPORT void doStartRequest(CdiRequestInfoBlk *pReq)
{ int i, j;
  CdiLineInfoBlk *pLine;
  CdiRequestInfoBlk *pR;

  if ( pReq->groupReq == NULL )
  {
    initStartRequest(pReq, 0x3);
    return;
  }
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      pR = (CdiRequestInfoBlk *)&(pLine->pReqList[i]);
      initStartRequest(pR, 0x1);
    }
  }
  for ( j = 0; j < numConnectedHandles; j++ )	/* search all devices belong to this group */
  {
    if ( busHandle[j] == NULL ) continue;
    pLine = busHandle[j];
    for ( i = 0; i < pLine->numberActiveRequest; i++ )
    {
      if ( pLine->pReqList[i].groupReq != pReq->groupReq ) continue;
      pR = (CdiRequestInfoBlk *)&(pLine->pReqList[i]);
      if ( pR == pReq ) continue;
      initStartRequest(pR, 0x2);
      /* Let background task start the rest of the group call */
      /*      if ( pReq->callback == NULL ) cdiExecRequest(pR, execIndex); */
    }
  }
  initStartRequest(pReq, 0x2);
}

/************************************/
CDI_EXPORT int cdi_StartRequest(int requestIndex)
{ CdiRequestInfoBlk *pReq;
  int retCode;

  if ( (retCode = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return retCode;
  if ( pReq->groupReq && pReq->groupReq != pReq ) return -invalid_index;
  if (cdiDebugLevel) dbglog("cdi_StartRequest, preq: %x,  call back == %x...\n", pReq, pReq->callback);
  if ( pReq->active == 0 ) return -database_not_loaded;	/* should call cdi_Request first or lock it... */
  if ( pReq->active != idleState ) return -operation_busy;	/* not finish yet */

//  cdiWaitSemaphore(&pReq->exec_SemId,0);	// moved to initStartRequest call 2010-1
  doStartRequest(pReq);
//  cdiPostSemaphore(&pReq->exec_SemId);

  if ( pReq->callback == NULL ) /* sync call */
  {
    while ( cdiWaitSemaphore(&pReq->sync_SemId,TRUE) == 0 ); /* clear semaphore, send request */
    cdiWaitSemaphore(&pReq->exec_SemId,0);
    if ( busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].deviceReadyFcn != NULL )
    {
      busList[((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->busIndex].deviceReadyFcn(pReq, execIndex);
      cdiAssertHeapOkay("cdi_StartRequest (cdiBusIsReady)",pReq->pDev->devName,pReq->accessFlag);
    }
    cdiExecRequest(pReq, execIndex);
    cdiSetAlarms(pReq);
    cdiPostSemaphore(&pReq->exec_SemId);
  }
  /* start background task */
  postBackTaskSemaphore(pReq);

  if ( pReq->callback == NULL ) /* sync call return error number */
  {
    if ( cdiWaitSemaphore(&pReq->sync_SemId,0) ) 
    {
      if (cdiDebugLevel) dbglog("wait on sync semaphore ERROR! cdiLine: %d, requestNumber: %d", ((CdiLineInfoBlk *)pReq->pDev->pcdiLine)->cdiLine, pReq->reqIndex);
    }
    cdiWaitSemaphore(&pReq->exec_SemId,0);
    pReq->userCallDown++;
    pReq->devReadyTime = 0;
    retCode = setGroupReturnFlags(pReq, 0x5);
    if ( retCode == 0 ) retCode = -pReq->errorCode; /* if group call, return (Number of errors), else return errorCode */
    cdiPostSemaphore(&pReq->exec_SemId);
    return retCode;
  }
  /* asyn return with requestIndex */
  return requestIndex; /* reqLineNumber; */

}
/************************************/
CDI_EXPORT int cdi_StopRequest(int requestIndex)
{ CdiRequestInfoBlk *pReq;
  int retCode;

  if ( (retCode = get_RequestBufferPointer(requestIndex, &pReq)) < 0 ) return retCode;
  if ( pReq->groupReq && pReq->groupReq != pReq ) return -invalid_index;
  cdiStopCancelGroup(pReq, 0);
  return cdiStopCancel(pReq, 0);
}

/************************************/
CDI_EXPORT int cdi_getDeviceNumberOfErrors(char *devName)
{ int i, j;
  if ( findDevice(devName, &i, &j) < 0 ) return -device_not_connected;
  return busHandle[i]->pDevList[j].numberDeviceError;
}

/************************************/
CDI_EXPORT char *cdi_getBusNameForCdiLine(int cdiLine)
{
  if ( busHandle[cdiLine] == NULL ) return NULL;
  return busList[ busHandle[cdiLine]->busIndex ].busName;
}

/************************************/
CDI_EXPORT char *cdi_getBusNameForDevice(char *devName)
{ 
  int i, j;

  if ( findDevice(devName, &i, &j) < 0 ) return NULL;
  return busList[ busHandle[i]->busIndex ].busName;
}

/************************************/
CDI_EXPORT char *cdi_getFieldBusNameForDevice(char *devName)
{ 
  int i, j;

  if ( findDevice(devName, &i, &j) < 0 ) return NULL;
  return cdiGetFieldBusName(busList[busHandle[i]->busIndex].busName,busHandle[i]->busLine);
}

/************************************/
CDI_EXPORT int cdi_getBusNames(NAME16 *namebuffer,int namesize)
{ 
  int i;
  for (i=0; i<numBusList && i<namesize; i++) strncpy(namebuffer[i].name,busList[i].busName,CDI_BUSNAME_SIZE);
  return numBusList;
}

/************************************/
CDI_EXPORT DADDRESS *cdi_getBusAddressForDevice(char *devName)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;
  static DADDRESS da;

  memset(&da,0,sizeof(da));
  if ( findDevice(devName, &i, &j) < 0 ) return &da;
  pDev = &busHandle[i]->pDevList[j];
  da.i1val = ((CdiLineInfoBlk *)pDev->pcdiLine)->cdiLine;
  da.i2val = pDev->cdiAddr[0];
  da.i3val = pDev->cdiAddr[1];
  da.i4val = pDev->cdiIndex;
  return &da;
}
CDI_EXPORT int cdi_getBusAddressParametersForDevice(char *devName,int *addrlist,char *addrmap)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  if (addrlist != NULL)
  {
    memcpy(addrlist,pDev->cdiAddr,CDI_ADDR_SIZE*sizeof(int));
  }
  if (addrmap != NULL)
  {
    memset(addrmap,0,CDI_DEVICE_LONGNAME_SIZE);
    strncpy(addrmap,pDev->lngName,CDI_DEVICE_LONGNAME_SIZE);
  }
  return 0;
}
CDI_EXPORT int cdi_getDeviceParametersForDevice(char *devName,int *paramlist)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  if (paramlist != NULL && pDev->devParameters != NULL)
  {
    memcpy(paramlist,pDev->devParameters,NUMBER_EXEC_PARAM*sizeof(int));
  }
  return 0;
}

/************************************/
CDI_EXPORT int cdi_getDescriptionForDevice(char *devName,char *devDesc)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devDesc == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  strncpy(devDesc,pDev->description,CDI_DEVICE_DESC_SIZE);
  return 0;
}
CDI_EXPORT int cdi_getUnitsForDevice(char *devName,char *devUnits)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devUnits == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  strncpy(devUnits,pDev->units,CDI_UNITS_SIZE);
  return 0;
}
CDI_EXPORT int cdi_getPrecisionForDevice(char *devName,char *devPrec)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devPrec == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  strncpy(devPrec,pDev->precision,CDI_UNITS_SIZE);
  return 0;
}
CDI_EXPORT int cdi_getRangeForDevice(char *devName,float *devRange)
{
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devRange == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  devRange[0] = pDev->minimum;
  devRange[1] = pDev->maximum;
  return 0;
}
CDI_EXPORT int cdi_getFormatForDevice(char *devName,int *devFmt)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devFmt == NULL) return argument_list_error;
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];

  *devFmt = pDev->dataFormat; /* the cdi format */
  return 0;
}
CDI_EXPORT int cdi_getOutputSizeForDevice(char *devName,int *devSiz)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j,siz,fmt,fsiz;

  if (devName == NULL || devSiz == NULL) return argument_list_error;
  *devSiz = 0; /* flag as 0 in case of error below */
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  siz = pDev->dataOuputSize; /* obtained from LIMIT column */
  fmt = getTineFormat((short)pDev->dataFormat);
  if ((fsiz=fmtsizeof(fmt)) < 1) return 0;
  if (siz == 0) siz = fsiz; /* not given explicitly */
  *devSiz = siz/fsiz;
  return 0;
}
CDI_EXPORT int cdi_getLimitSizeForDevice(char *devName,int *devOutputSiz,int *devInputSiz)
{ 
  CdiDeviceInfoBlk *pDev;
  int i,j;

  if (devName == NULL || devOutputSiz == NULL || devInputSiz == NULL) return argument_list_error;
  *devOutputSiz = 0; /* flag as 0 in case of error below */
  *devInputSiz = 0; /* flag as 0 in case of error below */
  if ( findDevice(devName, &i, &j) < 0 ) return device_not_connected;
  pDev = &busHandle[i]->pDevList[j];
  *devOutputSiz = pDev->dataOuputSize + pDev->dataOuput2ndSize; /* obtained from LIMIT column */
  *devInputSiz = pDev->dataInputSize; /* obtained from LIMIT column */
  return 0;
}
int gNumBitfieldDevices = 0;
int getNumFieldsInBitfield(DBITFIELD *db)
{
  struct bitfieldFormat *bf;
  int n = 0;
  if (db == NULL) return 0;
  for (bf=db->bfs; bf != NULL; bf=bf->nxt) n++;
  return n;
}
/************************************/
CDI_EXPORT int getCDIDeviceNames(NAME32 *namebuffer,int namesize,int cdiLine)
{ 
  static int discovered = FALSE;
  int i, j, n=0, istart, istop;
  DBITFIELD *db;
  if ( cdiLine >= 0 )
  {
    if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
    istart = cdiLine;
    istop = cdiLine + 1;
  }
  else
  {
    istart = 0;
    istop = numConnectedHandles;
  }
  for (i=istart; i< istop; i++ ) /* search through all bus lines */
  {
    if ( busHandle[i] == NULL ) continue;
    for ( j=0; j<busHandle[i]->numberDevices; j++ )
    {
      if ( busHandle[i]->pDevList[j].devName[0] == 0 ) continue;
      if (IsDiscoveryDevice(busHandle[i]->pDevList[j].devName))
      { /* a 'discovery' device instance (OPCua?) */
        char ddev[64];
        strncpy(ddev, &busHandle[i]->pDevList[j].devName[9], 32); ddev[32] = 0;
        if (strlen(ddev) > 0 && !discovered)
        {
          char *c = &ddev[1]; /* skip 1st ':' */
          int rstart = atoi(c), rstop = rstart+1;
          if ((c=strchr(c, ':')) != NULL) rstop = atoi(++c); else rstop = rstart, rstart = 0;
          if (rstop > rstart)
          { /* prepare place holders */
            if (!discovered) n += rstop - rstart;
          }
          if (busList[busHandle[i]->busIndex].execBusFcn != NULL)
          {
            cdiExecLinkEx(busHandle[i]->pDevList[j].devName, "RECV", NULL, NULL, CA_READ, 1000);
            discovered = TRUE;
          }
        }
      }
      if ( namebuffer != NULL )
      {
        if ( n < namesize )	strncpy(namebuffer[n].name, busHandle[i]->pDevList[j].devName, CDI_DEVICE_NAME_SIZE);
      }
      if ( !strcmp(busHandle[i]->pDevList[j].precision,"BITFIELD")) 
      {
        if ((db=findBitField(BITFIELD_LCL,busHandle[i]->pDevList[j].units)) != NULL)
        {
          gNumBitfieldDevices += getNumFieldsInBitfield(db);
        }
      }
      n++;
    }
  }
  return n;
}

/************************************/
CDI_EXPORT int cdi_getDeviceNames(NAME32 *namebuffer,int namesize)
{
  return getCDIDeviceNames(namebuffer,namesize,-1);
}

/************************************/
CDI_EXPORT int cdi_getDeviceNamesByLine(NAME32 *namebuffer, int namesize, int cdiLine)
{
  return getCDIDeviceNames(namebuffer,namesize,cdiLine);
}
/************************************/
CDI_EXPORT int cdi_GetNumberDevicesByLine(int cdiLine)
{
  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL ) return 0;
  return busHandle[cdiLine]->numberDevices;
}
/************************************/
/* Register Parameters with Name */
/* return index number */
/************************************/
CDI_EXPORT int cdi_RegisterParameters(char *paramName, int *parameters)
{ 
  int i;

  if (paramList == NULL) return -database_not_loaded;
  if (paramName == NULL || parameters == NULL || strlen(paramName) >= CDI_PARAM_NAME_SIZE ) return -argument_list_error;
  for (i=0; i<paramListSize; i++)
  {
    if ( strlen(paramList[i].paramName) == 0 ) break; /* empty slot */
    if ( strncmp(paramList[i].paramName, paramName, CDI_PARAM_NAME_SIZE) == 0 ) break; /* previous registered */
  }
  if (i == paramListSize) return -resources_exhausted;
  strncpy(paramList[i].paramName, paramName, CDI_PARAM_NAME_SIZE);
  memcpy(paramList[i].devParameters, parameters, NUMBER_EXEC_PARAM * sizeof(int));
  return i;
}

/****************************************/
/* get Registered Parameters with Name */
/****************************************/
CDI_EXPORT int *cdi_GetParameters(char *paramName)
{ 
  int i;

  if (paramList == NULL) return NULL;
  if (paramName == NULL || strlen(paramName) >= CDI_PARAM_NAME_SIZE ) return NULL;
  for (i=0; i<paramListSize; i++)
  {
    if ( strlen(paramList[i].paramName) == 0 ) break; /* empty slot */
    if ( strncmp(paramList[i].paramName, paramName, CDI_PARAM_NAME_SIZE) == 0 ) return paramList[i].devParameters; /* previous registered */
  }
  return NULL;
}

/************************************/
/* Register Devices */
/************************************/
int maxCdiRedirectedDevices = MAX_REDIRECTED_DEVICES;
int gCdiRedirectedDevicesHashSize = 0;
CDI_EXPORT int cdi_RegisterRedirectedDevice(char *devName, int devNumber, char *redirectedServerName)
{
  int idx;
  cdiRedirXRefStruct *xref;
  static int done = 0;

  if (!done)
  {
    if (maxCdiRedirectedDevices < CDI_PAYOFF_HASHSIZE) gCdiRedirectedDevicesHashSize = CDI_TINY_HASHSIZE;
    else if (maxCdiRedirectedDevices < CDI_SMALL_HASHSIZE) gCdiRedirectedDevicesHashSize = CDI_SMALL_HASHSIZE;
    else if (maxCdiRedirectedDevices < CDI_MEDIUM_HASHSIZE) gCdiRedirectedDevicesHashSize = CDI_MEDIUM_HASHSIZE;
    else gCdiRedirectedDevicesHashSize = CDI_LARGE_HASHSIZE;
    if ((cdiRedirXref=(cdiRedirXRefStruct **)calloc(gCdiRedirectedDevicesHashSize,sizeof(cdiRedirXRefStruct *))) == NULL)
    {
      cdiDebugLevel = 1;
      cdilog("cannot allocate redirected cross-reference buffer");
      exit(1);
    }
    done = TRUE;
  }
  idx = (int)(ElfHash((unsigned char *)devName) % gCdiRedirectedDevicesHashSize);
  for (xref=cdiRedirXref[idx]; xref && strncmp(devName,cdiRedirXref[idx]->devName,CDI_DEVICE_NAME_SIZE); xref=xref->nxt);
  if (xref == NULL) /* it's a new one*/
  {
    if ((xref=(cdiRedirXRefStruct *)SystemCalloc(1,sizeof(cdiRedirXRefStruct))) == NULL) 
    {
      cdilog("CDI: cannot allocate hash table entry !");
      return -out_of_server_memory;
    }
    strncpy(xref->devName,devName,CDI_DEVICE_NAME_SIZE);
    strncpy(xref->devRedirectedServerName,redirectedServerName,64);
    xref->devNumber = devNumber;
    xref->nxt = cdiRedirXref[idx];
    cdiRedirXref[idx] = xref;
    nCdiRedirectedDevices++;
  }
  return 0;
}
CDI_EXPORT int cdi_getRedirectedDeviceNames(NAME32 *namebuffer,int namesize)
{
  int i,n=0;
  cdiRedirXRefStruct *xref;
  if (namebuffer == NULL) return nCdiRedirectedDevices;
  for (i=0; i<gCdiRedirectedDevicesHashSize; i++)
  {
    for (xref=cdiRedirXref[i]; xref; xref=xref->nxt)
    {
      if (n >= namesize) return n;
      strncpy(namebuffer[n++].name,xref->devName,CDI_DEVICE_NAME_SIZE);
    }
  }
  return n;
}
int getNextFreeIndex(int cdiLine)
{
  int idx = 0;
    
  if (cdiLine < 0 || cdiLine >= NUMBER_LINES) return -illegal_line;
  idx = CdiNumIndiciesPerLine[cdiLine];
  if (idx >= busHandle[cdiLine]->numberDevices)
  {
	  return -resources_exhausted;
  }
  CdiNumIndiciesPerLine[cdiLine] = idx + 1;
  return idx;
}
int checkDevNumber(char *devName, int cdiLine, int cdiIndex, int devNumber)
{ 
  int i, j, idx;
  cdiHashXRefStruct *xref;

  if ( devNumber <= 0 )
  {
    cdilog("global device number %d invalid( >0 allowed)",devNumber);
    return invalid_index;
  }
  if (cdiIndex < 0 || cdiIndex >= busHandle[cdiLine]->numberDevices)
  {
    cdilog("cdi Index: %d invalid (0 to %d allowed)",cdiIndex,busHandle[cdiLine]->numberDevices-1);
    return invalid_index;
  }
  for (i=0; i<numConnectedHandles; i++)
  {
    if ( busHandle[i] == NULL ) continue;
    if (busHandle[i]->devHashSize == 0) /* no hash table ...*/
	  {
	    for ( j=0; j<busHandle[i]->numberDevices; j++)
	    {
        if (busHandle[i]->pDevList[j].devName[0] == 0) continue;
        if ( strncmp(devName, busHandle[i]->pDevList[j].devName, CDI_DEVICE_NAME_SIZE) == 0 )
        {
          cdilog("device Name %s (nr: %d) repeated at cdiLine %d cdiIndex %d (nr %d)",
            devName, devNumber, i, j, busHandle[i]->pDevList[j].devNumber);
          return invalid_index;
        }
        if ( busHandle[i]->pDevList[j].devNumber != devNumber ) continue;
      
		    cdilog("global device number %d invalid! In use by %s (cdiLine: %d cdiIndex: %d)", 
			    busHandle[i]->pDevList[j].devName, devNumber, i, j);
		    cdilog("try to use a device number in the range 0 to %d",cdiAddrDbSize+TEMPLATE_RECORD_GAP);
		    return invalid_index;
	    }
	  }
	  else
	  {
      idx = (int)(ElfHash((unsigned char *)devName) % busHandle[i]->devHashSize);
      for (xref=busHandle[i]->devHashList[idx]; xref; xref=xref->nxt)
      {
        j = xref->cdiIndex;
        if (busHandle[i]->pDevList[j].devName[0] == 0) continue;
        if (strncmp(devName, busHandle[i]->pDevList[j].devName,CDI_DEVICE_NAME_SIZE) == 0)
        {
	        cdilog("device Name %s (nr: %d) repeated at cdiLine %d cdiIndex %d (nr %d)",
	          devName, devNumber, i, j, busHandle[i]->pDevList[j].devNumber);
	        return invalid_index;
        }
        if (devNumber != xref->devNumber) continue;
        cdilog("global device number %d invalid! In use by %s (cdiLine: %d cdiIndex: %d)", 
        busHandle[i]->pDevList[j].devName, devNumber, i, j);
        cdilog("try to use a device number in the range 0 to %d",cdiAddrDbSize+TEMPLATE_RECORD_GAP);
        return invalid_index;
      }
	  }
  }
  return 0;
}

void parseDeviceDescription(char *devDesc,char *desc,char *prec,char *units,float *mn,float *mx)
{
  char *c0, *c1, *c, *p, *v, strBuffer[128], *tok = " \t";
  int len;

  desc[0] = 0;
  prec[0] = units[0] = 0;
  *mn = *mx = 0;
  strncpy(strBuffer,devDesc,96); strBuffer[96] = 0;
  strtrm(strBuffer);
  if ((c0=strchr(strBuffer,'[')) == NULL)
  { /* this one's easy: no display info to parse */
    strncpy(desc,strBuffer,64);
    return;
  }
  /* the [] are 'somewhere' in the description string (not necessarily at the beginning) */
  *c0++ = 0;
  strncpy(desc,strBuffer,64); /* first section (or empty string) */
  len = (int)strlen(desc);
  if (len > 64) len = 64;
  if ((c1=strchr(c0,']')) == NULL) 
  {
    cdilog("parseDeviceDescription : %s did not parse correctly!",devDesc);
    return;
  }
  *c1++ = 0;
  strncpy(strBuffer,c0,32);
  for (p=strtok(strBuffer,tok); p != NULL; p=strtok(NULL,tok))
  {
    if ((v=strchr(p,'=')) == NULL) continue;
    *v++ = 0;
    switch (tolower(*p))
    {
      case 'r': /* range */
        *mn = (float)strtod(v,NULL);
        if ((c=strchr(v,':')) != NULL)
        {
          *c++ = 0;
          *mx = (float)strtod(c,NULL);
        }
        break;
      case 'p': /* precision */
        strncpy(prec,v,CDI_UNITS_SIZE);
        break;
      case 'u': /* units */
        strncpy(units,v,CDI_UNITS_SIZE);
        break;
      default:
        break;
    }
  }
  /* second section */
  strncpy(strBuffer,c1,64); strBuffer[64] = 0;
  if ((c0=strchr(strBuffer,'[')) == NULL)
  { /* second set of display info : ignore it ! */
    strncat(desc,strBuffer,64-len);
    return;
  }
  *c0++ = 0;
  strncat(desc,strBuffer,64-len);
  len = (int)strlen(desc);
  if ((c1=strchr(c0,']')) == NULL) 
  {
    cdilog("parseDeviceDescription : %s did not parse correctly!",devDesc);
    return;
  }
  *c1++ = 0;
  strncat(desc,c1,64-len);
  return;
}

CDI_EXPORT int cdi_RegisterDevice(
  int cdiLine, char *devName, char *lngName, char *devDesc,
  int cdiIndex, int devNumber, int *cdiAddress, int cdiAddressLength
)
{ 
  CdiDeviceInfoBlk *pDev;
  cdiHashXRefStruct *xref;
  int i, idx;
  size_t slen;
  char devNrStr[32],tmp[1024];
  char desc[64], prec[CDI_UNITS_SIZE], units[CDI_UNITS_SIZE];
  float mx, mn;

  if (devName == NULL) return argument_list_error;

  if (cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL)
  {
    if (cdiDebugLevel) dbglog("line number %d not open",cdiLine);
    return illegal_line;
  }
  if (cdiIndex < 0) cdiIndex = getNextFreeIndex(cdiLine);
  if ( (idx=checkDevNumber(devName, cdiLine, cdiIndex, devNumber)) != 0 ) return idx;

  pDev = &(busHandle[cdiLine]->pDevList[cdiIndex]);
  strncpy(pDev->devName, devName, CDI_DEVICE_NAME_SIZE);
  if ( strlen(devName) > CDI_DEVICE_NAME_SIZE )
  {
    cdilog("truncating device name %s to %d characters !",devName,CDI_DEVICE_NAME_SIZE);
  }

  if (lngName == NULL) lngName = "";
  if ((slen=strlen(lngName)) > 0)
  { /* use extensible name */
    slen += (slen % 2 ? 1 : 2);
    pDev->extName = (char *)calloc(1, slen);
    if (pDev->extName != NULL)
    {
      strncpy(pDev->extName, lngName, slen);
    }
    if (slen > CDI_DEVICE_LONGNAME_SIZE)
      cdilog("long name %s length = %d -> use extensible name", lngName, slen);
  }
  strncpy(pDev->lngName, lngName, CDI_DEVICE_LONGNAME_SIZE);
  parseDeviceDescription(devDesc,desc,prec,units,&mn,&mx);
  strncpy(pDev->description, desc, CDI_DEVICE_DESC_SIZE);
  strncpy(pDev->precision,prec,CDI_UNITS_SIZE);
  strncpy(pDev->units,units,CDI_UNITS_SIZE);
  pDev->maximum = mx;
  pDev->minimum = mn;
  memcpy(pDev->cdiAddr,cdiAddress,cdiAddressLength*sizeof(int));
  pDev->cdiIndex = cdiIndex;
  pDev->devNumber = devNumber;
  pDev->pcdiLine = busHandle[cdiLine];
  pDev->numberDeviceError = 0;
  pDev->dataMaskStartIndex = pDev->dataMaskStopIndex = -1; /* not applicable */
  pDev->dataPatternStartIndex = pDev->dataPatternStopIndex = -1; /* not applicable */

  if (busHandle[cdiLine]->devHashList != NULL) /* there's a hash table */
  {
    idx = (int)(ElfHash((unsigned char *)busHandle[cdiLine]->pDevList[cdiIndex].devName) % busHandle[cdiLine]->devHashSize);
    for (xref=busHandle[cdiLine]->devHashList[idx]; 
         (xref != NULL) && (xref->cdiLine != cdiLine || xref->cdiIndex != cdiIndex); 
         xref=xref->nxt);
    if (xref == NULL) /* it's a new one*/
    {
      if ((xref=(cdiHashXRefStruct *)SystemCalloc(1,sizeof(cdiHashXRefStruct))) == NULL) 
      {
        cdilog("CDI: cannot allocate hash table entry !");
        return out_of_server_memory;
      }
      xref->cdiLine = cdiLine;
      xref->cdiIndex = cdiIndex;
      xref->devNumber = devNumber;
      xref->nxt = busHandle[cdiLine]->devHashList[idx];
      busHandle[cdiLine]->devHashList[idx] = xref;
      /* now add the '#<devnr>' representation as well */
      if ((xref=(cdiHashXRefStruct *)SystemCalloc(1,sizeof(cdiHashXRefStruct))) == NULL) 
      {
        cdilog("CDI: cannot allocate hash table entry !");
        return out_of_server_memory;
      }
      xref->cdiLine = cdiLine;
      xref->cdiIndex = cdiIndex;
      xref->devNumber = devNumber;
      sprintf(devNrStr,"#%d",devNumber);
      idx = (int)(ElfHash((unsigned char *)devNrStr) % busHandle[cdiLine]->devHashSize);
      xref->nxt = busHandle[cdiLine]->devHashList[idx];
      busHandle[cdiLine]->devHashList[idx] = xref;
    }
  }
  if (cdiDebugLevel) 
  {
    sprintf(tmp,"Register device name %.32s ( devNum %d, addr ",devName, devNumber);
    for (i=0; i<cdiAddressLength; i++) sprintf(&tmp[strlen(tmp)],"%d ",cdiAddress[i]);
    strcat(tmp,")");
    dbglog(tmp);
  }
  return 0;
}
CDI_EXPORT int cdi_assignDefaultDeviceParameters(char *devName,int *devParameters)
{
  int cc, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;
  char paramName[36];

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0)
  {
    if (cdiDebugLevel) dbglog("invalid device name: %s",devName);
    return device_not_connected;
  }
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];

  strncpy(paramName,devName,CDI_DEVICE_NAME_SIZE); paramName[CDI_DEVICE_NAME_SIZE] = 0;
  strcat(paramName,"default");
  if ((cc=cdi_RegisterParameters(paramName,devParameters)) < 0) return -cc;
  pDev->devParameters = cdi_GetParameters(paramName);
  return 0;
}
CDI_EXPORT int cdi_assignAllowedAccess(char *devName,int access)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->allowedAccess = access;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignAllowedAccess for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDefaultDataMaskAndPattern(char *devName,BYTE *dataMask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  if (dataMask != NULL) 
  {
    if ((pDev->dataMask=(BYTE *)calloc(8,1)) == NULL) ccerr(out_of_server_memory);
    memcpy(pDev->dataMask,dataMask,8);
  }
  if (maskRange != NULL)
  {
    pDev->dataMaskStartIndex = maskRange[0];
    pDev->dataMaskStopIndex = maskRange[1];
  }
  if (pattern != NULL)
  {
    if ((pDev->dataPattern=(BYTE *)calloc(CDI_PATTERN_FILTER_SIZE,1)) == NULL) ccerr(out_of_server_memory);
    memcpy(pDev->dataPattern,pattern,CDI_PATTERN_FILTER_SIZE);
    pDev->dataPatternTrigger = patternTrigger;
  }
  if (patternRange != NULL)
  {
    pDev->dataPatternStartIndex = patternRange[0];
    pDev->dataPatternStopIndex = patternRange[1] != -1 ? patternRange[1] : patternRange[0];
  }
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDefaultDataMaskAndPattern for %d -> %.32s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDataLimits(char *devName,int dataSizeOut,int dataSizeIn,int dataSize2ndOut)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->dataOuputSize = dataSizeOut;
  pDev->dataInputSize = dataSizeIn;
  pDev->dataOuput2ndSize = dataSize2ndOut;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDefaultLimits for %d -> %.32s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDefaultInput(char *devName,BYTE *pDataInput,int dataSize,int dataFormatIn)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  if (dataFormatIn < 0) ccerr(illegal_format);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->dataInput = pDataInput;
  pDev->dataInputSize = dataSize;
  pDev->dataFormat = dataFormatIn;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDefaultInput for %d -> %.32s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDataFormat(char *devName,int dataFormat)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->dataFormat = dataFormat;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDataFormat for %d -> %.32s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_assignDefaultInterval(char *devName,int interval)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->interval = interval;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDefaultInterval for %d -> %.32s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_getDefaultInterval(char *devName)
{
  int cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) return DEFAULT_POLLING_INTERVAL;
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  return pDev->interval;
}
CDI_EXPORT int cdi_assignDataCalibration(char *devName,CdiRule *clbrRules,CdiRule *rvrsRules,int dataBusFormat)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  pDev->dataFormat = dataBusFormat;
  pDev->clbrRules = clbrRules;
  pDev->rvrsRules = rvrsRules;
err:
  if (cc)
  {
    dbglog("CDI: cdi_assignDataFormat for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
CDI_EXPORT int cdi_eraseDefaultDataMaskAndPattern(char *devName,int erasedataMask, int erasepattern)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) ccerr(device_not_connected);
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  if ( erasedataMask && pDev->dataMask != NULL) 
  {
    free(pDev->dataMask);
    pDev->dataMask = NULL;
  }
  if (erasepattern && pDev->dataPattern != NULL)
  {
    free(pDev->dataPattern);
    pDev->dataPattern = NULL;
  }
err:
  if (cc)
  {
    dbglog("CDI: cdi_eraseDefaultDataMaskAndPattern for %d -> %s",devName,erlst[cc]);
  }
  return cc;
}
int getNumberLines(char *busName, int flag)
{
  int i,j,cnt;
  if ( flag == 1 )  /* check busName */
  {
    for (i=0; i<numBusList; i++) if (!strncmp(busName, busList[i].busName,CDI_BUSNAME_SIZE)) break;
    if (i == numBusList) return -not_implemented;
  }
  cnt=0;
  for ( j = 0; j < numConnectedHandles; j++ )
  {
    if ( busHandle[j] == NULL ) continue;
    if ( flag == 1 && busList[ busHandle[j]->busIndex ].busType != busList[i].busType ) continue;
    cnt++;
  }
  return cnt;
}
CDI_EXPORT int cdi_GetNumberOpenedLines(void)
{
  return getNumberLines(NULL, 0);
}
CDI_EXPORT int cdi_GetNumberOpenedLinesByBus(char *busName)
{
  return getNumberLines(busName, 1);
}
CDI_EXPORT int cdi_GetCdiLine(char *busName, int busLine)
{
  int i,cdiLine;
  for (i=0; i<numBusList; i++) if (!strncmp(busName, busList[i].busName, CDI_BUSNAME_SIZE)) break;
  if (i == numBusList) return -not_implemented;
  if ((cdiLine=checkLine(busList[i].busType,busLine)) >= 0) return cdiLine;
  return -no_such_line;
}
/************************************/
/* return CDI line index */
/************************************/
CDI_EXPORT int cdi_Open(char *busName, int busLine, char *parameter, int numberDevices, int numberRequestBuffers)
{ int i, cdiLine = -1, numBuffers;
  int shared, value;
  int fileHandle, retSem = 0, retSemCall = 0, cc = 0;
  static int insideCdiOpen = FALSE;
  struct timespec taskSleepTime;

  if (insideCdiOpen) ccerr(operation_busy);
  insideCdiOpen = TRUE;
  if (cdiInitFlag == 0 && (i=cdiInitialize()) < 0) ccerr(-i);

  for (i=0; i<numBusList; i++) if (strncmp(busName, busList[i].busName, CDI_BUSNAME_SIZE) == 0) break;
  if (i == numBusList)
  {
    cdilog("CDI: BusName not supported! %.16s",busName);
    ccerr(not_implemented);
  }

  if ((cdiLine = checkLine(busList[i].busType, busLine)) >= 0) ccerr(0);

  cdiLine = findFreeLine();
  if ( cdiLine < 0 )
  {
    cdilog("CDI: cannot find available slot!; maximum slots = %d", numConnectedHandles);
    ccerr(resources_exhausted);
  }
  cdilog("CDI: using free cdi line %d", cdiLine);

  if ( numberDevices < 1 ) numberDevices = 1; /* at least one device */
  fileHandle = connectToBus(busList[i].busType, busLine, parameter, cdiLine, numberDevices);
  if ( fileHandle < 0 )
  {
    if (cdiDebugLevel) dbglog("open line error; line Index = %d", busLine);
    ccerr(code_failure);
  } 
  if ( numberRequestBuffers <= 0 ) numBuffers = MULTI_N_DEVICE * numberDevices;
  else if ( numberRequestBuffers > MULTI_MAX_DEVICE * numberDevices ) numBuffers = MULTI_MAX_DEVICE * numberDevices;
  else numBuffers = numberRequestBuffers;
  cdilog("CDI: number of Buffers allocated: %d with this cdi line %d", numBuffers, cdiLine);
  /* allocate memory... */
  if ((busHandle[cdiLine]=(CdiLineInfoBlk *)calloc((size_t)LINE_INFOR_SIZE,1)) == NULL) 
    ccerr(out_of_server_memory);
  if ((busHandle[cdiLine]->pDevList=(CdiDeviceInfoBlk *)calloc((size_t)DEVICE_INFOR_SIZE,numberDevices)) == NULL) 
    ccerr(out_of_server_memory);
  if ((busHandle[cdiLine]->pReqList=(CdiRequestInfoBlk *)calloc((size_t)REQUEST_INFOR_SIZE,numBuffers)) == NULL) 
    ccerr(out_of_server_memory);

  busHandle[cdiLine]->busIndex = i;
  busHandle[cdiLine]->cdiLine = cdiLine;
  busHandle[cdiLine]->busLine = busLine;
  busHandle[cdiLine]->fileHandle = fileHandle;
  busHandle[cdiLine]->numberDevices = numberDevices;
  busHandle[cdiLine]->numberRequest = numBuffers;
  busHandle[cdiLine]->backReq.tv_sec = 0;
  busHandle[cdiLine]->backReq.tv_nsec = BACK_THREAD_TIME * 1000;
  cdiMutexInit(&busHandle[cdiLine]->lineMutex, cdiLine);

  retSem = 0;
  for (i=0; i<busHandle[cdiLine]->numberRequest; i++)
  {
    shared = value = 0;
    if ((retSem=cdiGetSemaphore(&busHandle[cdiLine]->pReqList[i].sync_SemId,shared,value)) != 0) ccerr(semaphore_error);
    shared = 0;
    value = 1;
    if ((retSem=cdiGetSemaphore(&busHandle[cdiLine]->pReqList[i].exec_SemId,shared,value)) != 0) ccerr(semaphore_error);
  }

  shared = value = 0;
  if ((retSem=cdiGetSemaphore(&busHandle[cdiLine]->back_SemId,shared,value)) != 0) ccerr(semaphore_error);
  shared = value = 0;
  if ((retSemCall=cdiGetSemaphore(&busHandle[cdiLine]->call_SemId,shared,value)) != 0) ccerr(semaphore_error);
  taskSleepTime.tv_sec = 0;
  taskSleepTime.tv_nsec = 400000 * 1000;  /* 400 ms  */
  cdi_nanosleep( &taskSleepTime, NULL );
  cdilog("ThreadCallback starting: cdi line: %d, hardware line number: %d", cdiLine, busLine);
  if ((busHandle[cdiLine]->tIdCallback=(pthread_t)cdiCreateThread(busHandle[cdiLine], THREAD_INDEX_CALLBACK)) == 0) ccerr(software_error);
  cdilog("ThreadMaster starting: cdi line: %d, hardware bus line number: %d", cdiLine, busLine);
  if ((busHandle[cdiLine]->tIdMaster=(pthread_t)cdiCreateThread(busHandle[cdiLine], THREAD_INDEX_MASTER)) == 0) ccerr(software_error);
  busHandle[cdiLine]->threadActive = TRUE;

  if ( cdiLine == numConnectedHandles ) numConnectedHandles++;
  cdilog("CDI: cdi_Open -> cdi Line number %d success",cdiLine);

   makeCdiHashTable(cdiLine,numberDevices); /* will do nothing if already allocated */
err:
  if (cc && cdiLine >= 0 && busHandle[cdiLine] != NULL)
  {
    if (busHandle[cdiLine]->pDevList != NULL) free (busHandle[cdiLine]->pDevList);
    else cdilog("CDI: could not allocate memory for busHandle %d device list",cdiLine);
    if (busHandle[cdiLine]->pReqList != NULL) free (busHandle[cdiLine]->pReqList);
    else cdilog("CDI: could not allocate memory for busHandle %d request list",cdiLine);
    if (busHandle[cdiLine] != NULL) free(busHandle[cdiLine]);
    else cdilog("CDI: could not allocate memory for busHandle %d",cdiLine);
    if (busHandle[cdiLine]->tIdCallback != 0) cdiCancelThread(busHandle[cdiLine]->tIdCallback);
    else cdilog("CDI: could not create thread for busHandle %d",cdiLine);
    if (busHandle[cdiLine]->tIdMaster != 0) cdiCancelThread(busHandle[cdiLine]->tIdMaster);
    else cdilog("CDI: could not create thread for busHandle %d",cdiLine);
    if (retSemCall != 0) cdiDestroySemaphore(&busHandle[cdiLine]->call_SemId);
    else cdilog("CDI: could not create semaphore for busHandle %d",cdiLine);
    if (retSem != 0) cdiDestroySemaphore(&busHandle[cdiLine]->back_SemId);
    else cdilog("CDI: could not create semaphore for busHandle %d",cdiLine);
    cdilog("CDI: cdi_Open -> %s",erlst[cc]); 
    disconnectFromBus(busList[i].busType, busLine, fileHandle);
  }
  insideCdiOpen = FALSE;
  return cc ? -cc : cdiLine;
}

/**************************************/
CDI_EXPORT void cdi_Cleanup(void)
{ 
  static int done = FALSE;
  int i;
  struct timespec ts = { 0, 10000 * 1000 }; // 10 msec
  if (done) return;
  isCleaningUp = TRUE;
  done = TRUE;
  cdilog("cleanup: stop all listeners");
  freeListenerMemory();
  cdi_nanosleep( &ts, NULL );
  cdilog("cleanup: remove all bus plugs");
  cdiInitFlag = -1;
  cdilog("cleanup: close down all lines");
  for (i=0; i<numConnectedHandles; i++) cdi_Close(i);
  rmvBusLibs();
  return;
}
/* below: need to worry about template produced devices and repeat references */
#define FLUSHREFENCE(ib,f) for (k=i+1; k<(ib)->numberDevices; k++) {\
 if ((ib)->pDevList[k].f == (ib)->pDevList[i].f) (ib)->pDevList[k].f = NULL;\
}
CDI_EXPORT int cdi_Close(int cdiLine)
{ 
  int status, i, k, cc, tTmp;

  if ( cdiLine >= numConnectedHandles || busHandle[cdiLine] == NULL )
  {
    cdilog("CDI: cdi_Close -> cdi line %d invalid",cdiLine);
    return -argument_list_error;
  }
  status = disconnectFromBus(busList [ busHandle[cdiLine]->busIndex ].busType, busHandle[cdiLine]->busLine, busHandle[cdiLine]->fileHandle);
  if ( status >= 0 || busHandle[cdiLine]->threadActive )
  {
    cc = busHandle[cdiLine]->threadActive;
    busHandle[cdiLine]->threadActive = FALSE;
    if ( cc )
    {
      tTmp = getTimeMilliseconds();
      while (busHandle[cdiLine]->threadBusy)
      {
        cdi_nanosleep( &busHandle[cdiLine]->backReq, NULL );
        if ( cdiInitFlag == -1 && getTimeMilliseconds() - tTmp > 2000 ) break;	// wait for maximum 2 seconds
      }
    }
    cc = cdiCancelThread(busHandle[cdiLine]->tIdCallback); 
    cdilog("cdiCancelThread callback thread return code: %d, cdi line: %d, hardware line number: %d",cc, busHandle[cdiLine]->cdiLine, busHandle[cdiLine]->busLine);
    cc = cdiCancelThread(busHandle[cdiLine]->tIdMaster);
    cdilog("cdiCancelThread background thread return code: %d, cdi line: %d, hardware line number: %d",cc, busHandle[cdiLine]->cdiLine, busHandle[cdiLine]->busLine);
    cdiDestroySemaphore(&busHandle[cdiLine]->back_SemId);
    cdiDestroySemaphore(&busHandle[cdiLine]->call_SemId);
    for (i=0; i<busHandle[cdiLine]->numberDevices; i++)
    {
      if (busHandle[cdiLine]->pDevList[i].dataMask != NULL) free(busHandle[cdiLine]->pDevList[i].dataMask);
      FLUSHREFENCE(busHandle[cdiLine], dataMask);
      if (busHandle[cdiLine]->pDevList[i].dataPattern != NULL) free(busHandle[cdiLine]->pDevList[i].dataPattern);
      FLUSHREFENCE(busHandle[cdiLine], dataPattern);
      if (busHandle[cdiLine]->pDevList[i].dataInput != NULL) free(busHandle[cdiLine]->pDevList[i].dataInput);
      FLUSHREFENCE(busHandle[cdiLine], dataInput);
      if (busHandle[cdiLine]->pDevList[i].clbrRules != NULL) free(busHandle[cdiLine]->pDevList[i].clbrRules);
      FLUSHREFENCE(busHandle[cdiLine], clbrRules);
      if (busHandle[cdiLine]->pDevList[i].rvrsRules != NULL) free(busHandle[cdiLine]->pDevList[i].rvrsRules);
      FLUSHREFENCE(busHandle[cdiLine], rvrsRules);
      if (busHandle[cdiLine]->pDevList[i].extName != NULL) free(busHandle[cdiLine]->pDevList[i].extName);
      FLUSHREFENCE(busHandle[cdiLine], extName);
    }
    free ( busHandle[cdiLine]->pDevList );
    free ( busHandle[cdiLine]->pReqList );
    free ( busHandle[cdiLine] );
    busHandle[cdiLine] = NULL;
    if ( cdiLine == numConnectedHandles - 1 )
    {
      i = cdiLine;
      for ( --i; i >= 0 && busHandle[i] == NULL ; --i )
      numConnectedHandles = i + 1;
    }
  }
  if (cdiDebugLevel) dbglog("cdi_Close OK! line number: %d; status: %d",cdiLine, status);

  return status;
}
CDI_EXPORT int cdiRegisterCycleFunction(char *busName, int interval, void (*fcn)(void))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName, -1);
  for (i = 0; i < numBusList && strncmp(busName, busList[i].busName, CDI_BUSNAME_SIZE); i++);
  if (i == numBusList)
  {
    cdilog("Cannot register cycle function for bus %.16s : non existent!", busName);
    return non_existent;
  }
  busList[i].interval = interval;
  busList[i].cycleFcn = fcn;
  cdilog("cdi cycle function for bus %.16s registered", busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusFilter(char *busName,int (*fcn)(int *,int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName, CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register filter function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].filterBusFcn = fcn;
  cdilog("cdi filter function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusConfiguration(char *busName,int (*fcn)(void))
{
  int i, cc;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register configuration function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].cfgBusFcn = fcn;
  /* don't call this in cdi.c !! */
  if (busList[i].cfgBusFcn != NULL && gSimulationFlag == FALSE )
  {
    cc = busList[i].cfgBusFcn();
    cdiAssertHeapOkay("cdiRegisterBusConfiguration",busName,i);
    if (cc < 0)
    {
      dbglog("failure configuring %s",busList[i].busName);
    }
  }
  cdilog("cdi configuration function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusInitialization(char *busName,int (*fcn)(int,int,int,char *))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register initialization function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].initBusFcn = fcn;
  cdilog("cdi initialization function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusBlockFunction(char *busName,int (*fcn)(int *,int,void **,int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register BusBlock function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].blockFcn = fcn;
  cdilog("cdi block Transfer function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusCancelRequest(char *busName,int (*fcn)(CdiRequestInfoBlk *, int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register bus cancel Request function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].cancelRequestFcn = fcn;
  cdilog("cdi bus cancel Request function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusLoadRequest(char *busName,int (*fcn)(CdiRequestInfoBlk *, int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register bus load Request function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].loadRequestFcn = fcn;
  cdilog("cdi bus load Request function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusDeviceReady(char *busName,int (*fcn)(CdiRequestInfoBlk *, int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register bus Device Ready function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].deviceReadyFcn = fcn;
  cdilog("cdi bus Device Ready for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusReqTimeout(char *busName,int (*fcn)(CdiRequestInfoBlk *, int, int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register bus timwout function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].timeoutFcn = fcn;
  cdilog("cdi bus timeout function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusHandler(char *busName,void (*fcn)(CdiRequestInfoBlk *))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register bus handler for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].execBusFcn = fcn;
  cdilog("cdi bus handler for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBusCleanup(char *busName,int (*fcn)(int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register cleanup function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].freeBusFcn = fcn;
  cdilog("cdi cleanup function for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterBus(char *busName,int busType)
{
  int id = busType; /* if positive, use it */
  if (id < 0) id = cdiGetBusType(busName); /* is this bus already registered ? */
  if (id >= 0) return id;  /* if registered already, just return the busType */
  id = cdiGetNextAvailableBusType(); /* it's a new one ! */
  return cdi_RegisterBus(busName,id,NULL,NULL,NULL,NULL,NULL);

}
CDI_EXPORT int cdiScanBus(char *busName,char *target,char *result,int resultBufferSize)
{
  int i, cc;
  if (result == NULL) return argument_list_error;
  if (resultBufferSize < 32) return buffer_too_small;
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
#   ifdef WIN32
    _snprintf(result,32,"%s not registered",busName);
#   else
    snprintf(result,32,"%s not registered",busName);
#   endif
    return non_existent;
  }
  if (busList[i].scanBusFcn == NULL || gSimulationFlag != FALSE )
  {
#   ifdef WIN32
    _snprintf(result,32,"no bus scanner registered");
#   else
    snprintf(result,32,"no bus scanner registered");
#   endif
    return not_implemented;
  }
  cc = busList[i].scanBusFcn(target,result,resultBufferSize);
  cdiAssertHeapOkay("cdiScanBus",busName,resultBufferSize);
  return cc;
}
CDI_EXPORT int cdiRegisterBusScanner(char *busName,int (*fcn)(char *,char *,int))
{
  int i;
  if (cdiGetBusType(busName) < 0) cdiRegisterBus(busName,-1);
  for (i=0; i<numBusList && strncmp(busName,busList[i].busName,CDI_BUSNAME_SIZE); i++);
  if (i == numBusList) 
  {
    cdilog("Cannot register initialization function for bus %.16s : non existent!",busName);
    return non_existent;
  }
  busList[i].scanBusFcn = fcn;
  cdilog("cdi bus scanner for bus %.16s registered",busName);
  return 0;
}
CDI_EXPORT int cdiRegisterCalibrationFunction(char *fcnName,double (*fcn)(double))
{
  return registerClbrConversionFunction(fcnName,fcn);
}
CDI_EXPORT int cdi_RegisterCalibrationFunction(char *fcnName,double (*fcn)(double))
{
  return registerClbrConversionFunction(fcnName,fcn);
}
CdiTemplate *getCdiTemplate(char *templateName)
{
  int i;
 
  for (i=0; i<CdiTemplateTblSize; i++)
    if (!strncmp(templateName,CdiTemplateTbl[i]->name,CDI_TEMPLATE_NAME_SIZE)) 
      return CdiTemplateTbl[i];

  if (i >= MAX_NUM_TEMPLATES)
  {
    cdilog("no resources to allocate new template for %.16s",templateName);
    return NULL;
  }
  cdilog("adding new template %.16s to the template table",templateName);
  if ((CdiTemplateTbl[i]=(CdiTemplate *)calloc(1,sizeof(CdiTemplate))) == NULL)
  {
    cdilog("cannot allocate new template %.16s",templateName);
    return NULL;
  }
  CdiTemplateTblSize++;
  CdiTemplateTbl[i]->length = 0;
  strncpy(CdiTemplateTbl[i]->name,templateName,CDI_TEMPLATE_NAME_SIZE);
  return CdiTemplateTbl[i];
}
CdiTemplateList *findFieldInTemplate(CdiTemplate *tmpt,char *field)
{
  CdiTemplateList *t;
  if (tmpt == NULL || field == NULL) return NULL;
  for (t=tmpt->cdiTemplateList; t != NULL; t=t->next)
  {
    if (!strncmp(t->field,field,CDI_TEMPLATE_FIELDNAME_SIZE)) return t;
  }
  return NULL;
}

int addFieldToTemplate(
  CdiTemplate *tmpt, char *field, char *lngfield, char *fieldDesc,
  int *addrParameters,int naddrParameters,
  BYTE *mask,int *maskRange,BYTE *pattern,int *patternRange,int patternTrigger,
  BYTE *dataIn,int maxBytesIn,int maxBytesOut,int maxBytes2ndOut,int dataFmt,
  int access,
  int interval,
  CdiRule *rules_rcv,CdiRule *rules_snd,
  DBITFIELD *bf
)
{
  CdiTemplateList *t;
  if (tmpt == NULL || field == NULL) return argument_list_error;
  if (findFieldInTemplate(tmpt,field) != NULL) return already_assigned;
  if ((t=(CdiTemplateList *)calloc(1,sizeof(CdiTemplateList))) == NULL) return out_of_local_memory;
  strncpy(t->field,field,CDI_TEMPLATE_FIELDNAME_SIZE);
  strncpy(t->lngfield,lngfield,64);
  strncpy(t->description,fieldDesc,96);
  if (addrParameters != NULL) memcpy(t->addrParameters,addrParameters,NUMBER_EXEC_PARAM*sizeof(int));
  t->naddrParameters = naddrParameters;
  if (mask != NULL)
  {
    if ((t->dataMask=(BYTE *)calloc(8,1)) == NULL) return out_of_server_memory;
    memcpy(t->dataMask,mask,8);
  }
  if (maskRange != NULL) memcpy(t->dataMaskRange,maskRange,2*sizeof(int));
  if (pattern != NULL)
  {
    if ((t->dataPattern=(BYTE *)calloc(CDI_PATTERN_FILTER_SIZE,1)) == NULL) return out_of_server_memory;
    memcpy(t->dataPattern,pattern,CDI_PATTERN_FILTER_SIZE);
  }
  if (patternRange != NULL) memcpy(t->dataPatternRange,patternRange,2*sizeof(int));
  t->dataPatternTrigger = patternTrigger;
  t->dataInput = dataIn;
  t->dataInputSize = maxBytesIn;
  t->dataOutputSize = maxBytesOut;
  t->dataOutput2ndSize = maxBytes2ndOut;
  t->dataFormat = dataFmt;
  t->allowedAccess = access;
  t->interval = interval;
  t->clbrRules = rules_rcv;
  t->rvrsRules = rules_snd;
  t->bitfield = bf;
  /* add this one to the list */
  t->next = tmpt->cdiTemplateList;
  tmpt->cdiTemplateList = t;
  tmpt->length++;  
  return 0;
}

CDI_EXPORT int cdi_getTemplateNames(NAME16 *namebuffer,int namesize)
{ 
  int i;
  for (i=0; i<CdiTemplateTblSize && i<namesize; i++)
  {
    strncpy(namebuffer[i].name,CdiTemplateTbl[i]->name,CDI_TEMPLATE_NAME_SIZE);
  }
  return CdiTemplateTblSize;
}
CDI_EXPORT int cdi_getTemplate(char *templateName,NAME16II *cdiTemplate,int *namesize)
{ 
  int i;
  CdiTemplate *tmpl;
  CdiTemplateList *t;

  if (templateName == NULL || cdiTemplate == NULL || 
      namesize == NULL || *namesize == 0) return argument_list_error;
  for (i=0; i<CdiTemplateTblSize; i++)
    if (!strncmp(templateName,CdiTemplateTbl[i]->name,CDI_TEMPLATE_NAME_SIZE)) break;

  if (i == CdiTemplateTblSize) return non_existent;
  tmpl = CdiTemplateTbl[i];
  t = tmpl->cdiTemplateList;

  for (i=0; i<tmpl->length && i<*namesize && t != NULL; i++) 
  {
    strncpy(cdiTemplate[i].name,t->field,CDI_TEMPLATE_FIELDNAME_SIZE);
    cdiTemplate[i].i1val = t->addrParameters[0];
    cdiTemplate[i].i2val = t->addrParameters[1];
    t = t->next;
  }
  *namesize = tmpl->length;
  return 0;
}
