#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <math.h>
#include <errno.h>
#include "tine.h"
#include "cdi.h"
#include "tinecdi.h"

typedef void (__stdcall *cbTLink)(int id,int cc);

typedef struct CdiCbElemStruct
{
  int cdiId; // as returned from cdiAttachLink (needed for closing the link)
  int cbId;  // requested callback id from App
  cbTLink cb; // reference to callback function
} CdiCbElem;
#define MAX_CDI_LINK_ELEMS 1024
int gMaxCdiLinkElems = MAX_CDI_LINK_ELEMS;
CdiCbElem *cdiCbElems = NULL;

int initLinkElemTable(void)
{
  int i;
  static int done = FALSE;
  if (done) return 0;
  done = TRUE;
  if ((cdiCbElems=(CdiCbElem *)calloc(gMaxCdiLinkElems,sizeof(CdiCbElem))) == NULL)
  {
    cdilog("cannot allocate memory for callback elements table!");
    return out_of_server_memory;
  }
  for (i=0; i<gMaxCdiLinkElems; i++) cdiCbElems[i].cdiId = -1;
  return 0;
}
int getNextElemId(void)
{
  int i;
  if (cdiCbElems == NULL) return -out_of_server_memory;
  for (i=0; i<gMaxCdiLinkElems; i++) if (cdiCbElems[i].cdiId == -1) break;
  if (i == gMaxCdiLinkElems) return -resources_exhausted;
  return i;
}
int addLinkElem(int cdiId,cbTLink cb,int cbId)
{
  int i = getNextElemId();
  if (i >= 0)
  {
    cdiCbElems[i].cdiId = cdiId;
    cdiCbElems[i].cb = cb;
    cdiCbElems[i].cbId = cbId;
  }
  return i;
}
int rmvLinkElem(int linkId)
{
  if (cdiCbElems == NULL) return out_of_server_memory;
  if (linkId < 0 || linkId >= gMaxCdiLinkElems) return link_not_open;
  memset(&cdiCbElems[linkId],0,sizeof(CdiCbElem));
  cdiCbElems[linkId].cdiId = -1;
  return 0;
}

void qdCallback(int id,int cc) // quick and dirty callback (only one of these allowed for the VB apps at the moment
{
  if (id < 0 || id > gMaxCdiLinkElems) return; // this can't be !
  if (cdiCbElems[id].cb != NULL)
  {
    cdiCbElems[id].cb(cdiCbElems[id].cbId,cc);
  }
}

void hwndCallback(int id,int cc)
{
  if (id < 0 || id > gMaxCdiLinkElems) return; // this can't be !
  if (cdiCbElems[id].cb != NULL)
  {
    SendMessage((HWND)cdiCbElems[id].cb, WM_SYSCOMMAND, id, cc);
  }
}


__declspec(dllexport) int PASCAL cdiCloseLinkA(int id)
{
  int cc = 0;
  if (cdiCbElems[id].cdiId < 0) return link_not_open;
  cc = cdiCloseLink(cdiCbElems[id].cdiId);
  rmvLinkElem(id);
  cdilog("close cdi link %d : %s",id,erlst[cc]);
  return cc;
}
__declspec(dllexport) int cdiCloseLinkWnd(int id)
{
  int cc = 0;
  if (cdiCbElems[id].cdiId < 0) return link_not_open;
  cc = cdiCloseLink(cdiCbElems[id].cdiId);
  rmvLinkElem(id);
  cdilog("close cdi link %d : %s",id,erlst[cc]);
  return cc;
}
__declspec(dllexport) int PASCAL cdiAttachLinkExA(
  char *devName,     /* device name */
  char *devProperty, /* "READ", "WRITE", "READWRITE", "WRITEREAD" */
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  cbTLink callback,  /* callback routine */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 cbId        /* callback ID */
)
{
  int idCdi, id = -not_implemented;

  id = getNextElemId();
  idCdi = cdiAttachLinkEx(devName,devProperty,dout,din,access,pollingRate,qdCallback,mode,id);
  id = addLinkElem(idCdi,callback,cbId);

  return id;
}

/* attach link routine with Windows Message instead of callback */
__declspec(dllexport) int cdiAttachLinkWnd(
  char *devName,     /* device name */
  char *devProperty, /* "READ", "WRITE", "READWRITE", "WRITEREAD" */
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  HWND hwnd,         /* callback window */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 cbId        /* callback ID */
)
{
  int idCdi, id = -not_implemented;

  id = getNextElemId();
  idCdi = cdiAttachLinkEx(devName,devProperty,dout,din,access,pollingRate,hwndCallback,mode,id);
  id = addLinkElem(idCdi,(cbTLink)hwnd,cbId);

  return id;
}

__declspec(dllexport) int PASCAL cdiExecLinkExA(
  char *devName,     /* device name */
  char *devProperty, /* "READ", "WRITE", "READWRITE", "WRITEREAD" */
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  UINT16 timeout     /* timeout in msec. */
  )
{
  return cdiExecLinkEx(devName,devProperty,dout,din,access,timeout);
}


BOOL WINAPI DllMain(HINSTANCE hInstDll,DWORD fdwReason,LPVOID lpvReserved)
{
  switch (fdwReason)
  {
    case DLL_PROCESS_ATTACH:  // DLL being loaded
      initLinkElemTable();
      cdiInitialize();
      break;
    case DLL_THREAD_ATTACH:
      break;
    case DLL_THREAD_DETACH:
//      cdi_Cleanup();
      break;
    case DLL_PROCESS_DETACH:  // DLL being unloaded
      cdi_Cleanup();
      break;
  }
  return (TRUE);
}
