/** \file listener.h
 * \brief TINE static listener documentation 
 *
 */
#ifndef LISTENER_H_
#define LISTENER_H_

#include "project.def"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifdef WIN32
# define LSNR_STARTUP_WAITTIME 1000000
#else
# define LSNR_STARTUP_WAITTIME 100000
#endif
#define ALLOWED_IDLE_TIME 300
#define CDI_SCHED_NONE 0x00
#define CDI_SCHED_PRP  0x01
#define CDI_SCHED_DEV  0x02 
typedef struct cdiListenerStruct
{
  char *devName;/**< listener device name */ 
  char devProperty[PROPERTY_NAME_SIZE]; /**< listener device property */
  DTYPE dout;   /**< output data object */
  DTYPE din;    /**< input data object */
  BYTE *rbbuf;  /**< readback data buffer */
  BYTE *outbuf; /**< output data buffer */
  BYTE *inbuf;  /**< input data buffer */
  UINT32 outlen;/**< output data size for comparison */
  int mlid;     /**< listener table id */
  int lid;      /**< link id */
  int status;   /**< current status */
  int pollrate; /**< polling interval */
  int isSched;  /**< schedule flag: TRUE -> schedule the results on change */
  float atolSched; /**< absolute tolerance (scheduling criteria) */
  float ptolSched; /**< relative (percent) tolerance (scheduling criteria) */
  double timestamp; /**< current time stamp */
  struct cdiListenerStruct *nxt;
} cdiListener;

#ifndef MAX_NUM_LISTENERS
#define MAX_NUM_LISTENERS 1000
#endif

extern cdiListener **cdiListenerTable;
extern int nCdiListeners;

int stopListener(cdiListener *ml);
cdiListener *startListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,int pollrate,int flags);
int stopListener(cdiListener *ml);
cdiListener *getListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din);
int getDataFromListener(char *devName,char *devProperty,DTYPE *dout,DTYPE *din,double *dTime,int rate,int flags);
int cdiDumpListeners(int i1,int i2,int i3,int i4);
void freeListenerMemory(void);
void stopAllListeners(void);
extern cdiListener *cdiListeners;
extern int nListeners;
extern int startListenerCompCode;

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* LISTENER_H_ */
