#ifndef CDI_PROT_H_
#define CDI_PROT_H_

/********************************************************/
/* Common Device Interface protocol definition file */
/* ------------ cdi_prot.h --------------- */
/********************************************************/
#include "cdi.def"
#include "project.def"
#if defined(UNIX)

#include <pthread.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/time.h>  /* gethrtime() */
#include <poll.h>
#include <termios.h>
#include <semaphore.h>

#define THREAD_TYPE void *
#define THREAD_PARAM void *
typedef pthread_mutex_t cdi_mutex_t;
typedef pthread_t cdi_thread_t;

#elif defined(WIN32)

#include "windows.h"
typedef HANDLE sem_t;
typedef HANDLE pthread_t;
#define THREAD_TYPE DWORD WINAPI 
#define THREAD_PARAM LPVOID
#define CDI_SEM_MAXCOUNT 1000
typedef HANDLE cdi_mutex_t;
typedef HANDLE cdi_thread_t;

#elif defined(VXWORKS)

#endif /* UNIX, WIN32, VXWORKS */

#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

CDI_EXPORT int cdi_nanosleep(const struct timespec *req, struct timespec *rem);

CDI_EXPORT cdi_thread_t cdiCreateThread(void *cdiThreadParam, int index);
CDI_EXPORT int cdiCancelThread(cdi_thread_t cdiThreadId);
CDI_EXPORT int cdiGetSemaphore(sem_t *sem,int access,int count);
CDI_EXPORT int cdiDestroySemaphore(sem_t *sem);
CDI_EXPORT int cdiPostSemaphore(sem_t *sem);
CDI_EXPORT int cdiWaitSemaphore(sem_t *sem,int peek);
CDI_EXPORT int cdiMutexInit(cdi_mutex_t *mutex, int line);
CDI_EXPORT int cdiMutexLock(cdi_mutex_t *mutex);
CDI_EXPORT int cdiMutexUnlock(cdi_mutex_t *mutex);

#include "errors.h"
#include "tinetype.h"

/* open CAN parameter index */
enum accessNumeraters
{
  cdiReadFlag, /* read only, 2 steps: send and receive */
  cdiWriteFlag, /* write only, 2 steps: send and receive */
  cdiReadWriteFlag, /* read with condition flag then write, 4 steps */
  cdiWriteReadFlag, /* write then read, 4 steps */
  cdiWriteWriteFlag, /* write 1 then write 2 for sedac 2 steps */
  cdiWriteReadWriteFlag, /* write 1, read 1, then write 2 for sedac 3 steps */
  cdiReadReadFlag, /* read 1, then read 2 for sedac 2 steps */
  cdiWriteReadReadFlag /* write 1, read 1, then read 2 for sedac 3 steps */
};

enum readConditionNumeraters
{
  anyData = 0,			/* return with any data => callback	*/
  maskCompEqual,		/* 1: after mask compare with pattern equal = */
  maskCompNotEqual,	/* 2: after mask compare with pattern NOT equal != */
  maskChangeWindow,		/* 3: after mask on change, compare with last data, with range window */
  maskChangePercent,	/* 4: after mask on change, compare with last data, with percentage */
};

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* CDI_PROT_H_ */
