// pdoclient code
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "cdi.h"
#include "pdoclient.h"
#include "tinecdi.h"

#define INITDELAY 500

//int cdi_getDeviceParametersForDevice(char *devName,int *paramlist);
/*
 * PdoClient : version 23.09.10
 * 1. After TINE Server initialized, get device name list from it
 * 2. Pick out names containing ".pdoT" 
 * 3. For these, set TPDO refresh time interval via SDO 0x1800,1,2
 * (this is tuned to CAN modules of T Delfs)
 */
void PdoClientInit() {
  strcpy(tineAddress,"/localhost/cdi/");
  dDevSdo16.data.sptr = &devsdo16;
  dDevSdo16.dFormat = CF_SHORT;
  dDevSdo16.dArrayLength = 1;
  tpdo = 0x200;  // 0x200 = 512 msec (?)
  stateFlag = QDELAY;
  tcycle = 20; // msec
  tlast = 0;
  tstart = msectime(); 
}
// This uses memory-based call to cdi code
int GetCoDevices() {
  int nco, nrc; 
   nDevice = cdi_getDeviceNames(NULL,0); 
   cdilog("PdoClient: devices found %d\n",nDevice);
   printf("PdoClient: devices found %d\n",nDevice);
   if (nDevice>0) {
     devlist = (NAME32*) calloc(nDevice,32);
     if (devlist==0) {  // || devstatus == 0) {
	cdilog("PdoClient: Mem Alloc fails: stop pdoT init\n");
        if (devlist) free(devlist);
	nrc = 0;
     }
     else nrc = cdi_getDeviceNames(devlist,nDevice);  // returns ndev again
     if (nrc==0) return nrc;
     nco =CountCoDevices();
     if (nco == 0) { nCoDevice = 0; return 0; } // no 'CO devices' defined in cdiaddr file
     RegisterCoDevices(nDevice);
   }
   return nrc;
}
// 13.01.11 Check to see whether devices with "hbeat" are defined.  If not
// assume that we don't want to use automated pdo control system. In particular
// the NEG devices use a different implementation which is controlled by
// a middle layer program.  Sometime develop a common system including NEGS?
int CountCoDevices() {
  int i, nco=0;
  for (i=0; i<nDevice; i++) {
    if (strstr(devlist[i].name,"hbeat") != 0) nco++;
  }
  cdilog("PdoClient: Count hbeat devices: %d\n",nco);
  return nco;
}
// CDI Dev Address Parameters: ca[0] = (0-5) CANopen Type
// =0   SDO, ca[1] = Index, ca[2]=subindex, ca[4] = 0 (send needed for recv)
// =1-4 PDO, ca[1] = offset bytes, ca[4] = 1 (= no send needed for recv)
// =5   NMT, other indices unused
void RegisterCoDevices() {
  int i, j,  nodeId, ncod = 0, rc;
  DADDRESS* da;
  int ca[NUMBER_EXEC_PARAM];
  short pdo[MAXCANID];
  short pdodev[MAXCANID];
  for (i=0; i<MAXCANID; i++) { pdo[i] = -1;  pdodev[i] = -1; }
  for (i=0; i<nDevice; i++) { // scan to count #devices with at least one PDO defined
     rc = cdi_getDeviceParametersForDevice(devlist[i].name,ca);
     if (ca[0] < 1 || ca[0] > 4) continue; // only pdos 1-4
     da = cdi_getBusAddressForDevice(devlist[i].name);
     nodeId = da->i2val;
     if (pdo[nodeId] >= 0) continue; // already filled? (0 is valid index)
     pdo[nodeId] = ncod;  // node to co  index
     pdodev[nodeId] = i;  // node to cdi device#
     //printf("CO# %d :  %s %d %d : %d \n",ncod,devlist[i].name,ca[0],ca[1],nodeId);
     cdilog("CO# %d :  %s %d %d : %d \n",ncod,devlist[i].name,ca[0],ca[1],nodeId);
     ncod++;
  }
  nCoDevice = ncod;
  printf("#CAN Devs supporting PDOs %d\n",nCoDevice);
  cdilog("#CAN Devs supporting PDOs %d\n",nCoDevice);
  coInfo = (CoInfo*) calloc(nCoDevice,sizeof(CoInfo));
  ncod = 0;
  for (i=0; i<MAXCANID; i++) {         // set up CoInfo
    if (pdo[i] < 0) continue;
    ncod = pdo[i];                        // from above      
    coInfo[ncod].canId = i;               // nodeId
    coInfo[ncod].devPdoInd = pdodev[i];   // cdi dev index for pdo MOD 22.09.11
    coInfo[ncod].devHbInd = -1;           // if none
    coInfo[ncod].hbTimestamp = time(NULL);
    for (j=0; j<5; j++) coInfo[ncod].resetId[j] = -1; // note additional index here for logic
    coInfo[ncod].pdoTimestamp = time(NULL);
  }
  for (i=0; i<nDevice; i++) { // fill in device indices for heartbeats, Pdo time interval sdos 
     rc = cdi_getDeviceParametersForDevice(devlist[i].name,ca); 
     da = cdi_getBusAddressForDevice(devlist[i].name);
     nodeId = da->i2val;
     ncod = pdo[nodeId]; 
     cdilog("%d  CAN %d    %d %d %d %d %d   %s\n",i,nodeId,ca[0],ca[1],ca[2],ca[3],ca[4],devlist[i].name);
     if (ca[0] != 5 && ca[2] != 5) continue;   // ca[0] = 5 for NMT,  ca[2]=5 for 0x180z subindex 5
     if (ca[0] == 5) coInfo[ncod].devHbInd = i; // heartbeat nmt for nodeId
     else if (ca[2] == 5) {
       if (ca[1] == 0x1800) coInfo[ncod].resetId[0] = i;
       else if (ca[1] == 0x1801) coInfo[ncod].resetId[1] = i;
       else if (ca[1] == 0x1802) coInfo[ncod].resetId[2] = i;
       else if (ca[1] == 0x1803) coInfo[ncod].resetId[3] = i;
     } 
  }
  for (i=0; i<nCoDevice; i++) { 
    cdilog("COINFO %d CANID %d HB %d TI %d %d %d %d\n",i,coInfo[i].canId,coInfo[i].devHbInd,
            coInfo[i].resetId[0],coInfo[i].resetId[1],coInfo[i].resetId[2],coInfo[i].resetId[3]);
  }
  return;
}
// print to terminal canId and elapsed time since last time update via callback
int PdoStatus(int i1,int i2,int i3, int i4) {  // for status printout ?
  //char buf[80];
  int i, thb, tpdo, tnow = time(NULL);
  if (!nCoDevice) { dbglog("No Registered PDOs\n"); return 0; }
  else dbglog(" #  CanId  dt(HB) dt(PDO)\n");
  for (i=0; i<nCoDevice; i++) {
    thb = tnow - coInfo[i].hbTimestamp;
    if (thb>1000) thb = 1000; // for formatting
    tpdo = tnow - coInfo[i].pdoTimestamp;
    if (tpdo>1000) tpdo = 1000;
    dbglog("%2d %4d %6d %6d\n",i,coInfo[i].canId,thb,tpdo);
  }
  return 0;
}

void SubscribeCoDevices() {
  int coInd, devInd, hbInd, i;
  for (i=0; i<nCoDevice; i++) {
    coInd = i;
    hbInd = coInfo[i].devHbInd;
    devInd = coInfo[i].devPdoInd;
    if (hbInd >=0) SubscribeHb(hbInd,coInd);
    SubscribePdo(devInd,coInd);
  }
}

int SendDevSdo16(short data, int id) {
   int rc; 
   devsdo16 = data;
   rc = AttachLinkEx(devAddress,"SEND",&dDevSdo16,&dDevSdo16,CA_WRITE,1000,CallbackSdo,CM_SINGLE,id);
   rcCallback = (rc>=0)? not_ready : -rc;
   return rcCallback;
}

int RecvDevSdo16() {
  int rc;  
   rc = AttachLinkEx(devAddress,"RECV",&dDevSdo16,0,CA_READ,1000,CallbackSdo,CM_SINGLE,2);
   rc = (rc>=0)? not_ready : -rc;
   return rc;
}
int SetPdoTimeInt(int devInd) { 
  if (devInd >= nDevice) return at_limit;
  strcpy(devAddress,tineAddress);
  strcat(devAddress,(char*)&(devlist[devInd]));
  rcCallback = SendDevSdo16(tpdo, devInd);
  printf("SEND PDO TI %s : %d rets %d\n",devAddress, devInd, rcCallback);
  cdilog("SEND PDO TI %s : %d rets %d\n",devAddress, devInd, rcCallback);
  return rcCallback;
}
void CallbackSdo(int id,int rc) {
  rcCallback = rc;
  if (rc==0) printf("CallbackSdo %d %d\n",devsdo16,id);
  else  printf("CallbackSdo id  Err %d  %d\n",id, rc);
}
long64 msectime() {                           
  struct timeval tb;
  gettimeofday( &tb, NULL );
  long64 tsec = tb.tv_sec;
  return tsec * 1000 +  tb.tv_usec/1000;   // milliseconds
}
// ------------------------------------------------------------
// notes on pdo1 triggering for TDelfs CAN Modules
//  sdo 0x1800,2 = trigger mode, ff = all, seems to be default
//  sdo 0x1800,5 = timer trig msec, default = 0 => no triggers
// => ff includes Data Change Triggers, can stay at default 
// unless we want to suppress these ( fe does this?)
//-------------------------------------------------------------
int PdoInitCycle() {
  static int npdo=0, nterr=0;
  int cc=0, devInd, rc = 0;
  long64 tnow = msectime();
  static long64 interim = 0;
  if ((tnow - tlast) < tcycle) return cc;
  else tlast = tnow;
  // -- execute at rate tcycle --
  switch(stateFlag) {
  case QDEFAULT: 
    stateFlag = QDELAY;
    // fall-through
  case QDELAY:    
    if ((tnow - tstart) > INITDELAY) stateFlag = QUERYDEV;
    break; // delay for server startup
  case QUERYDEV: 
    rc = GetCoDevices();
    nextIndex(0); // init
    stateFlag = (rc>0)? SETTIME : QFIN; // if no devices, give up
    break;
  case SETTIME:
    devInd = nextIndex(1); 
    printf("DI %d\n",devInd);
    if (devInd == 0x1000) { stateFlag = QFIN; break; }
    rc = SetPdoTimeInt(devInd);
    npdo++;       
    if (rc == at_limit) stateFlag = QFIN;              // finished
    else if (rc == not_ready) 
    {
      stateFlag = RECVTIME;    // else repeat SetTime
      interim = tnow;
    }
    break;
  case RECVTIME:
    // default is to cycle here until callBack return: is timeout necessary?
    if (rcCallback != not_ready) 
    {
      stateFlag = SETTIME;  // cycle until sdo return
      if (rcCallback !=0) {
	nterr++;
	printf("%s SetTime Fails: %d  %d\n",devAddress,rcCallback,nterr);
      }
    }
    if (tnow > interim + 1000) stateFlag = SETTIME;
    break;
  case QFIN:
    printf("Pdo Client PDOTs %d Init Errs %d\n",npdo,nterr);
    cdilog("Pdo Client PDOTs %d Init Errs %d\n",npdo,nterr);
    stateFlag = QDEFAULT;
    cc = 1;
    break;
  default: printf("PdoClient Bad Case %d\n",stateFlag);
    stateFlag = QDEFAULT;
    cc = 1;
  }
  return cc;
}
int nextIndex(int init) {
  static int coInd, resetInd;
  if (init==0) { coInd = -1; resetInd = 4; return -1; }
  while (coInd < nCoDevice) {
    if (resetInd >= 4) { coInd++; resetInd = -1; }
    if (coInd >= nCoDevice) { printf("CO %d %d\n",coInd,nCoDevice);  return 0x1000; }
    resetInd++;
    if (resetInd<10) printf("Ind %d %d\n",coInd,resetInd);
    while (coInfo[coInd].resetId[resetInd] < 0) {
      resetInd++;
      if (resetInd == 4) break;
    }
    if (resetInd <4) break;
  }
  printf("DEV %d %d\n",coInd,resetInd); 
  return coInfo[coInd].resetId[resetInd];
}
void SetPdoStateflag(int sf) {
  stateFlag = sf;
}
// cycle through PDO devices and go into RESET cycle if trigger satisfied
// note that since SystemCycle sleeps, max call rate here will be low, so
// may want to set rate up somewhat? What to do with different #s of PDOs?
void PdoMonitorCycle() {
  static int coInd, nterr=0, resetInd, tlast;
  int canId, dtPdo, dtHb, ptInd, tnow; // canId
  static long64 interim = 0;
  if (nCoDevice==0) return;
   tnow = time(NULL);
   if ((tnow - tlast) < 2) return; // limit access rate
   tlast = tnow;  
  switch(stateFlag) {
  case QDEFAULT:
    coInd = -1;
    stateFlag = QUERYDEV;
    break;
  case QUERYDEV:
    coInd = (coInd == nCoDevice-1)? 0 : coInd+1; // to next Pdo device
    dtPdo = (tnow - coInfo[coInd].pdoTimestamp);
    dtHb = tnow - coInfo[coInd].hbTimestamp;
    //printf("PdoMon %d %d %d\n",coInd,dtPdo,dtHb);
    if (dtPdo < 60) return;   // refresh ~ok?
    if (dtHb > 20) return;    // Hb NOT ok?
    nextResetIndex(0,coInd);    // init
    // fall-through
  case SETTIME:
    canId = coInfo[coInd].canId;  // add 22.09.11
    resetInd = nextResetIndex(1,coInd);
    if (resetInd == at_limit) { stateFlag = QUERYDEV; break; } 
    ptInd = coInfo[coInd].resetId[resetInd];
    SetPdoTimeInt(ptInd);  // note dtPdo,dtHb not meaningful except on fallthrough case !!
    cdilog("Pdo ReInit CanId %d : %d %d ,dt %d %d\n",ptInd, resetInd, dtPdo, dtHb);
    coInfo[coInd].pdoTimestamp = tnow;  // inhibit repeats
    stateFlag = RECVTIME;
    interim = tnow;
    break;
  case RECVTIME:  
    if (rcCallback != not_ready) {
      stateFlag = SETTIME;  // next
      if (rcCallback !=0) {
	nterr++;
	printf("%s SetTime Fails: %d  %d\n",devAddress,rcCallback,nterr);
      }
    }  
    if (tnow > interim + 1000) stateFlag = SETTIME;
    break;
  case QFIN:
    break;  
  default:  
    printf("PdoMonitorCycle Bad Case %d\n",stateFlag);
  } 
}
int nextResetIndex(int init, int coInd) {
  static int rind;
  if (init==0) { rind = -1; return rind; }
  rind++;
  while (coInfo[coInd].resetId[rind] < 0) {
    rind++;
    if (rind==4) return at_limit;
  }
  return rind;
}
// each callback sets timestamp
int SubscribePdo(int devPdoInd, int coInd) {
  int rc;
  if (devPdoInd >= nDevice) return at_limit;
  strcpy(devAddress,tineAddress);
  strcat(devAddress,(char*)&(devlist[devPdoInd]));
  pdOut.dArrayLength=1;
  pdOut.dFormat = CF_LONG;  // valid for i8,i16,i32 (?)
  pdOut.data.lptr = &pdoData;
  rc = AttachLinkEx(devAddress,"RECV",&pdOut,0,CA_READ,5000,CallbackPdo,CM_POLL,coInd);
  printf("Pdo %s  AttLnk %d %d rets %d\n",devAddress,devPdoInd,coInd,rc);
  return rc;
}
void CallbackPdo(int id,int rc)  {
  if (rc==0 && (0 <= id && id < nDevice)) {
    //printf("CallbackPdo %d %d\n",id,pdoData);
    coInfo[id].pdoTimestamp = time(NULL);
  }
  else  printf("CallbackPdo id  err %d  %d\n",id, rc);
}  

int SubscribeHb(int hbInd, int coInd) { 
  int rc;
  hbOut.dArrayLength = 1;
  hbOut.dFormat = CF_LONG;
  hbOut.data.lptr = &hbData;
  strcpy(devAddress,tineAddress);
  strcat(devAddress,(char*)&(devlist[hbInd]));  // call-back index should match that for Pdo Device
  rc = AttachLinkEx(devAddress,"RECV",&hbOut,0,CA_READ,5000,CallbackHb,CM_POLL,coInd);
  printf("HB %s  AttLnk %d %d rets %d\n",devAddress,hbInd,coInd,rc);
  return rc;
}
void CallbackHb(int id,int rc)  {
  if (rc==0 && (0 <= id && id < nDevice)) {
    //printf("CallbackHb %d val= %d\n",id, hbOut.data.lptr[0]);
    coInfo[id].hbTimestamp = time(NULL);
  }
  else  printf("CallbackHb id  err %d  %d\n",id, rc);
}  

