/** \file cdiTineLib.c
 * \brief CDI-TINE API library routines
 *      
 */
#ifndef NO_DOXYGEN
#include "project.def"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#ifdef UNIX
#include <dlfcn.h>
#endif
#include "tine.h"
#include "cdi.h"
#include "tinecdi.h"
#include "cdiListener.h"
#include "cdidb.h"
#endif /* NO_DOXYGEN */

extern char *lineinfFieldNames[];
extern int nLineinfFieldNames;
extern int gSimulationFlag;
extern int gNumBitfieldDevices;
extern char cdiDeviceFilter[64];
int isPastDeviceFilter(char* str);
CDI_EXPORT int cdiDebugLevel = 0;
CDI_EXPORT int cdiBusPlugDebugLevel = 0;
int cdiStandAlone = FALSE;
int gCdiIdleState = 0; /* not idle */
int gSetToIdleState = FALSE;
int gDefaultQueryPollingRate = DEFAULT_QUERY_POLLING_INTERVAL;
static BYTE cdiCurrentByteMask[8];
static int cdiHasByteMask = 0;
static int devCounter = 0;
char CdiDBpath[132];
char CdiLogPath[132];
int errListMemPoolSize = MAX_ERROR_LIST_SIZE;
int *errListMemPool = NULL;
NAME32 *nameListMemPool = NULL;
int gMaxTemplateLength = 1;
char gCdiExportName[EXPORT_NAME_SIZE] = "";
char gCdiExportContext[CONTEXT_NAME_SIZE];
char gCdiExportPrefix[96];
int gCdiExportPrefixLength = 0;
int gCdiEqmInitialized = FALSE;
int gCdiEqmInitDone = FALSE;
int gCdiHasXRead = FALSE;
int gNumHistories = 0;
int isCdiStandAlone(void)
{
  char *ptr = getenv("CDI_STANDALONE");
  if (ptr == NULL || strnicmp(ptr,"TRUE",4)) return FALSE;
  cdilog("using cdi in standalone mode");
  return TRUE;
}
int gGenerateAlarms = TRUE;
int gIoErrorThreshold = 0;
int isCdiGenerateAlarms(void)
{
  char *ptr = getenv("CDI_GENERATEALARMS");
  if (ptr != NULL && strnicmp(ptr,"FALSE",5) == 0)
  {
    cdilog("cdi alarm generation turned off by environment");
    gGenerateAlarms = FALSE;
  }
  return gGenerateAlarms;
}
int getCdiIoErrorThresholdFromEnvironment(void)
{
  char *ptr = getenv("CDI_ALARMTHRESHOLD");
  int at = 0;
  if (ptr != NULL)
  {
    if ((at=atoi(ptr)) > 0)
    {
      cdilog("cdi alarm threshold set by environment");
      cdiSetHardwareAlarmThreshold(at);
    }
  }
  return gGenerateAlarms;
}
int gCdiEnforceMcaAccess = FALSE;
int isCdiEnforceMcaAccess(void)
{
  char *ptr = getenv("CDI_FORCE_MCA_ACCESS");
  if (ptr != NULL && strnicmp(ptr,"TRUE",4) == 0)
  {
    cdilog("cdi force mca access turned ON by environment");
    gCdiEnforceMcaAccess = TRUE;
  }
  return gCdiEnforceMcaAccess;
}
CDI_EXPORT void cdiSetEnforceMcaAccess(int value)
{
  gCdiEnforceMcaAccess = value;
  cdilog("cdi force mca access turned %s by API",value ? "ON" : "OFF");
}
int isCdiAllowRemoteManagement(void)
{
  int value = FALSE;
  char *ptr = getenv("CDI_ALLOW_REMOTE_MANAGMENT");
  if (ptr != NULL && strnicmp(ptr,"TRUE",4) == 0)
  {
    cdilog("cdi allow remote management turned ON by environment");
    SetAllowRemoteManagement(TRUE);
    value = TRUE;
  }
  return value;
}
CDI_EXPORT void cdiSetAllowRemoteManagement(int value)
{
  cdilog("cdi allow remote management turned %s by API",value ? "ON" : "OFF");
  SetAllowRemoteManagement(value);
}
void cdiAllocateDeviceStringBuffer(void)
{
  if (cdiDevStringBuffer != NULL) return;
  cdiDevStringBufferSize = 4096;
  if ((cdiDevStringBuffer=calloc(cdiDevStringBufferSize,1)) == NULL)
  {
    cdilog("no memory to allocate the cdi device string buffer space!");
    exit(1);
  }
}
int cdiSystemInitialization(void)
{
  static int cc = database_not_loaded;
  static int initDone = FALSE;
  static int inside = FALSE;
  int force_mca = FALSE;
  if (initDone) return cc;
  if (inside) return operation_busy;
  inside = TRUE;
  fixCdiRepository();
  isCdiAllowRemoteManagement();
  force_mca = isCdiEnforceMcaAccess();
  SetEnforceMultiChannelAccess(force_mca);
  if (force_mca) cdilog("cdi is enabled to force multi-channel acquisition of 'extended' properties");
  if ((cc=readCdiManifest()) == 0) 
  {
    cdiStandAlone = isCdiStandAlone();
    if (!cdiStandAlone) initTineCDI();
    readCdiDataBase();
  }
  if (isCdiGenerateAlarms()) 
  {
    cdilog("cdi is enabled to generate bus alarms on bus error");
    cdilog("the alarm threshold is set to %d consecutive io errors at the same address",gIoErrorThreshold);
  }
  cdilog("cdi library version : %s",cdiGetVersionAsString());
  readCdiSchedulingInfo();
  registerDbStructs();
  initDone = TRUE;
  inside = FALSE;
  return cc;
}  
CDI_EXPORT BYTE *cdiVersion(void)
{
  static BYTE ver[4];
  ver[0] = CDI_MAJOR_VERSION; ver[1] = CDI_MINOR_VERSION;
  ver[2] = CDI_VERSION_REVISION/256; ver[3] = CDI_VERSION_REVISION%256;
  return ver;
}
CDI_EXPORT char *cdiGetVersionAsString(void)
{
  static char s[32];
  BYTE *ver = cdiVersion();
  sprintf(s,"%d.%02d.%04d",ver[0],ver[1],ver[2]*256 + ver[3]);
  return s;
}
time_t gCdiStartupTime = 0;
CDI_EXPORT int cdiInitialize(void)
{ 
  gCdiStartupTime = time(NULL);
  cdiSystemInitialization();
  cdiInitFlag = 1;
  return 0;
}
CDI_EXPORT int cdiGetGenerateAlarms(void)
{
  return gGenerateAlarms;
}
CDI_EXPORT void cdiSetGenerateAlarms(int value)
{
  gGenerateAlarms = value;
}
CDI_EXPORT void cdiSetHardwareAlarmThreshold(int value)
{
  if (value < 0) value = 0;
  gIoErrorThreshold = value;
  cdilog("cdi alarm threshold set to %d",value);
}
CDI_EXPORT int cdiGetHardwareAlarmThreshold(void)
{
  return gIoErrorThreshold;
}
CDI_EXPORT void rotateLogFile(void)
{
  static char fn[148] = "", fb[148] = "";
  if (strlen(fn) == 0) sprintf(fn,"%scdi.log",CdiLogPath);
  if (strlen(fb) == 0) sprintf(fb,"%scdi.bak",CdiLogPath);
  remove(fb);
  rename(fn,fb);
  return;
}
cdi_mutex_t logMutex;
int cdilogInit(void)
{
  return cdiMutexInit(&logMutex, -1);
}
CDI_EXPORT void cdilog(char *text,...)
{ /* TODO:  there's no check against overwriting the (enormously long) string buffer 's' */
# define CDILOG_MAXLINELENGTH (1024 * 5)
  double ts = makeDataTimeStamp();
  FILE *fp = NULL, *fpb = NULL;
  long flen=0;
  char s[CDILOG_MAXLINELENGTH], *ptr;
  int logged = FALSE;
  static char fn[148] = "";
  static char sufn[148] = "";
  static int init_done = FALSE;
  va_list args;

  if (strlen(fn) == 0) sprintf(fn,"%scdi.log",CdiLogPath);
  if (strlen(sufn) == 0) sprintf(sufn,"%scdi-startup.log",CdiLogPath);
  if (!init_done)
  {
    cdilogInit();
    remove(sufn);
    init_done = TRUE;
  }

  if (cdiMutexLock(&logMutex) != 0) 
  {
    printf("cannot get log file mutex !");
    return;
  }
  if ((fp=fopen(fn,"a")) == NULL)
  {
    perror(fn);
    if ((ptr=getenv("CDI_HOME")) == NULL) ptr = getenv("CDIDB");
    printf("CDI_HOME set to : %s\n",(ptr == NULL) ? "(not set)" : ptr);
    if ((ptr=getenv("FEC_HOME")) == NULL) ptr = getenv("FECDB");
    printf("FEC_HOME set to : %s\n",(ptr == NULL) ? "(not set)" : ptr);
    printf("please check consistency\n");
    goto error;
  }
  va_start(args,text);
  sprintf(s,"%s : ",getDataTimeString(ts,0));
  vsprintf(&s[strlen(s)],text,args);
  va_end(args);

  if (s[strlen(s)-1] != '\n') strcat(s,"\n");
  fprintf(fp,"%s",s);
 
  fseek(fp, 0L, SEEK_END);
  flen = ftell(fp);
  logged = TRUE;

  if (cdiInitFlag == 0 || gCdiEqmInitialized == FALSE)
  { /* not yet initialized */
    FILE *sufp = NULL;
    if ((sufp=fopen(sufn,"a")) != NULL)
    {
      fprintf(sufp,"%s",s);
      fclose(sufp);
    }
  }
error:
  if (fp) fclose(fp);
  if (flen > 64000) rotateLogFile();
  cdiMutexUnlock(&logMutex);
  if (logged && cdiDebugLevel) dbglog(s);
  return;
}

int cdiDumpBusList(int line,int i2,int i3,int i4)
{
  int i;
  i2 = i2; i3 = i3; i4 = i4;

  for (i=0; i<numBusList; i++)
  {
    ttyoutput("\tbus name %.16s (id %d)",busList[i].busName,busList[i].busType);
  }
  return 0;
}

int cdiDumpDevices(int line,int i2,int i3,int i4)
{
  int start, stop;
  int i, j, idx, cc = 0;
  cdiHashXRefStruct *xref;

  i2 = i2; i3 = i3; i4 = i4;
  if (line < 1) 
  {
    start = 0; stop = numConnectedHandles - 1;
  }
  else
  {
    start = stop = line - 1;
  }

  for (i=start; i<=stop; i++)
  {
    if (busHandle[i] == NULL) 
    {
      dbglog("bus line %d : not configured",i);
      continue;
    }
    if (busHandle[i]->devHashSize == 0) /* no hash table ...*/
    {
      dbglog("bus line %d : not hashed (%d devices)",i,busHandle[i]->numberDevices);
      for (j=0; j<busHandle[i]->numberDevices; j++)
      {
        ttyoutput("\tdevice %d : %s",j,busHandle[i]->pDevList[j].devName);
      }
    }
    else
    {
      dbglog("bus line %d : hashed (%d devices)",i,busHandle[i]->numberDevices);
      for (idx = 0; idx < busHandle[i]->devHashSize; idx++)
      {
        for (xref=busHandle[i]->devHashList[idx]; xref; xref=xref->nxt)
        {
          j = xref->cdiIndex;
          ttyoutput("\tdevice %d (hash %d, line %d srv dev %d): %.32s (%d msec)",
            j,idx,xref->cdiLine,xref->devNumber,busHandle[i]->pDevList[j].devName,
            busHandle[i]->pDevList[j].interval);
        }
      }
    }
  }
  return 0;
}

CDI_EXPORT void cdiSetDataByteMask(void *dataMask,int sizeInBytes)
{
  memcpy(cdiCurrentByteMask,dataMask,MIN(8,sizeInBytes));
  cdiHasByteMask = TRUE;
}
CDI_EXPORT void *cdiGetDataByteMask(void)
{
  void *bmptr = NULL;
  if (cdiHasByteMask) bmptr = cdiCurrentByteMask; 
  cdiHasByteMask = 0; /* always clear */
  return bmptr;
}
CDI_EXPORT int cdiCloseLink(int id)
{
  struct timespec ts;
  int i;
  if (cdi_isValidRequestIndex(id)) 
  {
    cdi_CancelRequest(id);
    ts.tv_sec = 0;
    ts.tv_nsec = 10000000;
    for (i=0; !cdi_RequestTerminated(id) && i<100; i++) cdi_nanosleep(&ts,NULL);
    if (i == 100) cdilog("close link %d did not terminate properly",id);
    return 0;
  }
  return invalid_index;
}
CDI_EXPORT void cdiCloseLinks(void)
{
  int i, j;

  cdilog("stopping all active links");
  for (i=0; i<numConnectedHandles; i++)
  {
    if ( busHandle[i] == NULL ) continue;
    for (j=0; j<busHandle[i]->numberActiveRequest; j++)
      cdi_CancelRequest(busHandle[i]->pReqList[j].requestNumber + 0x10000);
  }
  return;

}
int cdi_assignBitfieldToDevice(char *devName, cdiBitfieldXRef *xbf)
{
  int cc = 0, cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (xbf == NULL) ccerr(out_of_server_memory);
  if (findDevice(devName, &cdiLine, &cdiIndex) < 0)
  {
    if (cdiDebugLevel) dbglog("assignBitfieldToDevice : invalid device name: %s",devName);
    ccerr(device_not_connected);
  }
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  /* units field contains the bf tag name */
  strncpy(pDev->units,xbf->bf->name,16);
  strcpy(pDev->precision,"BITFIELD");
err:
  if (cc) cdilog("cdi_assignBitfieldToDevice %s : %s",devName,erlst[cc]);
  return cc;
}
DBITFIELD *cdiGetBitfieldFromDevice(char *devName)
{
  int cdiLine, cdiIndex;
  CdiDeviceInfoBlk *pDev;

  if (findDevice(devName, &cdiLine, &cdiIndex) < 0) return NULL;
  pDev = &busHandle[cdiLine]->pDevList[cdiIndex];
  if (strcmp(pDev->precision,"BITFIELD") != 0) return NULL;
  return findBitField(BITFIELD_LCL,pDev->units);
}
cdiBitfieldXRef *gCdiBitfieldXRefLst = NULL;
cdiBitfieldXRef *cdiGetBitfieldXRef(char *devName,DBITFIELD *bf)
{
  cdiBitfieldXRef *xbf=NULL;

  for (xbf=gCdiBitfieldXRefLst; xbf != NULL; xbf=xbf->nxt)
  {
    if (!strnicmp(xbf->device,devName,CDI_DEVICE_NAME_SIZE)) return xbf;
  }
  if ((xbf=(cdiBitfieldXRef *)calloc(1,sizeof(cdiBitfieldXRef))) != NULL)
  {
    xbf->bf = bf;
    strncpy(xbf->device,devName,CDI_DEVICE_NAME_SIZE);
    xbf->nxt = gCdiBitfieldXRefLst;
    gCdiBitfieldXRefLst = xbf;
  }
  return xbf;
}

CDI_EXPORT int cdiRegisterDevice(
  char *devName, int devNumber, char *lngName, char *devDesc,
  int cdiLine, int cdiIndex, int *cdiAddress, int cdiAddressLength,int *devParameters
)
{
  int cc;
  if ((cc=cdi_RegisterDevice(cdiLine,devName,lngName,devDesc,cdiIndex,devNumber,cdiAddress,cdiAddressLength)) != 0) return cc;
  return devParameters == NULL ? 0 : cdi_assignDefaultDeviceParameters(devName,devParameters);
}

CDI_EXPORT int cdiRegisterDeviceNames(void)
{
  int i, n, nlen, ndevs, nredir, nbf = 0;
  NAME32 *devs = NULL;
  char dname[DEVICE_NAME_SIZE+4];
  DBITFIELD *bf;
  struct bitfieldFormat *bff;

  ndevs = cdi_getDeviceNames(NULL,0); /* first get the number of devices */
  if (ndevs > 0)
  {
    if ((devs=(NAME32 *)alloca(ndevs*32)) == NULL) return out_of_local_memory;
    cdi_getDeviceNames(devs,ndevs);
    for (i=0, n=0; i<ndevs; i++) 
    {
      if (IsDiscoveryDevice(devs[i].name)) continue;
      strncpy(dname,devs[i].name,CDI_DEVICE_NAME_SIZE); dname[CDI_DEVICE_NAME_SIZE] = 0;
      RegisterDeviceName(CDIEQM_TAG,dname,n);
      if ((bf=cdiGetBitfieldFromDevice(dname)) != NULL)
      {
        nlen = (int)strlen(dname);
        for (bff=bf->bfs; bff != NULL; bff=bff->nxt)
        {
          sprintf(&dname[nlen],".%.16s",bff->field);
          RegisterDeviceName(CDIEQM_TAG,dname,ndevs + nbf);
          nbf++; /* running count of the bitfields registered */
        }
      }
      n++;
    }
  }
  nredir = cdi_getRedirectedDeviceNames(NULL,0);
  if (nredir == 0) return 0;
  i += nbf; /* in case there were bitfields */
  if (nredir > ndevs)
  {
    if ((devs=(NAME32 *)alloca(nredir*32)) == NULL) return out_of_local_memory;
  }
  cdi_getRedirectedDeviceNames(devs,nredir);
  for (i=0; i<nredir; i++)
  {
    strncpy(dname,devs[i].name,32); dname[32] = 0;
    RegisterDeviceName(CDIEQM_TAG,dname,i+ndevs);
  }
  return 0;
}

ExtCdiProperty *extCdiPrps;
int gNumExtCdiProperties = 0;
char *cdiDevStringBuffer = NULL;
int cdiDevStringBufferSize = 0;

int splitCdiDeviceName(char *cdiname,char *dev,char *prp)
{
  char cname[DEVICE_NAME_SIZE+4], *c, *tok = ".";
  if (cdiname == NULL || dev == NULL || prp == NULL) return argument_list_error;
  strncpy(cname,cdiname,DEVICE_NAME_SIZE); cname[DEVICE_NAME_SIZE] = 0;
  if ((c=strstr(cname,tok)) == NULL)
  { /* not splittable */
    strncpy(prp,cname,DEVICE_NAME_SIZE); prp[DEVICE_NAME_SIZE] = 0;
    strncpy(dev,cname,DEVICE_NAME_SIZE);
  }
  else
  {
    *c++ = 0;
    strncpy(prp,c,DEVICE_NAME_SIZE); prp[DEVICE_NAME_SIZE] = 0;
    strncpy(dev,cname,DEVICE_NAME_SIZE);
  }
  return 0;
}
CDI_EXPORT int cdiRegisterExtendedProperties(void)
{
  int i, k, n, ndevs, nxprps = 0, plen = 1, prpid, cc = 0, dfmt=0, atype, maxdevs=0;
  int cdinr_last, cdinr_this, ln, ix, iscont = TRUE;
  short accRead = gCdiHasXRead ? CA_XREAD : CA_READ;
  NAME32 *devs = NULL; /* cdi device name list */
  NAME64 *pdevs = NULL; /* for the call into AssignDeviceListToProperty() */
  char dname[DEVICE_NAME_SIZE+4],pname[PROPERTY_NAME_SIZE+4],tname[PROPERTY_NAME_SIZE+4],devDesc[64];
  DTYPE dout, din;
  DTYPEZERO(dout); DTYPEZERO(din);
  ndevs = cdi_getDeviceNames(NULL,0); /* first get the number of devices */
  cdilog("looking for extended properties");
  if (ndevs > 0)
  {
    if ((extCdiPrps=(ExtCdiProperty *)calloc(ndevs,sizeof(ExtCdiProperty))) == NULL)
      ccerr(out_of_local_memory);
    if ((devs=(NAME32 *)malloc(ndevs*32)) == NULL) ccerr(out_of_local_memory);
    if ((pdevs=(NAME64 *)malloc(ndevs*64)) == NULL) ccerr(out_of_local_memory);
    cdi_getDeviceNames(devs,ndevs);
    for (i=0; i<ndevs; i++) 
    {
      if (IsDiscoveryDevice(devs[i].name)) continue;
      cdi_getOutputSizeForDevice(devs[i].name,&plen);
      cdi_getFormatForDevice(devs[i].name,&dfmt);
      splitCdiDeviceName(devs[i].name,dname,pname);
      /* is property registered ? */
      if ((prpid=GetPropertyId(CDIEQM_TAG,pname)) < 0)
      { /* find them all : */
        strncpy(pdevs[0].name,dname,32);
        cdinr_last = findDevice(devs[i].name,&ln,&ix);
        iscont = TRUE;
        for (n=1,k=i+1; k<ndevs; k++)
        {
          splitCdiDeviceName(devs[k].name,dname,tname);
          if (strncmp(pname,tname,PROPERTY_NAME_SIZE)) continue;
          cdinr_this = findDevice(devs[k].name,&ln,&ix);
          if (cdinr_this != cdinr_last + 1) iscont = FALSE;
          cdinr_last = cdinr_this;
          strncpy(pdevs[n].name,dname,32);
          n++;
        }
        if (cdi_getDescriptionForDevice(devs[i].name,devDesc) != 0) strcpy(devDesc,"extended property");
        strncpy(extCdiPrps[nxprps].prp,pname,PROPERTY_NAME_SIZE);
        extCdiPrps[nxprps].ndevs = n;
        extCdiPrps[nxprps].isContiguous = iscont;
        extCdiPrps[nxprps].fmt = CF_FLOAT;
        memset(&dout,0,sizeof(DTYPE));
        if (plen > 1)
        {
          dout.dFormat = getTineFormat((short)dfmt);
          dout.dArrayLength = plen;
          din.dFormat = dout.dFormat;
          din.dArrayLength = plen;
          atype = AT_COLLECTION;
          extCdiPrps[nxprps].isArray = TRUE;
        }
        else
        { /* use multi-channel array if possible */
          plen = n;
          dout.dFormat = CF_FLOAT;
          dout.dArrayLength = plen;
          din.dFormat = CF_FLOAT;
          din.dArrayLength = 1;
          atype = n > 1 ? AT_CHANNEL : AT_SCALAR;
        }
        extCdiPrps[nxprps].id = RegisterPropertyInformation(CDIEQM_TAG,pname,&dout,&din,
                                       (short)(accRead|CA_WRITE),(short)atype,(UINT16)plen,devDesc,
                                       PRP_EXTENDED_OFFSET+nxprps,NULL);
        /* Note: the following call takes a NAME16 list (which should be okay following a split of a 32-char CDI dev name */
        AssignDeviceListToProperty(CDIEQM_TAG,pname,pdevs,n);
        if ((extCdiPrps[nxprps].devs=(NAME32 *)calloc(n,sizeof(NAME32))) == NULL) ccerr(out_of_local_memory);
        for (k=0; k<n; k++) memcpy(extCdiPrps[nxprps].devs[k].name,pdevs[k].name,sizeof(NAME32));
        if (n > maxdevs) maxdevs = n;
        nxprps++;
      }
    }
  }
  gNumExtCdiProperties = nxprps;
  cdilog("registered %d extended properties",nxprps);
  if (cdiDevStringBufferSize < maxdevs*16)
  {
    char *tmp;
    n = maxdevs * 16;
    if ((tmp=realloc(cdiDevStringBuffer,n)) != NULL)
    {
      cdiDevStringBuffer = tmp;
      cdiDevStringBufferSize = n;
    }
  }
err:
  if (devs) free(devs);
  if (pdevs) free(pdevs);
  return cc ;
}

CDI_EXPORT char *cdiGetFullDeviceName(char *devName)
{
  int cdiLine,cdiIndex;
  char *c = cdiGetRedirectedDeviceServer(devName);
  static char localDeviceServerName[96];
  if (c != NULL) return c;
  if (findDevice(devName,&cdiLine,&cdiIndex))
  {
    sprintf(localDeviceServerName,"/%s/%s.CDI/%s",gDeviceContext,gFecName,devName);
    return localDeviceServerName;
  }
  return NULL;
}
NAME16 *cdiBusNames = NULL;
int nCdiBusNames = 0;
char **cdiHardwareTypes;
int nCdiHardwareTypes = 0;
CDI_EXPORT NAME16 *cdiGetBusNames(void)
{
  int i;
  if (cdiBusNames == NULL)
  {
    nCdiBusNames = cdi_getBusNames(NULL,0);
    if ((cdiBusNames=(NAME16 *)calloc(nCdiBusNames,16)) == NULL)
    {
      cdilog("cdiGetBusNames : out of memory!");
      return NULL;
    }
    if ((cdiHardwareTypes=(char **)malloc((nCdiBusNames+1)*sizeof(char *))) == NULL)
    {
      cdilog("cdiGetBusNames : out of memory!");
      return NULL;
    }
    nCdiHardwareTypes = nCdiBusNames + 1;
    cdi_getBusNames(cdiBusNames,nCdiBusNames);
    cdiHardwareTypes[0] = "CDI"; /* zeroth position is general ! */
    for (i=0; i<nCdiBusNames; i++)
    {
      cdiHardwareTypes[i+1] = cdiBusNames[i].name;
    }
  }
  return cdiBusNames;
}
CDI_EXPORT char *getCdiLinkName(char *devName,char *cdiDevName)
{
  int i;
  char *lclDevName;
  if (devName == NULL || strlen(devName) < 8) return NULL;
  cdiGetBusNames(); /* make sure it's filled in ... */
  if (!strnicmp(&devName[1],"LOCAL",5) && (lclDevName=strchr(&devName[1],'/')) != NULL)
  {
    lclDevName++;
    for (i=0; i<nCdiHardwareTypes; i++)
    {
      if (!strnicmp(lclDevName,cdiHardwareTypes[i],strlen(cdiHardwareTypes[i])))
      {
        strncpy(cdiDevName,&lclDevName[strlen(cdiHardwareTypes[i])+1],1024);
        return cdiDevName; /* return the pointer given */
      }
    }
  }
  if (gCdiExportPrefixLength > 0 && !strnicmp(gCdiExportPrefix,devName,gCdiExportPrefixLength))
  {
    strncpy(cdiDevName,&devName[gCdiExportPrefixLength],1024);
    return cdiDevName; /* return the pointer given */
  }
  return NULL;
}
#define cdiParsePropertyForAccess(prp) (\
  !strcmp(prp,"RECV") ? cdiReadFlag :\
  !strcmp(prp,"SEND") ? cdiWriteFlag :\
  !strcmp(prp,"RECV.SEND.ATOM") ? cdiReadWriteFlag :\
  !strcmp(prp,"RECV.SEND") ? cdiReadWriteFlag :\
  !strcmp(prp,"SEND.RECV.ATOM") ? cdiWriteReadFlag :\
  !strcmp(prp,"SEND.RECV") ? cdiWriteReadFlag :\
  !strcmp(prp,"SEND.SEND.ATOM") ? cdiWriteWriteFlag :\
  !strcmp(prp,"SEND.SEND") ? cdiWriteWriteFlag :\
  !strcmp(prp,"SEND.RECV.SEND.ATOM") ? cdiWriteReadWriteFlag : \
  !strcmp(prp,"SEND.RECV.SEND") ? cdiWriteReadWriteFlag : \
  !strcmp(prp,"RECV.RECV.ATOM") ? cdiReadReadFlag :\
  !strcmp(prp,"RECV.RECV") ? cdiReadReadFlag :\
  !strcmp(prp,"SEND.RECV.RECV.ATOM") ? cdiWriteReadReadFlag :\
  !strcmp(prp,"SEND.RECV.RECV") ? cdiWriteReadReadFlag : -1\
  )

short getBaseFormat(short tineFormat,int *needsErrList,int *needsNameList)
{
  *needsNameList = FALSE;
  *needsErrList = FALSE;
  switch (tineFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL: *needsNameList = TRUE;
    case CF_DBLDBL: *needsErrList = TRUE;
    case CF_DOUBLE: return CF_DOUBLE;
    case CF_NAME16FI: *needsNameList = TRUE;
    case CF_FLTINT: *needsErrList = TRUE;
    case CF_FLOAT:  return CF_FLOAT;
    case CF_BITFIELD16:
    case CF_INT16: return CF_INT16;
    case CF_NAME32I: *needsErrList = TRUE;
    case CF_NAME32: return CF_NAME32;
    case CF_NAME16II: *needsNameList = TRUE;
    case CF_LNGINT: *needsErrList = TRUE;
    case CF_BITFIELD32:
    case CF_INT32: return CF_INT32;
    case CF_BITFIELD8:
    case CF_INT8:  return CF_INT8;
    case CF_TEXT:   return CF_TEXT;
    default: break;
  }
  return -1;
}

short getCdiBaseFormat(short tineFormat)
{
  switch (tineFormat)
  {
    case CF_NAME16DBLDBL:
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
    case CF_DBLDBL:
    case CF_DOUBLE: return cdiDouble;
    case CF_NAME16FI:
    case CF_FLTINT:
    case CF_FLOAT:  return cdiFloat;
    case CF_BITFIELD16:
    case CF_INT16: return cdiShort;
    case CF_NAME16II:
    case CF_LNGINT:
    case CF_BITFIELD32:
    case CF_INT32: return cdiInteger;
    case CF_BITFIELD8:
    case CF_INT8:  return cdiByte;
    case CF_TEXT:   return cdiChar;
    case CF_NAME32I:
    case CF_NAME32: return cdiChar32;
    default: break;
  }
  return -1;
}
CDI_EXPORT short getCdiFormat(short tineFormat)
{
  switch (tineFormat)
  {
    case CF_DOUBLE: return cdiDouble;
    case CF_FLOAT:  return cdiFloat;
    case CF_INT16: return cdiShort;
    case CF_INT32: return cdiInteger;
    case CF_INT8:  return cdiByte;
    case CF_TEXT:   return cdiChar;
    case CF_NAME32: return cdiChar32;
    default: break;
  }
  return -1;
}
CDI_EXPORT short getTineFormat(short cdiFormat)
{
  switch (cdiFormat)
  {
    case cdiDouble: return CF_DOUBLE;
    case cdiFloat:  return CF_FLOAT;
    case cdiShort: return CF_INT16;
    case cdiInteger: return CF_INT32;
    case cdiByte:  return CF_INT8;
    case cdiChar:   return CF_TEXT;
    case cdiChar32: return CF_NAME32;
    default: break;
  }
  return CF_NULL;
}
CDI_EXPORT int getCdiAllowedAccess(char *accessString)
{
  char access[256], *s=NULL, *c=NULL, *tok = " +|";
  int allowedAccess = 0;
  if (accessString == NULL || strlen(accessString) == 0) return 0xffff;
  strncpy(access,accessString,255); access[255] = 0;
  strupr(access);
  for (s=strtok_r(access,tok,&c); s != NULL; s=strtok_r(NULL,tok,&c))
  {
    if (!strcmp(s,"READ") || !strcmp(s,"RD")) allowedAccess |= 0x01;
    if (!strcmp(s,"WRITE") || !strcmp(s,"WR")) allowedAccess |= 0x02;
    if (!strcmp(s,"READWRITE") || !strcmp(s,"RDWR")) allowedAccess |= 0x04;
    if (!strcmp(s,"WRITEREAD") || !strcmp(s,"WRRD")) allowedAccess |= 0x08;
    if (!strcmp(s,"WRITEWRITE") || !strcmp(s,"WRWR")) allowedAccess |= 0x10;
    if (!strcmp(s,"WRITEREADWRITE") || !strcmp(s,"WRRDWR")) allowedAccess |= 0x20;
    if (!strcmp(s,"READREAD") || !strcmp(s,"RDRD")) allowedAccess |= 0x40;
    if (!strcmp(s,"WRITEREADREAD") || !strcmp(s,"WRRDRD")) allowedAccess |= 0x80;
  }
  if ((allowedAccess & 0x03) == 0x03) allowedAccess = 0xffff; /* READ & WRITE separately => everything's okay */
  return allowedAccess;
}
int cdiRemapDataSets(DTYPE *dout,BYTE *dataout,int *errLst,NAME32 *nameLst)
{
  int i, namlen = 0, cpysiz = 32, fmtsiz = CF_NULL, lstsiz = errListMemPoolSize;
  int *errout = NULL, *itmp = NULL;
  NAME32 *nameout = NULL, *ntmp = NULL;
  double dStatus;
  BYTE *dataptr = NULL;
  char buf[64];

  if (dout == NULL) return argument_list_error;
  fmtsiz = fmtsizeof(dout->dFormat);
  dataptr = dout->data.bptr;

  if (lstsiz < (int)dout->dArrayLength) lstsiz = (int)dout->dArrayLength;
  if (errListMemPool == NULL)
  { /* has not been allocated yet */
    if ((errListMemPool=(int *)calloc(lstsiz,sizeof(int))) == NULL) return out_of_local_memory;
  }
  else if (lstsiz > errListMemPoolSize)
  {     
    if ((itmp=(int *)realloc(errListMemPool,lstsiz*sizeof(int))) == NULL) return out_of_local_memory;    
    errListMemPool = itmp;
  }
  if (nameListMemPool == NULL)
  { /* has not been allocated yet */
    if ((nameListMemPool=(NAME32 *)calloc(errListMemPoolSize,sizeof(NAME32))) == NULL) return out_of_local_memory;
  }
  else if (lstsiz > errListMemPoolSize)
  {
    if ((ntmp=(NAME32 *)realloc(nameListMemPool,lstsiz*sizeof(NAME32))) == NULL) return out_of_local_memory;
    nameListMemPool = ntmp;
  }
  errListMemPoolSize = lstsiz;

  /* copy error array out to the errList memory pool (must remap e.g. INTINT structs into 2 arrays) */
  if (errLst != NULL)
  {
    errout = errListMemPool;
    memcpy(errout,errLst,dout->dArrayLength*sizeof(int));
  }
  if (nameLst != NULL)
  {
    nameout = nameListMemPool;
    memcpy(nameout,nameLst,dout->dArrayLength*sizeof(NAME32));
  }
  switch (dout->dFormat)
  {
    case CF_NAME16DBLDBL:
      cpysiz = 16;
    case CF_NAME32DBLDBL:
    case CF_NAME64DBLDBL:
      if (nameout == NULL) return argument_list_error;
      namlen = fmtsiz - 16;
    case CF_DBLDBL:
      if (errout == NULL) return argument_list_error;
      for (i=dout->dArrayLength-1; i >= 0;i--)
      { /* dataptr = dataout => traverse the list in reverse */
        dStatus = (double)errout[i];        
        memcpy(dataptr + i*fmtsiz + namlen + 8,&dStatus,8);  /* status */
        memcpy(dataptr + i*fmtsiz + namlen,dataout + i*8,8); /* data */
        if (namlen > 0) memcpy(dataptr + i*fmtsiz,&nameout[i],cpysiz); /* name */
      }
    case CF_NAME16FI:
    case CF_NAME16II:
      if (nameout == NULL) return argument_list_error;
      cpysiz = 16;
      namlen = 16;
    case CF_FLTINT:
    case CF_LNGINT:
      if (errout == NULL) return argument_list_error;
      for (i=dout->dArrayLength-1; i >= 0;i--)
      { /* dataptr = dataout => traverse the list in reverse */
        memcpy(dataptr + i*fmtsiz + namlen + 4,&errout[i],4);  /* status */
        memcpy(dataptr + i*fmtsiz + namlen,dataout + i*4,4); /* data */
        if (namlen > 0) memcpy(dataptr + i*fmtsiz,&nameout[i],cpysiz); /* name */
      }
      break;
    case CF_NAME32I:
      if (errout == NULL) return argument_list_error;
      for (i=dout->dArrayLength-1; i >= 0;i--)
      { /* dataptr = dataout => traverse the list in reverse */
        memcpy(dataptr + i*fmtsiz + 32,&errout[i],4);  /* status */
        memcpy(buf,dataout + i*32,32);
        /* memcpy(dataptr + i*fmtsiz,dataout + i*32,32); < this doesn't compile correctly on VC 6.0 @ 8byte boundaries */
        memcpy(dataptr + i*fmtsiz,buf,32); /* data */
      }
      break;
    default:
      break;
  }
  return 0;
}
/*
 * synchronous call :
 * e.g. ExecLinkEx("/LOCALHOST/CDI/CanTestDevNumber_4/CanParamSet_0","WRITE",NULL,&din,CA_WRITE,3000);
 * context LOCAL => either cdiExecLinkEx() or execLocalLink()
 * if group == CDI, or explicitly CAN, SEDAC, etc. => jump to cdiExecLinkEx()
 */
CDI_EXPORT int cdiExecLinkEx(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  UINT16 timeout     /* timeout in msec. */
)
{
  char *c=NULL,devPropertyStr[PROPERTY_NAME_SIZE+4];
  char *dn = NULL,*pn = NULL;
  int *cdiParameters = NULL;
  void *cdiDataMask = NULL;
  int cc = 0, id, rc, cdiAccessFlag, cdiReturnCondition = 0, cdiPollingFlag = 0, cdiNeedsCalibration = 0;
  UINT32 doutlen=0, dinlen=0, doutlenInBytes=0;
  void *doutptr = NULL, *dinptr = NULL;
  short dinFormat=0, doutFormat=0, baseFormat = CF_NULL;
  int hasErrLst = FALSE, hasNameLst = FALSE;
  int *errLst = NULL, *errLstWr = NULL, *errLstAlloc = NULL, *perrLst = NULL;
  NAME32 *nameLst = NULL;
  CdiRequestInfoBlk *pReq;
  static DTYPE dinZero;
  static short dinZeroValue = 0;

  if (cdiDebugLevel) dbglog("cdiExecLinkEx: %s %s with timeout %d",devName,devProperty,timeout);
  if (devName == NULL || strlen(devName) == 0) ccerr(argument_list_error);
  if (devProperty == NULL || strlen(devProperty) == 0) ccerr(argument_list_error);
  strncpy(devPropertyStr,devProperty,PROPERTY_NAME_SIZE); devPropertyStr[PROPERTY_NAME_SIZE] = 0;
  /* parse devName for hdw device name and hdw parameter set */
  if (cdiDevStringBuffer == NULL) cdiAllocateDeviceStringBuffer();
  strncpy(cdiDevStringBuffer,devName,cdiDevStringBufferSize); 
  cdiDevStringBuffer[cdiDevStringBufferSize-1] = 0;
  remapCdiDeviceAndProperty(cdiDevStringBuffer,devPropertyStr,dout,din,access);
  if ((dn=strtok_r(cdiDevStringBuffer,DELIMITER,&c)) == NULL) ccerr(argument_list_error);
  if ((pn=strtok_r(NULL,"<",&c)) != NULL)
  { 
    if ((c=strchr(pn,'>')) != NULL) *c = 0;
    if ((cdiParameters=cdi_GetParameters(pn)) == NULL) cdilog("input parameter list not found");
  }
  if (dout != NULL && dout->dArrayLength > 0) 
  {
    baseFormat = getBaseFormat(dout->dFormat,&hasErrLst,&hasNameLst);
    doutlen = dout->dArrayLength;
    doutlenInBytes = dout->dArrayLength*fmtsizeof(baseFormat);
    doutptr = dout->data.vptr;
    doutFormat = getCdiFormat(baseFormat);
    if ( doutFormat < 0 ) ccerr(illegal_format)
    if (hasErrLst) 
    {
      errLst = (int *)((int *)doutptr + doutlen);
      if (dout->dFormat == CF_NAME32I) errLst = (int *)((BYTE *)doutptr + doutlen * 32);
    }
    perrLst = errLst;
  }
  if ((c=strstr(devPropertyStr,"SEND.CLBR")) != NULL)
  { /* does SEND call need reverse calibration ? */
    cdiReturnCondition |= INPUT_CALIBRATED;
    *c = 0;
    strcat(devPropertyStr,"SEND");
    if (din == NULL || din->dArrayLength == 0)
    { /* no input given with call -> send a 0 */
      /* TODO: dinZero assumes bus format = CF_SHORT */
      memset(&dinZero,0,sizeof(DTYPE));
      dinZero.dArrayLength = 1;
      dinZero.dFormat = CF_SHORT;
      dinZero.data.sptr = &dinZeroValue;
      din = &dinZero;
    }
  }
  if (din != NULL && din->dArrayLength > 0) 
  { 
    dinlen = din->dArrayLength*fmtsizeof(din->dFormat);
    dinptr = din->data.vptr;
    dinFormat = getCdiFormat(din->dFormat);
    if ( dinFormat < 0 ) ccerr(illegal_format)
  }
  /* is it a calibration property ? */
  if ((c=strstr(devPropertyStr,".CLBR")) != NULL)
  {
    cdiNeedsCalibration = TRUE;
    *c = 0;
  }
  /* has dataMask been set ? */
  /* cdi_applyDataByteMask()  */
  cdiDataMask = cdiGetDataByteMask();
  cdiAccessFlag = cdiParsePropertyForAccess(devPropertyStr);
  if (cdiAccessFlag < 0) ccerr(illegal_property);
  if (cdiAccessFlag > cdiReadFlag && cdiAccessFlag != cdiReadReadFlag && (access & CA_WRITE) == 0) ccerr(illegal_read_write);
  if (cdiAccessFlag == cdiWriteFlag && doutptr && dout->dFormat == CF_INT32) errLstWr = (int *)doutptr;
  if (perrLst == NULL) perrLst = errLstWr;
  if (perrLst == NULL && doutlen > 0)
  { /* if no error list at this stage allocate it ! */
    if ((errLstAlloc=(int *)calloc(doutlen,sizeof(int))) == NULL) ccerr(out_of_local_memory);
    perrLst = errLstAlloc;
    cdiReturnCondition |= FREE_ONCLOSE;
  }
  /* synchronous link will not look for pattern or tolerance ! */
  /* returnCondition = maskChange since there's no callback */
  id = cdi_Request(dn,doutptr,doutFormat,doutlenInBytes,dinptr,dinFormat,dinlen,
                   NULL,cdiDataMask,cdiParameters,cdiAccessFlag,cdiReturnCondition,
                   timeout,NULL,cdiPollingFlag,perrLst, -1, 0);
  
  if (id < 0) 
  {
    if (errLstAlloc != NULL) free(errLstAlloc);
    return -id;
  }
  if ((rc=get_RequestBufferPointer(id, &pReq)) < 0 ) return -rc;
  rc = assignGroupInOutFormatCode(pReq, din != NULL ? din->dFormat : CF_NULL, dout != NULL ? dout->dFormat : CF_NULL);

  rc = cdi_StartRequest(id);
/*  if (rc < 0 && doutlen == 1) cc = hardware_error;*/
  if (rc < 0 ) cc = -rc;
  else if ( rc > 0 ) cc = hardware_error; /* in fact return # of errors for group call  */

  if (cc == 0 && doutlen > 0)
  {
    rc = cdi_CalibrateRequest(id,cdiNeedsCalibration);
    if (rc < 0) cc = -rc;
  }
  if (hasNameLst && doutptr != NULL)
  {
    nameLst = (NAME32 *)((int *)doutptr + doutlen*2);
    cdi_GetRequestDeviceNames(id,nameLst);
  }
  cdi_CancelRequest(id); /* will free errLstAlloc if necessary */
  cdiRemapDataSets(dout,(BYTE *)doutptr,errLst,nameLst);
err:
  if (cdiDebugLevel) dbglog("cdiExecLinkEx: finished with completion code %d",cc);
  if (cc && dout != NULL && dout->dArrayLength > 0 && !isPrimitiveFormat(dout->dFormat)) cc |= CE_SENDDATA;
  return cc;
}

CDI_EXPORT int cdiExecLinkHookFcn(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  UINT16 timeout     /* timeout in msec. */
)
{
  char *dn;
  int cc;

  cc = lclcdieqm(devName,devProperty,dout,din,access);
  if (cc != illegal_property) return cc;
  cc = cdiExecLinkEx(devName,devProperty,dout,din,access,timeout);
  if (cc == server_redirection)
  {
    if ((dn=cdiGetFullDeviceName(devName)) == NULL) return non_existent_fec;
    strncpy(devName,dn,512);
  }
  return cc;
}
/* asynchronous call:
 * e.g. AttachLinkEx("/LOCALHOST/CDI/CanTestDevNumber_4/CanParamSet_0","WRITE",&dout,&din,CA_READ,200,cbCan,CM_REFRESH,id);
 * context LOCAL => cdiAttachLinkEx()
 * if group == CDI, or explicitly CAN, SEDAC, etc. => jump to cdiAttachLinkEx()
 */
CDI_EXPORT int cdiAttachLinkEx(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  void (*callback)(int,int), /* callback routine */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 callbackID  /* callback ID */
)
{
  char *c=NULL;
  char devPropertyStr[PROPERTY_NAME_SIZE+4];
  char *dn = NULL,*pn = NULL;
  int *cdiParameters = NULL;
  void *cdiDataMask = NULL;
  void *cdiDataPatternWindow = NULL;
  int cc = 0, cdiAccessFlag, cdiReturnCondition = -1, cdiPollingFlag = 0, cdiNeedsCalibration = 0, rc;
  UINT32 doutlen=0, dinlen=0, doutlenInBytes=0;
  void *doutptr = NULL, *dinptr = NULL;
  short dinFormat=0, doutFormat=0, baseFormat = CF_NULL;
  int hasErrLst = FALSE, hasNameLst = FALSE, isGrouped = FALSE, isWait = FALSE;
  int *errLst = NULL, *errLstWr = NULL, *errLstAlloc = NULL, *perrLst = NULL;
  CdiRequestInfoBlk *pReq;

  if (devName == NULL || strlen(devName) == 0) ccerr(-argument_list_error);
  if (devProperty == NULL || strlen(devProperty) == 0) ccerr(-argument_list_error);
  strncpy(devPropertyStr,devProperty,PROPERTY_NAME_SIZE); devPropertyStr[PROPERTY_NAME_SIZE] = 0;
  /* parse devName for hdw device name and hdw parameter set */
  if (cdiDevStringBuffer == NULL) cdiAllocateDeviceStringBuffer();
  strncpy(cdiDevStringBuffer,devName,cdiDevStringBufferSize); 
  cdiDevStringBuffer[cdiDevStringBufferSize-1] = 0;
  remapCdiDeviceAndProperty(cdiDevStringBuffer,devPropertyStr,dout,din,access);
  if ((dn=strtok_r(cdiDevStringBuffer,DELIMITER,&c)) == NULL) ccerr(-argument_list_error);
  if ((pn=strtok_r(NULL,"<",&c)) != NULL)
  { 
    if ((c=strchr(pn,'>')) != NULL) *c = 0;
    if ((cdiParameters=cdi_GetParameters(pn)) == NULL) cdilog("input parameter list not found");
  }
  if (dout != NULL && dout->dArrayLength > 0) 
  {
    baseFormat = getBaseFormat(dout->dFormat,&hasErrLst,&hasNameLst);
    doutlen = dout->dArrayLength;
    doutlenInBytes = dout->dArrayLength*fmtsizeof(baseFormat);
    doutptr = dout->data.vptr;
    doutFormat = getCdiFormat(baseFormat);
    if ( doutFormat < 0 ) ccerr(-illegal_format)
    if (hasErrLst) 
    {
      errLst = (int *)((int *)doutptr + doutlen);
      if (dout->dFormat == CF_NAME32I) errLst = (int *)((BYTE *)doutptr + doutlen * 32);
    }
    perrLst = errLst;
  }
  if (din != NULL && din->dArrayLength > 0) 
  {
    /* does SEND call need reverse calibration ? */
    dinlen = din->dArrayLength*fmtsizeof(din->dFormat);
    dinptr = din->data.vptr;
    dinFormat = getCdiFormat(din->dFormat);
    if ( dinFormat < 0 ) ccerr(-illegal_format)
  }
  /* is it a calibration property ? */
  if ((c=strstr(devPropertyStr,".CLBR")) != NULL)
  {
    cdiNeedsCalibration = TRUE;
    *c = 0;
  }
  if (mode & CM_GROUPED) isGrouped = TRUE;
  if (mode & CM_WAIT) isWait = TRUE;
  /* has dataMask been set ? */
  /* cdi_applyDataByteMask()  */
  cdiDataMask = cdiGetDataByteMask();
  /* is there a data pattern or tolerance ? */
  if (mode & CM_SEEKPATTERN) cdiReturnCondition = maskCompEqual;
  else if (mode & CM_MISSPATTERN) cdiReturnCondition = maskCompNotEqual;
  else if (mode & CM_NODUPLICATES) cdiReturnCondition = maskChangeWindow;
  if (cdiReturnCondition < 0)
  {
    cdiReturnCondition = 0;
  }
  else
  {
    if ((cdiDataPatternWindow=calloc(doutlenInBytes,1)) == NULL) ccerr(-out_of_local_memory);
    memcpy(cdiDataPatternWindow,dout->data.vptr,doutlenInBytes);
  }
  /* accessFlag */
  cdiAccessFlag = cdiParsePropertyForAccess(devPropertyStr);
  if (cdiAccessFlag < 0) ccerr(-illegal_property);
  if (cdiAccessFlag > cdiReadFlag && cdiAccessFlag != cdiReadReadFlag && (access & CA_WRITE) == 0) ccerr(-illegal_read_write);
  if (BASEMODE(mode) > CM_SINGLE) cdiPollingFlag = 1;
  if (cdiNeedsCalibration) cdiReturnCondition |= RETURN_CALIBRATED;
  if (hasErrLst) cdiReturnCondition |= RETURN_STATUS;
  if (hasNameLst) cdiReturnCondition |= RETURN_NAMES;
  if (cdiAccessFlag == cdiWriteFlag && doutptr && dout->dFormat == CF_INT32) errLstWr = (int *)doutptr;
  if (perrLst == NULL) perrLst = errLstWr;
  if (perrLst == NULL && doutlen > 0)
  { /* if no error list at this stage allocate it ! */
    if ((errLstAlloc=(int *)calloc(doutlen,sizeof(int))) == NULL) ccerr(out_of_local_memory);
    perrLst = errLstAlloc;
    cdiReturnCondition |= FREE_ONCLOSE;
  }
  /* returned handle includes high word! */
  /* how to use the user's preferred callbackID ? */
  cc = cdi_Request(dn,doutptr,doutFormat,doutlenInBytes,dinptr,dinFormat,dinlen,
                   cdiDataPatternWindow,cdiDataMask,cdiParameters,cdiAccessFlag,
                   cdiReturnCondition,pollingRate,callback,cdiPollingFlag,perrLst,callbackID, 0);
  if (cc < 0) 
  {
    if (errLstAlloc != NULL) free(errLstAlloc);
    ccerr(cc);
  }
  if ((rc=get_RequestBufferPointer(cc, &pReq)) < 0 ) ccerr(rc);
  rc = assignGroupInOutFormatCode(pReq, din == NULL ? CF_NULL : din->dFormat, dout == NULL ? CF_NULL : dout->dFormat);
  cdi_StartRequest(cc);
err:
  return cc;  /* cc > 0 => id */
}
CDI_EXPORT int cdiAttachLinkHookFcn(
  char *devName,     /* device name */
  char *devProperty, 
  DTYPE *dout,       /* output data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  DTYPE *din,        /* input data pointer, type, and length (type is always UINT8, UINT16, UINT32 ?) */
  short access,      /* READ, WRITE are the only TINE access bits relevant here */
  int pollingRate,   /* polling rate in msec. */
  void (*callback)(int,int), /* callback routine */
  int mode,          /* polling mode (SINGLE, POLL, REFRESH, etc.) */
  UINT32 callbackID  /* callback ID */
)
{
  char *dn;
  int cc = cdiAttachLinkEx(devName,devProperty,dout,din,access,pollingRate,callback,mode,callbackID);
  if (cc == -server_redirection)
  {
    if ((dn=cdiGetFullDeviceName(devName)) == NULL) return -non_existent_fec;
    strncpy(devName,dn,512);
  }
  return cc;
}
CDI_EXPORT void AssignCdiSystemInitHook(void (*hook)(void))
{
  cdiSystemInitHook = hook;
}
CDI_EXPORT void AssignGetCdiLinkNameHook(char *(*hook)(char *,char *))
{
  getCdiLinkNameHook = hook;
}
CDI_EXPORT void AssignCdiCloseLinkHook(int (*hook)(int id))
{
  cdiCloseLinkHook = hook;
}
CDI_EXPORT void AssignCdiExecLinkHook(int (*hook)(char *,char *,DTYPE *,DTYPE *,short,UINT16))
{
  cdiExecLinkHook = hook;
}
CDI_EXPORT void AssignCdiAttachLinkHook(int (*hook)(char *,char *,DTYPE *,DTYPE *,short,int,void (*)(int,int),int,UINT32))
{
  cdiAttachLinkHook = hook;
}
CDI_EXPORT void initTineCDI(void)
{
  AssignCdiSystemInitHook(cdiPostSystemInit);
  AssignGetCdiLinkNameHook(getCdiLinkName);
  AssignCdiCloseLinkHook(cdiCloseLink);
  AssignCdiExecLinkHook(cdiExecLinkHookFcn);
  AssignCdiAttachLinkHook(cdiAttachLinkHookFcn);
  cdilog("initTineCDI done!");
}
int gCdiHomeEstablished = FALSE;
CDI_EXPORT int SetCdiHome(char *cdiHome)
{
  int len;
  if (cdiHome == NULL) return invalid_parameter;
  memset(CdiDBpath,0,80);
  strncpy(CdiDBpath,cdiHome,80);
# ifdef FS_DELIMITER
  len = (int)strlen(CdiDBpath);
  if (CdiDBpath[len-1] != FS_DELIMITER) CdiDBpath[len] = FS_DELIMITER;
# endif
  cdilog("Set CDI HOME to %s",cdiHome);
  gCdiHomeEstablished = TRUE;
  return 0;
}
CDI_EXPORT char *GetCdiHome(void)
{
  static char lclCdiHome[80];
  strncpy(lclCdiHome,CdiDBpath,80);
  return lclCdiHome;
}
void fixCdiRepository(void)
{
  char *ptr;
  int len;
  static int done=0;
  if (done) return;
  done = TRUE;
  if (!gCdiHomeEstablished)
  { /* not set via API so use the envirnoment ... */
    memset(CdiDBpath,0,80);
    if ((ptr=getenv("CDI_HOME")) != NULL || (ptr=getenv("CDIDB")) != NULL ||
        (ptr=getenv("FEC_HOME")) != NULL || (ptr=getenv("FECDB")) != NULL)
    {
      strncpy(CdiDBpath,ptr,80);
#     ifdef FS_DELIMITER
      len = (int)strlen(CdiDBpath);
      if (CdiDBpath[len-1] != FS_DELIMITER) CdiDBpath[len] = FS_DELIMITER;
#     endif
    }
    gCdiHomeEstablished = TRUE;
  }
  memset(CdiLogPath,0,80);
  if ((ptr=getenv("CDI_LOG")) == NULL) 
  {
    strcpy(CdiLogPath,CdiDBpath);
  }
  else 
  {
    strncpy(CdiLogPath,ptr,80);
#   ifdef FS_DELIMITER
    len = (int)strlen(CdiLogPath);
    if (CdiLogPath[len-1] != FS_DELIMITER) CdiLogPath[len] = FS_DELIMITER;
#   endif
  }
  rotateLogFile();
  cdilog("cdi initializing");
  cdilog("CDI_HOME : %s",CdiDBpath);
}
HANDLE cdiLoadLib(const char *libpath)
{
# if defined(UNIX)
  return dlopen(libpath, RTLD_LAZY);
# elif defined(WIN32)
  return LoadLibrary(libpath);
# else
  return NULL;
# endif
}
int cdiUnloadLib(HANDLE hLib)
{
  int cc = -1;
  AssignCdiSystemInitHook(NULL);
  AssignGetCdiLinkNameHook(NULL);
  AssignCdiCloseLinkHook(NULL);
  AssignCdiExecLinkHook(NULL);
  AssignCdiAttachLinkHook(NULL);
# if defined(UNIX)
  cc = dlclose(hLib);
  return cc;
# elif defined(WIN32)
  cc = FreeLibrary(hLib);
  return (cc == 0) ? GetLastError() : 0;
# else
  return cc;
# endif
}
char *cdiLoadLibError(void)
{
# if defined(UNIX)
  return dlerror();
# elif defined(WIN32)
  LPVOID lpMsgBuf;
  static char lclStrBuf[256] = "";
  FormatMessage( 
    FORMAT_MESSAGE_ALLOCATE_BUFFER | 
    FORMAT_MESSAGE_FROM_SYSTEM | 
    FORMAT_MESSAGE_IGNORE_INSERTS,
    NULL,
    GetLastError(),
    MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), 
    (LPTSTR) &lpMsgBuf,
    0,
    NULL 
  );
  strncpy(lclStrBuf,lpMsgBuf,256);
  LocalFree(lpMsgBuf);
  return lclStrBuf;
# else
  return "not implemented";
# endif
}

CdiBusLib *buslibs = NULL;

int addBusLib(char *bus,HANDLE *hlib,int flags)
{
  CdiBusLib *lib;
  for (lib=buslibs; lib != NULL; lib=lib->nxt)
  {
    if (!strncmp(bus,lib->bus,32)) return already_assigned;
  }
  if ((lib=(CdiBusLib *)calloc(1,sizeof(CdiBusLib))) == NULL) return out_of_local_memory;
  strncpy(lib->bus,bus,32);
  lib->hlib = hlib;
  lib->flags = flags;
  lib->nxt = buslibs;
  buslibs = lib;
  return 0;
}
void rmvBusLibs(void)
{
  CdiBusLib *lib, *libnxt;
  int cc;
  for (lib=buslibs; lib != NULL; lib=libnxt)
  {
    libnxt = lib->nxt;
    cc = (lib->flags & BUSLIB_UNLOAD_UNSAFE) ? not_accepted : cdiUnloadLib(lib->hlib);
    cdilog("unload busplug %.32s returned %d",lib->bus,cc);
    free(lib);
  }
  buslibs = NULL;
}
int cdiSetIdleState(int idleState)
{
  if (idleState)
  { /* set to idle */
    FlushContractTable();
    stopAllListeners();
  }
  gCdiIdleState = idleState;
  return gCdiIdleState;
}

typedef struct cdiCsvMfStruct
{
  char bus[32];
  char lib[64];
  char env[128];
  int id;
  int simulation;
  int almThreshold;
} cdiCsvMfStructType;
cdiCsvMfStructType gCdiMf;
CSVCOLUMN csvCdiMfColumns[] =
{
  { "BUS","",0,CF_TEXT,32,(BYTE *)gCdiMf.bus,NULL,0 },
  { "LIBRARY","",0,CF_TEXT,64,(BYTE *)gCdiMf.lib,NULL,0 },
  { "BUS_ENV","",0,CF_TEXT,128,(BYTE *)gCdiMf.env,NULL,0 },
  { "BUS_ID","",0,CF_INT32,1,(BYTE *)&gCdiMf.id,NULL,0 },
  { "SIMULATION","",0,CF_INT32,1,(BYTE *)&gCdiMf.simulation,NULL,0 },
  { "ALARM_THRESHOLD","0",0,CF_INT32,1,(BYTE *)&gCdiMf.almThreshold,NULL,0 },
};
int csvCdiMfIterator(struct CSVDBStruct *db)
{
  int cc = 0, flags = 0, id;
  char libPath[256],*ccstr = "success";
  HANDLE hLib=NULL;
# if defined(WIN32)
  char *libExt = ".dll";
  char *libPre = NULL;
# elif defined(UNIX)
  char *libExt = ".so";
  char *libPre = "lib";
# endif

  if (gCdiMf.almThreshold > 0) 
  {
    cdilog("set alarm threshold %d via manifest",gCdiMf.almThreshold);
    cdiSetHardwareAlarmThreshold(gCdiMf.almThreshold);
  }
  if (strlen(gCdiMf.bus) > 0)
  { /* explicitly named in the manifest file */
    if ((id=cdiRegisterBus(gCdiMf.bus,gCdiMf.id)) < 0) 
    {
      cc = hardware_error;
      ccstr = erlst[cc];
      goto err;
    }
  }
  /* the loaded library will register its function handlers ... */
  memset(libPath,0,256);
  if (libPre != NULL && strncmp(libPre,gCdiMf.lib,strlen(libPre))) strcpy(libPath,libPre);
  strncat(libPath,gCdiMf.lib,200);
  if (strcmp(&libPath[strlen(libPath)-strlen(libExt)],libExt)) strcat(libPath,libExt);

  /* use bus environment to set port parameters */
  /* e.g. CANJANZ_CDI=1:500;2:300 */
  /* e.g. RS232_CDI=1:9600:0:1:1:0:0:0;2:3200:0:1:1:0:0:0 */
  if (strlen(gCdiMf.env) > 0) 
  {
    if (strchr(gCdiMf.env,'=') != NULL)
    {
      cdilog("set environment  %s : %s", gCdiMf.env, putenv(gCdiMf.env) ? "failed" : "success");
    }
    else
    {
      if (stricmp(gCdiMf.env,"UNSAFE") == 0) flags |= BUSLIB_UNLOAD_UNSAFE;
    }
  }
  if (gCdiMf.simulation) gSimulationFlag = TRUE; 
  if (!strnicmp(libPath,"NULL",4))
  {
    cc = 0;
    ccstr = "process environment only";
    goto err;
  }
  if ((hLib=cdiLoadLib(libPath)) == NULL)
  {
    cc = software_error;
    ccstr = cdiLoadLibError();
    goto err;
  }
  addBusLib(gCdiMf.bus,hLib,flags);
err:
  cdilog("load bus library %s : %s",libPath,ccstr);
  db->cnt++;
  return cc;
}
CSVDB csvCdiMfDb =
{
  csvCdiMfColumns,
  sizeof(csvCdiMfColumns)/sizeof(csvCdiMfColumns[0]),
  0,
  NULL,
  csvCdiMfIterator,
  NULL,
  0,
  NULL,
  0,
  0
};
CDI_EXPORT int readCdiManifest(void)
{
  int cc=0;
  fixCdiRepository();
  cc = csvReadFile(CdiDBpath,CDIMF_FILE,&csvCdiMfDb,NULL);
  cdilog("read database file %s%s : %s",CdiDBpath,CDIMF_FILE,erlst[cc]);
  return cc;
}
CdiSchdLstType *gCdiSchdLst = NULL;
int sizCdiSchdLst = 0;
CSVCOLUMN csvCdiSchedColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiSchdLstType,prp),NULL,0 },
  { "DEVICE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiSchdLstType,dev),NULL,0 },
  { "TOLERANCE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(CdiSchdLstType,tolstr),NULL,0 },
};
CSVDB csvCdiSchedDb =
{
  csvCdiSchedColumns,
  sizeof(csvCdiSchedColumns)/sizeof(csvCdiSchedColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(CdiSchdLstType),
  0
};
CDI_EXPORT int readCdiSchedulingInfo(void)
{
  int i,lin,idx,cc=0;
  float f = 0;
  fixCdiRepository();
  cc = csvReadFile(CdiDBpath,CDISCHED_FILE,&csvCdiSchedDb,(void **)&gCdiSchdLst);
  if (cc == no_such_file) return 0; /* there usually isn't one */
  cdilog("read scheduling info file %s%s : %s",CdiDBpath,CDISCHED_FILE,erlst[cc]);
  if (cc == 0)
  {
    sizCdiSchdLst = csvCdiSchedDb.siz;
    for (i=0; i<sizCdiSchdLst; i++)
    {
      if (cdiParsePropertyForAccess(gCdiSchdLst[i].prp) != -1)
      { /* if we don't get '-1' here then the property is a direct BUS access property ! */
        if (strchr(gCdiSchdLst[i].dev,'*') == NULL && findDevice(gCdiSchdLst[i].dev,&lin,&idx) < 0)
        { /* for BUS properties: device must be given as a single CDI device */
          cdilog("scheduled property '%.64s' is a direct BUS access property and '%.64s' is not a single CDI device",
            gCdiSchdLst[i].prp,gCdiSchdLst[i].dev);
          continue;
        }
      }
      gCdiSchdLst[i].enabled = TRUE;
      f = (float)strtod(gCdiSchdLst[i].tolstr,NULL);
      if (strchr(gCdiSchdLst[i].tolstr,'%') == NULL)
      {
        gCdiSchdLst[i].atol = f;
      }
      else
      {
        gCdiSchdLst[i].ptol = (float)(f * 0.01);
      }
    }
  }
  return cc;
}
typedef struct cdiCsvServStruct
{
  char prp[PROPERTY_NAME_SIZE];  /**< target extended property */
  char dev[DEVICE_NAME_SIZE];  /**< target extended device */
  char fmtStr[16];  /**< format string */
  char tolStr[64];  /**< schedule tolerance: empty(default) => no scheduling */
  int  fmt; /**< target format: empty => registered format CF_FLOAT */
  int  siz; /**< target size: empty => registered size */
  int  tmr; /**< interval timer: empty => 1000 msec */
  int  hasHst; /**< TRUE: flag for a stock history else only use history.csv settings */
  int  isScheduled; /**< TRUE: schedule property upon change */
  int  enabled; /**< property is an allowable exported CDI extended property */
  int  lid; /**< link id */
  int  sts; /**< link status */
  int  ndevs; /**< size of device list of extended property */
  float atol; /**< schedule tolerance (absolute) */
  float ptol; /**< schedule tolerance (relative) */
  double ts; /**< timestamp of data */
  NAME64 *devs; /**< device list of extended property */
  BYTE *rbdata; /**< readback data buffer */
  BYTE *outdata; /**< output data buffer for clients */
} cdiCsvServType;
cdiCsvServType *gCdiSrvPrpLst = NULL;
int sizCdiSrvPrpLst = 0;
CSVCOLUMN csvCdiSrvPrpLstColumns[] =
{
  { "PROPERTY","",0,CF_TEXT,PROPERTY_NAME_SIZE,(BYTE *)OFFSETIN(cdiCsvServType,prp),NULL,0 },
  { "DEVICE","",0,CF_TEXT,DEVICE_NAME_SIZE,(BYTE *)OFFSETIN(cdiCsvServType,dev),NULL,0 },
  { "FORMAT","",0,CF_TEXT,16,(BYTE *)OFFSETIN(cdiCsvServType,fmtStr),NULL,0 },
  { "SCHEDULE_TOLERANCE","",0,CF_TEXT,64,(BYTE *)OFFSETIN(cdiCsvServType,tolStr),NULL,0 },
  { "SIZE","0",0,CF_INT32,1,(BYTE *)OFFSETIN(cdiCsvServType,siz),NULL,0 },
  { "INTERVAL","1000",0,CF_INT32,1,(BYTE *)OFFSETIN(cdiCsvServType,tmr),NULL,0 },
  { "KEEP_HISTORY","0",0,CF_BOOLEAN,1,(BYTE *)OFFSETIN(cdiCsvServType,hasHst),NULL,0 },
};
CSVDB csvCdiSrvPrpLstDb =
{
  csvCdiSrvPrpLstColumns,
  sizeof(csvCdiSrvPrpLstColumns)/sizeof(csvCdiSrvPrpLstColumns[0]),
  0,
  csvStdInit,
  csvStdIterator,
  csvStdExit,
  0,
  NULL,
  sizeof(cdiCsvServType),
  0
};
int gCdiDevSrvEqmRegistered = FALSE;
char gCdiDeviceServerName[EXPORT_NAME_SIZE+4] = "";
CDI_EXPORT int cdiSetDeviceServerExportName(char *expName)
{
  if (expName == NULL || strlen(expName) == 0) return invalid_parameter;
  strncpy(gCdiDeviceServerName,expName,EXPORT_NAME_SIZE);
  return 0;
}
CDI_EXPORT int readCdiDevSrvInfo(void)
{
  int i,pid,cc=0,rc,ndevs;
  float f = 0;
  ExportPropertyListStruct pls;
  char *env;
  if ((env=getenv("CDI_DEVICESERVERNAME")) != NULL)
  {
    strncpy(gCdiDeviceServerName,env,EXPORT_NAME_SIZE);
    cdilog("assign device server name %s from environment",gCdiDeviceServerName);
  }
  fixCdiRepository();
  cc = csvReadFile(CdiDBpath,CDISRV_FILE,&csvCdiSrvPrpLstDb,(void **)&gCdiSrvPrpLst);
  if (cc == no_such_file) return 0; /* there usually isn't one */
  cdilog("read device server info file %s%s : %.32s",CdiDBpath,CDISRV_FILE,erlst[cc]);
  if (cc == 0)
  {
    if (csvCdiSrvPrpLstDb.siz == 0)
    {
      cdilog("device server file has no entries: skipping");
      return 0;
    }
    sizCdiSrvPrpLst = csvCdiSrvPrpLstDb.siz;
    for (i=0; i<sizCdiSrvPrpLst; i++)
    {
      if ((pid=GetPropertyId(CDIEQM_TAG,gCdiSrvPrpLst[i].prp)) < 0)
      { /* not a registered property ! */
        cdilog("device server property %.64s is not a registered CDI property",gCdiSrvPrpLst[i].prp);
        continue;
      }
      if (pid < PRP_EXTENDED_OFFSET)
      {
        cdilog("device server property %.64s is not an extended property (direct BUS property?)",gCdiSrvPrpLst[i].prp);
        continue;
      }
      if ((rc=GetRegisteredPropertyListStruct(CDIEQM_TAG,gCdiSrvPrpLst[i].prp,&pls)) != 0)
      {
        cdilog("device server property %.64s : cannot get property information -> %.32s",gCdiSrvPrpLst[i].prp,erlst[rc]);
        continue;
      }
      gCdiSrvPrpLst[i].enabled = TRUE;
      gCdiSrvPrpLst[i].devs = pls.devNames;
      gCdiSrvPrpLst[i].ndevs = pls.numDevices;
      if (strlen(gCdiSrvPrpLst[i].fmtStr) > 0)
      { /* use the assigned format */
        gCdiSrvPrpLst[i].fmt = GetFormatFromString(gCdiSrvPrpLst[i].fmtStr);
      }
      else
      { /* use the registered data format (CF_FLOAT) */
        gCdiSrvPrpLst[i].fmt = pls.prpFormat;
      }
      if (gCdiSrvPrpLst[i].siz == 0)
      { /* use the registered data size */
        gCdiSrvPrpLst[i].siz = pls.prpSize;
      }
      cdilog("access property %.64s (as %d %.16s values) every %d msec",
        gCdiSrvPrpLst[i].prp,gCdiSrvPrpLst[i].siz, 
        GetFormatAsString(gCdiSrvPrpLst[i].fmt),gCdiSrvPrpLst[i].tmr);
      if (strlen(gCdiSrvPrpLst[i].tolStr) > 0)
      { /* need to schedule this property upon change in readback data */
        gCdiSrvPrpLst[i].isScheduled = TRUE;
        f = (float)strtod(gCdiSrvPrpLst[i].tolStr,NULL);
        if (strchr(gCdiSrvPrpLst[i].tolStr,'%') == NULL)
        {
          gCdiSrvPrpLst[i].atol = f;
        }
        else
        {
          gCdiSrvPrpLst[i].ptol = (float)(f * 0.01);
        }
        cdilog("schedule property %.64s when tolerance %.64s is exceeded",
          gCdiSrvPrpLst[i].prp,gCdiSrvPrpLst[i].tolStr);
      }
    }
    /* OK to register the device server property */
    if (!gCdiDevSrvEqmRegistered)
    {
      if (strlen(gCdiDeviceServerName) == 0)
      {
        sprintf(gCdiDeviceServerName,"%.28s.SRV",gCdiExportName);
      }
      cdilog("CDI device server export name : %.32s",gCdiDeviceServerName);
      ndevs = cdi_getDeviceNames(NULL,0);
      cdilog("register device server with %d devices",ndevs);
      if ((rc=RegisterEquipmentModule(gCdiDeviceServerName,CDIDSM_TAG,ndevs,dsmeqm,dsmeqm_ini,dsmeqm_bkg,CDIEQM_RATE,dsmeqm_exi)) != 0)
      {
        cdilog("could not export device server %.64s : %.32s",gCdiDeviceServerName,erlst[rc]);
        return rc; /* no reason to continue */
      }
      gCdiDevSrvEqmRegistered = TRUE;
    }
  }
  return cc;
}

CDI_EXPORT int dsmeqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int pid = GetPropertyId(CDIDSM_TAG,devProperty);
  int devnr = GetDeviceNumberEx(CDIDSM_TAG,devName,devProperty);
  int fsiz;
  char *c, tgt[256];

  if (pid < 0) return illegal_property;
  if (devnr < 0) return illegal_device;

  if (!stricmp(devProperty,"DESCRIPTION"))
  { /* foward description of CDI device */
    return cdieqm(devName,devProperty,dout,din,access);
  }
  if (!stricmp(devProperty,"RECV.CLBR"))
  { /* remap these to the corresponding extended property : */
    if (access & CA_WRITE) return illegal_read_write;
    strncpy(tgt,devName,DEVICE_NAME_SIZE);
    if ((c=strrchr(tgt,'.')) != NULL)
    { /* normal extended property case */
      *c++ = 0;
    }
    else
    { /* a single extended property (no template): e.g. 'status.status' */
      c = tgt;
    }
    if ((pid=GetPropertyId(CDIDSM_TAG,c)) < 0) return illegal_property;
    if ((devnr=GetDeviceNumberEx(CDIDSM_TAG,tgt,c)) < 0) return illegal_device;
  }
  fsiz = fmtsizeof(gCdiSrvPrpLst[pid].fmt);
  if (access & CA_WRITE)
  { /* forward along */
    sprintf(tgt,"/LOCALHOST/CDI/%.32s",devName);
    return ExecLinkEx(tgt,gCdiSrvPrpLst[pid].prp,dout,din,access,100);
  }
  if (gCdiSrvPrpLst[pid].sts != 0) return gCdiSrvPrpLst[pid].sts;
  dout->dTimeStamp = gCdiSrvPrpLst[pid].ts;
  return PutValuesFromAny(dout,gCdiSrvPrpLst[pid].outdata,(short)gCdiSrvPrpLst[pid].fmt,1,fsiz,gCdiSrvPrpLst[pid].siz,devnr);
}
#define isNumberFormat(f) (isPrimitiveFormat(f) && (f) != CF_CHAR)
void cbDsm(int id,int cc)
{
  int i, siz, fireSched = FALSE, dbg = FALSE;
  float t = 0, frb, fv;
  gCdiSrvPrpLst[id].sts = cc;
  if (cc == 0) gCdiSrvPrpLst[id].ts = MakeDataTimeStamp();
  siz = fmtsizeof(gCdiSrvPrpLst[id].fmt) * gCdiSrvPrpLst[id].siz;
  if (gCdiSrvPrpLst[id].isScheduled)
  { /* check scheduling criteria */
    if (isNumberFormat(gCdiSrvPrpLst[id].fmt) && 
        (gCdiSrvPrpLst[id].atol + gCdiSrvPrpLst[id].ptol) != 0)
    { /* is a number type and has a tolerance set */
      for (i=0; i<gCdiSrvPrpLst[id].siz; i++)
      { /* run thru the list */
        frb = getFloatAtIndex(gCdiSrvPrpLst[id].rbdata,i,gCdiSrvPrpLst[id].fmt);
        fv = getFloatAtIndex(gCdiSrvPrpLst[id].outdata,i,gCdiSrvPrpLst[id].fmt);
        t = (float)fabs(gCdiSrvPrpLst[id].atol + frb * gCdiSrvPrpLst[id].ptol);
        if (fabs(fv-frb) > t)
        { /* this can creep without ever firing! (successive changes just under tolerance) */
          fireSched = TRUE;
          break;
        }
      }
    }
    else
    {
      if (memcmp(gCdiSrvPrpLst[id].rbdata,gCdiSrvPrpLst[id].outdata,siz) != 0) 
      {
        fireSched = TRUE;
      }
    }
  }
  memcpy(gCdiSrvPrpLst[id].outdata,gCdiSrvPrpLst[id].rbdata,siz);
  if (fireSched) 
  {
    cdilog("schedule %.64s for immediate delivery", gCdiSrvPrpLst[id].prp);
    SystemScheduleProperty(CDIDSM_TAG,gCdiSrvPrpLst[id].prp);
    SystemSchedulePropertyEx(CDIDSM_TAG,"RECV.CLBR",CA_ALARM);
  }
  dbg = cdiDebugLevel > 1 ? TRUE : isPastDeviceFilter(gCdiSrvPrpLst[id].prp);
  if (dbg) ttyoutput("CDI Device Server callback %.64s : %d", gCdiSrvPrpLst[id].prp,cc);
}
CDI_EXPORT void dsmeqm_ini(void)
{
  int i, k, nd = 0, pid, atype, acc, lsiz, isdevprp;
  DTYPE dout,din;
  char tgt[256],dev[DEVICE_NAME_SIZE],devDesc[64];
  cdilog("registering device server properties");
  for (i=0; i<sizCdiSrvPrpLst; i++)
  { /* first register the property */
    memset(&dout,0,sizeof(DTYPE));
    dout.dFormat = gCdiSrvPrpLst[i].fmt;
    dout.dArrayLength = gCdiSrvPrpLst[i].siz;
    memset(&din,0,sizeof(DTYPE));
    din.dFormat = gCdiSrvPrpLst[i].fmt;
    din.dArrayLength = 1;
    atype = gCdiSrvPrpLst[i].siz > 1 ? AT_CHANNEL : AT_SCALAR;
    acc = CA_READ|CA_WRITE;
    if (gCdiSrvPrpLst[i].hasHst) acc |= CA_HIST;
    lsiz = gCdiSrvPrpLst[i].fmt == CF_TEXT ? 1 : gCdiSrvPrpLst[i].siz;
    memset(dev,0,DEVICE_NAME_SIZE);
    if (lsiz > 0) sprintf(dev,"%.16s.%.16s",gCdiSrvPrpLst[i].devs[0].name,gCdiSrvPrpLst[i].prp);
    if (cdi_getDescriptionForDevice(dev,devDesc) != 0) strcpy(devDesc,"extended property");
    if ((pid=RegisterPropertyInformation(CDIDSM_TAG,gCdiSrvPrpLst[i].prp,&dout,&din,
               (short)acc,(short)atype,(UINT16)gCdiSrvPrpLst[i].siz,devDesc,i,NULL)) < 0)
    {
      cdilog("could not register device server property %.64s : %.32s",gCdiSrvPrpLst[i].prp,erlst[-pid]);
    }
    /* assign the device names */
    cdilog("device server: assign %d devices to property %.64s",lsiz,gCdiSrvPrpLst[i].prp);
    AssignDeviceListToProperty(CDIDSM_TAG,gCdiSrvPrpLst[i].prp,gCdiSrvPrpLst[i].devs,lsiz);
    /* start the corresponding readback links */
    gCdiSrvPrpLst[i].rbdata = (BYTE *)calloc(gCdiSrvPrpLst[i].siz,fmtsizeof(gCdiSrvPrpLst[i].fmt));
    if (gCdiSrvPrpLst[i].rbdata == NULL)
    {
      cdilog("could not allocate memory for device server property %.64s !",gCdiSrvPrpLst[i].prp);
      continue;
    }
    gCdiSrvPrpLst[i].outdata = (BYTE *)calloc(gCdiSrvPrpLst[i].siz,fmtsizeof(gCdiSrvPrpLst[i].fmt));
    if (gCdiSrvPrpLst[i].outdata == NULL)
    {
      cdilog("could not allocate memory for device server property %.64s !",gCdiSrvPrpLst[i].prp);
      continue;
    }
    dout.data.bptr = gCdiSrvPrpLst[i].rbdata;
    sprintf(tgt,"/LOCALHOST/CDI/%.32s",gCdiSrvPrpLst[i].dev);
    if (gCdiSrvPrpLst[i].dev[0] == '#')
    { /* oops!: find the registered device name and try again */
      int _pid = GetPropertyId(CDIEQM_TAG,gCdiSrvPrpLst[i].prp);
      int _dnr = atoi(&gCdiSrvPrpLst[i].dev[1]);
      int _xid;
      if (_pid >= PRP_EXTENDED_OFFSET &&
          _pid < PRP_EXTENDED_OFFSET + gNumExtCdiProperties)
      {
        _xid = _pid - PRP_EXTENDED_OFFSET;
        if (_dnr >= 0 && _dnr < extCdiPrps[_xid].ndevs)
        { /* extended property device number supplied: work with real name below */
          strncpy(dev,extCdiPrps[_xid].devs[_dnr].name,16); dev[16] = 0;
          sprintf(tgt,"/LOCALHOST/CDI/%.32s",dev);
        }
      }
    }
    gCdiSrvPrpLst[i].lid = AttachLinkEx(tgt,gCdiSrvPrpLst[i].prp,&dout,NULL,CA_READ,gCdiSrvPrpLst[i].tmr,cbDsm,CM_TIMER,i);
    if (gCdiSrvPrpLst[i].lid < 0)
    {
      cdilog("could not start link for device server property %.64s : %.32s",gCdiSrvPrpLst[i].prp,erlst[-gCdiSrvPrpLst[i].lid]);
      continue;
    }
    cdilog("link to device server property %.192s[%.64s] started (monitor at %d msec)",tgt,gCdiSrvPrpLst[i].prp, gCdiSrvPrpLst[i].tmr);
    isdevprp = (gCdiSrvPrpLst[i].siz == 1 && !strncmp(gCdiSrvPrpLst[i].devs[0].name,gCdiSrvPrpLst[i].prp,32));
    if (gCdiSrvPrpLst[i].fmt == CF_TEXT) isdevprp = TRUE;
    for (k=0; k<gCdiSrvPrpLst[i].ndevs; k++)
    {
      if (isdevprp)
      {
        sprintf(dev,"%.32s",gCdiSrvPrpLst[i].devs[k].name);
      }
      else
      {
        sprintf(dev,"%.30s.%.30s",gCdiSrvPrpLst[i].devs[k].name,gCdiSrvPrpLst[i].prp);
      }
      RegisterDeviceName(CDIDSM_TAG,dev,nd++);
    }
  }
  dout.dFormat = CF_INT16;
  dout.dArrayLength = 1;
  if ((pid=RegisterPropertyInformation(CDIDSM_TAG,"RECV.CLBR",&dout,NULL,
             CA_READ,AT_SCALAR,1,"mapped calibrated RECV bus property",i,NULL)) < 0)
  {
    cdilog("could not register mapped bus property 'RECV.CLBR' : %.32s",erlst[-pid]);
  }
  if ((pid=RegisterPropertyEx(CDIDSM_TAG,"DESCRIPTION",PRP_DESCRIPTION_SIZE,CF_NAME64,CA_READ,"[0:0 ]description selected device",PRP_DESCRIPTION)) < 0)
  {
    cdilog("could not register cdi property 'DESCRIPTION' : %.32s",erlst[-pid]);
  }
}
CDI_EXPORT void dsmeqm_bkg(void)
{
}
CDI_EXPORT void dsmeqm_exi(void)
{
}
cdiHstItem *cdiHstLst = NULL;
void parseHistInfo(char *accStr,const char *dev,const char *prp,int fmt,int len,int tmr)
{
  char *c;
  int type = HIST_NONE;
  cdiHstItem *itm = NULL;
  if ((c=strstr(accStr,"LSN")) != NULL) type = HIST_LSN;
  else if ((c=strstr(accStr,"HIST")) != NULL) type = HIST_MIN;
  else return;
  if ((itm=(cdiHstItem *)calloc(1,sizeof(cdiHstItem))) == NULL) return;
  itm->type = type;
  strncpy(itm->prp,prp,32);
  if (dev != NULL) strncpy(itm->dev,dev,32); else itm->isTmpl = TRUE;
  if (fmt == CF_BYTE || fmt == CF_INT16) fmt = CF_INT32; /* elevate the data type */
  itm->fmt = fmt;
  itm->len = len;
  itm->tmr = tmr;
  itm->nxt = cdiHstLst;
  cdiHstLst = itm;
  return;
}
HistorySpecification hstNormSpec = 
{
  HRS_DEFAULT_POLL_RATE,HRS_DEFAULT_ARCHIVE_RATE,HRS_DEFAULT_DEPTHSHORT,HRS_DEFAULT_DEPTHLONG,
  HRS_DEFAULT_HEARTBEAT,HRS_DEFAULT_REL_TOLERANCE,HRS_DEFAULT_ABS_TOLERANCE,NULL,NULL
};
HistorySpecification hstLsnSpec = 
{
  HRS_DEFAULT_POLL_RATE,0,5,-1,0,0,0,NULL,NULL
};
int registerHistories(void)
{
  cdiHstItem *itm=NULL, *nxt=NULL;
  ExportPropertyListStruct pl;
  HistorySpecification *hspc;
  char *dev;
  int cc,len,n=1;
  for (itm=cdiHstLst; itm != NULL; itm=nxt)
  {
    nxt = itm->nxt;
    hspc = (itm->type == HIST_LSN) ? &hstLsnSpec : &hstNormSpec;
    hspc->pollingRate = itm->tmr;
    dev = itm->dev; len = itm->len;
    if (itm->isTmpl)
    { /* a template entry */
      dev = NULL; len = -1;
      if ((cc=GetRegisteredPropertyListStruct(CDIEQM_TAG,itm->prp,&pl)) == 0)
      {
        dev = pl.devNames[0].name; len = pl.prpSize;
      }
      else
      {
        cdilog("CDI: registerHistories : property %.32s not registered %.32s",itm->prp,erlst[cc]);
      }
    }
    if (len > 0)
    {
      if ((cc=AppendHistoryInformation(CDIEQM_TAG,itm->prp,dev,len,itm->fmt,n,hspc)) == 0) 
      {
        cdilog("CDI: registerHistories : added property %.32s device %.32s to %s histories (interval: %d ms)",
          itm->prp,dev,(itm->type == HIST_LSN ? "listening" : "monthly"),itm->tmr);
        n++;
      }
      else
      {
        cdilog("CDI: registerHistories : property %.32s device %.32s : %.32s",itm->prp,dev,erlst[cc]);
      }
    }
    free(itm); /* return the memory */
  }
  cdiHstLst = NULL;
  return n-1;
}
typedef struct cdiCsvAddrStruct
{
  char bus[132];
  char addr[192];
  char base[96];
  char params[96];
  char access[64];
  char name[32];
  char lngname[64];
  char map[64];
  char description[96];
  int line;
  int index;
  int number;
  int interval;
  char mask[64];
  char pattern[64];
  char tolerance[16];
  char dataformat[32];
  char datainput[132];
  char datalimit[64];
  char clbrrule[132];
  char clbrrule_rcv[132];
  char clbrrule_snd[132];
} cdiCsvAddrStructType;
cdiCsvAddrStructType gCdiCsvAddr;
CSVCOLUMN csvCdiAddrColumns[] =
{
  { "BUS","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.bus,NULL,0 },
  { "LINE","",0,CF_INT32,1,(BYTE *)&gCdiCsvAddr.line,NULL,0 },
  { "ADDRESS","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.addr,NULL,0 },
  { "ADDRESS_BASE","",0,CF_TEXT,96,(BYTE *)gCdiCsvAddr.base,NULL,0 },
  { "ADDRESS_PARAMETERS","",0,CF_TEXT,96,(BYTE *)gCdiCsvAddr.params,NULL,0 },
  { "MASK","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.mask,NULL,0 },
  { "PATTERN","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.pattern,NULL,0 },
  { "ACCESS","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.access,NULL,0 },
  { "TOLERANCE","",0,CF_TEXT,16,(BYTE *)gCdiCsvAddr.tolerance,NULL,0 },
  { "NUMBER","",0,CF_INT32,1,(BYTE *)&gCdiCsvAddr.number,NULL,0 },
  { "INTERVAL","1000",0,CF_INT32,1,(BYTE *)&gCdiCsvAddr.interval,NULL,0 },
  { "NAME","",0,CF_TEXT,32,(BYTE *)gCdiCsvAddr.name,NULL,0 },
  { "FORMAT","INT16",0,CF_TEXT,32,(BYTE *)gCdiCsvAddr.dataformat,NULL,0 },
  { "INPUT","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.datainput,NULL,0 },
  { "LIMIT","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.datalimit,NULL,0 },
  { "RULE","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.clbrrule,NULL,0 },
  { "RULE_RECV","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.clbrrule_rcv,NULL,0 },
  { "RULE_SEND","",0,CF_TEXT,132,(BYTE *)gCdiCsvAddr.clbrrule_snd,NULL,0 },
  { "LONG_NAME","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.lngname,NULL,0 },
  { "ADDRESS_MAP","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.map,NULL,0 },
  { "DESCRIPTION","",0,CF_TEXT,64,(BYTE *)gCdiCsvAddr.description,NULL,0 },
};
BYTE *parseInputData(char *str,int dataFormat,int *dataSize)
{
  char *s, *c;
  int cc=0, i, n, sizeInBytes=0;
  BYTE *inputData=NULL;

  if (str == NULL || dataSize == NULL) ccerr(argument_list_error);
  if ((sizeInBytes=fmtsizeof(dataFormat)) == 0) ccerr(illegal_format);
  /* find out how many ...*/
  for (c=str, n=1; (c=strstr(c,CDI_PARAMETER_DELIMITER)) != NULL; c++, n++);
  if (dataFormat == CF_TEXT)
  {
    char *ccr=NULL,*clf=NULL;
    int rlen;
    if (n != 1) ccerr(string_expected);
    sizeInBytes = (int)strlen(str) + 1; /* add space for trailing '0' */
    while ((clf=strstr(str,"\\10")) != NULL)
    {
      c = clf + 3; 
      rlen = (int)strlen(c);
      *clf++ = (char)10;
      strncpy(clf,c,rlen); 
      clf[rlen] = 0;
      sizeInBytes -= 2;
    }
    while ((ccr=strstr(str,"\\13")) != NULL)
    {
      c = ccr + 3; 
      rlen = (int)strlen(c);
      *ccr++ = (char)13;
      strncpy(ccr,c,rlen);
      ccr[rlen] = 0;
      sizeInBytes -= 2;
    }
  }
  if ((inputData=(BYTE *)calloc(n,sizeInBytes)) == NULL) ccerr(out_of_server_memory);
  c = NULL;
  if ((s=strtok_r(str,CDI_PARAMETER_DELIMITER,&c)) == NULL) ccerr(string_expected);
  *dataSize = n * sizeInBytes;
  switch (dataFormat)
  {
    case CF_BYTE:
      for (i=0; i<n && s != NULL; i++) 
      {
        inputData[i] = (BYTE)strtol(s,NULL,0);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c);
      }
      break;
    case CF_SHORT:
      for (i=0; i<n; i++) 
      {
        ((short *)inputData)[i] = (short)strtol(s,NULL,0);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c);
      }
      break;
    case CF_INT32:
      for (i=0; i<n; i++) 
      {
        ((UINT32 *)inputData)[i] = (UINT32)strtol(s,NULL,0);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c);
      }
      break;
    case CF_FLOAT:
      for (i=0; i<n; i++) 
      {
        ((float *)inputData)[i] = (float)strtod(s,NULL);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c);
      }
      break;
    case CF_DOUBLE:
      for (i=0; i<n; i++) 
      {
        ((double *)inputData)[i] = (double)strtod(s,NULL);
        s = strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c);
      }
      break;
    case CF_TEXT:
      *dataSize = sizeInBytes - 1;
      strncpy(inputData,s,n * sizeInBytes);
      break;
  }
err:
  if (cc != 0)
  {
    cdilog("CDI: parseInputData : %s",erlst[cc]);
    if (inputData != NULL) free(inputData);
    inputData = NULL;
  }
  return inputData;
}
int cdiParseBusString(char *str,char *bus,int *parameters,int *nparameters)
{
  char *s,*c=NULL;
  int i;
  if (str == NULL || bus == NULL) return argument_list_error;
  if ((s=strtok_r(str,CDI_PARAMETER_DELIMITER,&c)) == NULL) return string_expected;
  strncpy(bus,s,32);
  if (parameters == NULL) return 0;
  memset(parameters,0,NUMBER_EXEC_PARAM*sizeof(int));
  for (i=0; i<NUMBER_EXEC_PARAM && 
           (s=strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c)) != NULL; i++)
  {
    parameters[i] = strtol(s,NULL,0);
  }
  if (nparameters != NULL) *nparameters = i;
  return 0;
}
int cdiParseAddressString(char *str,int *addrlist,int *addrlistsize,int *parameters,int *nparameters)
{
  char *s,*as,*c=NULL,*cp=NULL;
  int i;
  if (str == NULL || addrlist == NULL) return argument_list_error;
  /* break at the first delimiter (old-style ADDRESS column) */
  c = strtok_r(str,CDI_PARAMETER_DELIMITER,&cp);
  memset(addrlist,0,CDI_ADDR_SIZE*sizeof(int));
  as = strtok_r(str,CDI_ADDRESS_DELIMITER,&c);
  addrlist[0] = strtol(as,NULL,0);
  for (i=1; i<CDI_ADDR_SIZE && 
           (s=strtok_r(NULL,CDI_ADDRESS_DELIMITER,&c)) != NULL; i++)
  {
    addrlist[i] = strtol(s,NULL,0);
  }
  if (addrlistsize != NULL) *addrlistsize = i;
  if (parameters == NULL) return 0;
  /* parameter list wished for */
  memset(parameters,0,NUMBER_EXEC_PARAM*sizeof(int));
  i = 0;
  if (cp != NULL && strlen(cp) > 0)
  {
    s = strtok_r(cp,CDI_PARAMETER_DELIMITER,&c);
    parameters[0] = strtol(s,NULL,0);
    for (i=1; i<NUMBER_EXEC_PARAM && 
             (s=strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c)) != NULL; i++)
    {
      parameters[i] = strtol(s,NULL,0);
    }
  }
  if (nparameters != NULL) *nparameters = i;
  return 0;
}
CdiClbrFcn *cdiClbrFcnLst = NULL;

int registerClbrConversionFunction(char *fcnName,double (*fcn)(double))
{
  CdiClbrFcn *f = NULL;
  if (fcnName == NULL) return argument_list_error;
  for (f=cdiClbrFcnLst; f != NULL; f=f->next)
  {
    if (strncmp(f->fcnName,fcnName,64)) continue;
    break;
  }
  if (f != NULL)
  {
    cdilog("replacing conversion function for %s",fcnName);
    f->fcn = fcn;
    return 0;
  }
  if ((f=(CdiClbrFcn *)calloc(1,sizeof(CdiClbrFcn))) == NULL) return out_of_server_memory;
  strncpy(f->fcnName,fcnName,64);
  f->fcn = fcn;
  f->next = cdiClbrFcnLst;
  cdiClbrFcnLst = f;
  cdilog("register conversion function %s",fcnName);
  return 0;
}
int assignClbrConversionFunction(char *fcnName,CdiRule *rule)
{
  CdiClbrFcn *f = NULL;
  if (fcnName == NULL || rule == NULL) return argument_list_error;
  for (f=cdiClbrFcnLst; f != NULL; f=f->next)
  {
    if (strncmp(f->fcnName,fcnName,64)) continue;
    break;
  }
  rule->fcn = f == NULL ? NULL : f->fcn;
  return 0;
}
int chrtorule(char c)
{
  switch (c)
  {
    case CDICHAR_ADD: return cdiRuleAdd;
    case CDICHAR_SUB: return cdiRuleSub;
    case CDICHAR_MUL: return cdiRuleMult;
    case CDICHAR_DIV: return cdiRuleDiv;
    case CDICHAR_PWR: return cdiRulePwr;
    case CDICHAR_LOG: return cdiRuleLog;
    case CDICHAR_MOD: return cdiRuleMod;
    case CDICHAR_RGH: return cdiRuleShiftRight;
    case CDICHAR_LFT: return cdiRuleShiftLeft;
    case CDICHAR_UNS: return cdiRuleUnsigned;
    case CDICHAR_SGN: return cdiRuleSigned;
    case CDICHAR_MSG: return cdiRuleMessage;
    case CDICHAR_EXP: return cdiRuleExp;
    case CDICHAR_FCNA:
    case CDICHAR_FCN: return cdiRuleFcn;
    case CDICHAR_XOR: return cdiRuleXor;
    case CDICHAR_NOT: return cdiRuleNot;
    case CDICHAR_AND: return cdiRuleAnd;
    case CDICHAR_FIX: return cdiRuleFix;
    case CDICHAR_OR:  return cdiRuleOr;
  }
  return -1;
}
int getMsgStr(const char *source,char *target,int pos,int len)
{
  char str[128], *c=str;
  int i;

  strncpy(str,source,128);
  for (i=0; i<pos; i++)
  {
    if ((c=strchr(c,'<')) == NULL) break;
    c++;
  }
  if (c == NULL) 
  {
    *target = 0;
  }
  else
  {
    strncpy(target,c,len);
    if ((c=strchr(target,'>')) != NULL) *c = 0;
  }
  return 0;
}
CdiRule *parseCdiRules(char *rulesString)
{
  int cc = 0, r;
  CdiRule *rule = NULL, *next = NULL, *clbrRules = NULL;
  char *s,*c=NULL,fnam[64];
  int val;

  if (rulesString == NULL || strlen(rulesString) == 0) return NULL;
  for (s=strtok_r(rulesString,CDI_PARAMETER_DELIMITER,&c); s != NULL; s=strtok_r(NULL,CDI_PARAMETER_DELIMITER,&c))
  {
    if (strlen(s) == 0) continue;
    while (*s == ' ' || *s == '\t') s++; /* remove white space */
    if ((r=chrtorule(*s++)) < 0) ccerr(illegal_operator);
    while (isalpha(*s)) s++; /* skip to end of string */
    if ((next=(CdiRule *)calloc(1,sizeof(CdiRule))) == NULL) ccerr(out_of_local_memory);
    next->operation = r;
    switch (r)
    {
      case cdiRuleSigned:
        next->isSigned = TRUE;
        break;
      case cdiRuleUnsigned:
        next->isSigned = FALSE;
        break;
      case cdiRuleMessage:
        next->msgOnValue = strtol(s,&s,0);
        /* parse for <msgOn><msgOff> */
        getMsgStr(s,next->msgOn,1,32);
        getMsgStr(s,next->msgOff,2,32);
        break;
      case cdiRuleFcn:
        getMsgStr(s,fnam,1,64);
        assignClbrConversionFunction(fnam,next);
        break;
      case cdiRuleExp:
        val = 0;
        if (strlen(s) > 0) val = (int)strtod(s,NULL);
        if (val == 0) val = 10;
        next->operand = (double) val;
        break;
      case cdiRuleOr:
      case cdiRuleAnd:
      case cdiRuleXor:
        next->operand = (double)strtol(s,NULL,0);
        break;
      default:
        next->operand = strtod(s,NULL);
        break;
    }
    if (clbrRules == NULL) 
    { /* first in the list */
      clbrRules = next;
      rule = next;
    }
    else
    { /* advance it */
      rule->nxt = next;
      rule = next;
    }
  }

err:
  if (cc)
  {
    cdilog("cannot parse rules string : %s",erlst[cc]);
    if (clbrRules != NULL)
    {
      for (rule=clbrRules; rule != NULL; rule=next)
      {
         next = rule->nxt;
         free(rule);
      }
      clbrRules = NULL;
    }
  }
  return clbrRules;
}
void csvCdiAddrInitializer(struct CSVDBStruct *db)
{
  if (db->siz == 0)
  {
    cdilog("cdi database empty");
    return;
  }
  if ((paramList=(PARAM_STRUCT *)SystemCalloc(db->siz*gMaxTemplateLength,sizeof(PARAM_STRUCT))) == NULL)
  {
    cdilog("unable to allocate device parameter list space");
    return;
  }
  paramListSize = db->siz * gMaxTemplateLength;
}

int cdiRegisterTemplateDevices(int ln,cdiCsvAddrStructType *gCdiCsvAddr,CdiTemplate *tmpl,int nrecords)
{
  int cc, i, k, devnr, devidx, lnidx, *prms;
  static int nparams, params[NUMBER_EXEC_PARAM];
  char addrstr[256],s[512], dname[40], lname[128], desc[96];
  int cdiaddr[16], naddr=0;
  CdiTemplateList *t;
  CdiTemplateInstanceName *tin, *tn;

  if (tmpl == NULL || gCdiCsvAddr == NULL) return argument_list_error;
  if ((cc=cdiParseAddressString(gCdiCsvAddr->addr,cdiaddr,&naddr,params,&nparams)) != 0)
  {
    cdilog("cannot parse address string -> %.32s",erlst[cc]);
    return cc;
  }
  lnidx = getNextFreeIndex(ln);
  if (lnidx < 0)
  {
    cc = -lnidx;
    cdilog("cannot acquire line index for line %d -> %.32s",ln,erlst[cc]);
    return cc;
  }
  tmpl->numInstances++;
  if ((tin=(CdiTemplateInstanceName *)calloc(1,sizeof(CdiTemplateInstanceName))) != NULL)
  { /* keep record of template instance names */
    strncpy(tin->name,gCdiCsvAddr->name,32);
    if ((tn=tmpl->instanceNames) == NULL)
    { /* nothing there yet ... */
      tmpl->instanceNames = tin;
    }
    else
    { /* scan to the end for forward ordering ... */
      while (tn->nxt != NULL) tn = tn->nxt;
      tn->nxt = tin;
    }
  }
  sprintf(addrstr,"(%d",cdiaddr[0]);
  for (k=1; k<naddr; k++) sprintf(&addrstr[strlen(addrstr)],".%d",cdiaddr[k]);
  for (i=0,t=tmpl->cdiTemplateList; i<tmpl->length && t != NULL; i++, t=t->next)
  {
    snprintf(dname,32,"%.16s.%.16s",gCdiCsvAddr->name,t->field);
    snprintf(lname,64,"%.64s.%.64s",gCdiCsvAddr->lngname,t->lngfield);
    snprintf(desc,96,"%.96s (%.96s)",gCdiCsvAddr->description,t->description);
    if (!strcmp(lname,".")) lname[0] = 0; /* no long name or field given in template */
    strcpy(s,addrstr);
    if (t->naddrParameters > 0)
    {
      sprintf(&s[strlen(s)],"->%d",t->addrParameters[0]);
      for (k=1; k<t->naddrParameters; k++) sprintf(&s[strlen(s)],":%d",t->addrParameters[k]);
    }
    strcat(s,")");

    devidx = lnidx + nrecords * i;
    devnr = gCdiCsvAddr->number + (nrecords + TEMPLATE_RECORD_GAP)* i;

    if ((cc=cdi_RegisterDevice(ln,dname,lname,desc,devidx,devnr,cdiaddr,naddr)) != 0)
    {
      cdilog("cannot register device %.40s at devIndex %d and devNr %d on line %d -> %.32s",dname,devidx,devnr,ln,erlst[cc]);
      return cc;
    }
    cdilog("%s %s line %d line index %d %s : cdi dev nr %d on cdi bus #%d",
      gCdiCsvAddr->bus, dname, gCdiCsvAddr->line, devidx, s, devnr, ln);
    if ( ++devCounter > 1000 && (devCounter % 500) == 0 ) printf(".%d", devCounter);
     
    prms = t->naddrParameters > 0 ? t->addrParameters : params;
    if ((cc=cdi_assignDefaultDeviceParameters(dname,prms)) != 0)
    {
      cdilog("assign device parameters to %.40s -> %.32s",dname,erlst[cc]);
      return cc;
    }
    if ((cc=cdi_assignDefaultDataMaskAndPattern(dname,t->dataMask,t->dataMaskRange,t->dataPattern,t->dataPatternRange,t->dataPatternTrigger)) != 0)
    {
      cdilog("assign mask/pattern to %.40s -> %.32s",dname,erlst[cc]);
      return cc;
    }
    if ((cc=cdi_assignDataFormat(dname,t->dataFormat)) != 0)
    {
      cdilog("data format for %.40s could not be assigned : %.32s",dname,erlst[cc]);
      return cc;
    }
    if ((cc=cdi_assignDefaultInterval(dname,t->interval)) != 0)
    {
      cdilog("default update interval for %.40s could not be assigned : %.32s",dname,erlst[cc]);
      return cc;
    }
    if (t->bitfield != NULL)
    { /* is registered as a bitfield */
      cdi_assignBitfieldToDevice(dname,cdiGetBitfieldXRef(dname,t->bitfield));
    }
    if (t->dataInput != NULL)
    {
      if (t->dataFormat == CF_NULL)
      {
        cdilog("data input provided without a valid format!");
        return illegal_format;
      }
      if ((cc=cdi_assignDefaultInput(dname,t->dataInput,t->dataInputSize,t->dataFormat)) != 0)
      {
        cdilog("data input for %.40s could not be assigned : %.32s",dname,erlst[cc]);
        return cc;
      }
    }
    if ((cc=cdi_assignDataLimits(dname,t->dataOutputSize,t->dataInputSize,t->dataOutput2ndSize)) != 0)
    {
      cdilog("data limits for %.40s could not be assigned : %.32s",dname,erlst[cc]);
      return cc;
    }
    if (t->clbrRules != NULL || t->rvrsRules != NULL)
    {
      /*if ((cc=cdi_assignDataCalibration(dname,t->clbrRules,getCdiFormat((short)t->dataFormat))) != 0) */
      if ((cc=cdi_assignDataCalibration(dname,t->clbrRules,t->rvrsRules,t->dataFormat)) != 0)
      {
        cdilog("data calibration rule for %.40s could not be assigned : %.32s",dname,erlst[cc]);
        return cc;
      }
    }
    cdi_assignAllowedAccess(dname,t->allowedAccess);
  }
  return 0;
}

DBITFIELD *parseFormatForBitfield(char *fmtstr)
{ /* look for format strings of the form "bitfield:<tag>" */
  char *c, tag[32],fmt[32];
  DBITFIELD *bf = NULL;

  if ((c=strstr(fmtstr,":<")) != NULL)
  {
    *c = 0; c += 2;
    strncpy(fmt,fmtstr,32);
    strncpy(tag,c,32);
    if ((c=strstr(tag,">")) == NULL) return NULL; /* not well formed !*/
    *c = 0;
    return findBitField(BITFIELD_LCL,tag);
  }
  return NULL;
}

int csvCdiAddrIterator(struct CSVDBStruct *db)
{
  int cc, i, fmt, fsiz, len, rblen=1, siz=0, sizin=0, sizout2nd=0, sizout=0, ln, dw, access=0xffff;
  static int nparams, params[NUMBER_EXEC_PARAM], autoDevNumFlag = 0, autoDevNumber = 0;
  char bus[32], busParams[132];
  BYTE *pPattern=NULL, *pDataMask=NULL, *pDataInput=NULL;
  char *c, *c1, *hstp, s[512], fmtstr[32];
  CdiTemplate *tmpl;
  int patRange[2],mskRange[2],patTrigger=CM_SEEKPATTERN;
  int cdiaddr[CDI_ADDR_SIZE],naddr=0;
  DBITFIELD *bf=NULL;

  if (strlen(gCdiCsvAddr.bus) == 0) return 0;
  /* gCdiCsvAddr has been filled in */
  /* if a TEMPLATE entry then do nothing here ... */
  if (!stricmp(gCdiCsvAddr.bus,"TEMPLATE")) return 0;
  /* if a BITFIELD entry then likewise do nothing here ... */
  if (!stricmp(gCdiCsvAddr.bus,"BITFIELD")) return 0;
  /* if a FIELDBUS entry then likewise do nothing here ... */
  if (!stricmp(gCdiCsvAddr.bus,"FIELDBUS")) return 0;
  /* parameters from address and bus strings ... */
  strncpy(s,gCdiCsvAddr.bus,132);
  if ((c=strstr(s,CDI_PARAMETER_DELIMITER)) != NULL)
  {
    *c++ = 0;
    strncpy(busParams,c,132);
  }
  else
  {
    busParams[0] = 0;
  }
  strncpy(bus,s,32);
  siz = db->siz;
  if (IsDiscoveryDevice(gCdiCsvAddr.name))
  { /* if first item is a discovery then remap the expected DB size */
    char ddev[64];
    strncpy(ddev, &gCdiCsvAddr.name[9], 32); ddev[32] = 0;
    if (strlen(ddev) > 0)
    {
      char *c = &ddev[1]; /* skip 1st ':' */
      int rstart = atoi(c), rstop = rstart + 1;
      if ((c=strchr(c, ':')) != NULL) rstop = atoi(++c); else rstop = rstart, rstart = 0;
      if (rstop - rstart > 0) siz += rstop - rstart;
    }
  }
  /* number of devices -> twice the database size should allow for enough dynamic entries */
  /* repeated calls will just return the line index ! */
  if ((ln=cdi_Open(bus, gCdiCsvAddr.line, busParams, siz*gMaxTemplateLength, 0)) < 0)
  {
    cdilog("cannot open bus %s -> %s",bus,erlst[-ln]);
    return -ln;
  }
  siz = 0; /* reuse variable */
  /* check device number it might have side effect for template devices ???? */
  if ( autoDevNumFlag == 0  && gCdiCsvAddr.number == 0 ) autoDevNumFlag = 1;
  if ( autoDevNumFlag ) gCdiCsvAddr.number = ++autoDevNumber;
  if (strlen(gCdiCsvAddr.map) > 0) strncpy(gCdiCsvAddr.lngname,gCdiCsvAddr.map,64);
  /* combine addr and params into a single string for the parsing routine */
  if (strlen(gCdiCsvAddr.base) > 0) strncpy(gCdiCsvAddr.addr,gCdiCsvAddr.base,96);
  if ((len=(int)strlen(gCdiCsvAddr.params)) > 0)
  { /* there are parameters from the parameters column */
    if (len > (int)(191 - strlen(gCdiCsvAddr.addr)))
    { /* oops ! */
      cdilog("cannot concatenate %s to %s : string too long",gCdiCsvAddr.params,gCdiCsvAddr.addr);
    }
    else
    { /* add it on ... */
      strcat(gCdiCsvAddr.addr,":");
      strncat(gCdiCsvAddr.addr,gCdiCsvAddr.params,len);
    }
  }
  /* check address for template */
  if ((c=strstr(gCdiCsvAddr.addr,":<")) != NULL)
  {
    *c = 0; /* peel it off */
    strncpy(s,c+2,32); s[32] = 0;
    if ((c=strstr(s,">")) != NULL) 
    {
      *c++ = 0;
      if (strlen(c) > 0) strcat(gCdiCsvAddr.addr,c); /* splice any remainder back in */
    }
    if ((tmpl=getCdiTemplate(s)) == NULL)
    {
      cdilog("could not get template %s for %s!",s,gCdiCsvAddr.name);
      return invalid_structure_tag;
    }
    return cdiRegisterTemplateDevices(ln,&gCdiCsvAddr,tmpl,db->siz);
  }
  /* no template assigned:  just this record */
  nparams = 0; s[0] = 0;
  if ((cc=cdiParseAddressString(gCdiCsvAddr.addr,cdiaddr,&naddr,params,&nparams)) != 0)
  {
    cdilog("cannot parse address string -> %s",erlst[cc]);
    return cc;
  }
  sprintf(s,"(%d",cdiaddr[0]);
  for (i=1; i<naddr; i++) sprintf(&s[strlen(s)],".%d",cdiaddr[i]);
  if (nparams > 0)
  {
    sprintf(&s[strlen(s)],"->%d",params[0]);
    for (i=1; i<nparams; i++) sprintf(&s[strlen(s)],":%d",params[i]);
  }
  strcat(s,")");
  if ((gCdiCsvAddr.index=getNextFreeIndex(ln)) < 0)
  {
    cc = -gCdiCsvAddr.index;
    cdilog("cannot acquire index for device %s, number: %d on line %d -> %s",
      gCdiCsvAddr.name,gCdiCsvAddr.number,ln,erlst[cc]);
    return cc;
  }
  if ((cc=cdi_RegisterDevice(ln,gCdiCsvAddr.name,gCdiCsvAddr.lngname,gCdiCsvAddr.description,gCdiCsvAddr.index,gCdiCsvAddr.number,cdiaddr,naddr)) != 0)
  {
    cdilog("cannot register device %s, number: %d at index %d on line %d -> %s",
      gCdiCsvAddr.name,gCdiCsvAddr.number,gCdiCsvAddr.index,ln,erlst[cc]);
    return cc;
  }
  cdilog("%s %s %s line %d line index %d %s : cdi dev nr %d on cdi bus #%d",
    bus,busParams,gCdiCsvAddr.name, gCdiCsvAddr.line, gCdiCsvAddr.index, s, gCdiCsvAddr.number, ln);
  if ( ++devCounter > 1000 && (devCounter % 500) == 0 ) printf(".%d", devCounter);
  if ((cc=cdi_assignDefaultDeviceParameters(gCdiCsvAddr.name,params)) != 0)
  {
    cdilog("assign device parameters to %s -> %s",gCdiCsvAddr.name,erlst[cc]);
    return cc;
  }
  /*
    returnCondition established when making the call to AttachLink() or ExecLink()
    at which time the pattern is extracted from the dout data array ....
    supplying a pattern below is in the present state useless ...
  */
  patRange[0] = patRange[1] = -1;
  if (strlen(gCdiCsvAddr.pattern) > 0)
  { /* 32 bytes */
    if (!strncmp(gCdiCsvAddr.pattern,"!=",2))
    { /* looking for pattern not equal to that given */
      patTrigger = CM_MISSPATTERN;
      memset(s,0,64);
      strncpy(s,&gCdiCsvAddr.pattern[2],62);
      strtrm(s);
      memcpy(gCdiCsvAddr.pattern,s,62);
    }
    else if (!strncmp(gCdiCsvAddr.pattern,"=",1))
    { /* looking for pattern equal to that given */
      patTrigger = CM_SEEKPATTERN;
      memset(s,0,64);
      strncpy(s,&gCdiCsvAddr.pattern[1],63);
      strtrm(s);
      memcpy(gCdiCsvAddr.pattern,s,63);
    }
    if ( gCdiCsvAddr.pattern[0] >= '0' && gCdiCsvAddr.pattern[0] <= '9' )
    { /* interpret as a byte pattern (up to 4 bytes, i.e. 8 chars) */
      strncpy(s,gCdiCsvAddr.pattern,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          patRange[1] = strtol(c1,NULL,0);
        }
        patRange[0] = strtol(c,NULL,0);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,0);
      memset(gCdiCsvAddr.pattern,0,32);
      memcpy(gCdiCsvAddr.pattern,&dw,4);
      /* *(int *)gCdiCsvAddr.pattern = dw; */
    }
    pPattern = gCdiCsvAddr.pattern;
  }
  mskRange[0] = mskRange[1] = -1;
  if (strlen(gCdiCsvAddr.mask) > 0)
  { /* 8 bytes */
    if (!strncmp(gCdiCsvAddr.mask,"0x",2) || !strncmp(gCdiCsvAddr.mask,"0X",2)) 
    { /* interpret as a byte pattern (up to 8 bytes, i.e. 16 chars) */
      strncpy(s,gCdiCsvAddr.mask,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          mskRange[1] = atoi(c1);
        }
        mskRange[0] = atoi(c);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,16);
      memset(gCdiCsvAddr.mask,0,32);
      memcpy(gCdiCsvAddr.mask,&dw,8); /* was 4 ! */
    }
    pDataMask = gCdiCsvAddr.mask;
  }
  if ((cc=cdi_assignDefaultDataMaskAndPattern(gCdiCsvAddr.name,pDataMask,mskRange,pPattern,patRange,patTrigger)) != 0)
  {
    cdilog("assign mask/pattern to %s -> %s",gCdiCsvAddr.name,erlst[cc]);
    return cc;
  }
  strncpy(fmtstr,gCdiCsvAddr.dataformat,32);
  fmt = getSimpleFormatType(strupr(gCdiCsvAddr.dataformat));
  if (fmt == CF_NULL)
  {
    if ((bf=parseFormatForBitfield(fmtstr)) != NULL)
    {
      fmt = (int)GetIntegerFormat((short)bf->fmt);
      cdi_assignBitfieldToDevice(gCdiCsvAddr.name,cdiGetBitfieldXRef(gCdiCsvAddr.name,bf));
    }
  }
  if (fmt == CF_NULL)
  {
    cdiDebugLevel = 1;
    cdilog("%.32s format entry %s not recognized!",gCdiCsvAddr.name,gCdiCsvAddr.dataformat);
    cdilog("please provide a valid data format for %s in cdiaddr.csv before restarting CDI",gCdiCsvAddr.name);
    cdilog("valid format strings are :\n\t%s\n\t%s",
      "\"INT8\" (or \"BYTE\"), \"INT16\" (or \"SHORT\"), \"INT32\" (or \"LONG\"),",
      "\"FLOAT\" (or \"SINGLE\"), \"DOUBLE\", \"CHAR\" (or \"TEXT\"), \"CHAR32\" (or \"NAME32\")");
    exit(1);
  }
  if ((cc=cdi_assignDataFormat(gCdiCsvAddr.name,getCdiFormat((short)fmt))) != 0)
  {
    cdilog("%.32s data format could not be assigned : %.32s",gCdiCsvAddr.name,erlst[cc]);
    return cc;
  }
  if ((cc=cdi_assignDefaultInterval(gCdiCsvAddr.name,gCdiCsvAddr.interval)) != 0)
  {
    cdilog("%.32s default update interval could not be assigned : %.32s",gCdiCsvAddr.name,erlst[cc]);
    return cc;
  }
  if (strlen(gCdiCsvAddr.datainput) > 0)
  {
    if (fmt == CF_NULL)
    {
      cdilog("%.32s data input provided without a valid format!",gCdiCsvAddr.name);
      return illegal_format;
    }
    if ((pDataInput=parseInputData(gCdiCsvAddr.datainput,fmt,&siz)) == NULL)
    {
      cdilog("data input could not be parsed!");
      return invalid_data;
    }
    if ((cc=cdi_assignDefaultInput(gCdiCsvAddr.name,pDataInput,siz,getCdiFormat((short)fmt))) != 0)
    {
      cdilog("%.32s data input could not be assigned : %.32s",gCdiCsvAddr.name,erlst[cc]);
      return cc;
    }
  }
  if (strlen(gCdiCsvAddr.datalimit) > 0)
  {
    if (fmt == CF_NULL)
    {
      cdilog("%.32s data input provided without a valid format!",gCdiCsvAddr.name);
      return illegal_format;
    }
    fsiz = fmtsizeof(fmt);
    strncpy(s,gCdiCsvAddr.datalimit,16);
    if ((c=strstr(s,CDI_PARAMETER_DELIMITER)) != NULL) 
    {
      if ((c1=strstr(c+1,CDI_PARAMETER_DELIMITER)) != NULL) 
      {
        *c1++ = 0;
        sizout2nd = strtol(c1,NULL,10) * fsiz;
      }
      *c++ = 0;
      sizin = strtol(c,NULL,10) * fsiz;
      if (siz > 0 && sizin > siz) sizin = siz;
    }
    rblen = strtol(s,NULL,10);
    sizout = rblen * fsiz;
    if ((cc=cdi_assignDataLimits(gCdiCsvAddr.name,sizout,sizin, sizout2nd)) != 0)
    {
      cdilog("data limits could not be assigned : %.32s",erlst[cc]);
      return cc;
    }
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) == 0 && strlen(gCdiCsvAddr.clbrrule) > 0)
  { /* support legacy csv column */
    strncpy(gCdiCsvAddr.clbrrule_rcv,gCdiCsvAddr.clbrrule,132);
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) > 0 || strlen(gCdiCsvAddr.clbrrule_snd) > 0 || strlen(gCdiCsvAddr.clbrrule) > 0)
  {
    if ((cc=cdi_assignDataCalibration(gCdiCsvAddr.name,parseCdiRules(gCdiCsvAddr.clbrrule_rcv),
      parseCdiRules(gCdiCsvAddr.clbrrule_snd),getCdiFormat((short)fmt))) != 0)
    {
      cdilog("%.32s data calibration rule could not be assigned : %.32s",gCdiCsvAddr.name,erlst[cc]);
      return cc;
    }
  }
  hstp = strlen(gCdiCsvAddr.clbrrule_rcv) > 0 ? "RECV.CLBR" : "RECV";
  parseHistInfo(gCdiCsvAddr.access,gCdiCsvAddr.name,hstp,fmt,rblen,gCdiCsvAddr.interval);
  access = getCdiAllowedAccess(gCdiCsvAddr.access);
  cdi_assignAllowedAccess(gCdiCsvAddr.name,access);
  db->cnt++;
  return 0;
}
CSVDB csvCdiAddrDb =
{
  csvCdiAddrColumns,
  sizeof(csvCdiAddrColumns)/sizeof(csvCdiAddrColumns[0]),
  0,
  csvCdiAddrInitializer,
  csvCdiAddrIterator,
  NULL,
  0,
  NULL,
  0,
  0
};

int csvCdiTemplateIterator(struct CSVDBStruct *db)
{
  int cc, i, fmt, len, fsiz, rblen=1, siz=0, sizin=0, sizout2nd=0, sizout=0, dw, access=0xffff;
  static int nparams, params[NUMBER_EXEC_PARAM];
  char templateName[32], templateField[32], templateLongField[64], fieldDescription[96];
  BYTE *pPattern=NULL, *pDataMask=NULL, *pDataInput=NULL;
  char *c, *c1, s[512], fmtstr[32];
  CdiTemplate *tmpl;
  CdiRule *rule_rcv=NULL,*rule_snd=NULL;
  int mskRange[2], patRange[2], patTrigger=CM_SEEKPATTERN;
  int cdiaddr[CDI_ADDR_SIZE], naddr=0;
  DBITFIELD *bf=NULL;

  /* gCdiCsvAddr has been filled in */
  /* if not a TEMPLATE entry then do nothing ... */
  if (stricmp(gCdiCsvAddr.bus,"TEMPLATE")) return 0;
  /* template name is in the device name string */
  strncpy(s,gCdiCsvAddr.name,132);
  if ((c=strstr(s,CDI_PARAMETER_DELIMITER)) == NULL)
  { /* badly formed ! */
    cdilog("template entry without field (%s)!",gCdiCsvAddr.name);
    return argument_list_error;
  }
  *c++ = 0;
  strncpy(templateName,s,16);
  if ((tmpl=getCdiTemplate(templateName)) == NULL)
  {
    cdilog("could not get template for %s!",templateName);
    return resources_exhausted;
  }
  if (strlen(gCdiCsvAddr.map) > 0) strncpy(gCdiCsvAddr.lngname,gCdiCsvAddr.map,64);

  strncpy(templateField,c,16);
  strncpy(templateLongField,gCdiCsvAddr.lngname,64);
  strncpy(fieldDescription,gCdiCsvAddr.description,96);

  /* combine addr and params into a single string for the parsing routine */
  if (strlen(gCdiCsvAddr.base) > 0) strncpy(gCdiCsvAddr.addr,gCdiCsvAddr.base,96);
  if ((len=(int)strlen(gCdiCsvAddr.params)) > 0)
  { /* there are parameters from the parameters column */
    if (len > (int)(191 - strlen(gCdiCsvAddr.addr)))
    { /* oops ! */
      cdilog("cannot concatenate %s to %s : string too long",gCdiCsvAddr.params,gCdiCsvAddr.addr);
    }
    else
    { /* add it on ... */
      strcat(gCdiCsvAddr.addr,strlen(gCdiCsvAddr.addr) > 0 ? ":" : "0:");
      strncat(gCdiCsvAddr.addr,gCdiCsvAddr.params,len);
    }
  }

  nparams = 0; s[0] = 0;
  if ((cc=cdiParseAddressString(gCdiCsvAddr.addr,cdiaddr,&naddr,params,&nparams)) != 0)
  {
    cdilog("cannot parse address string -> %s",erlst[cc]);
    return cc;
  }
  for (i=0; i<nparams; i++) sprintf(&s[strlen(s)],i == 0 ? "(%d" : " %d",params[i]);
  if (nparams > 0) strcat(s,")");

  cdilog("Template %s field %s : parameters %s",templateName,templateField,s);

  patRange[0] = patRange[1] = -1;
  if (strlen(gCdiCsvAddr.pattern) > 0)
  { /* 32 bytes */
    if (!strncmp(gCdiCsvAddr.pattern,"!=",2))
    { /* looking for pattern not equal to that given */
      patTrigger = CM_MISSPATTERN;
      memset(s,0,64);
      strncpy(s,&gCdiCsvAddr.pattern[2],62);
      strtrm(s);
      memcpy(gCdiCsvAddr.pattern,s,62);
    }
    else if (!strncmp(gCdiCsvAddr.pattern,"=",1))
    { /* looking for pattern equal to that given */
      patTrigger = CM_SEEKPATTERN;
      memset(s,0,64);
      strncpy(s,&gCdiCsvAddr.pattern[1],63);
      strtrm(s);
      memcpy(gCdiCsvAddr.pattern,s,63);
    }
    if ( gCdiCsvAddr.pattern[0] >= '0' && gCdiCsvAddr.pattern[0] <= '9' )
    { /* interpret as a byte pattern (up to 4 bytes, i.e. 8 chars) */
      strncpy(s,gCdiCsvAddr.pattern,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          patRange[1] = strtol(c1,NULL,0);
        }
        patRange[0] = strtol(c,NULL,0);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,0);
      memset(gCdiCsvAddr.pattern,0,32);
      *(int *)gCdiCsvAddr.pattern = dw;
    }
    pPattern = gCdiCsvAddr.pattern;
  }
  mskRange[0] = mskRange[1] = -1;
  if (strlen(gCdiCsvAddr.mask) > 0)
  { /* 8 bytes */
    if (!strncmp(gCdiCsvAddr.mask,"0x",2) || !strncmp(gCdiCsvAddr.mask,"0X",2)) 
    { /* interpret as a byte pattern (up to 8 bytes, i.e. 16 chars) */
      strncpy(s,gCdiCsvAddr.mask,64); s[63] = 0;
      if ((c=strchr(s,':')) != NULL)
      {
        *c++ = 0;
        if ((c1=strchr(c,':')) != NULL)
        {
          *c1++ = 0;
          mskRange[1] = atoi(c1);
        }
        mskRange[0] = atoi(c);
      }
      else
      {
        s[10] = 0;
      }
      dw = strtol(s,NULL,16);
      memset(gCdiCsvAddr.mask,0,32);
      memcpy(gCdiCsvAddr.mask,&dw,8); /* <<< this was a 4 ! */
    }
    pDataMask = gCdiCsvAddr.mask;
  }
  strncpy(fmtstr,gCdiCsvAddr.dataformat,32);
  fmt = getSimpleFormatType(strupr(gCdiCsvAddr.dataformat));
  if (fmt == CF_NULL)
  {
    if ((bf=parseFormatForBitfield(fmtstr)) != NULL)
    {
      fmt = (int)GetIntegerFormat((short)bf->fmt);
    }
  }
  if (strlen(gCdiCsvAddr.datainput) > 0)
  {
    if (fmt == CF_NULL)
    {
      cdilog("data input provided without a valid format!");
      return illegal_format;
    }
    if ((pDataInput=parseInputData(gCdiCsvAddr.datainput,fmt,&siz)) == NULL)
    {
      cdilog("data input could not be parsed!");
      return invalid_data;
    }
  }
  if (strlen(gCdiCsvAddr.datalimit) > 0)
  {
    if ((fmt=getSimpleFormatType(strupr(gCdiCsvAddr.dataformat))) == CF_NULL)
    {
      cdilog("data input provided without a valid format!");
      return illegal_format;
    }
    fsiz = fmtsizeof(fmt);
    strncpy(s,gCdiCsvAddr.datalimit,16);
    if ((c=strstr(s,CDI_PARAMETER_DELIMITER)) != NULL) 
    {
      if ((c1=strstr(c+1,CDI_PARAMETER_DELIMITER)) != NULL) 
      {
        *c1++ = 0;
        sizout2nd = strtol(c1,NULL,10) * fsiz;
      }
      *c++ = 0;
      sizin = strtol(c,NULL,10) * fsiz;
      if (siz > 0 && sizin > siz) sizin = siz;
    }
    rblen = strtol(s,NULL,10);
    sizout = rblen * fsiz;
  }
  if (fmt == CF_NULL)
  {
    cdiDebugLevel = 1;
    cdilog("%s format entry %s not recognized!",templateName,gCdiCsvAddr.dataformat);
    cdilog("please provide a valid data format for %s in cdiaddr.csv before restarting CDI",templateName);
    cdilog("valid format strings are :\n\t%s\n\t%s",
      "\"INT8\" (or \"BYTE\"), \"INT16\" (or \"SHORT\"), \"INT32\" (or \"LONG\"),",
      "\"FLOAT\" (or \"SINGLE\"), \"DOUBLE\", \"CHAR\" (or \"TEXT\"), \"CHAR32\" (or \"NAME32\")");
    exit(1);
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) == 0 && strlen(gCdiCsvAddr.clbrrule) > 0)
  { /* support legacy csv column */
    strncpy(gCdiCsvAddr.clbrrule_rcv,gCdiCsvAddr.clbrrule,132);
  }
  if (strlen(gCdiCsvAddr.clbrrule_rcv) > 0)
  {
    rule_rcv = parseCdiRules(gCdiCsvAddr.clbrrule_rcv);
  }
  if (strlen(gCdiCsvAddr.clbrrule_snd) > 0)
  {
    rule_snd = parseCdiRules(gCdiCsvAddr.clbrrule_snd);
  }
  parseHistInfo(gCdiCsvAddr.access,NULL,templateField,fmt,rblen,gCdiCsvAddr.interval);
  access = getCdiAllowedAccess(gCdiCsvAddr.access);

  if ((cc=addFieldToTemplate(tmpl,templateField,templateLongField,fieldDescription,
            params,nparams,pDataMask,mskRange,pPattern,patRange,patTrigger,
            pDataInput,sizin,sizout,sizout2nd,getCdiFormat((short)fmt),
            access,gCdiCsvAddr.interval,rule_rcv,rule_snd,bf)) != 0)
  {
    cdilog("cannot add template field %s to template %s : %.32s",templateField,templateName,erlst[cc]);
  }

  db->cnt++;
  return 0;
}

CSVDB csvCdiTmplDb =
{
  csvCdiAddrColumns,
  sizeof(csvCdiAddrColumns)/sizeof(csvCdiAddrColumns[0]),
  0,
  NULL,
  csvCdiTemplateIterator,
  NULL,
  0,
  NULL,
  0
};
int cdiAddrDbSize = 0;
int getMaxTemplateLength(void)
{
  int i, maxsiz=1;

  for (i=0; i<CdiTemplateTblSize; i++)
  {
    if (CdiTemplateTbl[i] == NULL) continue;
    if (CdiTemplateTbl[i]->length > maxsiz) maxsiz = CdiTemplateTbl[i]->length;
  }
  return maxsiz;
}

int csvCdiBitfieldIterator(struct CSVDBStruct *db)
{
  int cc, fmt, msk;
  char bitfieldName[32], bitfieldField[32];
  char *c, s[512];

  /* gCdiCsvAddr has been filled in */
  /* if not a BITFIELD definition entry then do nothing ... */
  if (stricmp(gCdiCsvAddr.bus,"BITFIELD")) return 0;
  /* bitfield name and field name are in the device name string */
  strncpy(s,gCdiCsvAddr.name,132);
  if ((c=strstr(s,CDI_PARAMETER_DELIMITER)) == NULL)
  { /* badly formed ! */
    cdilog("bitfield entry without field (%s)!",gCdiCsvAddr.name);
    return argument_list_error;
  }
  *c++ = 0;
  strncpy(bitfieldName,s,16);
  fmt = getSimpleFormatType(strupr(gCdiCsvAddr.dataformat));
  if (fmt == CF_NULL) fmt = CF_INT16;
  if ((cc=OpenBitField(BITFIELD_LCL,bitfieldName,fmt)) != 0)
  {
    cdilog("could not open bitfield %s : %s",bitfieldName,erlst[cc]);
    return cc;
  }
  if (strlen(c) > 16) cdilog("truncating %s field name %s to 16 characters!",bitfieldName,c);
  strncpy(bitfieldField,c,16); bitfieldField[16] = 0;
  msk = 0xffffffff;
  if (strlen(gCdiCsvAddr.mask) > 0)
  { /* 8 bytes */
    if (!strncmp(gCdiCsvAddr.mask,"0x",2) || !strncmp(gCdiCsvAddr.mask,"0X",2)) 
    { /* interpret as a byte pattern (up to 4 bytes, i.e. 8 chars) */
      strncpy(s,gCdiCsvAddr.mask,64); s[63] = 0;
      msk = strtol(s,NULL,16);
    }
  }
  AddFieldToBitField(BITFIELD_LCL,bitfieldName,msk,bitfieldField);
  return 0;
}

CSVDB csvCdiBfDb =
{
  csvCdiAddrColumns,
  sizeof(csvCdiAddrColumns)/sizeof(csvCdiAddrColumns[0]),
  0,
  NULL,
  csvCdiBitfieldIterator,
  NULL,
  0,
  NULL,
  0
};

FieldBus *gFieldBusList = NULL;
FieldBus *cdiGetFieldBus(char *busName)
{
  FieldBus *f = NULL;
  for (f=gFieldBusList; f != NULL; f = f->nxt)
  {
    if (strnicmp(f->busName,busName,32) == 0) break;
  }
  return f;
}
char *cdiGetFieldBusName(char *busType,int line)
{
  FieldBus *f = NULL;
  static char fb[64];
  for (f=gFieldBusList; f != NULL; f = f->nxt)
  {
    if (strnicmp(f->busType,busType,32) != 0) continue;
    if (f->line != line) continue;
    return f->busName;
  }
  sprintf(fb,"%s-line%d",busType,line);
  return fb;
}
int cdiAddFieldBus(char *busName,char *busType,int line,int fmt)
{
  FieldBus *f = cdiGetFieldBus(busName);
  if (f != NULL) return already_assigned;
  f = (FieldBus *)calloc(1,sizeof(FieldBus));
  if (f == NULL) return out_of_server_memory;
  strncpy(f->busName,busName,32);
  strncpy(f->busType,busType,16);
  f->line = line;
  f->fmt = fmt;
  f->nxt = gFieldBusList;
  gFieldBusList = f;
  return 0;
}
extern CdiBusInfoBlk busList[NUMBER_LINES];
CDI_EXPORT int cdiGetFieldBusNames(char *nameBuffer,int size,int format)
{
  FieldBus *f = NULL;
  int nlen, clen = 32;
  int n = 0;
  char bn[64];

  switch (format)
  {
    case CF_NAME16:
    case CF_NAME32:
    case CF_NAME64:
      nlen = fmtsizeof(format);
      break;
    default:
      return illegal_format;
  }
  if (nlen < 32) clen = nlen;
  for (f=gFieldBusList, n=0; f != NULL && n < size; f=f->nxt, n++)
  {
    strncpy(nameBuffer + n*nlen,f->busName,clen);
  }
  if (n == 0)
  {
    for (n=0; n<numConnectedHandles; n++)
    {
      sprintf(bn,"%s-line%d",busList[busHandle[n]->busIndex].busName,busHandle[n]->busLine);
      strncpy(nameBuffer + n*nlen,bn,clen);
    }
  }
  return 0;
}

int csvCdiFieldbusIterator(struct CSVDBStruct *db)
{
  int cc, fmt;
  char busName[32], fieldName[64];
  char *c, s[512];

  /* gCdiCsvAddr has been filled in */
  /* if not a FIELDBUS definition entry then do nothing ... */
  if (stricmp(gCdiCsvAddr.bus,"FIELDBUS")) return 0;
  /* generic bus name and fieldbus name are in the device name string */
  strncpy(s,gCdiCsvAddr.name,132);
  if ((c=strstr(s,CDI_PARAMETER_DELIMITER)) == NULL)
  { /* badly formed ! */
    cdilog("bus entry without fieldbus name (%s)!",gCdiCsvAddr.name);
    return argument_list_error;
  }
  *c++ = 0;
  strncpy(busName,s,16);
  strncpy(fieldName,c,32);
  fmt = getSimpleFormatType(strupr(gCdiCsvAddr.dataformat));

  cc = cdiAddFieldBus(fieldName,busName,gCdiCsvAddr.line,fmt);
  cdilog("cdiAddFieldBus %s (%s line %d) : %s",fieldName,busName,gCdiCsvAddr.line,erlst[cc]);

  return 0;
}

CSVDB csvCdiFldBusDb =
{
  csvCdiAddrColumns,
  sizeof(csvCdiAddrColumns)/sizeof(csvCdiAddrColumns[0]),
  0,
  NULL,
  csvCdiFieldbusIterator,
  NULL,
  0,
  NULL,
  0
};

CDI_EXPORT int readCdiDataBase(void)
{
  int cc=0;
  fixCdiRepository();
  /* scan for fieldbus names */
  cc = csvReadFile(CdiDBpath,CDIDB_FILE,&csvCdiFldBusDb,NULL);
  if (cc == 0 && csvCdiFldBusDb.siz > 0)
  {
    cdilog("field bus names registered");
  }
  /* scan for bitfields */
  cc = csvReadFile(CdiDBpath,CDIDB_FILE,&csvCdiBfDb,NULL);
  if (cc == 0 && csvCdiBfDb.siz > 0)
  {
    cdilog("bitfields registered");
  }
  /* scan for templates */
  cc = csvReadFile(CdiDBpath,CDIDB_FILE,&csvCdiTmplDb,NULL);
  if (cc == 0)
  {
    cdilog("templates registered");
    gMaxTemplateLength = getMaxTemplateLength();
    cdilog("largest template has %d fields",gMaxTemplateLength);
  }
  /* scan for records */
  cc = csvReadFile(CdiDBpath,CDIDB_FILE,&csvCdiAddrDb,NULL);
  cdiAddrDbSize = csvCdiAddrDb.siz;
  cdilog("read database file %s%s : %s",CdiDBpath,CDIDB_FILE,erlst[cc]);
  return cc;
}
#ifdef UNIX
/* prolog code for UNIX shared library (WINDOWS uses DllMain in cdi32.c) */
__attribute__((constructor)) void cdiLibMain(void)
{
  cdiInitialize();
}
#endif
/* ===============================================================
  Below are routines which provide the server functionality
 ================================================================= */
CDI_EXPORT int cdiSetExportName(char *expName)
{
  if (expName == NULL || strlen(expName) == 0) return invalid_parameter;
  strncpy(gCdiExportName,expName,EXPORT_NAME_SIZE);
  return 0;
}
/* call cdiPostSystemInit() following the call to SystemInit() ... */
CDI_EXPORT void cdiPostSystemInit(void)
{
  char *ptr, cdiExportName[EXPORT_NAME_SIZE+4];
  int ndevices;
  int devParameters[NUMBER_EXEC_PARAM];
  if (gCdiEqmInitialized) return;
  gRetardSingleContractRemoval = FALSE;
  cdiGetBusNames();
  if ((ptr=getenv("CDI_EXCLUSIVE_READ")) != NULL)
  {
    gCdiHasXRead = (!stricmp(ptr,"TRUE") || !stricmp(ptr,"YES")) ? TRUE : FALSE;
  }
  if (strlen(gFecName) == 0)
  {
    cdilog("fecname unknown at time of CDI initialization!");
    return;
  }
  memset(devParameters,0,NUMBER_EXEC_PARAM*sizeof(int));
  cdi_RegisterParameters("null", devParameters);
  if (strlen(gCdiExportName) == 0)
  {
    if ((ptr=getenv("CDI_SERVERNAME")) != NULL)
    {
      strncpy(gCdiExportName,ptr,EXPORT_NAME_SIZE);
      cdilog("assign export name %s from environment",gCdiExportName);
    }
    else
    {
      strncpy(cdiExportName,gFecName,EXPORT_NAME_SIZE-4); cdiExportName[EXPORT_NAME_SIZE-4] = 0;
      strcat(cdiExportName,".CDI");
      strncpy(gCdiExportName,cdiExportName,EXPORT_NAME_SIZE);
    }
  }
  cdilog("CDI export name : %.32s",gCdiExportName);
  ndevices = cdi_getDeviceNames(NULL,0) + nCdiRedirectedDevices + gNumBitfieldDevices;
  RegisterEquipmentModule(gCdiExportName,CDIEQM_TAG,ndevices,cdieqm,cdieqm_ini,cdieqm_bkg,CDIEQM_RATE,cdieqm_exi);
  strncpy(gCdiExportContext,GetExportedContext(CDIEQM_TAG),CONTEXT_NAME_SIZE);
  sprintf(gCdiExportPrefix,"/%.32s/%.32s/",gCdiExportContext,gCdiExportName);
  gCdiExportPrefixLength = (int)strlen(gCdiExportPrefix);
  cdilog("CDI remote server initialized");
  gCdiEqmInitialized = TRUE;
  //cdilog("CDI set server idle");
  //SetServerCycleIdle(TRUE); /* 19.2.21 */
}
CDI_EXPORT void cdieqm_bkg(void)
{
  int cc;
  static int devSrvDone = FALSE;
  if (!gCdiEqmInitDone) return;
  if (!devSrvDone)
  {
    if ((cc=readCdiDevSrvInfo()) != 0) cdilog("readCdiDevSrvInfo : %s",erlst[cc]);
    devSrvDone = TRUE;
    cdilog("CDI set server active");
    SetServerCycleIdle(FALSE); /* 19.2.21 */
  }
  if (gSetToIdleState)
  {
    cdiSetIdleState(gCdiIdleState);
    gSetToIdleState = FALSE;
  }
}
NAME64 dbdevs[4] = { "REPLACE", "UPDATE", "ROLLBACK", "DUMP" };
int ndbdevs = 4;
int getDbUpdateType(char *devStr)
{
  int i;
  if (devStr == NULL) return -1;
  for (i=0; i<ndbdevs; i++) if (!strnicmp(devStr,dbdevs[i].name,64)) return i;
  return -1;
}
CDI_EXPORT int cdiRegisterProperties(void)
{
  int cc = 0;
  short accRead = gCdiHasXRead ? CA_XREAD : CA_READ;
  DTYPE dout, din;

  DTYPEZERO(dout); DTYPEZERO(din);
  cdilog("register standard CDI properties");
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"IDLE",1,CF_INT32,CA_READ|CA_WRITE,"[0:1]get/set CDI Idle state",PRP_CDI_IDLE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"CDIVERSION",PRP_CDIVERSION_SIZE,CF_TEXT,CA_READ,"[0:65535 ]CDI library version",PRP_CDIVERSION)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"INSTANCES.NAM",PRP_TEMPLATE_NAM_SIZE,CF_NAME16,CA_READ,"[0:65535 ]registered template names",PRP_NTEMPLATEOBJECTS_NAM)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"INSTANCES",PRP_TEMPLATEOBJECTS_SIZE,CF_NAME16,CA_READ,"[0:65535 ]object names using template ",PRP_TEMPLATEOBJECTS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"NINSTANCES",PRP_NTEMPLATEOBJECTS_SIZE,CF_INT32,CA_READ,"[0:65535 ]number of objects using template ",PRP_NTEMPLATEOBJECTS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"NINSTANCES.NAM",PRP_TEMPLATE_NAM_SIZE,CF_NAME16,CA_READ,"[0:65535 ]registered template names",PRP_NTEMPLATEOBJECTS_NAM)) < 0) goto err;

  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSTYPE",PRP_BUS_TYPE_SIZE,CF_NAME16,CA_READ,"[0:65535 ]Bus type in use by selected device",PRP_BUS_TYPE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSTYPES",PRP_BUS_TYPES_SIZE,CF_NAME16,CA_READ,"[0:65535 ]Supported Bus types",PRP_BUS_TYPES)) < 0) goto err;

  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSDEVICES.NAM",PRP_CDIDEVNAMES_NAM_SIZE,CF_NAME16,CA_READ,"[0:65535 ]Line Names ",PRP_CDIDEVNAMES_NAM)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSDEVICES",PRP_CDIDEVNAMES_SIZE,CF_NAME32,CA_READ,"[0:65535 ]registered CDI devices for give bus line ",PRP_CDIDEVNAMES)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"LINESTATUS",PRP_CDIINFOR_SIZE,CF_NAME64I,CA_READ,"[0:65535 ]CDI line status(Number active request) ",PRP_CDIINFOR)) < 0) goto err;

  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSADDR",PRP_ADDR_SIZE,CF_ADDRESS,CA_READ,"[0:65535 ]HW address in use by selected device",PRP_ADDR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSNAME",PRP_BUS_NAME_SIZE,CF_NAME32,CA_READ,"[0:65535 ]Fieldbus in use by selected device",PRP_BUS_NAME)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSERRORS",PRP_BUS_ERRORS_SIZE,CF_INT32,CA_READ,"[0:65535 ]Number of bus errors since server start",PRP_BUS_ERRORS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSNAMES",PRP_BUS_NAMES_SIZE,CF_NAME32,CA_READ,"[0:65535 ]Supported Bus names",PRP_BUS_NAMES)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSSCAN",PRP_BUS_SCAN_SIZE,CF_TEXT,CA_READ,"[0:65535 ]Scan bus according to input given",PRP_BUS_SCAN)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"BUSSCAN.NAM",PRP_BUS_SCAN_NAM_SIZE,CF_NAME16,CA_READ,"[0:65535 ]registered bus names",PRP_BUS_SCAN_NAM)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"TEMPLATE",PRP_TEMPLATE_SIZE,CF_NAME16II,CA_READ,"[0:65535 ]template information",PRP_TEMPLATE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"TEMPLATE.NAM",PRP_TEMPLATE_NAM_SIZE,CF_NAME16,CA_READ,"[0:65535 ]registered template names",PRP_TEMPLATE_NAM)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.ATOM",PRP_WRITEREAD_SIZE,CF_SHORT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+Read Telegram Pair",PRP_WRITEREAD)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.SEND.ATOM",PRP_READWRITE_SIZE,CF_SHORT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Read+Write Telegram Pair",PRP_READWRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND",PRP_WRITE_SIZE,CF_SHORT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write Telegram",PRP_WRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV",PRP_READ_SIZE,CF_SHORT,accRead,"[0:65535 ]Issue CDI Read Telegram",PRP_READ)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.CLBR",PRP_READ_SIZE,CF_FLOAT,accRead,"[0:65535 ]Issue CDI Read Telegram with calibration",PRP_READ_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.CLBR",PRP_WRITEREAD_SIZE,CF_FLOAT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+Read Telegram Pair with calibration",PRP_WRITEREAD_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"DESCRIPTION",PRP_DESCRIPTION_SIZE,CF_NAME64,CA_READ,"[0:0 ]description selected device",PRP_DESCRIPTION)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.SEND.ATOM",PRP_WRITE_SIZE,CF_SHORT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+Write Telegram",PRP_WRITEWRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.SEND.ATOM",PRP_READ_SIZE,CF_SHORT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+READ+Write Telegram",PRP_WRITEREADWRITE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.RECV.ATOM",PRP_READ_SIZE,CF_SHORT,accRead,"[0:65535 ]Issue CDI READ+READ Telegram",PRP_READREAD)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.RECV.ATOM",PRP_READ_SIZE,CF_SHORT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+READ+READ Telegram",PRP_WRITEREADREAD)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"NUMBER",PRP_CDINUMBER_SIZE,CF_ADDRESS,CA_READ,"[0:1000 ]CDI device number (plus line and index) assigned to selected device",PRP_CDINUMBER)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RANGE",PRP_RANGE_SIZE,CF_FLOAT,CA_READ,"[0:1000 ]value range (calibrated) for selected device",PRP_RANGE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"UNITS",PRP_UNITS_SIZE,CF_NAME16,CA_READ,"[0:0 ]assigned units (calibrated) for selected device",PRP_UNITS)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"DISPLAYSIZE",1,CF_INT32,CA_READ,"[0:100 ]display (bus) size for selected device",PRP_SIZE)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"DISPLAYFORMAT",1,CF_INT32,CA_READ,"[0:255 ]display (bus) format for selected device",PRP_FORMAT)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"PRECISION",PRP_PRECISION_SIZE,CF_NAME16,CA_READ,"[0:0 ]assigned format precision (calibrated) for selected device",PRP_PRECISION)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.CLBR",PRP_WRITE_SIZE,CF_FLOAT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write Telegram",PRP_WRITE_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.SEND.CLBR",PRP_READWRITE_SIZE,CF_FLOAT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Read+Write Telegram Pair",PRP_READWRITE_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.SEND.CLBR",PRP_WRITE_SIZE,CF_FLOAT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+Write Telegram",PRP_WRITEWRITE_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.SEND.CLBR",PRP_READ_SIZE,CF_FLOAT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+READ+Write Telegram",PRP_WRITEREADWRITE_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"RECV.RECV.CLBR",PRP_READ_SIZE,CF_FLOAT,accRead,"[0:65535 ]Issue CDI READ+READ Telegram",PRP_READREAD_CLBR)) < 0) goto err;
  if ((cc=RegisterPropertyEx(CDIEQM_TAG,"SEND.RECV.RECV.CLBR",PRP_READ_SIZE,CF_FLOAT,(short)(accRead|CA_WRITE),"[0:65535 ]Issue CDI Write+READ+READ Telegram",PRP_WRITEREADREAD_CLBR)) < 0) goto err;
  /* database management properties */
  dout.dFormat = CF_STRUCT; dout.dArrayLength = DB_NICE_QUERY_SIZE; strncpy(dout.dTag,DBITEM_TAG,TAG_NAME_SIZE);
  if ((cc=RegisterPropertyInformation(CDIEQM_TAG,"CDIADDR.DB",
       &dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set address database entries",PRP_ADDRDB,NULL)) < 0) goto err;
  AssignDeviceListToProperty(CDIEQM_TAG,"CDIADDR.DB",dbdevs,ndbdevs);
  dout.dFormat = CF_STRUCT; dout.dArrayLength = MF_NICE_QUERY_SIZE; strncpy(dout.dTag,MFITEM_TAG,TAG_NAME_SIZE);
  if ((cc=RegisterPropertyInformation(CDIEQM_TAG,"CDIMF.DB",
       &dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set address database entries",PRP_MFDB,NULL)) < 0) goto err;
  AssignDeviceListToProperty(CDIEQM_TAG,"CDIMF.DB",dbdevs,ndbdevs);
  dout.dFormat = CF_STRUCT; dout.dArrayLength = SCH_NICE_QUERY_SIZE; strncpy(dout.dTag,SCHDITEM_TAG,TAG_NAME_SIZE);
  if ((cc=RegisterPropertyInformation(CDIEQM_TAG,"CDISCHED.DB",
       &dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set address database entries",PRP_SCHEDDB,NULL)) < 0) goto err;
  AssignDeviceListToProperty(CDIEQM_TAG,"CDISCHED.DB",dbdevs,ndbdevs);
  dout.dFormat = CF_STRUCT; dout.dArrayLength = SRV_NICE_QUERY_SIZE; strncpy(dout.dTag,SRVITEM_TAG,TAG_NAME_SIZE);
  if ((cc=RegisterPropertyInformation(CDIEQM_TAG,"CDISERV.DB",
       &dout,&dout,CA_READ|CA_WRITE,AT_UNKNOWN,0,"get/set address database entries",PRP_SRVDB,NULL)) < 0) goto err;
  AssignDeviceListToProperty(CDIEQM_TAG,"CDISERV.DB",dbdevs,ndbdevs);
  cdilog("finished registering standard CDI properties");
err:
  return cc < 0 ? -cc : 0;
}
char cdiDeviceFilter[64] = "";
int setCdiDeviceFilter(int argc, char *argv[])
{
  char *c, filter[512];
  if (argc < 1 || argv[0] == NULL || strlen(argv[0]) == 0)
  {
    cdiDeviceFilter[0] = 0;
    ttyoutput("cdi target cleared");
    return 0;
  }
  memset(filter, 0, 512);
  c = argv[0];
  if (*c == '"')
  {
    c++;
    strncpy(filter, c, 511);
    if ((c=strchr(filter, '"')) != NULL) *c = 0;
  }
  else
  {
    strncpy(filter, argv[0], 511);
  }
  strncpy(cdiDeviceFilter, filter, 32);
  ttyoutput("cdi target set to %.32s",cdiDeviceFilter);
  return 0;
}
int isPastDeviceFilter(char *str)
{
  if (str == NULL || strlen(cdiDeviceFilter) == 0) return FALSE; // none set 
  return strstr(str, cdiDeviceFilter) != NULL ? TRUE : FALSE;
}
CDI_EXPORT void cdieqm_ini(void)
{
  int cc;
  char *ptr;
  if ((ptr=getenv("CDI_DEFAULT_POLLING_RATE")) != NULL)
  {
    SetDefaultQueryPollingInterval(atoi(ptr));
  }
  cdiAllocateDeviceStringBuffer();
  if ((cc=cdiRegisterProperties()) != 0) cdilog("cdiRegisterProperties : %s",erlst[cc]);
  if ((cc=cdiRegisterDeviceNames()) != 0) cdilog("cdiRegisterDeviceNames : %s",erlst[cc]);
  if ((cc=cdiRegisterExtendedProperties()) != 0) cdilog("cdiRegisterExtendedProperties : %s",erlst[cc]);
  gNumHistories = registerHistories();
# if defined(UNIX) || defined(WIN32)
  RegisterUserCommand("cdidebug",NULL,(int *)&cdiDebugLevel,NULL,CA_READ|CA_WRITE);
  RegisterUserCommand("busplugdebug",NULL,(int *)&cdiBusPlugDebugLevel,NULL,CA_READ|CA_WRITE);
  RegisterUserCommand("cdilisteners",cdiDumpListeners,NULL,NULL,CA_READ);
  RegisterUserCommand("cdidevices",cdiDumpDevices,NULL,NULL,CA_READ);
  RegisterUserCommand("cdilineinfo",cdiDumpLineInfo,NULL,NULL,CA_READ);
  RegisterUserCommand("cdibuslist",cdiDumpBusList,NULL,NULL,CA_READ);
  RegisterUserFunction("cditarget", setCdiDeviceFilter);
# endif
  gCdiEqmInitDone = TRUE;
}
CDI_EXPORT void cdieqm_exi(void)
{
}
DBITFIELD *getBitfieldFromDeviceName(char *devName,char *field,int removeField)
{
  char *c=NULL, *c0, *c1, *pdn;
  static char dn[DEVICE_NAME_SIZE];
  if (strchr(devName,'.') == NULL) 
  { /* if it does point to a bitfield, this request wants the entire value */
    return NULL; 
  }
  if (!removeField)
  {
    strncpy(dn,devName,DEVICE_NAME_SIZE);
    pdn = dn;
  }
  else
  {
    pdn = devName;
  }
  for (c0=pdn; (c1=strpbrk(c0,".")) != NULL; c0=++c1) c = c1;
  if (c != NULL) *c++ = 0;
  if (field != NULL && c!= NULL) strncpy(field,c,16);
  return cdiGetBitfieldFromDevice(pdn);
}
int getTemplateIndex(char *name)
{
  int i;
  for (i=0; i<CdiTemplateTblSize; i++)
    if (!strnicmp(name,CdiTemplateTbl[i]->name,16)) return i;
  return -1;
}
CDI_EXPORT int cdiGetTemplates(NAME16 *names,int *num)
{
  int i;

  if (num == NULL || names == NULL) return argument_list_error;
  for (i=0; i<CdiTemplateTblSize && i< *num; i++)
  {
    strncpy(names[i].name,CdiTemplateTbl[i]->name,16);
  }
  *num = i;
  return 0;
}
CDI_EXPORT int cdiGetTemplateInstances(char *tmplName,NAME32 *names,int *num)
{
  int i, idx;
  CdiTemplate *tmpl;
  CdiTemplateInstanceName *tmpln;

  if (tmplName == NULL || num == NULL || names == NULL) return argument_list_error;
  if ((idx=getTemplateIndex(tmplName)) < 0) return un_allocated;
  if ((tmpl=CdiTemplateTbl[idx]) == NULL) return un_allocated;
  tmpln = tmpl->instanceNames;
  for (i=0; i<*num && i<tmpl->numInstances; i++)
  {
    if (tmpln == NULL) break;
    strncpy(names[i].name,tmpln->name,32);
    tmpln = tmpln->nxt;
  }
  *num = i;
  return 0;
}
CDI_EXPORT int cdiGetNumTemplateInstances(char *tmplName)
{
  int idx;
  CdiTemplate *tmpl;

  if (tmplName == NULL) return -argument_list_error;
  if ((idx=getTemplateIndex(tmplName)) < 0) return -un_allocated;
  if ((tmpl=CdiTemplateTbl[idx]) == NULL) return -un_allocated;
  return tmpl->numInstances;
}
int isCdiBusAccessProperty(char *prp)
{
  int prpid;
  if ((prpid=GetPropertyId(CDIEQM_TAG,prp)) < 0) return FALSE;
  if (prpid >= PRP_EXTENDED_OFFSET) return TRUE;
  switch (prpid)
  {
    case PRP_WRITEREADREAD_CLBR:
    case PRP_READREAD_CLBR:
    case PRP_WRITEREADWRITE_CLBR:
    case PRP_WRITEWRITE_CLBR:
    case PRP_READWRITE_CLBR:
    case PRP_WRITE_CLBR:
    case PRP_WRITEREAD_CLBR:
    case PRP_READ_CLBR:
    case PRP_WRITEREADREAD:
    case PRP_READREAD:
    case PRP_WRITEREADWRITE:
    case PRP_WRITEWRITE:
    case PRP_WRITEREAD:
    case PRP_READWRITE:
    case PRP_WRITE:
    case PRP_READ:
      return TRUE;
  }
  return FALSE;
}
int isCdiDeviceProperty(int prpid)
{
  switch (prpid)
  {
    case PRP_FORMAT:
    case PRP_SIZE:
    case PRP_RANGE:
    case PRP_UNITS:
    case PRP_PRECISION:
    case PRP_CDINUMBER:
    case PRP_CDIINFOR:
    case PRP_DESCRIPTION:
    case PRP_ADDR:
    case PRP_BUS_NAME:
    case PRP_BUS_TYPE:
    case PRP_BUS_ERRORS:
      return TRUE;
  }
  return FALSE;
}
CDI_EXPORT void SetDefaultQueryPollingInterval(int value)
{
  if (value <= 0) return;
  cdilog("set default query polling interval to %d",value);
  gDefaultQueryPollingRate = value;
}
static time_t gFirstEqmCall = 0;
static int gHistoryLinksStarted = FALSE;
static int gNumHistoryStarts = 0;
#define MAX_QUERYABLE_LIST_SIZE 1024
static NAME32 qryDevList[MAX_QUERYABLE_LIST_SIZE];
static short qryDevMca[MAX_QUERYABLE_LIST_SIZE];
CDI_EXPORT int cdieqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc=0,devnr=0,prpid,extid=-1,errcnt=0,i,len,namlen,issync,rate,drate,flags,iaddr[2*CDI_ADDR_SIZE];
  int isfirst = (access & CA_FIRST);
  int islast = (access & CA_LAST);
  int iswrite = (access & CA_WRITE);
  int isquery = (access & (CA_HIST|CA_ALARM));
  int cdiDevNr, cdiLine, cdiIndex, numDevices, devFmt, devSiz;
  int lineinf[PRP_CDIINFOR_SIZE];
  float range[2];
  char *c = NULL, *scanTarget = NULL, dn[1048], strbuffer[256], devDesc[64], field[32], cdiDevName[32], scratch[32];
  NAME16 *n16;
  DADDRESS da = { 0,0,0,0 }, *dap;
  CdiTemplate *tmpl;
  CdiTemplateInstanceName *tmpln;
  CdiTemplateList *tmplst;
  cdiListener *ml;
  double dt;
  DBITFIELD *bf;

  /* get device number from device name */
  devnr = GetDeviceNumberEx(CDIEQM_TAG,devName,devProperty);
  prpid = GetPropertyId(CDIEQM_TAG,devProperty);
  if (prpid >= PRP_EXTENDED_OFFSET &&
      prpid < PRP_EXTENDED_OFFSET + gNumExtCdiProperties)
  {
    extid = prpid - PRP_EXTENDED_OFFSET;
    if (devName[0] == '#' && devnr >= 0 && devnr < extCdiPrps[extid].ndevs)
    { /* extended property device number supplied: work with real name below */
      strncpy(scratch,extCdiPrps[extid].devs[devnr].name,16); scratch[16] = 0;
      devName = scratch;
    }
    snprintf(cdiDevName,CDI_DEVICE_NAME_SIZE,"%s.%s",devName,devProperty);
  }
  else
  {
    strncpy(cdiDevName,devName,CDI_DEVICE_NAME_SIZE);
  }

  /* prepare local device name */
  sprintf(dn,"/LOCALHOST/CDI/%s",devName);
  
  issync = (isfirst && islast) ? TRUE : FALSE; /* is this a synchronous request ? */

  if (isCdiDeviceProperty(prpid))
  { /* cdi device property versus a cdi bus property */
    cdiDevNr = findDevice(devName,&cdiLine,&cdiIndex);
    if (cdiDevNr < 0) 
    { /* not found as given : is it decorated with bitfield name ? */
      if (getBitfieldFromDeviceName(devName,NULL,TRUE) == NULL) return -cdiDevNr; /* TRUE => will strip off the field name */
      if ((cdiDevNr=findDevice(devName,&cdiLine,&cdiIndex)) < 0) return -cdiDevNr;
    }
  }
  switch (prpid)
  {
    case PRP_RANGE:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,qryDevList,&numDevices)) != 0) 
        {
          cdiAssertHeapOkay("cdieqm",devProperty,cc);          
          return cc;
        }
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        switch (dout->dFormat)
        {
          case CF_FLOAT:
          case CF_INT32:
          case CF_SHORT:
          case CF_DOUBLE:
            if (numDevices > 1) return illegal_format;
            cdi_getRangeForDevice(devName,range);
            cc = putValuesFromFloat(dout,range,2);
            cdiAssertHeapOkay("cdieqm",devProperty,cc);          
            return cc;
          case CF_FLTFLT:
           if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
            for (i=0; i<numDevices; i++)
            {
              cdi_getRangeForDevice(qryDevList[i].name,range);
              ((DPOINT *)dout->data.vptr)[i].x = range[0];
              ((DPOINT *)dout->data.vptr)[i].y = range[1];
            }
            break;
          case CF_DBLDBL:
            if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
            for (i=0; i<numDevices; i++)
            {
              cdi_getRangeForDevice(qryDevList[i].name,range);
              ((DBLDBL *)dout->data.vptr)[i].d1val = (double)range[0];
              ((DBLDBL *)dout->data.vptr)[i].d2val = (double)range[1];
            }
            break;
          default:
            return illegal_format;
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_UNITS:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,qryDevList,&numDevices)) != 0) 
        {
          cdiAssertHeapOkay("cdieqm",devProperty,cc);          
          return cc;
        }
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        len = fmtsizeof(dout->dFormat);
        switch (dout->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            namlen = 16;
            break;
          case CF_TEXT:
            if (dout->dArrayLength > 1) return illegal_format;
            namlen = MIN(dout->dArrayLength,16);
            break;
          default:
            return illegal_format;
        }
        for (i=0; i<numDevices; i++)
        {
          cdi_getUnitsForDevice(qryDevList[i].name,devDesc);
          strncpy(dout->data.cptr+i*namlen,devDesc,namlen);
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_PRECISION:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,qryDevList,&numDevices)) != 0) 
        {
          cdiAssertHeapOkay("cdieqm",devProperty,cc);          
          return cc;
        }
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        len = fmtsizeof(dout->dFormat);
        switch (dout->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            namlen = 16;
            break;
          case CF_TEXT:
            if (dout->dArrayLength > 1) return illegal_format;
            namlen = MIN(dout->dArrayLength,16);
            break;
          default:
            return illegal_format;
        }
        for (i=0; i<numDevices; i++)
        {
          cdi_getPrecisionForDevice(qryDevList[i].name,devDesc);
          strncpy(dout->data.cptr+i*namlen,devDesc,namlen);
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_FORMAT:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,qryDevList,&numDevices)) != 0) 
        {
          cdiAssertHeapOkay("cdieqm",devProperty,cc);          
          return cc;
        }
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        switch (dout->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            namlen = 16;
            break;
          default:
            namlen = 0;
            break;
        }
        for (i=0; i<numDevices; i++)
        {
          cdi_getFormatForDevice(qryDevList[i].name,&devFmt);
          qryDevMca[i] = getTineFormat((short)devFmt);
          if (qryDevMca[i] == CF_NULL) qryDevMca[i] = CF_INT16; /* if not explicitly specified */
          if (namlen > 0) strncpy(dout->data.cptr+i*namlen,GetFormatAsString(qryDevMca[i]),namlen);
        }
        if (namlen == 0) cc = PutValuesFromShortEx(dout,qryDevMca,numDevices,0);
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return cc;
    case PRP_SIZE:
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,qryDevList,&numDevices)) != 0) 
        {
          cdiAssertHeapOkay("cdieqm",devProperty,cc);
          return cc;
        }
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        for (i=0; i<numDevices; i++)
        {
          if ((cc=cdi_getOutputSizeForDevice(qryDevList[i].name,&devSiz)) != 0) 
          {
            cdiAssertHeapOkay("cdieqm",devProperty,cc);
            return cc;
          }
          qryDevMca[i] = devSiz;
        }
        cc = PutValuesFromShortEx(dout,qryDevMca,numDevices,0);
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return cc;
    case PRP_CDIDEVNAMES_NAM:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          switch (dout->dFormat)
          {
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME64:
              namlen = fmtsizeof(dout->dFormat);
              break;
            default: 
              return illegal_format;
          }
          if (numConnectedHandles < (int)dout->dArrayLength) dout->dArrayLength = numConnectedHandles;
          for (i=0; i<(int)dout->dArrayLength; i++)
          {
            sprintf(dout->data.cptr + i*namlen,"LINE%d",i);
          }
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_CDIDEVNAMES:
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dFormat != CF_NAME32) return illegal_format;
        if (dout->dArrayLength < 1) return dimension_error;
        /* sscanf(devName,"#%d",&devnr); <- not necessary if .NAM property is registered */
        namlen = cdi_getDeviceNamesByLine((NAME32 *)dout->data.cptr, dout->dArrayLength, devnr);
        if (namlen <= 0) return no_such_line;
        dout->dArrayLength = namlen;
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_CDINUMBER:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength < 1) return dimension_error;
        switch (dout->dFormat)
        {
          case CF_ADDRESS:
            ((DADDRESS *)dout->data.vptr)[0].i1val = cdiDevNr;
            ((DADDRESS *)dout->data.vptr)[0].i2val = cdiLine;
            ((DADDRESS *)dout->data.vptr)[0].i3val = cdiIndex;
            break;
          default:
            cc = putValuesFromLong(dout,&cdiDevNr,1);
            cdiAssertHeapOkay("cdieqm",devProperty,cc);          
            return cc;

        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_CDIINFOR:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength < PRP_CDIINFOR_SIZE) return dimension_error;
        memset(lineinf,0,PRP_CDIINFOR_SIZE*sizeof(int));
        lineinf[0] = cdi_GetNumberRequests(cdiLine);
        lineinf[1] = cdi_GetNumberActiveRequests(cdiLine);
        lineinf[2] = cdi_GetNumberRequestsMax(cdiLine);
        flags = 1;
        if ( din->dArrayLength > 0 ) GetValuesAsLong(din,&flags,1);
        if ( flags < 1 ) flags = 1;
        cc = cdi_GetActiveRequestInfor(cdiLine, &lineinf[5], PRP_CDIINFOR_SIZE - 5, flags);
        lineinf[4] = flags;
        if ( !cc || cc != flags )
        {
          lineinf[4] = 0;
	        for (i=5; i<PRP_CDIINFOR_SIZE; i++) lineinf[i] = 0;
	        cc = 0;
        }
        switch (dout->dFormat)
        {
          case CF_NAME64I:
            for (i=0; i<PRP_CDIINFOR_SIZE; i++)
            {
              ((NAME64I *)dout->data.vptr)[i].ival = lineinf[i];
              if (i<nLineinfFieldNames) 
                strncpy(((NAME64I *)dout->data.vptr)[i].name,lineinfFieldNames[i],64);
            }
            break;
          default:
            cc = PutValuesFromLongEx(dout,lineinf,PRP_CDIINFOR_SIZE,0);
            break;
        }
        dout->dArrayLength = nLineinfFieldNames;
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_SRVDB:
        populateSrvItemList();
        return dbCallHndlr(prpid,devName,SRVITEM_TAG,dout,din,access);
    case PRP_MFDB:
        populateMfItemList();
        return dbCallHndlr(prpid,devName,MFITEM_TAG,dout,din,access);
    case PRP_SCHEDDB:
        populateSchItemList(); 
        return dbCallHndlr(prpid,devName,SCHDITEM_TAG,dout,din,access);
    case PRP_ADDRDB:
        populateDbItemList();
        return dbCallHndlr(prpid,devName,DBITEM_TAG,dout,din,access);
    case PRP_DESCRIPTION:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        numDevices = MAX_QUERYABLE_LIST_SIZE;
        if ((cc=cdiParseDeviceString(devName,qryDevList,&numDevices)) != 0) return cc;
        if (numDevices > MAX_QUERYABLE_LIST_SIZE) numDevices = MAX_QUERYABLE_LIST_SIZE;
        if (dout->dArrayLength < 1) return dimension_error;
        if (dout->dArrayLength > (UINT32)numDevices) dout->dArrayLength = numDevices;
        len = fmtsizeof(dout->dFormat);
        switch (dout->dFormat)
        {
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME64:
            namlen = len;
            break;
          case CF_TEXT:
            if (dout->dArrayLength > 1) return illegal_format;
            namlen = MIN(dout->dArrayLength,64);
            break;
          default:
            return illegal_format;
        }
        for (i=0; i<numDevices; i++)
        {
          cdi_getDescriptionForDevice(qryDevList[i].name,devDesc);
          strncpy(dout->data.cptr+i*namlen,devDesc,namlen);
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_ADDR:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          switch (dout->dFormat)
          {
            case CF_ADDRESS:
              if (dout->dArrayLength > PRP_ADDR_SIZE) return dimension_error;
              if ((dap=cdi_getBusAddressForDevice(devName)) == NULL) dap = &da;
              memcpy(dout->data.vptr,dap,dout->dArrayLength*fmtsizeof(dout->dFormat));
              break;
            case CF_INT16:
            case CF_INT32:
              if ((cc=cdi_getBusAddressParametersForDevice(devName,iaddr,NULL)) != 0) return cc;
              if ((cc=cdi_getDeviceParametersForDevice(devName,&iaddr[CDI_ADDR_SIZE])) != 0) return cc;
              if ((cc=PutValuesFromLongEx(dout,iaddr,2*CDI_ADDR_SIZE,0)) != 0) return cc;
              break;
            case CF_TEXT:
            case CF_NAME32:
            case CF_NAME64:
              if ((cc=cdi_getBusAddressParametersForDevice(devName,NULL,strbuffer)) != 0) return cc;
              if ((cc=PutValuesFromString(dout,strbuffer,CF_TEXT,CDI_DEVICE_LONGNAME_SIZE,0)) != 0) return cc;
              break;
            default:
              return illegal_format;
          }
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_BUS_NAME:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          switch (dout->dFormat)
          {
            case CF_TEXT:
              namlen = MIN(dout->dArrayLength,32);
            case CF_NAME16:
            case CF_NAME32:
              namlen = fmtsizeof(dout->dFormat);
              break;
            case CF_NAME64:
              namlen = 32;
              break;
            default:
              return illegal_format;
          }
          if ((c=cdi_getFieldBusNameForDevice(devName)) == NULL) c = "[none]";
          strncpy(dout->data.cptr,c,namlen);
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_BUS_TYPE:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_BUS_TYPE_SIZE) return dimension_error;
          if (dout->dFormat != CF_NAME16) return illegal_format;
          if ((c=cdi_getBusNameForDevice(devName)) == NULL) c = "[none]";
          memcpy(dout->data.vptr,c,dout->dArrayLength*fmtsizeof(dout->dFormat));
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_BUS_ERRORS:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          if (dout->dArrayLength > PRP_BUS_ERRORS_SIZE) return dimension_error;
          if ((errcnt=cdi_getDeviceNumberOfErrors(devName)) < 0) return -errcnt;
          return putValuesFromLong(dout,&errcnt,1);
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_BUS_NAMES:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        cc = cdiGetFieldBusNames(dout->data.cptr,dout->dArrayLength,dout->dFormat);
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return cc;
    case PRP_BUS_SCAN_NAM:
    case PRP_BUS_TYPES:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          switch (dout->dFormat)
          {
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME64:
              namlen = fmtsizeof(dout->dFormat);
              break;
            default:
              return illegal_format;
          }
          if (nCdiBusNames < (int)dout->dArrayLength) dout->dArrayLength = nCdiBusNames;
          if ((n16=cdiGetBusNames()) == NULL)
          {
            strcpy(dout->data.cptr,"[none]");
            dout->dArrayLength = 1;
          }
          else
          {
            for (i=0; i<(int)dout->dArrayLength; i++)
            {
              strncpy(dout->data.cptr + i*namlen,n16[i].name,16);
            }
          }
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_BUS_SCAN:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (din->dArrayLength > 0)
        {
          if (din->dFormat != CF_TEXT) return illegal_format;
          strncpy(strbuffer,din->data.cptr,din->dArrayLength);
          strbuffer[din->dArrayLength] = 0;
          scanTarget = strbuffer;
        }
        if (dout->dArrayLength > 0)
        {
          if (dout->dFormat != CF_TEXT) return illegal_format;
          if ((cc=cdiScanBus(devName,scanTarget,dout->data.cptr,dout->dArrayLength)) != 0) return cc;
          if (strlen(dout->data.cptr) < (size_t)dout->dArrayLength) 
            dout->dArrayLength = (UINT32)strlen(dout->data.cptr);
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_NTEMPLATEOBJECTS_NAM:
    case PRP_TEMPLATE_NAM:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (dout->dArrayLength > 0)
        {
          switch (dout->dFormat)
          {
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME64:
              namlen = fmtsizeof(dout->dFormat);
              break;
            default:
              return illegal_format;
          }
          if (CdiTemplateTblSize < (int)dout->dArrayLength) dout->dArrayLength = CdiTemplateTblSize;

          for (i=0; i<(int)dout->dArrayLength; i++)
            strncpy(dout->data.cptr + i*namlen,CdiTemplateTbl[i]->name,16);
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;  
    case PRP_NTEMPLATEOBJECTS:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (devnr < 0 || devnr >= CdiTemplateTblSize) return illegal_equipment_number;
        if ((tmpl=CdiTemplateTbl[devnr]) == NULL) return un_allocated;
        return putValuesFromLong(dout,&tmpl->numInstances,1);
    case PRP_TEMPLATEOBJECTS:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (devnr < 0 || devnr >= CdiTemplateTblSize) return illegal_equipment_number;
        if ((tmpl=CdiTemplateTbl[devnr]) == NULL) return un_allocated;
        switch (dout->dFormat)
        {
          case CF_NAME8:
          case CF_NAME16:
          case CF_NAME32:
          case CF_NAME48:
          case CF_NAME64:
            namlen = fmtsizeof(dout->dFormat);
            break;
          default:
            return illegal_format;
        }
        len = MIN(namlen,32);
        tmpln = tmpl->instanceNames;
        for (i=0; i<(int)dout->dArrayLength && i<tmpl->numInstances; i++)
        {
          if (tmpln == NULL) break;
          strncpy(dout->data.cptr+i*namlen,tmpln->name,len);
          tmpln = tmpln->nxt;
        }
        dout->dArrayLength = (UINT32)i;
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_TEMPLATE:
        if (devnr < 0) return illegal_equipment_number;
        if (access&CA_WRITE) return illegal_read_write;
        if (devnr < 0 || devnr >= CdiTemplateTblSize) return illegal_equipment_number;
        tmpl = CdiTemplateTbl[devnr];
        tmplst = tmpl->cdiTemplateList;
        if (dout->dArrayLength > 0)
        {
          len = fmtsizeof(dout->dFormat);
          switch (dout->dFormat)
          {
            case CF_NAME16II:
              namlen = len - 8;
              for (i=0; i<tmpl->length && i<(int)dout->dArrayLength && tmplst != NULL; i++) 
              {
                strncpy(dout->data.cptr + i*len,tmplst->field,16);
                *(int *)(dout->data.bptr + i*len + namlen) = tmplst->addrParameters[0];
                *(int *)(dout->data.bptr + i*len + namlen + 4) = tmplst->addrParameters[1];
                tmplst = tmplst->next;
              }
              dout->dArrayLength = i;
              break;
            case CF_NAME16I:
            case CF_NAME32I:
            case CF_NAME64I:
              namlen = len - 4;
              for (i=0; i<tmpl->length && i<(int)dout->dArrayLength && tmplst != NULL; i++) 
              {
                strncpy(dout->data.cptr + i*len,tmplst->field,16);
                *(int *)(dout->data.bptr + i*len + namlen) = tmplst->addrParameters[0];
                tmplst = tmplst->next;
              }
              dout->dArrayLength = i;
              break;
            case CF_NAME16:
            case CF_NAME32:
            case CF_NAME64:
              namlen = len;
              for (i=0; i<tmpl->length && i<(int)dout->dArrayLength && tmplst != NULL; i++) 
              {
                strncpy(dout->data.cptr + i*len,tmplst->field,16);
                tmplst = tmplst->next;
              }
              dout->dArrayLength = i;
              break;
            default:
              return illegal_format;
          }
        }
        cdiAssertHeapOkay("cdieqm",devProperty,cc);          
        return 0;
    case PRP_CDI_IDLE:
      if (access&CA_WRITE)
      {
        int idleState = gCdiIdleState;
        if (din->dArrayLength > 0)
        { /* explicitly set with input */
          if ((cc=GetValuesAsLong(din,&idleState,1)) != 0) return cc;
        }
        else
        { /* toggle the current idle state */
          idleState = (idleState+1)%2;
        }
        gCdiIdleState = idleState;
        gSetToIdleState = TRUE;
        /* cdiSetIdleState(idleState); */
      }
      if (dout->dArrayLength > 0)
      {
        return PutValuesFromLong(dout,&gCdiIdleState,1);
      }
      return 0;
    case PRP_CDIVERSION:
      return PutValuesFromString(dout,cdiGetVersionAsString(),CF_TEXT,32,0);
    default:
        if (extid < 0) 
        {
          return illegal_property;
        }
    case PRP_WRITEREADWRITE_CLBR:
    case PRP_WRITEWRITE_CLBR:
    case PRP_READWRITE_CLBR:
    case PRP_WRITE_CLBR:
    case PRP_WRITEREADWRITE:
    case PRP_READWRITE:
    case PRP_WRITEWRITE:
    case PRP_WRITE:
        if (!iswrite && extid < 0) return illegal_read_write; /* catch early (avoid listener allocation) */
    case PRP_READREAD_CLBR:
    case PRP_WRITEREAD_CLBR:
    case PRP_READ_CLBR:
    case PRP_WRITEREADREAD:
    case PRP_WRITEREADREAD_CLBR:
    case PRP_READREAD:
    case PRP_WRITEREAD:
    case PRP_READ:
        if (gCdiIdleState) return server_idle;
        if (!gHistoryLinksStarted)
        {
          time_t tnow = time(NULL);
          if (gFirstEqmCall == 0) gFirstEqmCall = tnow;
          if (gNumHistories == 0) 
          {
            gHistoryLinksStarted = TRUE;
          }
          else if (access & CA_HIST)
          { /* count the history calls as they come in */
            if (++gNumHistoryStarts == gNumHistories) gHistoryLinksStarted = TRUE;
          }
          else if (tnow > gFirstEqmCall + 5)
          { /* up to 5 seconds to let this sort itself out, otherwise business as usual */
            gHistoryLinksStarted = TRUE;
          }
          else
          { /* not yet !*/
            return not_initialized;
          }
        }
        if (!iswrite)
        { /* not a WRITE request -> check for bitfields */
          if ((bf=getBitfieldFromDeviceName(devName,field,FALSE)) != NULL)
          { /* a bitfield registered device : reject */
            if (access & (CA_HIST|CA_ALARM))
            { /* local history or local alarm subsystem */
              if ((i = (int)(strlen(dn)-strlen(field))-1) < 0) return code_failure;
              dn[i] = 0; /* strip off the field */
              dout->dFormat = GetIntegerFormat(dout->dFormat);
              switch (dout->dFormat)
              { /* make the full call */
                case CF_INT8:
                case CF_INT16:
                case CF_INT32:
                  if ((cc=ExecLinkEx(dn,devProperty,dout,din,access,2000)) != 0) ccerr(cc);
                  break;
                default:
                  return illegal_format;
              }
              for (i=0; i<(int)dout->dArrayLength; i++)
              {
                switch (dout->dFormat)
                {
                  case CF_INT8:
                    dout->data.bptr[i] = GetFieldFromBitfield(BITFIELD_LCL,bf->name,field,dout->data.bptr[i]);
                    break;
                  case CF_INT16:
                    dout->data.sptr[i] = GetFieldFromBitfield(BITFIELD_LCL,bf->name,field,dout->data.sptr[i]);
                    break;
                  case CF_INT32:
                    dout->data.lptr[i] = GetFieldFromBitfield(BITFIELD_LCL,bf->name,field,dout->data.lptr[i]);
                    break;
                }
              }
              cdiAssertHeapOkay("cdieqm",devProperty,cc);          
              return 0;
            }
            /* normal remote access gets handshaking information : */
            SetEqmCompletion(CDIEQM_TAG,bf->name);
            return has_bitfield_tag;
          }
        }
        if (dout->dFormat > CF_BITFIELD8) dout->dFormat = GetIntegerFormat(dout->dFormat);
        if (issync || iswrite)
        { /* is synchronous or is a write request : just do it */
          cc = ExecLinkEx(dn,devProperty,dout,din,access,2000);
          cdiAssertHeapOkay("cdieqm",devProperty,cc);          
          ccerr(cc);
        }
        else
        { /* return buffered data */
          drate = cdi_getDefaultInterval(cdiDevName);
          rate = isquery ? gDefaultQueryPollingRate : getContractAccessRate(-1);
          if (drate < rate) rate = drate;
          if (rate == 0) rate = DEFAULT_POLLING_INTERVAL;
          flags = isfirst ? CM_WAIT : 0; /* coming into scope ?*/
          cc = getDataFromListener(dn,devProperty,dout,din,&dt,rate,flags);
          setDataTimeStamp(dt);
          if (islast) 
          { /* going out of scope: stop listener */
            if ((ml=getListener(dn,devProperty,dout,din)) != NULL)
              stopListener(ml);
          }
          cdiAssertHeapOkay("cdieqm",devProperty,cc);          
          ccerr(cc);
        }
err:
        if (cc)
        {
          NAME16 un[5];
          UINT32 ip[5];
          char addr[32];
          struct in_addr inip;
          int num = 5;
          if (GetCallerInfo(CDIEQM_TAG,un,NULL,ip,NULL,&num) == 0)
          {
            for (i=0; i<5 && i<num; i++)
            {
              memcpy(&inip,&ip[i],4);
              inet_ntoa_b(inip,addr);
              cdilog("%s %s from caller %s @ %s returned %s",
                dn,devProperty,un,addr,cc2str(cc));
            }
          }
        }
        return cc;
  }
}
int dbCallHndlr(int pid,char *devName,char *tag,DTYPE *dout,DTYPE *din,short access)
{
  int cc, type=-1;
  if (access&CA_WRITE)
  {
    if (!AssertIsAdministrator(GetCaller(CDIEQM_TAG))) return access_denied;
    type = getDbUpdateType(devName);
    if (din->dArrayLength > 0)
    { /* specific input given */
      if (din->dFormat != CF_STRUCT) return illegal_format;
      if (strncmp(din->dTag,tag,TAG_NAME_SIZE)) return invalid_structure_tag;
      cc = dbUpdate(pid,type,din->data.vptr,din->dArrayLength);
    }
    else
    { /* signal a dump of the current database */
      if (type != DB_ROLLBACK) type = -1;
      cc = dbUpdate(pid,type,NULL,0);
    }
    if (cc!= 0) return cc;
  }
  if (dout->dArrayLength > 0)
  {
    if (dout->dFormat != CF_STRUCT) return illegal_format;
    if (strncmp(dout->dTag,tag,TAG_NAME_SIZE)) return invalid_structure_tag;
    if ((cc=dbGet(pid,dout->data.vptr,&dout->dArrayLength)) != 0) return cc;
  }
  return 0;
}
int lclcdieqm(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int cc=0,i,n;
  int numDevices;
  char devDesc[64];

  if (stricmp(devProperty,"TEMPLATE.NAM") == 0)
  {
    if (dout->dArrayLength < 1) return dimension_error;
    if (CdiTemplateTblSize < (int)dout->dArrayLength) dout->dArrayLength = CdiTemplateTblSize;
    for (i=0; i<CdiTemplateTblSize; i++) strncpy(qryDevList[i].name,CdiTemplateTbl[i]->name,16);
    return PutValuesFromString(dout,qryDevList[0].name,CF_NAME32,CdiTemplateTblSize,0);
  }
  if (stricmp(devProperty,"NINSTANCES") == 0)
  {
    if ((n=cdiGetNumTemplateInstances(devName)) < 0) return -n;
    return PutValuesFromLong(dout,&n,1);
  }
  if (stricmp(devProperty,"INSTANCES") == 0)
  {
    if (dout->dArrayLength < 1) return dimension_error;
    numDevices = MAX_QUERYABLE_LIST_SIZE;
    cc = cdiGetTemplateInstances(devName,qryDevList,&numDevices);
    if (numDevices < (int)dout->dArrayLength) dout->dArrayLength = numDevices;
    return PutValuesFromString(dout,qryDevList[0].name,CF_NAME32,numDevices,0);
  }
  if (stricmp(devProperty,"DESCRIPTION") == 0)
  {
    if (dout->dArrayLength < 1) return dimension_error;
    cc = cdi_getDescriptionForDevice(devName,devDesc);
    return PutValuesFromString(dout,devDesc,CF_NAME64,1,0);
  }
  if (stricmp(devProperty,"BUSADDR") == 0)
  {
    if (dout->dArrayLength > 0)
    {
      DADDRESS da = { 0,0,0,0 }, *dap;
      int iaddr[CDI_ADDR_SIZE];
      char strbuffer[256];
      switch (dout->dFormat)
      {
        case CF_ADDRESS:
          if (dout->dArrayLength > PRP_ADDR_SIZE) return dimension_error;
          if ((dap=cdi_getBusAddressForDevice(devName)) == NULL) dap = &da;
          memcpy(dout->data.vptr,dap,dout->dArrayLength*fmtsizeof(dout->dFormat));
          break;
        case CF_SHORT:
        case CF_INT32:
          if ((cc=cdi_getBusAddressParametersForDevice(devName,iaddr,NULL)) != 0) return cc;
          if ((cc=PutValuesFromLongEx(dout,iaddr,CDI_ADDR_SIZE,0)) != 0) return cc;
          break;
        case CF_TEXT:
        case CF_NAME32:
        case CF_NAME64:
          if ((cc=cdi_getBusAddressParametersForDevice(devName,NULL,strbuffer)) != 0) return cc;
          if ((cc=PutValuesFromString(dout,strbuffer,CF_TEXT,CDI_DEVICE_LONGNAME_SIZE,0)) != 0) return cc;
          break;
        default:
          return illegal_format;
      }
    }
    return 0;
  }
  return illegal_property;
}
void remapCdiDeviceAndProperty(char *devName,char *devProperty,DTYPE *dout, DTYPE *din,short access)
{
  int devnr=0,prpid,extid=-1,i,k,isArray=FALSE;
  char *c = NULL, strbuffer[256];
  char *dn = devName;

  devnr = GetDeviceNumberEx(CDIEQM_TAG,devName,devProperty);
  prpid = GetPropertyId(CDIEQM_TAG,devProperty);
  if (prpid >= PRP_EXTENDED_OFFSET &&
      prpid < PRP_EXTENDED_OFFSET + gNumExtCdiProperties)
  {
    extid = prpid - PRP_EXTENDED_OFFSET;
  }
  if (extid < 0) return;
  if (cdiDebugLevel) dbglog("remapCdiDeviceAndProperty: input %s %s",devName,devProperty);
  isArray = extCdiPrps[extid].isArray;
  if (strncmp(devName,devProperty,DEVICE_NAME_SIZE))
  { /* a split extended property ! */
    if (devnr < 0) return;
    sprintf(dn,"%s.%s",devName,devProperty);
    if (access & CA_WRITE) 
    { /* a "SEND" to bus command */
      if (din->dArrayLength > 1 && !isArray)
      {
        k = devnr + (int)din->dArrayLength - 1;
        if (k >= extCdiPrps[extid].ndevs) k = extCdiPrps[extid].ndevs - 1;
        if (extCdiPrps[extid].isContiguous)
        { /* contiguous list ? */
          if (k > devnr)
          {
            sprintf(strbuffer," - %s.%s",extCdiPrps[extid].devs[k].name,devProperty);
            strcat(dn,strbuffer);
          }
        }
        else
        { /* group them one-by-one then */
          for (i=devnr+1; i<=k; i++)
          {
            sprintf(strbuffer,",%s.%s",extCdiPrps[extid].devs[i].name,devProperty);
            strcat(dn,strbuffer);
            if (strlen(dn) >= (size_t)(cdiDevStringBufferSize-16))
            {
              k = i; break;
            }
          }
        }
        din->dArrayLength = k - devnr + 1;
      }
    }
    else
    { /* a "RECV" from bus command */
      if (dout->dArrayLength > 1 && dout->dFormat != CF_TEXT && !isArray)
      {
        k = devnr + (int)dout->dArrayLength - 1;
        if (k >= extCdiPrps[extid].ndevs) k = extCdiPrps[extid].ndevs - 1;
        if (extCdiPrps[extid].isContiguous)
        { /* contiguous list ? */
          if (k > devnr)
          {
            sprintf(strbuffer," - %s.%s",extCdiPrps[extid].devs[k].name,devProperty);
            strcat(dn,strbuffer);
          }
        }
        else
        { /* group them one-by-one then */
          for (i=devnr+1; i<=k; i++)
          {
            sprintf(strbuffer,",%s.%s",extCdiPrps[extid].devs[i].name,devProperty);
            strcat(dn,strbuffer);
            if (strlen(dn) >= (size_t)(cdiDevStringBufferSize-16))
            {
              k = i; break;
            }
          }
        }
        dout->dArrayLength = k - devnr + 1;
      }
    }
  }        
  if (access & CA_WRITE) 
  {
    strcpy(devProperty,"SEND.CLBR");
  }
  else 
  {
    strcpy(devProperty,"RECV.CLBR");
  }
  if (cdiDebugLevel) dbglog("remapCdiDeviceAndProperty: output %s %s",devName,devProperty);
}
